#include "plot_2dtypes.h"
#include <numeric>
#include <algorithm>
#include <limits>
#include <QDebug>

double max(const QVector<double> &data)
{
    if (!data.size())
        return -std::numeric_limits<double>::max();

    return *std::max_element(data.constBegin(), data.constEnd());
}
double min(const QVector<double> &data)
{
    if (!data.size())
        return std::numeric_limits<double>::max();

    return *std::min_element(data.constBegin(), data.constEnd());
}
QPair<double, double> minmax(const QVector<double> &data)
{
    if (!data.size())
        return QPair<double, double>(std::numeric_limits<double>::max(), -std::numeric_limits<double>::max());

    std::pair<QVector<double>::const_iterator, QVector<double>::const_iterator> minmax;
    minmax = std::minmax_element(data.constBegin(), data.constEnd());

    return QPair<double, double>(*minmax.first, *minmax.second);
}
double rms(const QVector<double> &data)
{
    double disp = 0;
    double average = avg(data);

    for (double d:data)
        disp += pow(d - average, 2);

    if (data.size())
        disp /= data.size();

    return sqrt(disp);
}
double avg(const QVector<double> &data)
{
    double Average = std::accumulate(data.constBegin(), data.constEnd(), 0.);

    if (data.size())
        Average /= data.size();

    return Average;
}

namespace plot {

TCurveData::TCurveData()
{
    isVerticaled = false;
    Pen.setWidth(1);
}
void TCurveData::detach()
{
    XSeq.detach();
    YSeq.detach();
}
void TCurveData::clear()
{
    XSeq.clear();
    YSeq.clear();
}
void TCurveData::AddY(double y)
{
    YSeq.append(y);
}
void TCurveData::AddPoint(double x, double y)
{
    XSeq.append(x);
    YSeq.append(y);
}

int TCurveData::size() const
{
    return (!isCompressedXCurve() && XSeq.size() < YSeq.size()) ? XSeq.size() : YSeq.size();
}
bool TCurveData::isCompressedXCurve() const
{
    return XSeq.size() == 2;
}
bool TCurveData::isCompressedYCurve() const
{
    return YSeq.size() == 2;
}
void TCurveData::setXCalib(double x0, double dx)
{
    XSeq.resize(2);
    XSeq.squeeze();
    XSeq[0] = x0;
    XSeq[1] = dx;
}
double TCurveData::getXat(int i) const
{
    return !isCompressedXCurve() ? XSeq.at(i) : XSeq.at(0) + i * XSeq.at(1);
}
double TCurveData::getYat(int i) const
{
    return !isCompressedYCurve() ? YSeq.at(i) : YSeq.at(0) + i * YSeq.at(1);
}

TCurveData::~TCurveData()
{
    clear();
}

TCurveData* TCurveData::clone(TCurveData *orig, QPen *pen)
{
    TCurveData* c = new TCurveData();

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
    c->XSeq.append(orig->XSeq);
    c->YSeq.append(orig->YSeq);
#else
    for (int i=0; i<orig->XSeq.size(); i++)
        c->XSeq.append(orig->XSeq.at(i));
    for (int i=0; i<orig->YSeq.size(); i++)
        c->YSeq.append(orig->YSeq.at(i));
#endif
    if (pen)
    {
        c->Pen = QPen(pen->color(), pen->width());
    }
    else
        c->Pen = QPen(orig->Pen.color(), orig->Pen.width());

    return c;
}
TCurveData* TCurveData::clone(QPen *pen)
{
    return clone(this, pen);
}
TCurves::~TCurves()
{
    DeleteAllCurves();
}
void TCurves::SetActiveCurve(int Active)
{
    ActiveCurve = Active;
}
TCurves::TCurves(QwtPlot *_Plot)
{
    vertical = false;
    ActiveCurve = -1;
    Plot = _Plot;
    MCCurve = new QwtPlotCurve;
    MCCurve->setSymbol(new QwtSymbol(QwtSymbol::Ellipse, QBrush(Qt::gray), QPen(Qt::blue), QSize(8, 8)));
    MCCurve->attach(Plot);
}

void TCurves::VerifyActiveCurve()
{
    if (ActiveCurve >= QwtCurves.size()) {
        SetActiveCurve(-1);
        MCCurve->setSamples(QVector<QPointF>());
    }
}
void TCurves::GetClosestPoint(double *x, double *y, const QPoint &Pos, int N) const
{
    if (QwtCurves.count() == 0)
        return; //TODO bug, when touching 2dplot when 3dplot Point info is active
	auto curve = QwtCurves.at(N);
    int idx = curve->closestPoint(Pos, nullptr);
    if (idx < 0 || idx >= (int)curve->dataSize()) {
        *x = *y = 0;
        return;
    }

    *x = curve->sample(idx).x();
    *y = curve->sample(idx).y();
}

//requires ownership in gui thread
class TQwtCurveData : public QwtSeriesData<QPointF>
{
public:
    TQwtCurveData(const QWeakPointer<TCurveData> &_data) : data(_data) { }
    int size() const override {
        if (!data)
            return 0;

        return data.data()->size();
    }
    QPointF sample(int i) const override {
        return QPointF(data.data()->XSeq.at(i), data.data()->YSeq.at(i));
    }
    QRectF boundingRect() const override {
        if (!data)
            return QRectF();

        QPair<double, double> _minmaxx = minmax(data.data()->XSeq);
        double xmin = _minmaxx.first;
        double xmax = _minmaxx.second;
        QPair<double, double> _minmax = minmax(data.data()->YSeq);
        double ymin = _minmax.first;
        double ymax = _minmax.second;
        return QRectF(QPointF(xmin, ymin), QPointF(xmax, ymax));
    }

    QWeakPointer<TCurveData> data;
};

class TQwtCurveDataCompressedX : public TQwtCurveData
{
public:
    TQwtCurveDataCompressedX(const QWeakPointer<TCurveData> &_data) : TQwtCurveData(_data) { }
    QPointF sample(int i) const override {
        return QPointF(data.data()->XSeq.at(0) + i * data.data()->XSeq.at(1), data.data()->YSeq.at(i));
    }
    QRectF boundingRect() const override {
        if (!data)
            return QRectF();

        double xmin = data.data()->getXat(0);
        double xmax = data.data()->getXat(data.data()->size()-1);
        if (xmin > xmax)
            qSwap(xmin, xmax);
        QPair<double, double> _minmax = minmax(data.data()->YSeq);
        double ymin = _minmax.first;
        double ymax = _minmax.second;
        return QRectF(QPointF(xmin, ymin), QPointF(xmax, ymax));
    }
};
class TQwtCurveDataCompressedY : public TQwtCurveData
{
public:
    TQwtCurveDataCompressedY(const QWeakPointer<TCurveData> &_data) : TQwtCurveData(_data) { }
    QPointF sample(int i) const override {
        return QPointF(data.data()->YSeq.at(i), data.data()->XSeq.at(0) + i * data.data()->XSeq.at(1));
    }
    QRectF boundingRect() const override {
        if (!data)
            return QRectF();

        double ymin = data.data()->getYat(0);
        double ymax = data.data()->getYat(data.data()->size()-1);
        if (ymin > ymax)
            qSwap(ymin, ymax);
        QPair<double, double> _minmax = minmax(data.data()->XSeq);
        double xmin = _minmax.first;
        double xmax = _minmax.second;
        return QRectF(QPointF(xmin, ymin), QPointF(xmax, ymax));
    }
};

void TCurves::ReSyncDataAndQwtCurves()
{
    int target = PaintData.size();
    int curr = QwtCurves.size();

    if (!target)
        MCCurve->setSamples(QVector<QPointF>());

    for (int i = target; i < curr; i++)
        delete QwtCurves[i];

    QwtCurves.resize(target);
    LastPaintSize.resize(target);
    for (int i = curr; i < target; i++) {
        LastPaintSize[i] = 0;
        QwtCurves[i] = new QwtPlotCurve;
        QwtCurves[i]->setRenderHint(QwtPlotItem::RenderAntialiased, true);
        QwtCurves[i]->attach(Plot);
    }

    for (int i = 0; i < QwtCurves.size(); i++)
        if (PaintData.at(i)) {
            LastPaintSize[i] = PaintData.at(i).data()->size();
            if (PaintData.at(i).data()->isCompressedXCurve() || PaintData.at(i).data()->isCompressedYCurve())
            {
                if (vertical)
                {
//                    if (!PaintData.at(i).data()->isVerticaled)
//                    {
//                        TCurveData* curve = PaintData.at(i).data();
//                        auto xseq = curve->XSeq;
//                        curve->XSeq = curve->YSeq;
//                        curve->YSeq = xseq;
//                        curve->isVerticaled = true;
//                        QWeakPointer<TCurveData> weak(QWeakPointer<TCurveData>(*(new QSharedPointer<TCurveData>(curve))));
//                        QwtCurves[i]->setData(new TQwtCurveDataCompressedY(weak));
//                    }
//                    else
                        QwtCurves[i]->setData(new TQwtCurveDataCompressedY(PaintData.at(i)));
                }
                else
                    QwtCurves[i]->setData(new TQwtCurveDataCompressedX(PaintData.at(i)));
            }
            else
            {
                auto dt = new TQwtCurveData(PaintData.at(i));
                if (vertical && !dt->data.data()->isVerticaled)
                {
                    auto dt2 = new TQwtCurveData(PaintData.at(i));
					auto data = dt2->data.data();
                    auto xseq = data->XSeq;
					data->XSeq = data->YSeq;
					data->YSeq = xseq;
					data->isVerticaled = true;
                    QwtCurves[i]->setData(dt2);
                }
                else
                    QwtCurves[i]->setData(dt);
            }
            //QwtCurves[i]->setSamples(PaintData.at(i).data()->XSeq, PaintData.at(i).data()->YSeq);
            QwtCurves[i]->setPen(PaintData.at(i).data()->Pen);
        } else
            QwtCurves[i]->setSamples(QVector<QPointF>());
}
void TCurves::DeleteAllCurves()
{
    for (int i = 0; i < QwtCurves.size(); i++)
        delete QwtCurves[i];

    QwtCurves.clear();
    MCCurve->setSamples(QVector<QPointF>());
}
void TCurves::UnselectCurves()
{
    for (int i = 0; i < QwtCurves.size(); i++)
        if (PaintData.at(i))
            QwtCurves[i]->setPen(PaintData.at(i).data()->Pen);
}
void TCurves::SelectCurve(int N)
{
    for (auto &c:QwtCurves)
        c->setPen(DefaultPenForInactiveCurves);

    if (PaintData.at(N)) {
        QPen FatPen = PaintData.at(N).data()->Pen;
        FatPen.setWidth(3);
        QwtCurves[N]->setPen(FatPen);
    }
}
void TCurves::DefineLimits(QVector<double> &MaxX, QVector<double> &MaxY, QVector<double> &MinX, QVector<double> &MinY, int N)
{
    if (N < 0)
        for (int i = 0; i < PaintData.size(); i++)
            DefineLimits(MaxX, MaxY, MinX, MinY, i);
    else if (N < PaintData.size() && PaintData.at(N)) {
		auto data = PaintData.at(N).data();
        if (data->isCompressedXCurve() || data->isCompressedYCurve())
        {
			double xmin = data->getXat(0);
            double xmax = data->getXat(data->size()-1);

            /*double xmin;
            double xmax;

            if (vertical)
            {
                xmin = data->getYat(0);
                xmax = data->getYat(data->size()-1);
            }
            else
            {
                xmin = data->getXat(0);
                xmax = data->getXat(data->size()-1);
            }*/

            if (xmin > xmax)
                qSwap(xmin, xmax);
            MinX.push_back(xmin);
            MaxX.push_back(xmax);
        }
        else
        {
            QPair<double, double> _minmax = minmax(data->XSeq);
            MinX.push_back(_minmax.first);
            MaxX.push_back(_minmax.second);
        }
        QPair<double, double> _minmax = minmax(data->YSeq);
        MinY.push_back(_minmax.first);
        MaxY.push_back(_minmax.second);
    }
}

void TCurves::CleanData()
{
    TCurvesData data;

    for (const auto &d:SourceData)
        if (d)
            data.append(d);

    SourceData = data;
}

}
