#include "plot_core.h"

#include <QMouseEvent>
#include <QApplication>
#include <QDesktopWidget>
#include <QPainter>
#include <QPainterPath>
#include <QGestureEvent>
#include "widgets/plot_widgets.h"
#include <limits>
#include <qwt_plot_renderer.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_layout.h>
#include <qwt_plot_marker.h>
#include <qwt_plot_grid.h>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtMath>
#else
#include <qmath.h>
#include <qnumeric.h>
#include <QPinchGesture>
#endif

namespace plot {

QImage ImageItem::renderImage (const QwtScaleMap &xMap,
                               const QwtScaleMap &yMap, const QRectF &area, const QSize &imageSize) const
{
    if (targetRect.isEmpty())
        return image;

    QImage result(imageSize, QImage::Format_ARGB32);
    result.fill(Qt::black);

    QTransform t;
    double sx = xMap.transform(targetRect.width() / image.width()) - xMap.transform(0);
    double sy = -(yMap.transform(targetRect.height() / image.height()) - yMap.transform(0));
    t.scale(sx, sy);
    double left = (targetRect.left() - area.left()) * image.width() / targetRect.width();
    double top = -(targetRect.bottom() - area.bottom()) * image.height() / targetRect.height();
    t.translate(left, top);

    QPainter p(&result);
    p.setTransform(t);
    p.drawImage(0, 0, image);
    p.end();

    return result;
}

void TSIUnit::UpdateMultiplier(double Value)
{
    Value = fabs(Value);

    if (SIUnit.isEmpty() || !Value) {
        StrMultiplier.clear();
        Multiplier = 1;
        return;
    }

    typedef QPair<const char *, double> PairStrDbl;
    const PairStrDbl Multipliers[] = {
        PairStrDbl(QT_TR_NOOP("M"), 1e6), PairStrDbl(QT_TR_NOOP("k"), 1e3),
        PairStrDbl("", 1), PairStrDbl(QT_TR_NOOP("m"), 1e-3),
        PairStrDbl(QT_TR_NOOP("micro"), 1e-6), PairStrDbl(QT_TR_NOOP("n"), 1e-9),
        PairStrDbl(QT_TR_NOOP("p"), 1e-12)
    };
    const int size = sizeof(Multipliers) / sizeof(*Multipliers) - 1;
    const double Bound = 1.1;
    int FoundIndex = 2; //unit 1

    if (Value < Bound * Multipliers[size].second)
        FoundIndex = size;

    if (Value > Bound * Multipliers[0].second)
        FoundIndex = 0;

    for (int i = 0; i <= size; i++)
        if (Value / Multipliers[i].second <= 1000*Bound && Value / Multipliers[i].second >= Bound) {
            FoundIndex = i;
            break;
        }

    StrMultiplier = qApp->translate("plot::TSIUnit", Multipliers[FoundIndex].first);
    Multiplier = Multipliers[FoundIndex].second;
}

void TSIUnit::UpdateLegend()
{
    Legend = Description + (Description.isEmpty() || SIUnit.isEmpty() ? "" : ", ") + StrMultiplier + SIUnit;
}

//================================================================================================================================
void TValueInfoCreator::operator()(bool AutoLegendEnabled, const TSIUnit &UnitInfo, double Val)
{
    SrcValue = Val;
    if (AutoLegendEnabled) {
        TSIUnit Temp = UnitInfo; 
        Temp.UpdateLegend(Val);
        Val /= Temp.Multiplier;
        Unit = Temp.StrMultiplier + Temp.SIUnit;
    } else {
        Unit.clear();
    }
    Value = QString::number(Val, 'g', 5);
}

//================================================================================================================================
TInfoForAllPlots::TInfoForAllPlots() : AutoLegendEnabled(true)
{
    FontSize = qApp->font().pointSize() + 2; //12

    CurrentXLimits[0] = CurrentYLimits[0] = 0;
    CurrentXLimits[1] = CurrentYLimits[1] = 1;
    PlotsXLimits[0] = PlotsYLimits[0] = 0;
    PlotsXLimits[1] = PlotsYLimits[1] = 1;
    DefaultXLimits[0] = DefaultYLimits[0] = 0;
    DefaultXLimits[1] = DefaultYLimits[1] = 1;
}

//================================================================================================================================
void TPlotContainer::showEvent(QShowEvent *)
{
    resizeEvent(nullptr);
}
void TPlotContainer::resizeEvent(QResizeEvent *)
{
    ResizeQPlot(true);
    Replot();
}

QPoint TPlotContainer::mapToCanvas(const QPoint &Pos) const
{
    QPlot->plotLayout()->activate(QPlot, QPlot->rect());
    return QPlot->canvas()->mapFrom((QWidget *)this, Pos);
    /*int UpSpace = (QPlot->plotLayout()->titleRect().height() ? QPlot->plotLayout()->titleRect().height() + 5 : 0) + QPlot->plotLayout()->scaleRect(QwtPlot::xTop).height();
    int LeftSpace = QPlot->plotLayout()->scaleRect(QwtPlot::yLeft).width();
    return Pos - QPoint(LeftSpace, UpSpace);*/
}

void TPlotContainer::UpdateScalesDiv()
{
    QPlot->axisScaleDiv(QwtPlot::xBottom)->invalidate();
    QPlot->axisScaleDiv(QwtPlot::yLeft)->invalidate();
    QPlot->updateAxes(); //if we want to use QPlot->transform() or invTransform(), or QPlot->axisScaleDiv()->lowerBound() instead of CurrentXLimits[]
}


template<class T>
void TPlotDecorated<T>::paintEvent(QPaintEvent *)
{
    QPainter p(this);

    static_cast<T *>(this)->DrawBase(p);
    static_cast<T *>(this)->DrawSource(p);
}

//================================================================================================================================
TPlotSelectionData::TPlotSelectionData()
{
    Vertical = false;
    SelectRectEventOrigin = TSelectEvent::SRO_Unknown;
    SelectionModeEnabled = false;
    SelectAction = None;
    Rectangle = QRectF(200., 100., 80., 80.);
    link = nullptr;
}
void TPlotSelectionData::Updatekxky() const
{
    bool QPlotIsShown = link->QPlot->isVisible();
    if (!QPlotIsShown) link->QPlot->show();
    if (!link->QPlot->size().isNull()) {
        kx = fabs(link->ScreenXToPlotX(20) - link->ScreenXToPlotX(0)) / 20.;
        ky = fabs(link->ScreenYToPlotY(20) - link->ScreenYToPlotY(0)) / 20.;
    }
    if (!QPlotIsShown) link->QPlot->hide();
}
void TPlotSelectionData::UpdateCanvasRect()
{
    link->QPlot->plotLayout()->activate(link->QPlot, link->QPlot->rect());
    QRectF canvrect = link->QPlot->plotLayout()->canvasRect().adjusted(2, 3, -1, 0);
    if (canvrect.isValid())
        CanvasRectInArtFr = canvrect;
    Rectangle = Rectangle.normalized();
}
void TPlotSelectionData::DefineAction(QMouseEvent *Event)
{
    int x = Event->pos().x();
    int y = Event->pos().y();
    const int max_dist = 5;

    if (y > Rectangle.y() - max_dist &&
        y < Rectangle.bottom() + max_dist)
    {
        if (x > Rectangle.x() - max_dist && x < Rectangle.x() + max_dist)
        {
            SelectAction = MoveLeft;
            return;
        } else if (x > Rectangle.right() - max_dist && x < Rectangle.right() + max_dist)
        {
            SelectAction = MoveRight;
            return;
        }
    }
    if (x > Rectangle.x() - max_dist &&
        x < Rectangle.right() + max_dist)
    {
        if ( y > Rectangle.y() - max_dist && y < Rectangle.y() + max_dist)
        {
            SelectAction = MoveTop;
            return;
        } else if (y > Rectangle.bottom() - max_dist && y < Rectangle.bottom() + max_dist)
        {
            SelectAction = MoveBottom;
            return;
        }
    }
    if (Rectangle.contains(Event->pos()))
    {
        SelectAction = Panning;
        return;
    }
    SelectAction = None;
}
void TPlotSelectionData::ManageCursorShape(QMouseEvent *Event)
{
    Qt::CursorShape sh = Qt::ForbiddenCursor;
    if (SelectAction == MoveLeft || SelectAction == MoveRight) sh = Qt::SizeHorCursor;
    else if (SelectAction == MoveTop || SelectAction == MoveBottom) sh = Qt::SizeVerCursor;
    else if (SelectAction == Panning) sh = Event->buttons() != Qt::LeftButton ? Qt::OpenHandCursor : Qt::ClosedHandCursor;
    else if (SelectAction == None) sh = Qt::ForbiddenCursor;
    link->setCursor(sh);
}
template<class T>
void TPlotSelection<T>::mousePressEvent(QMouseEvent *Event)
{
    if (!SelectionModeEnabled) return;
    DefineAction(Event);
    ManageCursorShape(Event);
    Offset = Rectangle.center() - Event->pos();
}
template<class T>
void TPlotSelection<T>::mouseMoveEvent(QMouseEvent *Event)
{
    if (!SelectionModeEnabled) return;
    if (Event->buttons() != Qt::LeftButton)
    {
        DefineAction(Event);
        ManageCursorShape(Event);
    }
    else
    {
        if (SelectAction == Panning) {
            Rectangle = QRectF(Event->pos() + Offset, Rectangle.size());
            Rectangle.translate(-Rectangle.width() / 2., -Rectangle.height() / 2.);
        }
        if (SelectAction == MoveBottom)
        {
            qreal NewHeight = Event->posF().y() - Rectangle.y();
            if (NewHeight <= 0)
            { Rectangle.setHeight(.0); SelectAction = MoveTop; }
            else Rectangle.setHeight(Event->posF().y() - Rectangle.y());
        }
        else if (SelectAction == MoveLeft)
        {
            qreal NewWidth = Rectangle.right() - Event->posF().x();
            if (NewWidth <= 0)
            { Rectangle.setX(Rectangle.right()); SelectAction = MoveRight; }
            else Rectangle.setX(Event->posF().x());
        }
        else if (SelectAction == MoveRight)
        {
            qreal NewWidth = Event->posF().x() - Rectangle.x();
            if (NewWidth <= 0)
            { Rectangle.setWidth(.0); SelectAction = MoveLeft; }
            else Rectangle.setWidth(NewWidth);
        }
        else if (SelectAction == MoveTop)
        {
            qreal NewHeight = Rectangle.bottom() - Event->posF().y();
            if (NewHeight <= 0)
            { Rectangle.setY(Rectangle.bottom()); SelectAction = MoveBottom; }
            else Rectangle.setY(Event->posF().y());
        }

        if (SelectAction == MoveBottom && Rectangle.bottom() >= CanvasRectInArtFr.bottom()) Rectangle.setBottom(CanvasRectInArtFr.bottom());
        if (SelectAction == MoveLeft && Rectangle.left() <= CanvasRectInArtFr.left())       Rectangle.setLeft(CanvasRectInArtFr.left());
        if (SelectAction == MoveRight && Rectangle.right() >= CanvasRectInArtFr.right())    Rectangle.setRight(CanvasRectInArtFr.right());
        if (SelectAction == MoveTop && Rectangle.top() <= CanvasRectInArtFr.top())          Rectangle.setTop(CanvasRectInArtFr.top());
        if (SelectAction == Panning)
        {
            if (Rectangle.right() >= CanvasRectInArtFr.right()) Rectangle.moveRight(CanvasRectInArtFr.right());
            if (Rectangle.bottom() >= CanvasRectInArtFr.bottom()) Rectangle.moveBottom(CanvasRectInArtFr.bottom());
            if (Rectangle.left() <= CanvasRectInArtFr.left()) Rectangle.moveLeft(CanvasRectInArtFr.left());
            if (Rectangle.top() <= CanvasRectInArtFr.top()) Rectangle.moveTop(CanvasRectInArtFr.top());
        }

        ReDrawSelectedRectangle(TSelectEvent::SRO_Mouse);
    }
}
template<class T>
void TPlotSelection<T>::mouseReleaseEvent(QMouseEvent *Event)
{
    if (!SelectionModeEnabled) return;
    DefineAction(Event);
    ManageCursorShape(Event);
}
template<class T>
void TPlotSelection<T>::resizeEvent(QResizeEvent *e)
{
    if (SelectionModeEnabled) {
        QRectF OldSelRect = GetSelectedRect();
        T::resizeEvent(e);
        ResizeEvent(OldSelRect);
    } else
        T::resizeEvent(e);
}


void TPlotSelectionData::ReDrawSelectedRectangle(int from)
{
    link->update();
    SelectRectEventOrigin = from;
    SelectRectChanged(GetSelectedRect());
}
void TPlotSelectionData::ResizeEvent(QRectF prevsel)
{
    UpdateCanvasRect();
    Updatekxky();

    //restore old selection
    QPointF Point;
    Point.setX(prevsel.center().x() - (*GetCurXLims())[0]);
    Point.setY(prevsel.center().y() - (*GetCurYLims())[0]);

    QPointF center(CanvasRectInArtFr.left() + Point.x() / kx, CanvasRectInArtFr.bottom() - Point.y() / ky);

    SetSelect(center, fabs(prevsel.width()) / kx, fabs(prevsel.height()) / ky); //to pixel on screen
}

void TPlotSelectionData::SetSelect(QPointF center, qreal width, qreal height) //px
{
    qreal HalfW = width / 2.;
    qreal HalfH = height / 2.;

    HalfW = qMin(HalfW, CanvasRectInArtFr.right() - center.x());
    HalfW = qMin(HalfW, center.x() - CanvasRectInArtFr.left());
    Rectangle.setX(center.x() - HalfW);
    Rectangle.setWidth(2. * HalfW);

    HalfH = qMin(HalfH, CanvasRectInArtFr.bottom() - center.y());
    HalfH = qMin(HalfH, center.y() - CanvasRectInArtFr.top());
    Rectangle.setY(center.y() - HalfH);
    Rectangle.setHeight(2. * HalfH);
}
void TPlotSelectionData::SetSelectCenter(QPointF Point)
{
    if (!SelectionModeEnabled) return;

    Point.setX(Point.x() - (*GetCurXLims())[0]);
    Point.setY(Point.y() - (*GetCurYLims())[0]);

    QPointF center(CanvasRectInArtFr.left() + Point.x() / kx, CanvasRectInArtFr.bottom() - Point.y() / ky);

    SetSelect(center, Rectangle.width(), Rectangle.height());

    ReDrawSelectedRectangle(TSelectEvent::SRO_SetSelect);
}
void TPlotSelectionData::SetSelectWidthRelCenter(qreal PlotW)
{
    if (!SelectionModeEnabled) return;

    SetSelect(Rectangle.center(), fabs(PlotW) / kx, Rectangle.height()); //to pixel on screen

    ReDrawSelectedRectangle(TSelectEvent::SRO_SetSelect);
}
void TPlotSelectionData::SetSelectHeightRelCenter(qreal PlotH)
{
    if (!SelectionModeEnabled) return;

    SetSelect(Rectangle.center(), Rectangle.width(), fabs(PlotH) / ky); //to pixel on screen

    ReDrawSelectedRectangle(TSelectEvent::SRO_SetSelect);
}
QRectF TPlotSelectionData::TransformForeignRectToRectangle(QRectF fRect) const
{
    QRectF rect;
    rect.setLeft((fRect.x() - (*GetCurXLims())[0])/kx + CanvasRectInArtFr.left());
    rect.setBottom(CanvasRectInArtFr.bottom() - (fRect.y() - (*GetCurYLims())[0])/ky);
    rect.setRight(fRect.width()/kx + rect.left());
    rect.setTop(rect.bottom() - fRect.height()/ky);
    return rect;
}
QRectF TPlotSelectionData::GetSelectedRect() const
{
    QPointF ScrTopLeft(Rectangle.left() - CanvasRectInArtFr.left(), CanvasRectInArtFr.bottom() - Rectangle.top());
    QPointF ScrBottomRight(Rectangle.right() - CanvasRectInArtFr.left(), CanvasRectInArtFr.bottom() - Rectangle.bottom());
    QPointF TopLeft(ScrTopLeft.x() * kx, ScrTopLeft.y() * ky);
    QPointF BottomRight(ScrBottomRight.x() * kx, ScrBottomRight.y() * ky);
    QRectF rect = QRectF(TopLeft.x() + (*GetCurXLims())[0], BottomRight.y() + (*GetCurYLims())[0],
            BottomRight.x() - TopLeft.x(), TopLeft.y() - BottomRight.y());

    return rect;
}
void TPlotSelectionData::SetSelectionModeEnabled(const bool State)
{
    if (SelectionModeEnabled == State) return;
    link->setMouseTracking(State);
    if ((SelectionModeEnabled = State))
    {
        UpdateCanvasRect();
        Updatekxky();
        ReInitSelectRectGeometry(2.);
    }
}
void TPlotSelectionData::ReInitSelectRectGeometry(const double k)
{
    double dx = (*GetCurXLims())[1] - (*GetCurXLims())[0];
    double dy = (*GetCurYLims())[1] - (*GetCurYLims())[0];

    QPointF Point;
    Point.setX(dx / 2.);
    Point.setY(dy / 2.);

    QPointF center(CanvasRectInArtFr.left() + Point.x() / kx, CanvasRectInArtFr.bottom() - Point.y() / ky);

    SetSelect(center, fabs(dx / k) / kx, fabs(dy / k) / ky); //to pixel on screen
}
void TPlotSelectionData::SetSelectWidth(qreal PlotW)
{
    if (!SelectionModeEnabled) return;

    qreal W = fabs(PlotW) / kx; // to pixel on screen
    SetSelect(Rectangle.center(), W, Rectangle.height());

    if (Rectangle.width() < W)
    {  //Find free area - right or left?
        if (Rectangle.left() - CanvasRectInArtFr.left() > 0) //There is some free space at right hand
        {
            const qreal FreeSpace = Rectangle.left() - CanvasRectInArtFr.left();
            const qreal ExpandAt = qMin(W - Rectangle.width(), FreeSpace);
            Rectangle.setLeft(Rectangle.left() - ExpandAt);
        }
        else if (CanvasRectInArtFr.right() - Rectangle.right() > 0) //...left
        {
            const qreal FreeSpace = CanvasRectInArtFr.right() - Rectangle.right();
            const qreal ExpandAt = qMin(W - Rectangle.width(), FreeSpace);
            Rectangle.setRight(Rectangle.right() + ExpandAt);
        }
    }

    ReDrawSelectedRectangle(TSelectEvent::SRO_SetSelect);
}
void TPlotSelectionData::SetSelectHeight(qreal PlotH)
{
    if (!SelectionModeEnabled) return;

    const qreal H = fabs(PlotH) / ky; // to pixel on screen
    SetSelect(Rectangle.center(), Rectangle.width(), H);

    if (Rectangle.height() < H)
    {  //Find free area - up or down?
        if (Rectangle.top() - CanvasRectInArtFr.top() > 0) //There is some free space up
        {
            const qreal FreeSpace = Rectangle.top() - CanvasRectInArtFr.top();
            const qreal ExpandAt = qMin(H - Rectangle.height(), FreeSpace);
            Rectangle.setTop(Rectangle.top() - ExpandAt);
        }
        else if (CanvasRectInArtFr.bottom() - Rectangle.bottom() > 0) //...left
        {
            const qreal FreeSpace = CanvasRectInArtFr.bottom() - Rectangle.bottom();
            const qreal ExpandAt = qMin(H - Rectangle.height(), FreeSpace);
            Rectangle.setBottom(Rectangle.bottom() + ExpandAt);
        }
    }

    ReDrawSelectedRectangle(TSelectEvent::SRO_SetSelect);
}

template <class T>
void TPlotSelection<T>::DrawBase(QPainter &p)
{
    if (!SelectionModeEnabled) {
        T::DrawBase(p);
        return;
    }

    if (m_base_buffer.size() != size()) {
        m_base_buffer = QPixmap(size());
        m_base_buffer.fill(QColor(0, 0, 0, 0));

        QPainter p(&m_base_buffer);
        if (p.isActive()) {
            T::DrawBase(p);

            QwtPlotRenderer().render(QPlot, &p, QPlot->rect());
        }
    }

    p.drawPixmap(0, 0, m_base_buffer);
}
template <class T>
void TPlotSelection<T>::DrawSource(QPainter &p)
{
    T::DrawSource(p);

    if (!SelectionModeEnabled)
        return;

    p.setCompositionMode(QPainter::CompositionMode_HardLight);
    p.setPen(QPen(QColor(234, 117, 0), 2, Qt::SolidLine));
    p.setRenderHint(QPainter::Antialiasing);
    double m_circle_hue = 255., m_circle_alpha = 100.;
    QColor Colour = QColor::fromHsvF(m_circle_hue / 360.0, 1, 1, m_circle_alpha / 255.0);
    p.setBrush(Colour);
    p.drawRect(Rectangle);
}

//================================================================================================================================
template <class T>
void TPlotEnable<T>::DrawSource(QPainter &p)
{
    if (Enabled)
        T::DrawSource(p);
}
template <class T>
void TPlotEnable<T>::DrawBase(QPainter &p)
{
    if (Enabled) {
        T::DrawBase(p);
        return;
    }

    if (m_base_buffer.size() != size()) {
        m_base_buffer = QPixmap(size());
        m_base_buffer.fill(QColor(0, 0, 0, 0));

        QPainter p(&m_base_buffer);
        if (p.isActive()) {
            T::DrawBase(p);

            QwtPlotRenderer().render(QPlot, &p, QPlot->rect());

            p.setCompositionMode(QPainter::CompositionMode_Multiply);
            p.setPen(Qt::NoPen);
            p.setBrush(QColor(128, 128, 128, 200));
            p.drawRect(rect());
        }
    }

    p.drawPixmap(0, 0, m_base_buffer);
}
template <class T>
void TPlotActive<T>::DrawBase(QPainter &p)
{
    if (!PlotIsActive) {
        T::DrawBase(p);
        return;
    }

    p.setPen(Qt::NoPen);
    //p.setBrush(QColor::fromRgb(qApp->palette().color(QPalette::Window).darker(106).rgb() + 9));//QColor(227, 227, 236)); use palette as reference
    p.setBrush(QColor(240, 240, 240));
    p.drawRect(rect());
}
template <class T>
void TPlotBackground<T>::DrawBase(QPainter &p)
{
    T::DrawBase(p);

    p.setPen(Qt::NoPen);
    //p.setBrush(qApp->palette().color(QPalette::Window).darker(120)); //QColor(200, 200, 200)); use palette as reference
    p.setBrush(QColor(255, 255, 255));
    p.drawRect(rect());
}
template <class T>
void TPlotRoundedBorder<T>::DrawBase(QPainter &p)
{
    T::DrawBase(p);

    const int radius = 16;
    const int level = 180;
    static QPixmap cornerCache[4];
    if (cornerCache[0].isNull()) {
        cornerCache[0] = QPixmap(radius+2, radius+2);
        cornerCache[0].fill(QColor(0, 0, 0, 0));
        QPainter p(&cornerCache[0]);
        if (p.isActive()) {
            QPainterPath cornersFillPath;
            cornersFillPath.arcTo(0, 0, radius+2, radius+2, 180, -90);
            cornersFillPath.closeSubpath();

            p.setPen(Qt::NoPen);
            p.setBrush(qApp->palette().window());
            p.drawPath(cornersFillPath);

            p.setRenderHint(QPainter::Antialiasing);
            p.setPen(QPen(QColor(level, level, level), 2));
            p.drawArc(1, 1, radius+1, radius+1, 180*16, -90*16);
            p.end();
        }
        QImage tmp = cornerCache[0].toImage();
        cornerCache[1] = QPixmap::fromImage(tmp.mirrored(true, false));
        cornerCache[2] = QPixmap::fromImage(tmp = tmp.mirrored(false, true));
        cornerCache[3] = QPixmap::fromImage(tmp.mirrored(true, false));
    }

    p.setPen(QPen(QColor(level, level, level), 2));
    p.setBrush(Qt::NoBrush);
    p.drawRect(rect().adjusted(1, 1, -1, -1));
    //p.drawRoundedRect(rect().adjusted(1, 1, -1, -1), 16, 16);

    p.drawPixmap(rect().topLeft(), cornerCache[0]);
    p.drawPixmap(rect().topRight()-QPoint(radius+2-1, 0), cornerCache[1]);
    p.drawPixmap(rect().bottomLeft()-QPoint(0, radius+2-1), cornerCache[2]);
    p.drawPixmap(rect().bottomRight()-QPoint(radius+2-1, radius+2-1), cornerCache[3]);
}
//================================================================================================================================
TPlotCore::TPlotCore(QWidget *parent)
{
    setParent(parent);

    QPlot = new QwtPlot_(this);
    QPlot->canvas()->setStyleSheet("border: 1px solid gray; background-color: #E6FFE6;");
    QPlot->setCanvasBackground(QBrush(QColor(230, 255, 230)));

    TScale *sc;
    QPlot->setAxisScaleDraw(QwtPlot::xBottom, sc = new TScale);
    QPlot->setAxisScaleEngine(QwtPlot::xBottom, new TScaleDiv(sc));
    QPlot->setAxisScaleDraw(QwtPlot::yLeft, sc = new TScale);
    QPlot->setAxisScaleEngine(QwtPlot::yLeft, new TScaleDiv(sc));

    installEventFilter(this);

    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setMinimumSize(150, 150);

    setFocusPolicy(Qt::StrongFocus);

    InitQwtItems();
}

TPlotCore::~TPlotCore()
{
    delete CurveAB;
    delete CurveAC;
}

void TPlotCore::InitQwtItems()
{
    ReSetAxisScales();
    ReAssignPlotOptions();

    PlGrid = new QwtPlotGrid;
    PlGrid->enableXMin(true);
    PlGrid->enableYMin(true);
    PlGrid->setMajPen(QPen(Qt::gray, 0, Qt::SolidLine));
    PlGrid->setMinPen(QPen(Qt::lightGray, 0, Qt::DotLine));
    PlGrid->attach(QPlot); PlGrid->hide();

    MarkerA = new QwtPlotMarker;
    MarkerB = new QwtPlotMarker;
    MarkerC = new QwtPlotMarker;
    MarkerM = new QwtPlotMarker;
    QwtPlotMarker *marks[] = {MarkerA, MarkerB, MarkerC, MarkerM};
    for (int i = 0; i < 4; i++) {
        marks[i]->attach(QPlot);
        marks[i]->hide();
    }
    for (int i = 0; i < 3; i++) {
        marks[i]->setLineStyle(QwtPlotMarker::Cross);
        marks[i]->setLabelAlignment(Qt::AlignRight | Qt::AlignBottom);
        marks[i]->setLinePen(QPen(QColor(255, 160, 160), 2, Qt::DashLine));
        marks[i]->setSymbol(new QwtSymbol(QwtSymbol::Diamond, QColor(255, 255, 255), QColor(255, 255, 255), QSize(16, 16)));
    }

    CurveAB = new TLineSegment;
    CurveAB->attach(QPlot);
    CurveAB->hide();

    CurveAC = new TLineSegment;
    CurveAC->attach(QPlot);
    CurveAC->hide();

    CurveAB->setPen(QPen(QColor(255, 33, 33), 2, Qt::DashLine));
    CurveAC->setPen(QPen(QColor(255, 33, 33), 2, Qt::DashLine));
}

void TPlotCore::ReAssignPlotOptions()
{
    UpdateAxisLegends();

    QFont font = qApp->font(); font.setPointSize(FontSize);
    QwtText tt(Title); tt.setFont(font);
    QwtText tx(XLegend); tx.setFont(font);
    QwtText ty(YLegend); ty.setFont(font);

    QPlot->setTitle(tt);
    QPlot->setAxisTitle(QwtPlot::xBottom, tx);
    QPlot->setAxisTitle(QwtPlot::yLeft, ty);
    QPlot->setAxisFont(QwtPlot::xBottom, font);
    QPlot->setAxisFont(QwtPlot::yLeft, font);

    ((TScale *)QPlot->axisScaleDraw(QwtPlot::xBottom))->font = font;
    ((TScale *)QPlot->axisScaleDraw(QwtPlot::yLeft))->font = font;
}

void TPlotCore::ReSetAxisScales(const TPlotEnums::TAxis ApplyTo)
{
    if (ApplyTo == TPlotEnums::A_Both || ApplyTo == TPlotEnums::A_XAxis) {

        if (Vertical)
        {
            //qDebug()<<1<<PlotsYLimits[0]<<PlotsYLimits[1];
            CurrentYLimits = PlotsYLimits;
            //QPlot->setAxisScale(QwtPlot::yLeft, CurrentXLimits[0], CurrentXLimits[1]);
            QPlot->setAxisScale(QwtPlot::yLeft, CurrentYLimits[0], CurrentYLimits[1]);
            Replot();
        }
        else
        {
            //qDebug()<<0<<PlotsXLimits[0]<<PlotsXLimits[1];
            CurrentXLimits = PlotsXLimits;
            SetCurrentXScales();
        }
    }
    if (ApplyTo == TPlotEnums::A_Both || ApplyTo == TPlotEnums::A_YAxis)
    {
        if (Vertical)
        {
            //qDebug()<<1<<PlotsXLimits[0]<<PlotsXLimits[1];
            CurrentXLimits = PlotsXLimits;
            QPlot->setAxisScale(QwtPlot::xBottom, CurrentXLimits[0], CurrentXLimits[1]);
            Replot();
        }
        else
        {
            //qDebug()<<0<<PlotsYLimits[0]<<PlotsYLimits[1];
            CurrentYLimits = PlotsYLimits;
            SetCurrentYScales();
        }
    }

    HandleAxisMultiplier();
}

void TPlotCore::SetMarkerMCoords()
{
    MarkerM->setValue( (MarkerA->value() + MarkerB->value()) / 2. );
}
void TPlotCore::UpdateABCCurves()
{
    CurveAB->SetData(MarkerA->value(), MarkerB->value());
    CurveAC->SetData(MarkerA->value(), MarkerC->value());
}

void TPlotCore::ReSyncMarkers()
{
    UpdateABCCurves();
    SetMarkerMCoords();
}

//================================================================================================================================
void TCommonPlot::BAngle_clicked(bool IsChecked)
{
    MarkerC->setVisible(IsChecked);
    CurveAC->setVisible(IsChecked);

    if (IsChecked) {
        if (MarkerA->xValue() == MarkerB->xValue() || MarkerA->yValue() == MarkerB->yValue())
            MarkerC->setValue( (MarkerA->value() + MarkerB->value()) / 2. );
        else
            MarkerC->setValue(MarkerA->xValue(), MarkerB->yValue());

        ReSyncMarkers();
    }

    UpdateLabels();
    Replot();
    PlotWasChanged();
}
void TCommonPlot::BGrid_clicked(bool IsChecked)
{
    PlGrid->setVisible(IsChecked);
    Replot();
    PlotWasChanged();
}
void TCommonPlot::BMarkers_clicked(bool IsChecked)
{
    MarkerA->setVisible(IsChecked);
    MarkerB->setVisible(IsChecked);
    MarkerM->setVisible(IsChecked);
    CurveAB->setVisible(IsChecked);

    if (IsChecked)
    {
        QPointF aPoint(CurrentXLimits[0] + (CurrentXLimits[1] - CurrentXLimits[0])/5.,
                      CurrentYLimits[0] + (CurrentYLimits[1] - CurrentYLimits[0])/5.);
        QPointF bPoint(CurrentXLimits[1] - (CurrentXLimits[1] - CurrentXLimits[0])/5.,
                      CurrentYLimits[1] - (CurrentYLimits[1] - CurrentYLimits[0])/5.);

        if (Vertical)
        {
            MarkerA->setValue(bPoint);
            MarkerB->setValue(aPoint);
        }
        else
        {
            MarkerA->setValue(aPoint);
            MarkerB->setValue(bPoint);
        }


        ReSyncMarkers();
    }

    UpdateLabels();
    Replot();
    PlotWasChanged();
}
void TCommonPlot::SetAllYScales(const double Min, const double Max)
{
    SetDefaultYLimits(Min, Max);
    SetPlotsYScales(Min, Max);
    if (singleAutoScale)
    {
        singleAutoScaleYLimit[0] = Min;
        singleAutoScaleYLimit[1] = Max;
    }
}
void TCommonPlot::BRestoreScale_clicked()
{
    if (BFixXIsChecked && BFixYIsChecked)
        return;

    if (singleAutoScale)
    {
        CurrentXLimits[0] = 0;
        CurrentXLimits[1] = singleAutoScaleXLimit;
        PlotsXLimits[0] = 0;
        PlotsXLimits[1] = singleAutoScaleXLimit;
        DefaultXLimits[0] = 0;
        DefaultXLimits[1] = singleAutoScaleXLimit;

        CurrentYLimits[0] = singleAutoScaleYLimit[0];
        CurrentYLimits[1] = singleAutoScaleYLimit[1];
        PlotsYLimits[0] = singleAutoScaleYLimit[0];
        PlotsYLimits[1] = singleAutoScaleYLimit[1];
        DefaultYLimits[0] = singleAutoScaleYLimit[0];
        DefaultYLimits[1] = singleAutoScaleYLimit[1];
    }
    ReSetAxisScales();
    if (singleAutoScale)
    {
        CurrentXLimits[0] = 0;
        CurrentXLimits[1] = singleAutoScaleXLimit;
        PlotsXLimits[0] = 0;
        PlotsXLimits[1] = singleAutoScaleXLimit;
        DefaultXLimits[0] = 0;
        DefaultXLimits[1] = singleAutoScaleXLimit;

        CurrentYLimits[0] = singleAutoScaleYLimit[0];
        CurrentYLimits[1] = singleAutoScaleYLimit[1];
        PlotsYLimits[0] = singleAutoScaleYLimit[0];
        PlotsYLimits[1] = singleAutoScaleYLimit[1];
        DefaultYLimits[0] = singleAutoScaleYLimit[0];
        DefaultYLimits[1] = singleAutoScaleYLimit[1];
    }
    ReAssignPlotOptions();
    Replot();
    PlotWasChanged();
    if (SelectionModeEnabled) {
        UpdateCanvasRect();
        Updatekxky();
        SelectRectChanged(GetSelectedRect());
    }
}
void TCommonPlot::SetFontSize(int Size)
{
    FontSize = Size;

    ReAssignPlotOptions();
    ReplotChanged();
}
void TCommonPlot::SetClickParams(const QPoint &Pos)
{
    ClickPos = Pos;
    ClickMinX = CurrentXLimits[0];
    ClickMaxX = CurrentXLimits[1];
    ClickMinY = CurrentYLimits[0];
    ClickMaxY = CurrentYLimits[1];
}
void TCommonPlot::DefineClickInfo(const QPoint &Pos)
{
    SetClickParams(Pos);

    if (SelectionModeEnabled && SelectAction != TPlotSelectionData::None) {
        Action = ExternalAct;
        return;
    }

    Action = Panning;

    if (!MarkerA->isVisible())
        return;

    double Dist2A = (Pos - PlotToScreen(MarkerA->value())).manhattanLength();
    double Dist2B = (Pos - PlotToScreen(MarkerB->value())).manhattanLength();
    double Dist2C = (Pos - PlotToScreen(MarkerC->value())).manhattanLength();

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
    double scale = qApp->desktop()->logicalDpiX();
#else
    double scale = qApp->desktop()->physicalDpiX();
#endif

    int Range = 1.5*qMax<int>(20, 0.25 * scale);
    if     (Dist2A < Dist2B && Dist2A <= Dist2C && Dist2A < Range) Action = MoveA;
    else if (Dist2B < Dist2A && Dist2B <= Dist2C && Dist2B < Range) Action = MoveB;
    else if (Dist2C < Dist2B && Dist2C < Dist2A && Dist2C < Range) Action = MoveC;
}
void TCommonPlot::MoveMarker(QwtPlotMarker *Marker, double x, double y)
{
    if     (x < CurrentXLimits[0]) x = CurrentXLimits[0];
    else if (x > CurrentXLimits[1]) x = CurrentXLimits[1];
    if     (y < CurrentYLimits[0]) y = CurrentYLimits[0];
    else if (y > CurrentYLimits[1]) y = CurrentYLimits[1];
    Marker->setValue(x, y);
}
double TCommonPlot::CalcAngle() const
{
    double xA = MarkerA->xValue(), xB = MarkerB->xValue(), yA = MarkerA->yValue(), yB = MarkerB->yValue(), xC = MarkerC->xValue(), yC = MarkerC->yValue();
    return 180. / M_PI * acos(((xA - xB) * (xA - xC) + (yA - yB) * (yA - yC)) / (sqrt((xA - xB) * (xA - xB) + (yA - yB) * (yA - yB)) * sqrt((xA - xC) * (xA - xC) + (yA - yC) * (yA - yC))));
}
void TCommonPlot::MoveMarker(const double x, const double y, bool IsPosCoordInPlot)
{
    double X = x, Y = y;
    if (!IsPosCoordInPlot)
    {
        X = ScreenXToPlotX((int)x);
        Y = ScreenYToPlotY((int)y);
    }
    if (Action == MoveA)      {MoveMarker(MarkerA, X, Y); SetMarkerMCoords(); }
    else if (Action == MoveB) {MoveMarker(MarkerB, X, Y); SetMarkerMCoords(); }
    else if (Action == MoveC) {MoveMarker(MarkerC, X, Y); }
    UpdateLabels();
    UpdateABCCurves();

    Replot();
}
void TCommonPlot::ZoomPlot(const QPointF &Pos, int Increment, Range *XLims, Range *YLims)
{
    Range XLimits = XLims ? *XLims : PlotsXLimits, YLimits = YLims ? *YLims : PlotsYLimits;
    double x = qBound(CurrentXLimits[0], (double)Pos.x(), CurrentXLimits[1]);
    double y = qBound(CurrentYLimits[0], (double)Pos.y(), CurrentYLimits[1]);
    double MinX = 0., MaxX = 0., MinY = 0., MaxY = 0.;

    if (!Increment) return;
    else if (Increment > 0) { // zoom in
        MinX = x - fabs(CurrentXLimits[0] - x) * 100.0 / Increment;
        MaxX = x + fabs(CurrentXLimits[1] - x) * 100.0 / Increment;
        MinY = y - fabs(CurrentYLimits[0] - y) * 100.0 / Increment;
        MaxY = y + fabs(CurrentYLimits[1] - y) * 100.0 / Increment;
    } else { // zoom out
        MinX = x - fabs(CurrentXLimits[0] - x) * (-Increment) / 100.0;
        MaxX = x + fabs(CurrentXLimits[1] - x) * (-Increment) / 100.0;
        MinY = y - fabs(CurrentYLimits[0] - y) * (-Increment) / 100.0;
        MaxY = y + fabs(CurrentYLimits[1] - y) * (-Increment) / 100.0;
    }
    if (MinX < XLimits[0]) { MaxX += XLimits[0] - MinX; MinX = XLimits[0]; }
    if (MaxX > XLimits[1]) { MinX += XLimits[1] - MaxX; MaxX = XLimits[1]; }
    if (MinY < YLimits[0]) { MaxY += YLimits[0] - MinY; MinY = YLimits[0]; }
    if (MaxY > YLimits[1]) { MinY += YLimits[1] - MaxY; MaxY = YLimits[1]; }

    if (MinX < XLimits[0]) MinX = XLimits[0];
    if (MaxX > XLimits[1]) MaxX = XLimits[1];
    if (MinY < YLimits[0]) MinY = YLimits[0];
    if (MaxY > YLimits[1]) MaxY = YLimits[1];

    #ifdef _SCAN_SLICE
    if (sliceBounds)
    {
        if (MinX < sliceX.arr[0]) MinX = sliceX.arr[0];
        if (MinY < sliceY.arr[0]) MinY = sliceY.arr[0];
        if (MaxX > sliceX.arr[1]) MaxX = sliceX.arr[1];
        if (MaxY > sliceY.arr[1]) MaxY = sliceY.arr[1];
    }
#endif

    if (!BFixXIsChecked) SetCurrentXScales(MinX, MaxX);
    if (!BFixYIsChecked) SetCurrentYScales(MinY, MaxY);

    ReAssignPlotOptions();
    Replot();
    PlotWasChanged();
    if (SelectionModeEnabled) {
        UpdateCanvasRect();
        Updatekxky();
        SelectRectChanged(GetSelectedRect());
    }

    SetClickParams(PlotToScreen(Pos).toPoint());
}
void TCommonPlot::PanPlot(const QPointF &Delta, Range *XLims, Range *YLims)
{
    Range XLimits = XLims ? *XLims : PlotsXLimits, YLimits = YLims ? *YLims : PlotsYLimits;
    double MinX = ClickMinX - Delta.x(), MaxX = ClickMaxX - Delta.x();
    double MinY = ClickMinY - Delta.y(), MaxY = ClickMaxY - Delta.y();
    double XRange = ClickMaxX - ClickMinX, YRange = ClickMaxY - ClickMinY;

    if (MinX < XLimits[0]) { MinX = XLimits[0]; MaxX = MinX + XRange; }
    if (MaxX > XLimits[1]) { MaxX = XLimits[1]; MinX = MaxX - XRange; }
    if (MinY < YLimits[0]) { MinY = YLimits[0]; MaxY = MinY + YRange; }
    if (MaxY > YLimits[1]) { MaxY = YLimits[1]; MinY = MaxY - YRange; }

#ifdef _SCAN_SLICE
    if (sliceBounds)
    {
        if (MinX < sliceX.arr[0]) MinX = sliceX.arr[0];
        if (MinY < sliceY.arr[0]) MinY = sliceY.arr[0];
        if (MaxX > sliceX.arr[1]) MaxX = sliceX.arr[1];
        if (MaxY > sliceY.arr[1]) MaxY = sliceY.arr[1];
    }
#endif

    SetCurrentXScales(MinX, MaxX);
    SetCurrentYScales(MinY, MaxY);

    ReAssignPlotOptions();
    Replot();
    PlotWasChanged();
    if (SelectionModeEnabled) {
        UpdateCanvasRect();
        Updatekxky();
        SelectRectChanged(GetSelectedRect());
    }
}

QImage TCommonPlot::MakeRawPreview(double width, double height) {
    //if (!XLegend.isEmpty() && !XLegend.isEmpty() && XLegend.split(",").last() == YLegend.split(",").last())
    if (inherits("plot::TPure3dPlot"))
        FindSizeWithAspectRatio(width, height, &width, &height);
    QImage Preview(width, height, QImage::Format_ARGB32);
    Preview.fill(0);

    QPainter PlotPainter(&Preview);
    if (!PlotPainter.isActive())
        return Preview;

    QwtPlotRenderer rndr;
    QwtScaleMap maps[QwtPlot::axisCnt];
    for ( int axisId = 0; axisId < QwtPlot::axisCnt; axisId++ )
    {
        maps[axisId].setTransformation(
            QPlot->axisScaleEngine( axisId )->transformation() );

        const QwtScaleDiv &scaleDiv = *QPlot->axisScaleDiv( axisId );
        maps[axisId].setScaleInterval(
            scaleDiv.lowerBound(), scaleDiv.upperBound() );

        double from = 0, to = 0;
        if (axisId == QwtPlot::xTop || axisId == QwtPlot::xBottom)
            to = Preview.width();
        else
            from = Preview.height();
        maps[axisId].setPaintInterval(from, to);
    }
    maps[QwtPlot::xBottom].setScaleInterval(PlotsXLimits[0], PlotsXLimits[1]);
    maps[QwtPlot::yLeft].setScaleInterval(PlotsYLimits[0], PlotsYLimits[1]);
    rndr.renderCanvas(QPlot, &PlotPainter, Preview.rect(), maps);

    return Preview;
}
const QImage TCommonPlot::MakePreview(TCommonPlotControl *Control, double width, double height)
{
    if (Control) PreparePlotForPreview();
    QImage p = MakeRawPreview(width, height);
    if (Control) RestorePlotAfterPreview(Control);
    return p;
}
void TCommonPlot::PreparePlotForPreview()
{
    PlGrid->hide();   MarkerA->hide();     CurveAB->hide();
    MarkerB->hide();  MarkerC->hide();     CurveAC->hide();
    MarkerM->hide();
}
void TCommonPlot::RestorePlotAfterPreview(TCommonPlotControl *Control)
{
    if (Control->BGrid->isChecked())     PlGrid->show();
    if (Control->BMarkers->isChecked()) {MarkerA->show(); CurveAB->show(); MarkerB->show(); MarkerM->show(); }
    if (Control->BAngle->isChecked())   {MarkerC->show(); CurveAC->show(); }
    Replot();
}
void TCommonPlot::HandleZeroDiff(double &Max, double &Min) const
{
    const double Diff = qFabs(Max - Min);
    const double Epsilon = 4. * std::numeric_limits<double>::epsilon();
    if (qFuzzyCompare(Min, Max) || Diff < Epsilon)
    {
        const double Average = qFabs(Max + Min) / 2.;
        const double Value = Average > Epsilon ? Average : 1;
        const double Delta = 1e-2 * Value;
        Max += Delta; Min -= Delta;
    }
}

void TCommonPlot::FindSizeWithAspectRatio(const double DesirableWidth, const double DesirableHeight, double *ResultDeltaX, double *ResultDeltaY) const
{
    *ResultDeltaX = DesirableWidth;
    *ResultDeltaY = DesirableHeight;

    double k = (PlotsYLimits[1] - PlotsYLimits[0]) / (PlotsXLimits[1] - PlotsXLimits[0]);
    if (qIsNull(k) || !qIsFinite(k))
        return;
    //k = qBound(0.1, k, 10);

    if (*ResultDeltaY / *ResultDeltaX < k)
        *ResultDeltaX = *ResultDeltaY / k;
    else
        *ResultDeltaY = *ResultDeltaX * k;
}

void TCommonPlot::pinchEvent(QPoint center, double scale)
{
    int UpSpace = QPlot->plotLayout()->titleRect().height() ? QPlot->plotLayout()->titleRect().height() + 5 : 0;
    int LeftSpace = QPlot->plotLayout()->scaleRect(QwtPlot::yLeft).width();
    center += QPoint(LeftSpace, UpSpace);

    int inc = (scale > 1.) ? qCeil(100.*scale) : qFloor(-100./scale);
    QWheelEvent wev(center, inc, Qt::MiddleButton, Qt::NoModifier);
    eventFilter(this, &wev);
}

void TCommonPlot::setGestureActive(bool active)
{
    Action = active ? Gesture : ExternalAct;
}


bool QwtPlot_::event(QEvent *event)
{
    if (event->type() == QEvent::Gesture) {
        return gestureEvent(static_cast<QGestureEvent*>(event));
    }
    return QwtPlot::event(event);
}
bool QwtPlot_::gestureEvent(QGestureEvent *event)
{
    if (QGesture *pinch = event->gesture(Qt::PinchGesture)) {
        QPinchGesture *p = (QPinchGesture *)pinch;
        if (p->changeFlags() & QPinchGesture::ScaleFactorChanged)
            emit pinchSig(mapFromGlobal(p->centerPoint().toPoint()), p->scaleFactor());
    } else if (event->gesture(Qt::TapAndHoldGesture)) {
        emit doubleClick();
        return true;
    }

    bool active = false;
    for (int i = 0; i < event->activeGestures().size(); i++)  {
        if (event->activeGestures().at(i)->state() == Qt::GestureStarted ||
                event->activeGestures().at(i)->state() == Qt::GestureUpdated) {
            active = true;
            break;
        }
    }
    emit isGestureActive(active);

    return true;
}


void TScale::drawLabel( QPainter *painter, double value ) const
{
    font = painter->font();
    QwtScaleDraw::drawLabel(painter, value);
}


TScaleDiv::TScaleDiv(TScale *_scale)
{
    scale = _scale;
    prev_len = 0;
}

#define CACHE_DIV_UPDATES 5
QwtScaleDiv TScaleDiv::divideScale( double x1, double x2, int numMajorSteps, int numMinorSteps, double stepSize ) const
{
    if (stepSize == 0.0 && scale) {
        if (prev_len == scale->len && qFuzzyCompare(prev_range, qFabs(x2 - x1)) && prev_counter < CACHE_DIV_UPDATES) {
            prev_counter++;
            numMajorSteps = prev_steps;
            goto breaked;
        }
        stepSize = divideInterval( qFabs(x2 - x1), numMajorSteps );

        bool endminus = false, endminus1 = false; //when -1 > -1 > +2 sequence, break
        int length = 0;
        double prev_step = -1;
        while (true) {
            if (!qFuzzyCompare(prev_step, stepSize)) {
                int start = qCeil(x1 / stepSize);
                int end = qFloor(x2 / stepSize);
                if (x1 > x2) {
                    start = qCeil(x2 / stepSize);
                    end = qFloor(x1 / stepSize);
                }

                length = 0;
                for (int i = start; i <= end; i++) {
                    QString label = scale->label(i * stepSize / scale->Multiplier).text();

                    QSize sz = QFontMetrics(scale->font).boundingRect(label).size();
                    if (scale->orientation() == Qt::Horizontal)
                        length = qMax(length, sz.width());
                    else
                        length = qMax(length, sz.height());

                    if (i == start+1) //optimize: only first and last 2 ticks
                        i = qMax(start+1, end -2);
                }
                if (end == start || length == 0) {
                    numMajorSteps = 5;
                    stepSize = 0.0;
                    break;
                }

                length *= (end - start) * 2 + 1;
            }

            if (length > scale->len) {
                if (endminus)
                    endminus1 = true;
                endminus = true;

                if (numMajorSteps >= 3)
                    numMajorSteps--;
                else
                    break;
            } else {
                if (endminus && endminus1)
                    break;
                endminus = false;

                numMajorSteps+=2;
                if (numMajorSteps >= 100) {
                    numMajorSteps = 5;
                    stepSize = 0.0;
                    break;
                }
            }

            prev_step = stepSize;
            stepSize = divideInterval( qFabs(x2 - x1), numMajorSteps );
        }

        prev_counter = 0;
        prev_len = scale->len;
        prev_range = qFabs(x2 - x1);
        prev_steps = numMajorSteps;
    }
    breaked:

    if (!qIsFinite(x1)) x1 = 0;
    if (!qIsFinite(x2)) x2 = x1 + 1;
    return QwtLinearScaleEngine::divideScale(x1, x2, numMajorSteps, numMinorSteps, stepSize);
}


}
