#include "plot_SubtPlane.h"

#include <QVector>

#include <QtGlobal>
#include <math.h>
#ifndef Q_CC_CLANG
#include <malloc.h>
#else
#include <cstdlib>
#endif
#include <string.h>

#define _TRUE 1
#define _FALSE 0

#define SLi(a, i, j) a[(i)*((i) + 1)/2 + (j)]

namespace plot {
namespace math {

/*
   \brief mathCholeskiDecompose
   \note Decomposes a symmetric positive definite matrix in place.

   \param dim The dimension of a
   \param a Lower triangular part of a symmetric matrix, stored by rows, i.e., matrix = [a_00 a_10 a_11 a_20 a_21 a_22 a_30 ...]

   \return Whether the matrix was really positive definite.  If %_FALSE,
         the decomposition failed and @a does not contain any meaningful
         values.
*/
int mathCholeskiDecompose(int dim, double *a)
{
    for (int k = 0; k < dim; k++)
    {
        // diagonal element
        double s = SLi(a, k, k);
        for (int i = 0; i < k; i++)
            s -= SLi(a, k, i) * SLi(a, k, i);
        if (s <= 0.0)
            return _FALSE;
        SLi(a, k, k) = s = sqrt(s);

        // nondiagonal elements
        for (int j = k+1; j < dim; j++)
        {
            double r = SLi(a, j, k);
            for (int i = 0; i < k; i++)
                r -= SLi(a, k, i) * SLi(a, j, i);
            SLi(a, j, k) = r/s;
        }
    }

    return _TRUE;
}

/*
   \brief mathCholeskiSolve
   \note Solves a system of linear equations with predecomposed symmetric positive
        definite matrix @a and right hand side @b.

   \param dim The dimension of @a
   \param a Lower triangular part of Choleski decomposition as computed by mathCholeskiDecompose().
   \param b Right hand side vector.  Is is modified in place, on return it contains the solution.
*/
void mathCholeskiSolve(int dim, const double *a, double *b)
{
    // back-substitution with the lower triangular matrix
    for (int j = 0; j < dim; j++)
    {
        for (int i = 0; i < j; i++)
            b[j] -= SLi(a, j, i)*b[i];
        b[j] /= SLi(a, j, j);
    }

    // back-substitution with the upper triangular matrix
    for (int j = dim-1; j >= 0; j--)
    {
        for (int i = j+1; i < dim; i++)
            b[j] -= SLi(a, i, j)*b[i];
        b[j] /= SLi(a, j, j);
    }
}

template double* dataFieldAreaFitPolynom<>(QVector<QVector<double > > &data,
int col, int row,
int width, int height,
int col_degree, int row_degree,
double *coeffs);

template void dataFieldAreaSubtractPolynom<>(QVector<QVector<double > > &data,
int col, int row,
int width, int height,
int col_degree, int row_degree,
const double *coeffs);

/*
   \brief dataFieldAreaFitPolynom
   \note Fits a two-dimensional polynomial to a rectangular part of a data field.
        The coefficients are stored by row into @coeffs, like data in a datafield.
        Row index is y-degree, column index is x-degree.
        Note naive x^n y^m polynomial fitting is numerically unstable, therefore
        this method works only up to @col_degree = @row_degree = 6.

   \param data A data field
   \param col Lower-left column coordinate
   \param row Lower-left row coordinate
   \param width Area width (number of columns)
   \param height Area height (number of rows)
   \param col_degree Degree of polynomial to fit column-wise (x-coordinate)
   \param row_degree Degree of polynomial to fit row-wise (y-coordinate)
   \param coeffs An array of size (@row_degree+1)*(@col_degree+1) to store the
         coefficients to, or %NULL (a fresh array is allocated then)

   \return Either @coeffs if it was not %NULL, or a newly allocated array
         with coefficients.
*/
template <class T>
double* dataFieldAreaFitPolynom(T &data,
                                int col, int row,
                                int width, int height,
                                int col_degree, int row_degree,
                                double *coeffs)
{
    double *sums, *m;

    int size = (row_degree+1)*(col_degree+1);
    if (!coeffs)
        coeffs = new double[size];
    for (int i = 0; i < size; i++)
        coeffs[i] = 0;

    sums = new double[(2*row_degree+1)*(2*col_degree+1)];
    for (int i = 0; i < (2*row_degree+1)*(2*col_degree+1); i++)
        sums[i] = 0;

    for (int c = col; c < col + width; c++)
        for (int r = row; r < row + height; r++)
        {
            double ry = 1.0;
            double z = data[c][r];

            for (int i = 0; i <= 2*row_degree; i++)
            {
                double cx = 1.0;

                for (int j = 0; j <= 2*col_degree; j++)
                {
                    sums[i*(2*col_degree+1) + j] += cx*ry;
                    cx *= (c - col + 1);
                }
                ry *= (r - row + 1);
            }

            ry = 1.0;
            for (int i = 0; i <= row_degree; i++)
            {
                double cx = 1.0;

                for (int j = 0; j <= col_degree; j++)
                {
                    coeffs[i*(col_degree+1) + j] += cx*ry*z;
                    cx *= (c - col + 1);
                }
                ry *= (r - row + 1);
            }
        }

    m = (double *) malloc(sizeof(double)* (size*(size+1)/2));
    for (int i = 0; i < size; i++)
    {
        double *mrow = m + i*(i+1)/2;

        for (int j = 0; j <= i; j++)
        {
            int pow_x, pow_y;

            pow_x = i % (col_degree+1) + j % (col_degree+1);
            pow_y = i/(col_degree+1) + j/(col_degree+1);
            mrow[j] = sums[pow_y*(2*col_degree+1) + pow_x];
        }
    }

    if (!mathCholeskiDecompose(size, m))
        memset(coeffs, 0, size*sizeof(double));
    else
        mathCholeskiSolve(size, m, coeffs);

    free(m);
    delete [] sums;

    return coeffs;
}

/*
   \brief dataFieldAreaSubtractPolynom
   \note Subtracts a two-dimensional polynomial from a rectangular part of a data
        field.
   \param data A data field
   \param col Lower-left column coordinate
   \param row Lower-left row coordinate
   \param width Area width (number of columns)
   \param height Area height (number of rows)
   \param col_degree Degree of polynomial to subtract column-wise (x-coordinate)
   \param row_degree Degree of polynomial to subtract row-wise (y-coordinate)
   \param coeffs An array of size (@row_degree+1)*(@col_degree+1) with coefficients, see dataFieldAreaFitPolynom() for details
*/
//TODO qtconcurrent
template <class T>
void dataFieldAreaSubtractPolynom(T &data,
                                  int col, int row,
                                  int width, int height,
                                  int col_degree, int row_degree,
                                  const double *coeffs)
{
    for (int c = col; c < col + width; c++)
        for (int r = row; r < row + height; r++) //fight da powah
        {
            double ry = 1.0;
            double z = data[c][r];

            for (int i = 0; i <= row_degree; i++)
            {
                double cx = 1.0;

                for (int j = 0; j <= col_degree; j++)
                {
                    z -= coeffs[i*(col_degree+1) + j]*cx*ry;
                    cx *= (c - col + 1);
                }
                ry *= (r - row + 1);
            }

            data[c][r] = z;
        }

    double checkNegative;
    checkNegative = 0.0;

    for (int c = col; c < col + width; c++)
        for (int r = row; r < row + height; r++)
            if (data[c][r] < checkNegative)
                checkNegative = data[c][r];

    if (checkNegative < 0)
        for (int c = col; c < col + width; c++)
            for (int r = row; r < row + height; r++)
                data[c][r] += fabs(checkNegative);

}

/*
   \brief dataLinePartFitPolynom
   \note Fits a polynomial through a part of a data line.

   \param data A data line
   \param n Polynom degree
   \param coeffs An array of size @n+1 to store the coefficients to, or %NULL (a fresh array is allocated then)
   \param from Index the line part starts at
   \param to Index the line part ends at + 1

   \returns The coefficients of the polynomial (@coeffs when it was not %NULL, otherwise a newly allocated array).
*/
double* dataLinePartFitPolynom(double *data,
                               int n, double *coeffs,
                               int from, int to)
{
    double *sumx, *m;
    int i, j;

    sumx = new double[2*n+1];
    for (int i = 0; i < 2*n+1; i++)
        sumx[i] = 0;

    if (!coeffs)
        coeffs = new double[n+1];
    for (int i = 0; i < n+1; i++)
        coeffs[i] = 0;

    for (i = from; i < to; i++)
    {
        double x = i;
        double y = data[i];
        double xp;

        xp = 1.0;
        for (j = 0; j <= n; j++)
        {
            sumx[j] += xp;
            coeffs[j] += xp*y;
            xp *= x;
        }
        for (j = n+1; j <= 2*n; j++)
        {
            sumx[j] += xp;
            xp *= x;
        }
    }

    m = new double[(n+1)*(n+2)/2];
    for (i = 0; i <= n; i++) {
        double *row = m + i*(i+1)/2;

        for (j = 0; j <= i; j++)
            row[j] = sumx[i+j];
    }
    if (!mathCholeskiDecompose(n+1, m))
        memset(coeffs, 0, (n+1)*sizeof(double));
    else
        mathCholeskiSolve(n+1, m, coeffs);

    delete [] m;
    delete [] sumx;

    return coeffs;
}

/*
   \brief dataLinePartSubtractPolynom
   \note Subtracts a polynomial from a part of a data line.

   \param data A data line
   \param n Polynom degree
   \param coeffs An array of size @n+1 with polynomial coefficients to
   \param from Index the line part starts at
   \param to Index the line part ends at + 1
*/
void dataLinePartSubtractPolynom(double *data,
                                 int n,
                                 const double *coeffs,
                                 int from, int to)
{
    int i, j;
    double val;

    for (i = from; i < to; i++)
    {
        val = 0.0;
        for (j = n; j; j--)
        {
            val += coeffs[j];
            val *= i;
        }
        val += coeffs[0];

        data[i] -= val;
    }

    double checkNegative;
    checkNegative = 0.0;

    for (i = from; i < to; i++)
        if (data[i] < checkNegative)
            checkNegative = data[i];

    if (checkNegative < 0)
        for (i = from; i < to; i++)
            data[i] = data[i] + fabs(checkNegative);
}

}
}
