#include "plot_DoubleSlider.h"

#include "core/plot_core.h"

#include <qwt_scale_engine.h>
#include <QtEvents>
#include <QGraphicsSceneEvent>
#include <QGestureEvent>
#include <QGraphicsProxyWidget>
#include <QApplication>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtMath>
#else
#include <qmath.h>
#include <qnumeric.h>
#endif
#include <QVBoxLayout>
#include <QSplitter>

namespace plot {

TQwtScaleBar::TQwtScaleBar()
{
    map = nullptr;
    setScaleDraw(new TScale);
    setAlignment(QwtScaleDraw::RightScale);
    setColorBarEnabled(true);
}
void TQwtScaleBar::SetMinMaxValues(double Min, double Max)
{
    TScaleDiv ScaleEngine((TScale *)scaleDraw());
    ScaleDiv = ScaleEngine.divideScale(Min, Max, 5, 5);
    setScaleDiv(ScaleEngine.transformation(), ScaleDiv);

    SetColourMap(map);
}
void TQwtScaleBar::resizeEvent(QResizeEvent *e)
{
    QwtScaleWidget::resizeEvent(e);
    if (map)
        SetMinMaxValues(GetMin(), GetMax());
}

//-----------------------------------
TSlider::TSlider(TSlidersBar *_Parent, const QPixmap &Pm) : active(false), SliderImage(Pm), Parent(_Parent)
{
    setAcceptedMouseButtons(Qt::LeftButton | Qt::MiddleButton);
    imageRect = Pm.rect();
    if (Pm.isNull()) imageRect = QRectF(0, 0, 60, 30);
}
void TSlider::paint(QPainter *Painter, const QStyleOptionGraphicsItem * /*Option*/, QWidget * /*Widget*/)
{
    Painter->drawPixmap(imageRect, SliderImage, SliderImage.rect());
}
void TSlider::mousePressEvent(QGraphicsSceneMouseEvent *Event)
{
    active = (Event->button() == Qt::LeftButton) ||
             (Event->button() == Qt::MiddleButton);
    MouseDownPos = Event->pos();
    MouseDownPos.setY(MouseDownPos.y() - imageRect.center().y());

    Parent->setAutoMode(false);
}
TRightSlider::TRightSlider(TSlidersBar *Parent, const QPixmap &Pm)
    : TSlider(Parent, Pm)
{
    setAcceptedMouseButtons(Qt::LeftButton | Qt::RightButton);
}
void TRightSlider::mousePressEvent(QGraphicsSceneMouseEvent *Event)
{
    TSlider::mousePressEvent(Event);
    active = (Event->button() == Qt::LeftButton) ||
             (Event->button() == Qt::RightButton);
}
void TSlider::mouseReleaseEvent(QGraphicsSceneMouseEvent * /*Event*/)
{
    active = false;
}
void TSlider::mouseMoveEvent(QGraphicsSceneMouseEvent *Event)
{
    if (active)
        Parent->MouseMoveEvent(Event->scenePos() - MouseDownPos, this);
}
void TSlider::SetCPos(double Pos)
{
    RelPos = (Pos - Parent->BackbonePos.y()) / Parent->BackboneLength;
    setPos(0, Pos - imageRect.height() / 2.);
    update();
}
void TSlider::SetRelPos(double RP)
{
    RelPos = RP;
    setPos(0, RelPos*Parent->BackboneLength + Parent->BackbonePos.y() - imageRect.height() / 2.);
    update();
}
double TSlider::GetValue() const
{
    return (1 - RelPos);
}

//-----------------------------------
TSlidersBar::TSlidersBar(QWidget *Parent): QGraphicsView(Parent)
{
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);
    setFrameStyle(QFrame::NoFrame);

    //for transparent gray background
    QPalette P = palette();
    QColor C(P.color(QPalette::Base)); C.setAlpha(0);
    P.setColor(QPalette::Base, C);
    setPalette(P);

    LowSlider = new TRightSlider(this, QPixmap(":/plot/Slider_Min.png").scaledToWidth(40));
    HighSlider = new TSlider(this, QPixmap(":/plot/Slider_MAX.png").scaledToWidth(40));
    QwtBar = new TQwtScaleBar;
    QwtBar->setBorderDist(20, 25); //up, down border dist
    QwtBar->setAttribute(Qt::WA_OpaquePaintEvent);

    QGraphicsScene *Scene = new QGraphicsScene(this);

    bgItem = Scene->addPixmap(QPixmap());
    bgItem->setShapeMode(QGraphicsPixmapItem::BoundingRectShape);
    bgItem->moveBy(0, 20); //up border dist

    QColourBarProxy = Scene->addWidget(QwtBar);
    Scene->addItem(LowSlider);
    Scene->addItem(HighSlider);
    setScene(Scene);

    HighSlider->RelPos = 0;
    LowSlider->RelPos = 1;

    setMinimumHeight(QwtBar->minimumHeight());

    grabGesture(Qt::TapAndHoldGesture);
}
QSize TSlidersBar::sizeHint() const
{
    return QSize(QwtBar->sizeHint().width() + 20, QGraphicsView::sizeHint().height());
}
QSize TSlidersBar::minimumSizeHint() const
{
    return QSize(0, 0);
}

void TSlidersBar::setAutoMode(bool on)
{
    emit AutoChanged(on);
}

bool TSlidersBar::event(QEvent *event)
{
    if (event->type() == QEvent::Gesture) {
        return gestureEvent(static_cast<QGestureEvent*>(event));
    }
    return QGraphicsView::event(event);
}
void TSlidersBar::mouseDoubleClickEvent(QMouseEvent *)
{
    SetMinMaxRelSlidersValues(0, 1);
    setAutoMode(true);
}
bool TSlidersBar::gestureEvent(QGestureEvent *event)
{
    if (event->gesture(Qt::TapAndHoldGesture) &&
        !(HighSlider->active || LowSlider->active))
        mouseDoubleClickEvent(nullptr);

    return true;
}
void TSlidersBar::resizeEvent(QResizeEvent * /*Event*/)
{
    setSceneRect(rect());

    QColourBarProxy->setPos(10, 0);
    qreal scWidth = sceneRect().width();
    qreal scHeight = sceneRect().height();
    if (scWidth < 0) scWidth = 0;
    if (scHeight < 0) scHeight = 0;
    if (scWidth > 9999) scWidth = 9999;
    if (scHeight > 9999) scHeight = 9999;
    QColourBarProxy->resize(scWidth, //QColourBarProxy->geometry().width(),
                            scHeight);
    updateGeometry();

    if (bgItem) {
        QTransform m;
        m.scale(width() / (qreal)bgItem->boundingRect().width(),
                (height()-45) / (qreal)bgItem->boundingRect().height()); //up, down border dist
        bgItem->setTransform(m);
    }

    BackbonePos = QwtBar->scaleDraw()->pos + QwtBar->geometry().topLeft();
    BackboneLength = QwtBar->scaleDraw()->len;

    HighSlider->SetRelPos(HighSlider->RelPos);
    LowSlider->SetRelPos(LowSlider->RelPos);
}
void TSlidersBar::MouseMoveEvent(const QPointF &Pos, TSlider *Sender)
{
    if (Pos.y() < BackbonePos.y() || Pos.y() > BackbonePos.y() + BackboneLength)
        return;

    double RelPos = (Pos.y() - BackbonePos.y()) / BackboneLength;

    if ((Sender == HighSlider && RelPos >= LowSlider->RelPos) ||
       (Sender == LowSlider && RelPos <= HighSlider->RelPos))
        return;

    Sender->SetCPos(Pos.y());
    update();
    emit ValuesChanged(LowSlider->GetValue(), HighSlider->GetValue());
}
void TSlidersBar::SetColourMap(QwtLinearColorMap *Map)
{
    QwtBar->SetColourMap(Map);
    update();
}
void TSlidersBar::SetMinMaxValues(double Min, double Max)
{
    QwtBar->SetMinMaxValues(Min, Max);
    updateGeometry();
}
void TSlidersBar::SetMinMaxRelSlidersValues(double Min, double Max)
{
    if (Min > Max) qSwap(Min, Max);
    if (Min < 0) Min = 0;
    if (Max > 1) Max = 1;

    HighSlider->SetRelPos(1. - Max);
    LowSlider->SetRelPos(1. - Min);

    emit ValuesChanged(LowSlider->GetValue(), HighSlider->GetValue());
}
void TSlidersBar::SetMinMaxSlidersValues(double Min, double Max)
{
    if (!qIsFinite(Min)) Min = QwtBar->GetMin();
    if (!qIsFinite(Max)) Max = QwtBar->GetMax();

    if (qFuzzyCompare(qFabs(QwtBar->GetMax() - QwtBar->GetMin()) + 1., 1.))
        QwtBar->SetMinMaxValues(QwtBar->GetMax() - 0.01, QwtBar->GetMax() + 0.01);

    double Diff = QwtBar->GetMax() - QwtBar->GetMin();
    SetMinMaxRelSlidersValues((Min - QwtBar->GetMin()) / Diff, (Max - QwtBar->GetMin()) / Diff);
}

//-----------------------------------
TDoubleSlider::TDoubleSlider()
{
    Sliders = new TSlidersBar(this);
    Title = new QLabel;

    QVBoxLayout *MainLayout = new QVBoxLayout(this);
    MainLayout->setContentsMargins(0, 0, 0, 0);
    MainLayout->addWidget(Title, 0, Qt::AlignHCenter);
    MainLayout->addWidget(Sliders);

    auto sz = Sliders->minimumSize().height();
    if (sz < 0) sz = 0;
    if (sz > 9999) sz = 9999;
    setMinimumHeight(sz);
    setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);

    setAttribute(Qt::WA_Hover);
    installEventFilter(this);

    Sliders->HighSlider->hide();
    Sliders->LowSlider->hide();
}
QSize TDoubleSlider::sizeHint() const
{
    return QSize(QWidget::sizeHint().width(), maximumHeight());
}
bool TDoubleSlider::eventFilter(QObject *Object, QEvent *Event)
{

    if (Event->type() == QEvent::HoverEnter) {
        Sliders->HighSlider->show();
        Sliders->LowSlider->show();
    } else if (Event->type() == QEvent::HoverLeave) {
        Sliders->HighSlider->hide();
        Sliders->LowSlider->hide();
    }
    return QWidget::eventFilter(Object, Event);
}

}
