#include "videoplayer.h"
#include "gui/status.hpp"
#include "motors.hpp"
#include "remote_nst_dataid.hpp"
#include "oscilloscope.h"
#include "modules.hpp"
#include "mainwindow.h"
#include "scanplot.h"

#include "ratiolayout.hpp"
#include "flowlayout.hpp"
#include "gui/mslider.hpp"
#include "qwt_text_label.h"
#include <qwt_plot_marker.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_grid.h>
#include <QCheckBox>
#include <QApplication>
#include <QStyleFactory>
#include <QRubberBand>
#include <QComboBox>
#include <QGroupBox>
#include <QTableWidget>
#include <QHeaderView>
#include <QRadioButton>
#include <QList>
#include "scanplot.h"
#include <QDebug>
#include "../qwt/qwt_plot.h"
#include "../qwt/qwt_plot_layout.h"


#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtConcurrent>
    #ifdef USE_IVIDEO_MOTIC_1080
    #include "hardware/motic1080.hpp"
    #endif
#else
#include <QtCore>
#endif

#ifdef USE_IVIDEO_MOTIC
#include "hardware/motic.hpp"
#endif
#ifdef USE_IVIDEO_MOTIC_X
#include "hardware/moticx.hpp"
#endif

namespace VProfile {
    enum VideoProfileCollumn
    {
        Selector,
        Name,
        RangeX,
        RangeY,
        Height,
        Width,
        D0,
        Vst,
        Vpr,
        Crat,
        LeftStPr,
        Crat2,
        CratPresize,
        COUNT,
    };
}

namespace VSettings {
    enum VideoSettings
    {
        Name,
        RangeX, RangeY,
        Height, Width,
        FlipX, FlipY, //todo remove it
        StageHeadCantX, StageHeadCantY,
        StageLaserX, StageLaserY,
        D0,
        Vst,
        Vpr,
        Crat,
        LeftStPr,
        Crat2,
        CratPresize,
    };
}

void VideoPlayer::closeEvent(QCloseEvent *event)
{
    event->ignore();
}

bool VideoPlayer::eventFilter(QObject *obj, QEvent *Event)
{
    //If its Destroying, QPlot->canvas() already deleted => nullpoint exception
    if (Event->type() == QEvent::Destroy)
        return QWidget::eventFilter(obj, Event);

    videoPlotEvent(obj, Event);

	auto plotCanvas = QPlot->canvas();
    if (obj != plotCanvas)
        return QWidget::eventFilter(obj, Event);

    auto mouse = dynamic_cast<QMouseEvent *>(Event);

    if (Event->type() == QEvent::MouseButtonPress)
    {
        if (mouse->button() == Qt::RightButton)
        {
            plot_mode = 0;
            active = false;
            BCenterAndSet->change_state_man(false);
            BCenterAndSet2->change_state_man(false);
            BLine->change_state_man(false);
            BLineMeasure->change_state_man(false);
            markerA->hide();
            markerB->hide();
            lineAB->hide();
            markerC->setLabel(QwtText());
            QPlot->replot();

            return true;
        }
        else {
            posx = mouse->posF();
            active = plot_mode == 0;
            prev_xy_joy_output = QPointF();

            if (plot_mode == 0) {
                qreal x = QPlot->invTransform(QwtPlot::xBottom, posx.x());
                qreal y = QPlot->invTransform(QwtPlot::yLeft, posx.y());
				auto point = QPointF(x, y);
                lineAB->SetData(point, point);
                lineAB->show();
            }
            if (plot_mode == 3 || plot_mode == 4) {
                qreal x = QPlot->invTransform(QwtPlot::xBottom, posx.x());
                qreal y = QPlot->invTransform(QwtPlot::yLeft, posx.y());

                if (sqrt(pow(x - markerA->xValue(), 2) +
                         pow(y - markerA->yValue(), 2)) <
                        sqrt(pow(x - markerB->xValue(), 2) +
                             pow(y - markerB->yValue(), 2)))
                    marker = markerA;
                else
                    marker = markerB;

                qreal kx = fabs(QPlot->invTransform(QwtPlot::xBottom, 20) -
                                QPlot->invTransform(QwtPlot::xBottom, 0)) / 20.;
                qreal ky = fabs(QPlot->invTransform(QwtPlot::yLeft, 20) -
                                QPlot->invTransform(QwtPlot::yLeft, 0)) / 20.;
                const int XRange = 15, YRange = 15;
                if (fabs((x - marker->xValue()) / kx) < XRange &&
                        fabs((y - marker->yValue()) / ky) < YRange)
                    active = true;
            }
            return true;
        }
    }
    else if (Event->type() == QEvent::MouseButtonRelease) {
        if (plot_mode == 0) {
            lineAB->hide();
            QPlot->replot();
            goXYjoy(QPointF());
        } else
        if (plot_mode == 1) {
            QRectF rect;
            rect = head_cant->geometry();
            rect.moveCenter(mouse->posF());
            head_cant->setGeometry(rect.toRect());
            rect = stage_cant->geometry();
            rect.moveCenter(mouse->posF());
            stage_cant->setGeometry(rect.toRect());

            auto profile = vSettings.objProfiles.at(prev_idx);
            profile->stageHeadCantX = QVariant(QPlot->invTransform(QwtPlot::xBottom,
                                               head_cant->geometry().center().x())).toDouble();
            profile->stageHeadCantY = QVariant(QPlot->invTransform(QwtPlot::yLeft,
                                               head_cant->geometry().center().y())).toDouble();

            head_cant->show();
            stage_cant->show();
            BCantileverRange->change_state_man(true);
            sl_save();
        } else if (plot_mode == 2) {
            QRectF rect;
            rect = stage_laser->geometry();
            rect.moveCenter(mouse->posF());
            stage_laser->setGeometry(rect.toRect());

            auto profile = vSettings.objProfiles.at(prev_idx);
            profile->stageLaserX = QVariant(QPlot->invTransform(QwtPlot::xBottom,
                                            stage_laser->geometry().center().x())).toDouble();
            profile->stageLaserY = QVariant(QPlot->invTransform(QwtPlot::yLeft,
                                            stage_laser->geometry().center().y())).toDouble();

            stage_laser->show();
            BLaserRange->change_state_man(true);
            sl_save();
        } else if (plot_mode == 3 && active) {
            QPointF posy = mouse->posF();
            marker->setXValue(QPlot->invTransform(QwtPlot::xBottom, posy.x()));
            marker->setYValue(QPlot->invTransform(QwtPlot::yLeft, posy.y()));

            posx.rx() = QPlot->transform(QwtPlot::xBottom, markerA->xValue()) / plotCanvas->width();
            posx.ry() = QPlot->transform(QwtPlot::yLeft, markerA->yValue()) / plotCanvas->height();

            posy.rx() = QPlot->transform(QwtPlot::xBottom, markerB->xValue()) / plotCanvas->width();
            posy.ry() = QPlot->transform(QwtPlot::yLeft, markerB->yValue()) / plotCanvas->height();

            qreal kx = fabs(QPlot->invTransform(QwtPlot::xBottom, 20) -
                            QPlot->invTransform(QwtPlot::xBottom, 0)) / 20.;
            qreal ky = fabs(QPlot->invTransform(QwtPlot::yLeft, 20) -
                            QPlot->invTransform(QwtPlot::yLeft, 0)) / 20.;
            qreal dx = (markerB->xValue() - markerA->xValue()) / kx;
            qreal dy = (markerB->yValue() - markerA->yValue()) / ky;

            qreal dist = sqrt(dx*dx + dy*dy);
            qreal scale = ewsz->value() / dist;

            auto range_x = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::RangeX));
            auto range_y = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::RangeY));
            range_x->setValue(scale * plotCanvas->width());
            range_y->setValue(scale * plotCanvas->height());

            mkm_px_x = (qreal)plotCanvas->width() / range_x->value();
            mkm_px_y = (qreal)plotCanvas->height() / range_y->value();

            QPlot->setAxisScale(QwtPlot::xBottom, 0, range_x->value());
            QPlot->setAxisScale(QwtPlot::yLeft, 0, range_y->value());

            QRectF rect; QPointF center;
            rect = head_cant->geometry(); center = rect.center();
            rect.setSize(QSizeF(mkm_px_x * Xmax, mkm_px_y * Ymax)); rect.moveCenter(center);
            head_cant->setGeometry(rect.toRect());
            rect = stage_cant->geometry(); center = rect.center();
            rect.setSize(QSizeF(mkm_px_x * Xmax2, mkm_px_y * Ymax2)); rect.moveCenter(center);
            stage_cant->setGeometry(rect.toRect());
            rect = stage_laser->geometry(); center = rect.center();
            rect.setSize(QSizeF(mkm_px_x * Xmax2, mkm_px_y * Ymax2)); rect.moveCenter(center);
            stage_laser->setGeometry(rect.toRect());

            BAxes->change_state_man_with_sig(true);
            QPlot->replot();

            markerA->setXValue(QPlot->invTransform(QwtPlot::xBottom, posx.x() * plotCanvas->width()));
            markerA->setYValue(QPlot->invTransform(QwtPlot::yLeft, posx.y() * plotCanvas->height()));
            markerB->setXValue(QPlot->invTransform(QwtPlot::xBottom, posy.x() * plotCanvas->width()));
            markerB->setYValue(QPlot->invTransform(QwtPlot::yLeft, posy.y() * plotCanvas->height()));
            lineAB->SetData(QPointF(markerA->xValue(), markerA->yValue()),
                            QPointF(markerB->xValue(), markerB->yValue()));

            QPlot->replot();
            sl_save();
        }
        active = false;
        return true;
    } else if (Event->type() == QEvent::MouseMove && active) {
        if (plot_mode == 0) {
            qreal x = QPlot->invTransform(QwtPlot::xBottom, posx.x());
            qreal y = QPlot->invTransform(QwtPlot::yLeft, posx.y());
            qreal x2 = QPlot->invTransform(QwtPlot::xBottom, mouse->posF().x());
            qreal y2 = QPlot->invTransform(QwtPlot::yLeft, mouse->posF().y());
            lineAB->SetData(QPointF(x, y), QPointF(x2, y2));
            QPlot->replot();

            QPointF dxy = mouse->posF() - posx;
            dxy.rx() *= 2. / qMin(plotCanvas->width(), plotCanvas->height());
            dxy.ry() *= 2. / qMin(plotCanvas->width(), plotCanvas->height());
            goXYjoy(dxy);
        }
        if (plot_mode == 3 || plot_mode == 4) {
            posx = mouse->posF();
            marker->setXValue(QPlot->invTransform(QwtPlot::xBottom, posx.x()));
            marker->setYValue(QPlot->invTransform(QwtPlot::yLeft, posx.y()));

            lineAB->SetData(QPointF(markerA->xValue(), markerA->yValue()),
                            QPointF(markerB->xValue(), markerB->yValue()));

            if (plot_mode == 4) {
                qreal dx = markerB->xValue() - markerA->xValue();
                qreal dy = markerB->yValue() - markerA->yValue();

                qreal dist = sqrt(dx*dx + dy*dy);
                markerC->setXValue((markerA->xValue() +  markerB->xValue()) / 2);
                markerC->setYValue((markerA->yValue() +  markerB->yValue()) / 2);
                QwtText text;
                text.setText(QString::number(dist, 'f', 2) + tr("micro") + tr("m"));
                text.setColor(Qt::black);

                QFont font(QPlot->axisFont(0).family());
                font.setBold(true);
                font.setPointSize(QPlot->axisFont(0).pointSize());
                auto scan = ScanPlot::GetIt();
                if (scan)
                {
                    font.setPointSize(scan->Plot->TwoDPlot->Pure2dPlot->QPlot->axisFont(0).pointSize());
                }
                text.setFont(font);
                markerC->setLabel(text);
            }

            QPlot->replot();
        }
        return true;
    } else if (Event->type() == QEvent::Resize) {
        mkm_px_x = (qreal)plotCanvas->width() / sl_get_rangeX();
        mkm_px_y = (qreal)plotCanvas->height() / sl_get_rangeY();

        auto profile = vSettings.objProfiles.at(prev_idx);
        QRectF rect;
        rect = head_cant->geometry();
        rect.setSize(QSizeF(mkm_px_x * Xmax, mkm_px_y * Ymax));
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
        head_cant->setGeometry(rect.toRect());

        rect = stage_cant->geometry();
        rect.setSize(QSizeF(mkm_px_x * Xmax2, mkm_px_y * Ymax2));
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
        stage_cant->setGeometry(rect.toRect());

        rect = stage_laser->geometry();
        rect.setSize(QSizeF(mkm_px_x * Xmax2, mkm_px_y * Ymax2));
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageLaserX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageLaserY)));
        stage_laser->setGeometry(rect.toRect());

        QPlot->replot();
    }

    return false;
}

void VideoPlayer::sl_center_set(bool value)
{
    if (value) {
        BCenterAndSet2->change_state_man(false);
        BLine->change_state_man(false);
        BLineMeasure->change_state_man(false);
        markerA->hide();
        markerB->hide();
        lineAB->hide();
        markerC->setLabel(QwtText());
        QPlot->replot();
        plot_mode = 1; //cantilever center

        auto osc = Oscilloscope::GetIt();
        if (!osc)
            return;

        osc->DacsWidget->go_to_center(0);
        osc->DacsWidget->go_to_center(1);
    } else {
        plot_mode = 0;
    }
}
void VideoPlayer::sl_center_set2(bool value)
{
    if (value) {
        BCenterAndSet->change_state_man(false);
        BLine->change_state_man(false);
        BLineMeasure->change_state_man(false);
        markerA->hide();
        markerB->hide();
        lineAB->hide();
        markerC->setLabel(QwtText());
        QPlot->replot();
        plot_mode = 2; //laser center

        auto osc = Oscilloscope::GetIt();
        if (!osc)
            return;

        osc->DacsWidget->go_to_center(0);
        osc->DacsWidget->go_to_center(1);
    } else {
        plot_mode = 0;
    }
}
void VideoPlayer::sl_axes(bool value)
{
    QPlot->enableAxis(QwtPlot::yLeft, value);
    QPlot->enableAxis(QwtPlot::xBottom, value);
    QPlot->replot();
}
void VideoPlayer::sl_grid(bool value)
{
    QGrid->setVisible(value);
    QPlot->replot();
}
void VideoPlayer::sl_line(bool value)
{
    if (value) {
        BCenterAndSet->change_state_man(false);
        BCenterAndSet2->change_state_man(false);
        BLineMeasure->change_state_man(false);
        plot_mode = 3; //line mode

        auto range_x = sl_get_rangeX();
        auto range_y = sl_get_rangeY();
        markerA->setXValue(range_x * 0.4);
        markerA->setYValue(range_y * 0.4);
        markerB->setXValue(range_x * 0.6);
        markerB->setYValue(range_y * 0.6);
        markerC->setLabel(QwtText());

        markerA->show();
        markerB->show();
        lineAB->SetData(QPointF(markerA->xValue(), markerA->yValue()),
                        QPointF(markerB->xValue(), markerB->yValue()));
        lineAB->show();

        QPlot->replot();
    } else {
        plot_mode = 0;
        markerA->hide();
        markerB->hide();
        lineAB->hide();
        QPlot->replot();
    }
}
void VideoPlayer::sl_line_measure(bool value)
{
    QPlot->canvas()->setCursor(QCursor(value ? Qt::ArrowCursor : Qt::OpenHandCursor));
    if (value) {
        BCenterAndSet->change_state_man(false);
        BCenterAndSet2->change_state_man(false);
        BLine->change_state_man(false);
        plot_mode = 4; //line measure mode

        auto range_x = sl_get_rangeX();
        auto range_y = sl_get_rangeY();
        markerA->setXValue(range_x * 0.4);
        markerA->setYValue(range_y * 0.4);
        markerB->setXValue(range_x * 0.6);
        markerB->setYValue(range_y * 0.6);

        qreal dx = markerB->xValue() - markerA->xValue();
        qreal dy = markerB->yValue() - markerA->yValue();

        qreal dist = sqrt(dx*dx + dy*dy);
        markerC->useCompositeDrawMode = false;
        markerC->setXValue((markerA->xValue() +  markerB->xValue()) / 2);
        markerC->setYValue((markerA->yValue() +  markerB->yValue()) / 2);
        QwtText text;
        text.setText(QString::number(dist, 'f', 2) + tr("micro") + tr("m"));
        text.setColor(Qt::black);

        QFont font(QPlot->axisFont(0).family());
        font.setBold(true);
        font.setPointSize(QPlot->axisFont(0).pointSize());
        auto scan = ScanPlot::GetIt();
        if (scan)
        {
            font.setPointSize(scan->Plot->TwoDPlot->Pure2dPlot->QPlot->axisFont(0).pointSize());
        }
        text.setFont(font);
        markerC->setLabel(text);
        markerC->show();

        markerA->show();
        markerB->show();
        lineAB->SetData(QPointF(markerA->xValue(), markerA->yValue()),
                        QPointF(markerB->xValue(), markerB->yValue()));
        lineAB->show();

        QPlot->replot();
    } else {
        plot_mode = 0;
        markerA->hide();
        markerB->hide();
        lineAB->hide();
        markerC->setLabel(QwtText());
        QPlot->replot();
    }
}
void VideoPlayer::sl_add_config()
{
    auto *profile = new videoObjectiveProfile();
    profile->name = tr("Default");
    vSettings.objProfiles.append(profile);
    int new_idx = vSettings.objProfiles.size() - 1;

    TableVideoProfiles->setRowCount(vSettings.objProfiles.size());

    addNewProfileInTable(new_idx);
    sl_select_config(new_idx);
    sl_save();
}
void VideoPlayer::sl_del_config()
{
    int idx = prev_idx;
    prev_idx = -1;

    delete vSettings.objProfiles.at(idx);
    vSettings.objProfiles.removeAt(idx);
    TableVideoProfiles->removeRow(idx);

    if (vSettings.objProfiles.isEmpty())
        sl_add_config();
    else {
        idx--;
        sl_select_config((idx < 0) ? 0 : idx);
        sl_save();
    }
}
void VideoPlayer::sl_select_config(int val)
{
    if (prev_idx == val)
        return;

    if (val != -1)
    {
        prev_idx = val;
        auto rBtn = dynamic_cast<QRadioButton*>(TableVideoProfiles->cellWidget(val, VProfile::Selector));
        if (rBtn)
            rBtn->setChecked(true);
        auto profile = vSettings.objProfiles.at(val);
        QPlot->setTitle(tr("Configuration - ") + profile->name);

        QRectF rect;
        rect = head_cant->geometry();
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
        head_cant->setGeometry(rect.toRect());

        rect = stage_cant->geometry();
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
        stage_cant->setGeometry(rect.toRect());

        rect = stage_laser->geometry();
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageLaserX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageLaserY)));
        stage_laser->setGeometry(rect.toRect());

        autoFocusDist1 = profile->afD0;
        autoFocusV1 = profile->afVst;
        autoFocusV2 = profile->afVpr;
        autoFocusMagicMultiply[0] = profile->afCrat;
        autoFocusMagicMultiply[1] = profile->afCrat2;

        emit VideoProfileChanged(val);
    }

    sl_range();
    on_cwh_ch();
    autoFocusSettingsChanged();
    sl_save();
}
void VideoPlayer::SetCurrentVideoProfile()
{
    if (!sender())
        return;

    int numProfiles = vSettings.objProfiles.size();
    for (int i = 0; i < numProfiles; i++)
    {
        if (sender() == TableVideoProfiles->cellWidget(i, VProfile::Selector)) {
            sl_select_config(i);
            return;
        }
    }
}
QList<QString> *VideoPlayer::getVideoProfiles()
{
    auto *list = new QList<QString>();

    for (auto profile : vSettings.objProfiles)
    {
        list->push_back(profile->name);
    }

    return list;
}
void VideoPlayer::sl_save()
{
    int idx = prev_idx;

    QString suffix = inst_idx == 0 ? QString() : QString::number(inst_idx+1);
    QString file = QString("Param/video%1.conf").arg(suffix);
    QSettings settingsNew(file, QSettings::IniFormat);

    settingsNew.setValue("LastSelectedObjProfile", idx);
    settingsNew.setValue("xyMsel", vSettings.xyMsel = xyMsel->currentIndex());
    settingsNew.setValue("FlipX", vSettings.flipX = flipx->isChecked());
    settingsNew.setValue("FlipY", vSettings.flipY = flipy->isChecked());
    settingsNew.setValue("SwapXY", vSettings.swapXY = rotxy->isChecked());
    settingsNew.setValue("MotorZFocusSpeed", vSettings.motorZFocusSpeed = motorZFocusSpeedEdit->value());
    settingsNew.setValue("MotorXYStageSpeed", vSettings.motorXYStageSpeed = motorXYStageSpeedEdit->value());
    settingsNew.setValue("AutoFocusExitDistanceEdit", vSettings.afExitDistance = autoFocusExitDistanceEdit->value());
    settingsNew.setValue("TipThinkness", vSettings.tipThinkbness = tipThinkbnessEdit->value());
    settingsNew.setValue("MafPrcsEnabled", vSettings.mafPrcsEnabled = mafPrcsEnabled->isChecked());
    settingsNew.setValue("MafFastEnabled", vSettings.mafFastEnabled = mafFastEnabled->isChecked());
    settingsNew.setValue("PafEnabled", vSettings.pafEnabled = pafEnabled->isChecked());
    settingsNew.setValue("MafPaf", vSettings.mafPaf = mafPaf->isChecked());

    QList<QString> writtedGroups;
    for (int i = 0; i < vSettings.objProfiles.size(); i++)
    {
        auto profile = vSettings.objProfiles.at(i);
        profile->name = dynamic_cast<QLineEdit*>(TableVideoProfiles->cellWidget(i, VProfile::Name))->text();
        auto gpName = QString::number(i) + QChar(1111) + profile->name;

        writtedGroups.append(gpName);
        settingsNew.beginGroup(gpName);
        {
            profile->rangeX = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::RangeX))->text().replace(",", ".").toInt();
            profile->rangeY = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::RangeY))->text().replace(",", ".").toInt();
            profile->height = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::Height))->text().replace(",", ".").toInt();
            profile->width = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::Width))->text().replace(",", ".").toInt();
            settingsNew.setValue("RangeX", profile->rangeX);
            settingsNew.setValue("RangeY", profile->rangeY);
            settingsNew.setValue("Height", profile->height);
            settingsNew.setValue("Width", profile->width);

            settingsNew.setValue("StageHeadCantX", profile->stageHeadCantX);
            settingsNew.setValue("StageHeadCantY", profile->stageHeadCantY);
            settingsNew.setValue("StageLaserX", profile->stageLaserX);
            settingsNew.setValue("StageLaserY", profile->stageLaserY);

            settingsNew.beginGroup("AutoFocus");
            {
                profile->afD0 = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::D0))->text().toInt();
                profile->afVst = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::Vst))->text().toInt();
                profile->afVpr = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::Vpr))->text().toInt();
                profile->afCrat = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::Crat))->text().toInt();
                profile->afLeftStPr = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::LeftStPr))->text().toInt();
                profile->afCrat2 = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::Crat2))->text().toDouble();
                profile->afCratPrecize = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(i, VProfile::CratPresize))->text().toInt();
                settingsNew.setValue("D0", profile->afD0);
                settingsNew.setValue("Vst", profile->afVst);
                settingsNew.setValue("Vpr", profile->afVpr);
                settingsNew.setValue("Crat", profile->afCrat);
                settingsNew.setValue("LeftStPr", profile->afLeftStPr);
                settingsNew.setValue("Crat2", profile->afCrat2);
                settingsNew.setValue("CratPresize", profile->afCratPrecize);
            }
            settingsNew.endGroup();
        }
        settingsNew.endGroup();
    }

    //removing groups that no longer exists
    auto groups = settingsNew.childGroups();
    for (int i = 0; i<groups.count(); i++)
    {
        if (writtedGroups.contains(groups[i]))
            continue;
        settingsNew.beginGroup(groups[i]);
        settingsNew.remove("");
        settingsNew.endGroup();
    }

    QPlot->setTitle(tr("Configuration - ") + vSettings.objProfiles.at(idx)->name);
    //---------------------------------------

    auto scan = ScanPlot::GetIt();
    if (scan)
    {
        scan->updateVideoProfiles(this, idx);//todo make it as a connect-event
    }

    autoFocusScanVariant = autoFocusMode->currentIndex();
    if (autoFocusScanVariant == 0)
        autoFocusMagicMultiply[0] = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Crat))->value();
    else if (autoFocusScanVariant == 1)
        autoFocusMagicMultiply[0] = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::CratPresize))->value();

    autoFocusMagicMultiply[1] = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Crat2))->value();
    motorZFocusSpeed = motorZFocusSpeedEdit->value();
    motorXYStageSpeed = motorXYStageSpeedEdit->value();
    autoFocusExitDistance = autoFocusExitDistanceEdit->value();
    tipThinkbness = tipThinkbnessEdit->value();
}

void VideoPlayer::sl_set_XYZ_range(double x, double y, double )
{
    Xmax = x;
    Ymax = y;

    sl_range();
}
void VideoPlayer::sl_set_XYZ_range2(double x, double y, double )
{
    Xmax2 = x;
    Ymax2 = y;

    sl_range();
}
double VideoPlayer::sl_get_rangeX()
{
    return dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::RangeX))->value();
}
double VideoPlayer::sl_get_rangeY()
{
    return dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::RangeY))->value();
}
void VideoPlayer::sl_range()
{
    auto range_x = sl_get_rangeX();
    auto range_y = sl_get_rangeY();

    mkm_px_x = (qreal)QPlot->canvas()->width() / range_x;
    mkm_px_y = (qreal)QPlot->canvas()->height() / range_y;

    QPlot->setAxisScale(QwtPlot::xBottom, 0, range_x);
    QPlot->setAxisScale(QwtPlot::yLeft, 0, range_y);

    QRectF rect; QPointF center;
    rect = head_cant->geometry(); center = rect.center();
    rect.setSize(QSizeF(mkm_px_x * Xmax, mkm_px_y * Ymax)); rect.moveCenter(center);
    head_cant->setGeometry(rect.toRect());
    rect = stage_cant->geometry(); center = rect.center();
    rect.setSize(QSizeF(mkm_px_x * Xmax2, mkm_px_y * Ymax2)); rect.moveCenter(center);
    stage_cant->setGeometry(rect.toRect());
    rect = stage_laser->geometry(); center = rect.center();
    rect.setSize(QSizeF(mkm_px_x * Xmax2, mkm_px_y * Ymax2)); rect.moveCenter(center);
    stage_laser->setGeometry(rect.toRect());
    QPlot->replot();

    static bool first = true;
    if (first) {
        first = false;

        auto profile = vSettings.objProfiles.at(prev_idx);
        rect = head_cant->geometry();
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
        head_cant->setGeometry(rect.toRect());

        rect = stage_cant->geometry();
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
        stage_cant->setGeometry(rect.toRect());

        rect = stage_laser->geometry();
        rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageLaserX),
                                QPlot->transform(QwtPlot::yLeft, profile->stageLaserY)));
        stage_laser->setGeometry(rect.toRect());

        QPlot->replot();
    }

    auto profile = vSettings.objProfiles.at(prev_idx);
    profile->stageHeadCantX = QVariant(QPlot->invTransform(QwtPlot::xBottom,
                                       head_cant->geometry().center().x())).toDouble();
    profile->stageHeadCantY = QVariant(QPlot->invTransform(QwtPlot::yLeft,
                                       head_cant->geometry().center().y())).toDouble();
    profile->stageLaserX = QVariant(QPlot->invTransform(QwtPlot::xBottom,
                                    stage_laser->geometry().center().x())).toDouble();
    profile->stageLaserY = QVariant(QPlot->invTransform(QwtPlot::yLeft,
                                    stage_laser->geometry().center().y())).toDouble();

    sl_save();
}

int VideoPlayer::times_constructed = 0;
VideoPlayer::VideoPlayer(QWidget *parent)
    : QWidget(parent)
{
    frames = 0;
    osc_contrast = 0.;
    new_osc_contrast = false;
    playing = false;
    marker = nullptr;
    inst_idx = times_constructed++;
    QString suffix = inst_idx == 0 ? QString() : QString::number(inst_idx+1);

    active = false;
    refresh_limit_timer.invalidate();
    refresh_limit_last_elapsed = 0;

    configPath = QString("Param/video%1.conf").arg(suffix);
    isInterrupted = false;

    plot_mode = 0;
    mkm_px_x = 1;
    mkm_px_y = 1;
    prev_idx = 0;
    cw = 1;
    ch = 1;


    QPlot = new QwtPlot(this);
    QPlot->canvas()->setCursor(Qt::OpenHandCursor);
    QPlot->enableAxis(QwtPlot::yLeft, false);
    QPlot->enableAxis(QwtPlot::xBottom, false);
    QPlot->setAxisScale(QwtPlot::xBottom, 0, 1024);
    QPlot->setAxisScale(QwtPlot::yLeft, 0, 768);
    QFont font = qApp->font(); font.setPointSize(font.pointSize() + 2);
    QPlot->titleLabel()->setFont(font);
    QwtText tx(tr("mkm")); tx.setFont(font);
    QPlot->setAxisTitle(QwtPlot::xBottom, tx);
    QPlot->setAxisTitle(QwtPlot::yLeft, tx);
    QPlot->setAxisFont(QwtPlot::xBottom, font);
    QPlot->setAxisFont(QwtPlot::yLeft, font);

    QGrid = new QwtPlotGrid();
    QGrid->setMajPen(QPen(Qt::black));
    QGrid->attach(QPlot);
    QGrid->hide();

    BAddConfig = new QPushButton(makeHugeIcon(":/USBVideo/ConfPlus.png"), QString());
    BAddConfig->setToolTip(tr("Add new configuration"));
    connect(BAddConfig, SIGNAL(clicked(bool)), this, SLOT(sl_add_config()));


    BDelConfig = new QPushButton(makeHugeIcon(":/USBVideo/ConfMinus.png"), QString());
    BDelConfig->setToolTip(tr("Remove current configuration"));
    connect(BDelConfig, SIGNAL(clicked(bool)), this, SLOT(sl_del_config()));

    markerA = new QwtPlotMarker();
    markerB = new QwtPlotMarker();
    markerC = new QwtPlotMarker();
    markerA->attach(QPlot); markerA->hide();
    markerB->attach(QPlot); markerB->hide();
    markerC->attach(QPlot); markerC->hide();
    markerA->setXValue(10); markerA->setYValue(10);
    markerB->setXValue(20); markerB->setYValue(20);
    markerB->setXValue(15); markerB->setYValue(15);
    markerA->setLineStyle(QwtPlotMarker::Cross); markerA->setLinePen(QPen(Qt::white, 1, Qt::DashLine));
    markerA->setSymbol(new QwtSymbol(QwtSymbol::Diamond, QColor(255,255,255), QColor(255,255,255), QSize(9, 9)));
    markerB->setLineStyle(QwtPlotMarker::Cross); markerB->setLinePen(QPen(Qt::white, 1, Qt::DashLine));
    markerB->setSymbol(new QwtSymbol(QwtSymbol::Diamond, QColor(255,255,255), QColor(255,255,255), QSize(9, 9)));

    lineAB = new plot::TLineSegment;
    lineAB->attach(QPlot);
    lineAB->hide();
    lineAB->setPen(QPen(QColor(255, 33, 33),   1, Qt::DashLine));

    head_cant = new QRubberBand(QRubberBand::Rectangle, QPlot->canvas());
    stage_cant = new QRubberBand(QRubberBand::Rectangle, QPlot->canvas());
    stage_laser = new QRubberBand(QRubberBand::Rectangle, QPlot->canvas());
    head_cant->hide();
    stage_cant->hide();
    stage_laser->hide();

    QPalette palette;
    palette.setBrush( QPalette::Base, QBrush( Qt::cyan ) );
    head_cant->setPalette(palette);
    palette.setBrush( QPalette::Base, QBrush( Qt::yellow ) );
    stage_cant->setPalette(palette);
    palette.setBrush( QPalette::Base, QBrush( Qt::green ) );
    stage_laser->setPalette(palette);

    head_cant->setStyle(QStyleFactory::create("Motif"));
    stage_cant->setStyle(QStyleFactory::create("Motif"));
    stage_laser->setStyle(QStyleFactory::create("Motif"));

    Xmax = Ymax = Xmax2 = Ymax2 = 100;

    QRectF rect;
    rect = head_cant->geometry();
    rect.setSize(QSizeF(mkm_px_x*50, mkm_px_y*50));
    head_cant->setGeometry(rect.toRect());
    rect = stage_cant->geometry();
    rect.setSize(QSizeF(mkm_px_x*100, mkm_px_y*100));
    stage_cant->setGeometry(rect.toRect());
    rect = stage_laser->geometry();
    rect.setSize(QSizeF(mkm_px_x*100, mkm_px_y*100));
    stage_laser->setGeometry(rect.toRect());



    QGroupBox *GBsettings = new QGroupBox(tr("Objective settings"), this);
    GBsettings->setVisible(true);

    QGroupBox *GrBControl = new QGroupBox(this);

    RatioLayout *lusbcamlay = new RatioLayout;
    lusbcamlay->setAlignment(Qt::AlignRight);
    lusbcamlay->addWidget(QPlot);

    QBoxLayout *layout = new QHBoxLayout(this);
    layout->addWidget(GrBControl);
    layout->addWidget(GBsettings);
    //layout->addWidget(GBExtraSettings);
    layout->addLayout(lusbcamlay, 1);

    mainWidget = new QWidget;
    mainWidget->setLayout(layout);

    BAddConfig->setParent(GBsettings);
    BDelConfig->setParent(GBsettings);
    BAddConfig->setFixedSize(21, 21);
    BDelConfig->setFixedSize(21, 21);

    BDelConfig->move(20 + QFontMetrics(GBsettings->font()).width(GBsettings->title()), 0);
    BAddConfig->move(BDelConfig->x() + BDelConfig->width(), 0);

    activeCamera = nullptr;
#ifdef USE_IVIDEO_MOTIC
    CameraList.append(new Motic());
#endif
#ifdef USE_IVIDEO_MOTIC_X
    CameraList.append(new MoticX());
#endif
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#ifdef USE_IVIDEO_MOTIC_1080
    CameraList.append(new motic1080());
#endif
#endif

    foreach (IVideo *v, CameraList)
        connect(v->getObject(), SIGNAL(ImgReceived(QImage)), this, SLOT(on_Camera_ImgReceived(QImage)), Qt::QueuedConnection);

    deviceStatus = new StatusWidget1(tr("Video") + (inst_idx == 0 ? QString() : (" " + suffix)));
    setDeviceStatus(EDno);
    connect(&resetWatch, SIGNAL(finished()), this, SLOT(resetCamerasCommit()));

    QAction *a;
    deviceStatus->setContextMenuPolicy(Qt::ActionsContextMenu);
    deviceStatus->addAction(a = new QAction(tr("Reinit camera"), deviceStatus));
    connect(a, SIGNAL(triggered()), this, SLOT(resetCameras()));

    m_BStart = new TwoStateButton(":start.png",":stop.png");
    connect(m_BStart, SIGNAL(state_changed(bool)), this, SLOT(StartStop(bool)));

    BAutoFocusOsc = new QPushButton(makeHugeIcon(":/SPMTuning_window/AutoFocusPMT.png"), QString());
    BAutoFocusOsc->setToolTip(tr("Laser Auto Focus"));
    connect(BAutoFocusOsc, SIGNAL(clicked()), SLOT(on_BAutoFocusOsc_clicked()));

    BAutoFocus = new QPushButton(makeHugeIcon(":/SPMTuning_window/AutoFocus.png"), QString());
    BAutoFocus->setToolTip(tr("Auto Focus"));
    connect(BAutoFocus, SIGNAL(clicked()), SLOT(on_BAutoFocus_clicked()));

    BToData = new QPushButton(makeHugeIcon(":/plot/Attach.png"), QString());
    BToData->setToolTip(tr("Convert to grayscale & Send to Data as 3d") + " (Ctrl+S)");
    BToData->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_S));
    connect(BToData, SIGNAL(clicked()), SLOT(on_BToData_clicked()));

    saveB = new QPushButton(makeHugeIcon(":/snap_shot.png"), QString());
    saveB->setToolTip(tr("Save image"));
    saveB->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_P));
    connect(saveB, SIGNAL(clicked()), this, SLOT(sl_saveToFile()));

    BCenterAndSet = new TwoStateButton(":/USBVideo/TipAlignCenterOff.png",":/USBVideo/TipAlignCenterOn.png");
    BCenterAndSet->setToolTip(tr("Tip scan area center alignment"));
    connect(BCenterAndSet, SIGNAL(state_changed(bool)), this, SLOT(sl_center_set(bool)));

    BCenterAndSet2 = new TwoStateButton(":/USBVideo/LaserSpotAlignCenterOff.png",":/USBVideo/LaserSpotAlignCenterOn.png");
    BCenterAndSet2->setToolTip(tr("Laser scan area center alignment"));
    connect(BCenterAndSet2, SIGNAL(state_changed(bool)), this, SLOT(sl_center_set2(bool)));

    BLine = new TwoStateButton(":/USBVideo/sizeByLineOff.png",":/USBVideo/sizeByLineOn.png");
    BLine->setToolTip(tr("Set video size"));
    connect(BLine, SIGNAL(state_changed(bool)), this, SLOT(sl_line(bool)));

    BLineMeasure = new TwoStateButton(":/USBVideo/sizeByLineOff.png",":/USBVideo/sizeByLineOn.png");
    BLineMeasure->setToolTip(tr("Measure distance"));
    connect(BLineMeasure, SIGNAL(state_changed(bool)), this, SLOT(sl_line_measure(bool)));
    BLineMeasure->setParent(QPlot);
    BLineMeasure->setFlat(true);
    BLineMeasure->setStyleSheet( "background-color: rgba(255,255,255,10);" );

    BCantileverRange = new TwoStateButton(":/USBVideo/TipAreaShowOff.png",":/USBVideo/TipAreaShowOn.png");
    BCantileverRange->setToolTip(tr("Show tip scan area"));
    connect(BCantileverRange, SIGNAL(state_changed(bool)),head_cant, SLOT(setVisible(bool)));
    connect(BCantileverRange, SIGNAL(state_changed(bool)),stage_cant, SLOT(setVisible(bool)));

    BLaserRange = new TwoStateButton(":/USBVideo/LaserAreaShowOff.png",":/USBVideo/LaserAreaShowOn.png");
    BLaserRange->setToolTip(tr("Show laser spot scan area"));
    connect(BLaserRange, SIGNAL(state_changed(bool)),stage_laser, SLOT(setVisible(bool)));

    BGrid = new TwoStateButton(":/USBVideo/GridOff.png",":/USBVideo/GridOn.png");
    BGrid->setToolTip(tr("Show grid"));
    connect(BGrid, SIGNAL(state_changed(bool)), this, SLOT(sl_grid(bool)));
    BGrid->setParent(QPlot);
    BGrid->setFlat(true);
    BGrid->setStyleSheet( "background-color: rgba(255,255,255,10);" );

    BAxes = new TwoStateButton(":/USBVideo/ScaleOff.png",":/USBVideo/ScaleOn.png");
    BAxes->setToolTip(tr("Show Axes"));
    connect(BAxes, SIGNAL(state_changed(bool)), this, SLOT(sl_axes(bool)));
    BAxes->setParent(QPlot);
    BAxes->setFlat(true);
    BAxes->setStyleSheet( "background-color: rgba(255,255,255,10);" );

    TRectWidget *clrHead = new TRectWidget(Qt::cyan);
    TRectWidget *clrStage = new TRectWidget(Qt::yellow);
    TRectWidget *clrLaser = new TRectWidget(Qt::green);
    clrHead->setFixedSize(m_BStart->sizeHint().width(), 7);
    clrStage->setFixedSize(m_BStart->sizeHint().width(), 7);
    clrLaser->setFixedSize(m_BStart->sizeHint().width(), 7);

    QPushButton *icHead = new QPushButton(makeHugeIcon(":Head_page.png"), QString());
    QPushButton *icStage = new QPushButton(makeHugeIcon(":stage_page.png"), QString());
    QPushButton *icLaser = new QPushButton(makeHugeIcon(":/SPMTuning_window/laser_on.png"), QString());
    icHead->setToolTip(tr("Head"));
    icStage->setToolTip(tr("Stage"));
    icLaser->setToolTip(tr("Laser"));
    icHead->setFlat(true);
    icStage->setFlat(true);
    icLaser->setFlat(true);
    icHead->setEnabled(false);
    icStage->setEnabled(false);
    icLaser->setEnabled(false);

    GrBControl->hide();
    HFlowLayout *LControls = new HFlowLayout(GrBControl);
    LControls->addWidget(m_BStart);
    LControls->addWidget(saveB);
    LControls->addWidget(BToData);
    addDelimeterH(LControls, m_BStart->sizeHint().width());
    LControls->addWidget(BAutoFocus);
    LControls->addWidget(BAutoFocusOsc);
    addDelimeterH(LControls, m_BStart->sizeHint().width());
    LControls->addWidget(BCantileverRange);
    LControls->addWidget(BLaserRange);
    addDelimeterH(LControls, m_BStart->sizeHint().width());
    addDelimeterH(LControls, m_BStart->sizeHint().width());
    LControls->addWidget(icHead);
    LControls->addWidget(clrHead);
    LControls->addWidget(icStage);
    LControls->addWidget(clrStage);
    LControls->addWidget(icLaser);
    LControls->addWidget(clrLaser);



    LControls->spaceX = LControls->spaceY = 5;
    LControls->setContentsMargins(3, 3, 3, 3);

    zPos = new DoubleChanger(0, 10, 4, false);
    zPos->edit->hide();

    QGroupBox *box_Zpos = new QGroupBox(tr("Obj.Z, ")+tr("mkm"));
    QHBoxLayout *layout_Zpos = new QHBoxLayout(box_Zpos);
    layout_Zpos->addWidget(zPos, 1);

    MotorSlider *mzslider = new MotorSlider(true);
    connect(mzslider, SIGNAL(newPos(int)), this, SLOT(goZjoy(int)));

    QGroupBox *box_mZpos = new QGroupBox(tr("Z motor"));
    QHBoxLayout *layout_mZpos = new QHBoxLayout(box_mZpos);
    layout_mZpos->addWidget(mzslider, 1);

    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        BAxes->change_state_man_with_sig(settings.value("video/axes", true).toBool());
        BGrid->change_state_man_with_sig(settings.value("video/grid", true).toBool());

        if (!settings.value("display/spectroscopy", false).toBool()) {
            BLaserRange->hide();
            BCenterAndSet2->hide();
            icLaser->hide();
            clrLaser->hide();
        }
        bool head_on = settings.value("display/oscilloscope/head", true).toBool();
        bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
        if (!head_on) {
            icHead->hide();
            clrHead->hide();
        }
        if (!stage_on) {
            icStage->hide();
            clrStage->hide();
        }
    }

    ewsz = new Edit(1, 5000, 1);
    flipx = new QCheckBox("");
    flipy = new QCheckBox("");
    rotxy = new QCheckBox("");
    xyMsel = new QComboBox;
    {
        xyMsel->addItem(tr("Off"));
        int idxx, idxy;
        getXYMidx(0, &idxx, &idxy);
        xyMsel->addItem(tr("Head: ") + QString::number(idxx) + ", " + QString::number(idxy));
        getXYMidx(1, &idxx, &idxy);
        xyMsel->addItem(tr("Stage: ") + QString::number(idxx) + ", " + QString::number(idxy));
    }
    mafPaf = new QCheckBox("");
    autoFocusMode = new QComboBox();
    autoFocusMode->addItem(tr("Fast"));
    autoFocusMode->addItem(tr("Presize"));
    motorZFocusSpeedEdit = new Edit(1, 9999);
    motorXYStageSpeedEdit = new Edit(1, 9999);
    autoFocusExitDistanceEdit = new Edit(1, 1000000);
    tipThinkbnessEdit = new Edit(0, 10000);
    mafPrcsEnabled = new QCheckBox("");
    mafFastEnabled = new QCheckBox("");
    pafEnabled = new QCheckBox("");
    
    motorZFocusSpeedEdit->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    motorXYStageSpeedEdit->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    ewsz->setValue(35);

    loadVideoConfiguration();

    connect(flipx, SIGNAL(clicked()), this, SLOT(sl_save()));
    connect(flipy, SIGNAL(clicked()), this, SLOT(sl_save()));
    connect(rotxy, SIGNAL(clicked()), this, SLOT(sl_save()));
    connect(mafPaf, SIGNAL(clicked()), this, SLOT(sl_save()));
    connect(autoFocusMode, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_save()));
    //connect(xyMsel, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_save()));
    connect(motorZFocusSpeedEdit, SIGNAL(newValue()), this, SLOT(sl_save()));
    connect(motorXYStageSpeedEdit, SIGNAL(newValue()), this, SLOT(sl_save()));
    connect(autoFocusExitDistanceEdit, SIGNAL(newValue()), this, SLOT(sl_save()));
    connect(tipThinkbnessEdit, SIGNAL(newValue()), this, SLOT(sl_save()));
    connect(mafPrcsEnabled, SIGNAL(clicked()), this, SLOT(sl_save()));
    connect(mafFastEnabled, SIGNAL(clicked()), this, SLOT(sl_save()));
    connect(pafEnabled, SIGNAL(clicked()), this, SLOT(sl_save()));

    QGridLayout *BGslay = new QGridLayout(GBsettings);
    //LControls->addWidget(BAddConfig);
    //LControls->addWidget(BDelConfig);
    LControls->addWidget(BLine);
    LControls->addWidget(BCenterAndSet);
    LControls->addWidget(BCenterAndSet2);

    int numProfiles = vSettings.objProfiles.size();
    TableVideoProfiles = new QTableWidget(numProfiles, VProfile::COUNT);
    TableVideoProfiles->verticalHeader()->setVisible(false);
    TableVideoProfiles->setDragEnabled(false);
    TableVideoProfiles->setFocusPolicy(Qt::NoFocus);
    TableVideoProfiles->horizontalHeader()->setClickable(false);
    TableVideoProfiles->horizontalHeader()->setMovable(false);
    TableVideoProfiles->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
    TableVideoProfiles->horizontalHeader()->setResizeMode(0, QHeaderView::Stretch);
    TableVideoProfiles->setSortingEnabled(false);

    QStringList headList;
    //header
    headList.append("");
    headList.append(tr("Name"));
    headList.append(tr("Range X") + ", " + tr("mkm"));
    headList.append(tr("Range Y") + ", " + tr("mkm"));
    headList.append(tr("Width") + ", " + tr("% max"));
    headList.append(tr("Height") + ", " + tr("% max"));
    headList.append(tr("D0") + ", " + tr("mkm"));
    headList.append(tr("Vst") + ", " + tr("mkm/s"));
    headList.append(tr("Vpr") + ", " + tr("mkm/s"));
    headList.append(tr("Crat (fast)"));
    headList.append(tr("LeftStPr") + ", " + tr("mkm"));
    headList.append(tr("Crat2"));
    headList.append(tr("Crat Presize"));
    TableVideoProfiles->setHorizontalHeaderLabels(headList);

    TableVideoProfiles->setAlternatingRowColors(true);
    for (int i = 0; i < numProfiles; i++)
    {
        addNewProfileInTable(i);
    }

    TableVideoProfiles->horizontalHeader()->resizeSections(QHeaderView::ResizeToContents);
    TableVideoProfiles->horizontalHeader()->resizeSection(VProfile::Selector, 1);
    int minWidth = 0;
    for (int i=0; i < headList.size(); i++)
        minWidth += TableVideoProfiles->horizontalHeader()->sectionSize(i);

    minWidth += TableVideoProfiles->verticalHeader()->width();
    TableVideoProfiles->setMinimumWidth(minWidth+5);
    TableVideoProfiles->setMaximumWidth(minWidth+5);
    GBsettings->setMinimumWidth(minWidth+25);
    GBsettings->setMaximumWidth(minWidth+25);

    //box_mZpos->setMaximumHeight(90);
    mzslider->setMinimumHeight(50);
    //mzslider->setMaximumHeight(50);

    QHBoxLayout *extraSettsLay = new QHBoxLayout();

    BGslay->addWidget(TableVideoProfiles, 1, 0, 1, 1);
    BGslay->addLayout(extraSettsLay, 4, 0, 1, -1);
    BGslay->addWidget(box_mZpos, 5, 0, 1, -1);
    BGslay->addWidget(box_Zpos, 6, 0, 1, -1);

    QGroupBox *gbImgSetts = new QGroupBox(tr("Image settings"));
    QGroupBox *gbMAFSetts = new QGroupBox(tr("MAF settings"));
    QGroupBox *gbPositSetts = new QGroupBox(tr("Positioning settings"));

    extraSettsLay->addWidget(gbImgSetts);
    extraSettsLay->addWidget(gbMAFSetts);
    extraSettsLay->addWidget(gbPositSetts);

    QGridLayout *glImgSetts = new QGridLayout(gbImgSetts);
    QGridLayout *glMAFSetts = new QGridLayout(gbMAFSetts);
    QGridLayout *glPositSetts = new QGridLayout(gbPositSetts);

    glImgSetts->addWidget(new QLabel(tr("Flip x")), 0, 0);
    glImgSetts->addWidget(new QLabel(tr("Flip y")), 1, 0);
    glImgSetts->addWidget(new QLabel(tr("Swap X/Y")), 2, 0);
    glImgSetts->addWidget(flipx, 0, 1);
    glImgSetts->addWidget(flipy, 1, 1);
    glImgSetts->addWidget(rotxy, 2, 1);

    glMAFSetts->addWidget(new QLabel(tr("Enable MAF Prcs:")), 0, 0);
    glMAFSetts->addWidget(new QLabel(tr("Enable MAF Fast:")), 1, 0);
    glMAFSetts->addWidget(new QLabel(tr("Enable PAF:")), 2, 0);
    glMAFSetts->addWidget(new QLabel(tr("MAF + PAF")), 3, 0);
    //glMAFSetts->addWidget(new QLabel(tr("Exit distance [mkm]:")), 4, 0);
    //glMAFSetts->addWidget(new QLabel(tr("Tip thickness [mkm]:")), 5, 0);
    glMAFSetts->addWidget(mafPrcsEnabled, 0, 1);
    glMAFSetts->addWidget(mafFastEnabled, 1, 1);
    glMAFSetts->addWidget(pafEnabled, 2, 1);
    glMAFSetts->addWidget(mafPaf, 3, 1);
    //glMAFSetts->addWidget(autoFocusExitDistanceEdit, 4, 1);
    //glMAFSetts->addWidget(tipThinkbnessEdit, 5, 1);

    glPositSetts->addWidget(new QLabel(tr("Focus speed [mkm/s]:")), 0, 0);
    glPositSetts->addWidget(new QLabel(tr("VxyPos [mkm/s]:")), 1, 0);
    glPositSetts->addWidget(motorZFocusSpeedEdit, 0, 1);
    glPositSetts->addWidget(motorXYStageSpeedEdit, 1, 1);

    BGslay->setAlignment(Qt::AlignTop);

    ritem = new plot::ImageItem;
    ritem->image = QImage(":/USBVideo/USB_cam_win.png");
    ritem->attach(QPlot);
    QPlot->setMinimumSize(10, 8);
    QPlot->canvas()->installEventFilter(this);

    setWindowTitle(tr("Video - NSpec"));
    autoFocusSettingsChanged();

    setVisiblePlotButtons(false);
}
void VideoPlayer::loadVideoConfiguration()
{
    {
    QString suffix = inst_idx == 0 ? QString() : QString::number(inst_idx+1);
    QString file = QString("Param/video%1.conf").arg(suffix);
    QSettings settingsNew(file, QSettings::IniFormat);

    vSettings.lastSelectedObjProfile = settingsNew.value("LastSelectedObjProfile", vSettings.lastSelectedObjProfile).toInt();
    vSettings.xyMsel = settingsNew.value("xyMsel", vSettings.xyMsel).toInt();
    vSettings.flipX = settingsNew.value("FlipX", vSettings.flipX).toBool();
    vSettings.flipY = settingsNew.value("FlipY", vSettings.flipY).toBool();
    vSettings.swapXY = settingsNew.value("SwapXY", vSettings.swapXY).toBool();
    vSettings.motorZFocusSpeed = settingsNew.value("MotorZFocusSpeed", vSettings.motorZFocusSpeed).toInt();
    vSettings.motorXYStageSpeed = settingsNew.value("MotorXYStageSpeed", vSettings.motorXYStageSpeed).toInt();
    vSettings.afExitDistance = settingsNew.value("AutoFocusExitDistanceEdit", vSettings.afExitDistance).toInt();
    vSettings.tipThinkbness = settingsNew.value("TipThinkness", vSettings.tipThinkbness).toInt();
    vSettings.mafPrcsEnabled = settingsNew.value("MafPrcsEnabled", vSettings.mafPrcsEnabled).toBool();
    vSettings.mafFastEnabled = settingsNew.value("MafFastEnabled", vSettings.mafFastEnabled).toBool();
    vSettings.pafEnabled = settingsNew.value("PafEnabled", vSettings.pafEnabled).toBool();
    vSettings.mafPaf = settingsNew.value("MafPaf", vSettings.mafPaf).toBool();

    auto groups = settingsNew.childGroups();
    for (int i = 0; i<groups.count(); i++)
    {
        auto *profile = new videoObjectiveProfile();
        QString name = groups[i];
        int unicodeIndex = name.indexOf(QChar(1111));
        if (unicodeIndex != -1)
            name = name.mid(unicodeIndex + 1);
        if (name.isEmpty())
            name = profile->name;
        profile->name = name;

        settingsNew.beginGroup(groups[i]);
        {
            profile->rangeX = settingsNew.value("RangeX", profile->rangeX).toInt();
            profile->rangeY = settingsNew.value("RangeY", profile->rangeY).toInt();
            profile->height = settingsNew.value("Height", profile->height).toInt();
            profile->width = settingsNew.value("Width", profile->width).toInt();
            profile->stageHeadCantX = settingsNew.value("StageHeadCantX", profile->stageHeadCantX).toDouble();
            profile->stageHeadCantY = settingsNew.value("StageHeadCantY", profile->stageHeadCantY).toDouble();
            profile->stageLaserX = settingsNew.value("StageLaserX", profile->stageLaserX).toDouble();
            profile->stageLaserY = settingsNew.value("StageLaserY", profile->stageLaserY).toDouble();
            settingsNew.beginGroup("AutoFocus");
            {
                profile->afD0 = settingsNew.value("D0", profile->afD0).toInt();
                profile->afVst = settingsNew.value("Vst", profile->afVst).toInt();
                profile->afVpr = settingsNew.value("Vpr", profile->afVpr).toInt();
                profile->afCrat = settingsNew.value("Crat", profile->afCrat).toInt();
                profile->afLeftStPr = settingsNew.value("LeftStPr", profile->afLeftStPr).toInt();
                profile->afCrat2 = settingsNew.value("Crat2", profile->afCrat2).toDouble();
                profile->afCratPrecize = settingsNew.value("CratPresize", profile->afCratPrecize).toInt();
            }
            settingsNew.endGroup();
        }
        settingsNew.endGroup();

        vSettings.objProfiles.append(profile);
    }

    if (vSettings.objProfiles.empty())
        vSettings.objProfiles.append(new videoObjectiveProfile());
    }
    //---------------------------------------

    int idx = vSettings.lastSelectedObjProfile;
    idx = qBound(0, idx, vSettings.objProfiles.size()-1);
    int xymidx = vSettings.xyMsel;
    xymidx = qBound(0, xymidx, xyMsel->count()-1);
    prev_idx = idx;
    QPlot->setTitle(tr("Configuration - ") + vSettings.objProfiles.at(idx)->name);
    xyMsel->setCurrentIndex(xymidx);

    flipx->setChecked(vSettings.flipX);
    flipy->setChecked(vSettings.flipY);
    rotxy->setChecked(vSettings.swapXY);
    mafPaf->setChecked(vSettings.mafPaf);
    //autoFocusMode->setCurrentIndex(_autoFocusMode);
    motorZFocusSpeedEdit->setValue(vSettings.motorZFocusSpeed);
    motorXYStageSpeedEdit->setValue(vSettings.motorXYStageSpeed);
    autoFocusExitDistanceEdit->setValue(vSettings.afExitDistance);
    tipThinkbnessEdit->setValue(vSettings.tipThinkbness);
    mafPrcsEnabled->setChecked(vSettings.mafPrcsEnabled);
    mafFastEnabled->setChecked(vSettings.mafFastEnabled);
    pafEnabled->setChecked(vSettings.pafEnabled);

    auto profile = vSettings.objProfiles.at(idx);
    QRectF rect = head_cant->geometry();
    rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                            QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
    head_cant->setGeometry(rect.toRect());

    rect = stage_cant->geometry();
    rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageHeadCantX),
                            QPlot->transform(QwtPlot::yLeft, profile->stageHeadCantY)));
    stage_cant->setGeometry(rect.toRect());

    rect = stage_laser->geometry();
    rect.moveCenter(QPointF(QPlot->transform(QwtPlot::xBottom, profile->stageLaserX),
                            QPlot->transform(QwtPlot::yLeft, profile->stageLaserY)));
    stage_laser->setGeometry(rect.toRect());
}
void VideoPlayer::setVisiblePlotButtons(bool visible)
{
    BLineMeasure->setVisible(visible);
    BAxes->setVisible(visible);
    BGrid->setVisible(visible);
}
void VideoPlayer::videoPlotEvent(QObject*, QEvent *event)
{
    auto mouse = static_cast<QMouseEvent *>(event);
    if (mouse != nullptr)
    {
        if (mouse->type() == QEvent::Enter || mouse->type() == QEvent::MouseMove)
        {
            setVisiblePlotButtons(true);
        }
    }
    if (event->type() == QEvent::Leave)
    {
		auto rect = QPlot->rect();
        int boundH = rect.height()*0.09;
        int boundW = rect.width()*0.09;
        QRect smallerRect(rect.left()+boundW*2, rect.top()+boundH,
			rect.width()-boundW*2, rect.height()-boundH*3);
        auto pos = QPlot->mapFromGlobal(QCursor::pos());
        setVisiblePlotButtons(smallerRect.contains(pos));
    }
    if (event->type() == QEvent::Resize)
    {
        int leftX = QPlot->plotLayout()->scaleRect(QwtPlot::yLeft).width()*1.15;
        int upY = QPlot->height()*0.07;
        int videoBtnsSize = (QPlot->rect().width() - leftX)*0.2;
        int UpButtonSize = videoBtnsSize*0.8;
        if (UpButtonSize < 0) UpButtonSize = 0;
        if (UpButtonSize > 9999) UpButtonSize = 9999;
        BLineMeasure->setFixedSize(UpButtonSize, UpButtonSize);
        BAxes->setFixedSize(UpButtonSize, UpButtonSize);
        BGrid->setFixedSize(UpButtonSize, UpButtonSize);

        BLineMeasure->setIconSize(QSize(UpButtonSize, UpButtonSize));
        BAxes->setIconSize(QSize(UpButtonSize, UpButtonSize));
        BGrid->setIconSize(QSize(UpButtonSize, UpButtonSize));

        double mulBtn = 0.05;
        BLineMeasure->move(leftX, upY);
        BAxes->move(leftX + videoBtnsSize + videoBtnsSize*mulBtn, upY);
        BGrid->move(leftX + videoBtnsSize*2 + videoBtnsSize*mulBtn*2, upY);
    }
}
void VideoPlayer::addNewProfileInTable(int index)
{
    QRadioButton *radBtn = new QRadioButton();
    QLineEdit *name = new QLineEdit(tr("Default"));
    Edit *rangeX = new Edit(1, 5000, 1);
    Edit *rangeY = new Edit(1, 5000, 1);
    Edit *height = new Edit(10, 100);
    Edit *width = new Edit(10, 100);
    Edit *D0 = new Edit(0, 99999);
    Edit *Vst = new Edit(1, 1000);
    Edit *Vpr = new Edit(1, 1000);
    Edit *Crat = new Edit(1, 1000000);
    Edit *LeftStPr = new Edit(1, 9999);
    Edit *Crat2 = new Edit(0, 1000000, 3);
    Edit *CratPresize = new Edit(1, 1000000);

    auto profile = vSettings.objProfiles.at(index);
    name->setText(profile->name);
    name->setStyleSheet("QLineEdit {border: 0px solid white;} ");
    rangeX->setValue(profile->rangeX);
    rangeY->setValue(profile->rangeY);
    height->setValue(profile->height);
    width->setValue(profile->width);
    D0->setValue(profile->afD0);
    Vst->setValue(profile->afVst);
    Vpr->setValue(profile->afVpr);
    Crat->setValue(profile->afCrat);
    LeftStPr->setValue(profile->afLeftStPr);
    Crat2->setValue(profile->afCrat2);
    CratPresize->setValue(profile->afCratPrecize);

    if (prev_idx == index)
    {
        radBtn->setChecked(true);
        cw = width->value() / 100;
        ch = height->value() / 100;
        mkm_px_x = (qreal)QPlot->canvas()->width() / rangeX->value();
        mkm_px_y = (qreal)QPlot->canvas()->height() / rangeY->value();

        QPlot->setAxisScale(QwtPlot::xBottom, 0, rangeX->value());
        QPlot->setAxisScale(QwtPlot::yLeft, 0, rangeY->value());
    }

    TableVideoProfiles->setCellWidget(index, VProfile::Selector, radBtn);
    TableVideoProfiles->setCellWidget(index, VProfile::Name, name);
    TableVideoProfiles->setCellWidget(index, VProfile::RangeX, rangeX);
    TableVideoProfiles->setCellWidget(index, VProfile::RangeY, rangeY);
    TableVideoProfiles->setCellWidget(index, VProfile::Height, height);
    TableVideoProfiles->setCellWidget(index, VProfile::Width, width);
    TableVideoProfiles->setCellWidget(index, VProfile::D0, D0);
    TableVideoProfiles->setCellWidget(index, VProfile::Vst, Vst);
    TableVideoProfiles->setCellWidget(index, VProfile::Vpr, Vpr);
    TableVideoProfiles->setCellWidget(index, VProfile::Crat, Crat);
    TableVideoProfiles->setCellWidget(index, VProfile::LeftStPr, LeftStPr);
    TableVideoProfiles->setCellWidget(index, VProfile::Crat2, Crat2);
    TableVideoProfiles->setCellWidget(index, VProfile::CratPresize, CratPresize);

    TableVideoProfiles->setColumnHidden(VProfile::Crat2, true);

    connect(radBtn, SIGNAL(clicked()), this, SLOT(SetCurrentVideoProfile()));
    connect(name, SIGNAL(textEdited(QString)), this, SLOT(sl_save()));
    connect(rangeX, SIGNAL(newValue()), this, SLOT(sl_range()));
    connect(rangeY, SIGNAL(newValue()), this, SLOT(sl_range()));
    connect(height, SIGNAL(newValue()), this, SLOT(on_cwh_ch()));
    connect(width, SIGNAL(newValue()), this, SLOT(on_cwh_ch()));

    connect(D0, SIGNAL(newValue()), this, SLOT(autoFocusSettingsChanged()));
    connect(Vst, SIGNAL(newValue()), this, SLOT(autoFocusSettingsChanged()));
    connect(Vpr, SIGNAL(newValue()), this, SLOT(autoFocusSettingsChanged()));
    connect(Crat, SIGNAL(newValue()), this, SLOT(autoFocusSettingsChanged()));
    connect(LeftStPr, SIGNAL(newValue()), this, SLOT(autoFocusSettingsChanged()));
    connect(Crat2, SIGNAL(newValue()), this, SLOT(autoFocusSettingsChanged()));
    connect(CratPresize, SIGNAL(newValue()), this, SLOT(sl_save()));
}

void VideoPlayer::autoFocusSettingsChanged()
{
    autoFocusDist1 = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::D0))->value();
    autoFocusV1 = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Vst))->value();
    autoFocusV2 = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Vpr))->value();
    if (autoFocusScanVariant == 0)
        autoFocusMagicMultiply[0] = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Crat))->value();
    autoFocusMagicMultiply[1] = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Crat2))->value();
    autoFocusVLeftStPr = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::LeftStPr))->value();

    sl_save();
}

VideoPlayer::~VideoPlayer()
{
    if (activeCamera)
        activeCamera->Stop();
    activeCamera = nullptr;

    if (inst_idx == 0) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        settings.setValue("video/grid", BGrid->isOn);
        settings.setValue("video/axes", BAxes->isOn);
    }

    foreach (IVideo *v, CameraList) {
        while (qobject_cast<const QThread *>(v->getObject()) && qobject_cast<const QThread *>(v->getObject())->isRunning()) {
            qApp->processEvents();
            SleepExport::msleep2(3);
        }
        delete v;
    }
}

void VideoPlayer::on_cwh_ch()
{
    cw = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Width))->value() / 100;
    ch = dynamic_cast<Edit*>(TableVideoProfiles->cellWidget(prev_idx, VProfile::Height))->value() / 100;

    sl_save();
}


void VideoPlayer::getXYMidx(int set, int *x, int *y)
{
    if (set == 0) {
        *x = Motors::getRoleMotor(Motors::RoleX1);
        *y = Motors::getRoleMotor(Motors::RoleY1);
    } else {
        *x = Motors::getRoleMotor(Motors::RoleStageX);
        *y = Motors::getRoleMotor(Motors::RoleStageY);
    }
}

void VideoPlayer::goXYjoy(QPointF speedxy)
{
    if (xyMsel->currentIndex() <= 0)
        return;

    auto motors = Motors::GetIt();
    if (!motors)
        return;

    QLineF line(0, 0, speedxy.x(), speedxy.y());
    double len = line.length();
    double angle = line.angle();
    double da = angle - prev_xy_joy_output.y();
    while (da >= 180)
        da -= 180;
    while (da <= -180)
        da += 180;

    if (speedxy.isNull() || fabs(da) >= 10 ||
            fabs(len - prev_xy_joy_output.x()) >= 0.2) {
        prev_xy_joy_output.rx() = qRound(len * 5) / 5.;
        prev_xy_joy_output.ry() = qRound(angle / 10.) * 10;
        int idxx, idxy;
        getXYMidx(xyMsel->currentIndex() - 1, &idxx, &idxy);
        double speedx = len > 0.1 ? qFabs(speedxy.x()) / len * motors->getShuttleSpeed(len, idxx) : 0;
        double speedy = len > 0.1 ? qFabs(speedxy.y()) / len * motors->getShuttleSpeed(len, idxy) : 0;
        motors->go(idxx, speedxy.x() == 0 ? 0 : (speedxy.x() > 0 ? 1 : -1), speedx, 0);
        motors->go(idxy, speedxy.y() == 0 ? 0 : (speedxy.y() > 0 ? 1 : -1), speedy, 0);
    }
    QPlot->canvas()->setCursor(QCursor(!speedxy.isNull() ? Qt::SizeAllCursor : Qt::OpenHandCursor));
}
void VideoPlayer::goZjoy(int speed)
{
    auto motors = Motors::GetIt();
    if (!motors)
        return;

    int idx = Motors::getRoleMotor(Motors::RoleZ2Focus);

    motors->goShuttle(speed, idx);
}


void VideoPlayer::setDeviceStatus(int idx)
{
    playing = false;
    switch (idx) {
    case EDno:
        deviceStatus->status->set(StatusWidget::Cross, Qt::red);
        deviceStatus->status->setToolTip(tr("No device"));
        break;
    case EDconn:
        deviceStatus->status->set(StatusWidget::Circle, QColor(255, 128, 0));
        deviceStatus->status->setToolTip(tr("Connecting"));
        break;
    case EDok:
        deviceStatus->status->set(StatusWidget::Circle, Qt::green);
        deviceStatus->status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        playing = true;
        deviceStatus->status->set(StatusWidget::Play, Qt::green);
        deviceStatus->status->setToolTip(tr("Updating"));
        break;
    }
}

void VideoPlayer::resetCameras()
{
    setDeviceStatus(EDconn);

    if (activeCamera)
        activeCamera->closeCamera();
    activeCamera = nullptr;

    m_BStart->setEnabled(false);
    BAutoFocus->setEnabled(false);
    BAutoFocusOsc->setEnabled(false);

    QFuture<void> f = QtConcurrent::run(this, &VideoPlayer::resetCamerasImpl);
    ManagedThreadedCleaner::registerManagedFuture(f);
    resetWatch.setFuture(f);
}
static QMutex videoreset_threadlock;
void VideoPlayer::resetCamerasImpl()
{
    QMutexLocker lck(&videoreset_threadlock); //don't try simultaneous video driver init

    foreach (IVideo *v, CameraList)
        if (v->openCamera(1)) {
            activeCamera = v;
            break;
        }
}
void VideoPlayer::resetCamerasCommit()
{
    ManagedThreadedCleaner::unregisterManagedFuture(resetWatch.future());

    bool opened = activeCamera != nullptr;
    if (opened) {
        deviceStatus->label->setToolTip(activeCamera->getName());

        m_BStart->setEnabled(true);
        BAutoFocus->setEnabled(true);
        BAutoFocusOsc->setEnabled(true);
    }

    setDeviceStatus(opened ? EDok : EDno);
    refresh_limit_timer.invalidate();
}

bool VideoPlayer::StartStop(bool value)
{
    if (!activeCamera)
    {
        if (!value)
            return false;
        resetCameras();

        if (!activeCamera)
            return false;
    }

    if (value) {
        activeCamera->Start();

        frames = 0;
    } else {
        activeCamera->Stop();
    }
    setDeviceStatus(value ? EDplay : EDok);

    return true;
}

void VideoPlayer::on_BToData_clicked()
{
    auto data = DataPlot::GetIt();
    if (!data)
        return;

    plot::TMetaStorage attributes;

    attributes.insert("Name", "Video");
    attributes.insert("Source", "Video");

    attributes.insert("XMin", 0);
    attributes.insert("XMax", 1e-6 * sl_get_rangeX());
    attributes.insert("YMin", 0);
    attributes.insert("YMax", 1e-6 * sl_get_rangeY());

    attributes.insert("XYLabel", QString());
    attributes.insert("XYUnit", tr("m"));

    data->Container->AddDataToContainer(ritem->image, attributes);
}
void VideoPlayer::sl_saveToFile()
{
    QString SelectedFilter;
    QString FileName = QFileDialog::getSaveFileName(this, tr("Save to file"), QString(), "*.png;;*.jpg;;*.tiff;;*.bmp", &SelectedFilter, QFileDialog::DontUseNativeDialog);

    if (FileName.isEmpty()) return;
    if (!FileName.endsWith(SelectedFilter.remove(0, 1))) FileName += SelectedFilter;

    QPixmap::grabWidget(QPlot).save(FileName, SelectedFilter.remove(0, 1).toAscii().data());
}

VideoPlayer* VideoPlayer::GetIt() noexcept
{
    auto video = ModuleManager::instance->find1VoidPtr("video");
    return static_cast<VideoPlayer*>(video);
}
