#include "autocalib.hpp"
#include "oscilloscope.h"
#include "scanplot.h"
#include "gui/status.hpp"
#include "modules.hpp"
#include <QApplication>
#include <QSettings>
#include <QCheckBox>
#include <QMessageBox>
#include <QDebug>
#include "options_window.h"


AutoCalib::AutoCalib()
{
    isInterrupted = false;
    use_direct_get_data = false;
    sensorNoZ = false;
    for (int i = 0; i < 3; i++)
        is_optical_scanner[i] = false;
}

bool skip_pid_error = false;
void AutoCalib::s_pidError(int idx)
{
    return;//TODO removing that message

    static bool is_shown = false;
    static bool skiperr_2 = false;
    if (is_shown || skip_pid_error || skiperr_2)
        return;
    is_shown = true;

    int b1 = QMessageBox::warning(nullptr, tr("Controller"),
                                  tr("Setpoint x/y/z PID error.") + "\n" +
                                  tr("Can't reach setpoint. Recalibrate sensors.") + "\n" +
                                  tr("Do you want to recalibrate sensors now?"),
                                  QMessageBox::Yes, QMessageBox::No, QMessageBox::NoToAll);

    skiperr_2 = (QMessageBox::NoToAll == b1);

    bool ok = (QMessageBox::Yes == b1);
    if (!ok) {
        is_shown = false;
        return;
    }

    if (true/*show_additional_gui() || show_advanced_gui()*/)
    {
        QString scanner = (idx == 0) ? tr("Head") : tr("Stage");
        bool ok1 = (QMessageBox::Yes ==
                    QMessageBox::warning(nullptr, tr("Controller"),
                                         tr("Do you want to skip Z axis calibration?") + "\n" +
                                         scanner, QMessageBox::Yes, QMessageBox::No));

        if (ok1)
            sensorNoZ = true;
    } else
        sensorNoZ = true;

    SensorReCalc(idx);

    is_shown = false;
}
void AutoCalib::doXYcalib(int idx)
{
    if (skip_pid_error)
        return;
    skip_pid_error = true;

    sensorNoZ = true;
    bool ok = SensorReCalc(idx);
    emit calibDone(ok);

    skip_pid_error = false;
}


void AutoCalib::cancelOperation()
{
    destructStatus(barRecalc);
    isInterrupted = true;
}

void AutoCalib::capBits(int n, long bits)
{
    if (n == 0) {
        use_direct_get_data = bool(bits & 1);
        capbits_updated++;
    }
    if (n >= 2 && n <= 4) {
        is_optical_scanner[n-2] = bool(bits & 1);
        capbits_updated++;
    }
}
void AutoCalib::reset()
{
    capbits_updated = 0;
}

void AutoCalib::SetOscChannel(int ch)
{
    want_osc_ch = ch;
    osc->CBChannel[0]->setCurrentIndex(osc->CBChannel[0]->findData(ch));
}

bool AutoCalib::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD )
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_GET_DATA) {
        EG_GetData *Get_data = (EG_GetData *)data;
        if (Get_data->Source == want_osc_ch) {
            osc_data[0] = Get_data->Data;
            new_osc_data = true;
        }
        return false;
    }

    return false;
}

void AutoCalib::recvOscData(QByteArray data)
{
    const EG_OscData *sdata = (const EG_OscData *)data.constData();

    if (sdata->Channel >= 4 || sdata->Source != want_osc_ch)
        return;

    memcpy(osc_data, sdata->Data, sizeof(WORD)*OSC_LINE_SIZE);

    new_osc_data = true;
}

double AutoCalib::ReadOscLine(WORD *data_line, int timeout)
{
    double ret = 0;
    QElapsedTimer timer;
    timer.start();

    new_osc_data = false;
    if (!use_direct_get_data || data_line) {
        connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(recvOscData(QByteArray)));
        osc->start->change_state_man_with_sig(true);
    } else
        osc->get_curr_data(want_osc_ch);

    while (!new_osc_data && !timer.hasExpired(timeout)) {
        SleepExport::msleep2(2);
        qApp->processEvents();
        if (isInterrupted) {
            if (!use_direct_get_data || data_line) {
                osc->start->change_state_man_with_sig(false);
                disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(recvOscData(QByteArray)));
            }
            return 0;
        }
    }

    if (new_osc_data) {
        ret = osc_data[0];
        if (data_line != nullptr)
            memcpy(data_line, osc_data, sizeof(WORD)*OSC_LINE_SIZE);
    }

    if (!use_direct_get_data || data_line) {
        disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(recvOscData(QByteArray)));
        osc->start->change_state_man_with_sig(false);
    }

    return ret;
}
double AutoCalib::ReadOscFirst(int timeout)
{
    return ReadOscLine(nullptr, timeout);
}
double AutoCalib::ReadOscAverage(int timeout)
{
    if (use_direct_get_data)
        return ReadOscLine(nullptr, timeout);

    WORD temp[OSC_LINE_SIZE];

    ReadOscLine(temp, timeout);

    return LineCalcAverage(temp);
}

double AutoCalib::LineCalcAverage(WORD *data_line)
{
    long int avg = 0;
    for (int i = 0; i < OSC_LINE_SIZE; i++)
        avg += data_line[i];

    return avg / OSC_LINE_SIZE; //fixme losing fractional part
}
double AutoCalib::LineCalcRms(WORD *data_line)
{
    double avg = LineCalcAverage(data_line);

    double rms = 0;
    for (int i = 0; i < OSC_LINE_SIZE; i++)
        rms += pow((double)data_line[i] - avg, 2);

    return sqrt(rms / OSC_LINE_SIZE);
}
long AutoCalib::LineCalcMaximums(WORD *data_line)
{
    long int maximums = 0;
    for (int i = 1; i < OSC_LINE_SIZE-2; i++) {
        if (data_line[i-1] < data_line[i] &&
                data_line[i] > data_line[i+1]) maximums++;
        else if (data_line[i-1] < data_line[i] &&
                 data_line[i] == data_line[i+1] && data_line[i+1] > data_line[i+2]) maximums++;
    }

    if (maximums == 0) maximums++;
    qDebug() << "maximums" << maximums;
    return maximums;
}


void AutoCalib::go_osc_xyz(double xpos, double ypos, double zpos)
{
    if (sensorNoZ)
        zpos = 0;

    if (scanner == scHead) {
        osc->DacsWidget->X->setValue(xpos);
        osc->DacsWidget->Y->setValue(ypos);
        osc->DacsWidget->Z->setValue(zpos);
        osc->DacsWidget->X->setNewValue();
        osc->DacsWidget->Y->setNewValue();
        osc->DacsWidget->Z->setNewValue();
    } else {
        osc->DacsWidget->X2->setValue(xpos);
        osc->DacsWidget->Y2->setValue(ypos);
        osc->DacsWidget->Z2->setValue(zpos);
        osc->DacsWidget->X2->setNewValue();
        osc->DacsWidget->Y2->setNewValue();
        osc->DacsWidget->Z2->setNewValue();
        auto scan = ScanPlot::GetIt();
        if (scan)
        {
            scan->deviceSchemeWid->ZPiezoPosEdit[0]->setValue(xpos);
            scan->deviceSchemeWid->ZPiezoPosEdit[1]->setValue(ypos);
            scan->deviceSchemeWid->ZPiezoPosEdit[2]->setValue(zpos);
            scan->deviceSchemeWid->ZPiezoPosEdit[0]->setNewValue();
            scan->deviceSchemeWid->ZPiezoPosEdit[1]->setNewValue();
            scan->deviceSchemeWid->ZPiezoPosEdit[2]->setNewValue();
        }
    }
}

double AutoCalib::read_direct(int ch)
{
    if (!use_direct_get_data)
        return 0;

    want_osc_ch = ch;
    return ReadOscAverage();
}

void AutoCalib::read_osc(double *v, int ch)
{
    want_osc_ch = ch;
    if (!use_direct_get_data)
        SetOscChannel(ch);

    *v = ReadOscAverage();
}
void AutoCalib::read_osc_xyz(double *x, double *y, double *z)
{
    if (x) read_osc(x, scanner == scHead ? srcCX : srcCX2);
    if (y) read_osc(y, scanner == scHead ? srcCY : srcCY2);
    if (z) read_osc(z, scanner == scHead ? srcCZ : srcCZ2);
}
void AutoCalib::read_osc_xyz_v(double *x, double *y, double *z)
{
    if (x) read_osc(x, scanner == scHead ? srcDACX : srcDACX2);
    if (y) read_osc(y, scanner == scHead ? srcDACY : srcDACY2);
    if (z) read_osc(z, scanner == scHead ? srcDACZ : srcDACZ2);
}

bool AutoCalib::read_osc_xyz_no_outer_drift(double *x, double *y, double *z, int x_sign, int y_sign, int z_sign)
{
    double x1, y1, z1;
    read_osc_xyz(x ? &x1 : nullptr, y ? &y1 : nullptr, z ? &z1 : nullptr);

    bool updated = false;

    if (x && x_sign * (x1 - *x) > 0) {
        *x = x1;
        updated = true;
    }

    if (y && y_sign * (y1 - *y) > 0) {
        *y = y1;
        updated = true;
    }

    if (z && z_sign * (z1 - *z) > 0) {
        *z = z1;
        updated = true;
    }

    return updated;
}
bool AutoCalib::read_osc_xyz_v_no_outer_drift(double *x, double *y, double *z, int x_sign, int y_sign, int z_sign)
{
    double x1, y1, z1;
    read_osc_xyz_v(x ? &x1 : nullptr, y ? &y1 : nullptr, z ? &z1 : nullptr);

    bool updated = false;

    if (x && x_sign * (x1 - *x) > 0) {
        *x = x1;
        updated = true;
    }

    if (y && y_sign * (y1 - *y) > 0) {
        *y = y1;
        updated = true;
    }

    if (z && z_sign * (z1 - *z) > 0) {
        *z = z1;
        updated = true;
    }

    return updated;
}

void AutoCalib::read_osc_xyz(double xpos, double ypos, double zpos,
                                 double *x, double *y, double *z)
{
    go_osc_xyz(xpos, ypos, zpos);

    if (!x && !y && !z)
        return;

    double x_range = scanner == scHead ? osc->CalibWidget->Scanner_data_pid_cal_range[45] : osc->CalibWidget->Scanner2_data_pid_cal_range[45];//mkm
    double y_range = scanner == scHead ? osc->CalibWidget->Scanner_data_pid_cal_range[46] : osc->CalibWidget->Scanner2_data_pid_cal_range[46];//mkm
    double z_range = scanner == scHead ? osc->CalibWidget->Scanner_data_pid_cal_range[47] : osc->CalibWidget->Scanner2_data_pid_cal_range[47];//mkm
    double dist = qMax(x_range, qMax(y_range, z_range));
    int time = qMin(30., 2 * dist / osc->CalibWidget->free_move_speed) * 1000;
    QElapsedTimer timer;

    timer.start();
    while (!timer.hasExpired(time)) {
        SleepExport::msleep2(10);
        qApp->processEvents();
        if (isInterrupted) return;
    }

    read_osc_xyz(x, y, z);
}

void AutoCalib::read_osc_xyz_no_outer_drift(double xpos, double ypos, double zpos,
                                                double *x, double *y, double *z,
                                                int x_sign, int y_sign, int z_sign)
{
    go_osc_xyz(xpos, ypos, zpos);

    if (!x && !y && !z)
        return;

    double x_range = scanner == scHead ? osc->CalibWidget->Scanner_data_pid_cal_range[45] : osc->CalibWidget->Scanner2_data_pid_cal_range[45];//mkm
    double y_range = scanner == scHead ? osc->CalibWidget->Scanner_data_pid_cal_range[46] : osc->CalibWidget->Scanner2_data_pid_cal_range[46];//mkm
    double z_range = scanner == scHead ? osc->CalibWidget->Scanner_data_pid_cal_range[47] : osc->CalibWidget->Scanner2_data_pid_cal_range[47];//mkm
    double dist = qMax(x_range, qMax(y_range, z_range));
    int time = qMin(30., 2 * dist / osc->CalibWidget->free_move_speed) * 1000;
    QElapsedTimer timer;

    QElapsedTimer timer2;
    double x1 = x_sign > 0 ? 66000 : -1;
    double y1 = y_sign > 0 ? 66000 : -1;
    double z1 = z_sign > 0 ? 66000 : -1;

    timer.start();
    timer2.start();
    while (!timer.hasExpired(time)) {
        SleepExport::msleep2(2);
        qApp->processEvents();
        if (isInterrupted) return;
        if (timer2.hasExpired(use_direct_get_data ? 10 : 100)) {
            timer2.start();
            if (!read_osc_xyz_v_no_outer_drift(&x1, &y1, &z1, x_sign*-1, y_sign*-1, z_sign*-1))
                break;
        }
    }
    read_osc_xyz_no_outer_drift(x, y, z, x_sign, y_sign, z_sign);
    timer2.start();
    while (!timer.hasExpired(time)) {
        SleepExport::msleep2(2);
        qApp->processEvents();
        if (isInterrupted) return;
        if (timer2.hasExpired(use_direct_get_data ? 10 : 100)) {
            timer2.start();
            if (!read_osc_xyz_no_outer_drift(x, y, z, x_sign, y_sign, z_sign))
                return;
        }
    }
}

//----------------------------------------------------------

bool AutoCalib::restoreSensorReCalc(double x_range, double y_range, double z_range,
        bool old_fb, bool old_fbz, bool old_nl,
        DWORD Scanner_data_scale[3], long Scanner_data_Shift[3],
        int old_ch[4], QWidgetList cwl)
{
    sensorNoZ = false;

    osc->CalibWidget->setSensorOverflowMode(scanner, 0);

    if (scanner == scHead) {
        for (int i = 0; i < 3; i++) {
            osc->CalibWidget->lineSensorScale[i]->setValue(Scanner_data_scale[i]);
            osc->CalibWidget->lineSensorScale0[i]->setValue(Scanner_data_Shift[i]);
        }
    } else {
        for (int i = 0; i < 3; i++) {
            osc->CalibWidget->lineSensorScale2[i]->setValue(Scanner_data_scale[i]);
            osc->CalibWidget->lineSensorScale02[i]->setValue(Scanner_data_Shift[i]);
        }
    }
    if (old_fb != osc->CalibWidget->fb_on_off_xy->isOn)
    {
        osc->CalibWidget->fb_on_off_xy->change_state_man_with_sig(old_fb);
    }
    if (old_fbz != osc->CalibWidget->fb_on_off_z->isOn)
        osc->CalibWidget->fb_on_off_z->change_state_man_with_sig(old_fbz);
    if (old_nl != osc->CalibWidget->nl_on_off_ol->isChecked())
        osc->CalibWidget->nl_on_off_ol->click();
    osc->CalibWidget->s_sendData();


    emit calibInProgress(false);

    go_osc_xyz(x_range/2, y_range/2, z_range/2);


    if (!use_direct_get_data)
        for (int i = 0; i < 4; i++)
            osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    foreach (QWidget *wid, cwl)
        wid->setEnabled(true);

    return false;
}

bool AutoCalib::fixScaleAtBounds(double x, double y, double z, double *scale)
{
    bool restart = false;
    if (x < 1 || x > 65534) {
        scale[0] = qFloor(scale[0] / 2);
        restart = true;
    }
    if (y < 1 || y > 65534) {
        scale[1] = qFloor(scale[1] / 2);
        restart = true;
    }
    if (z < 1 || z > 65534) {
        scale[2] = qFloor(scale[2] / 2);
        restart = true;
    }
    return restart;
}

//V2008 Cyclomatic complexity:   101
bool AutoCalib::SensorReCalc(BYTE _scanner, bool silent)
{
    scanner = _scanner;
    if (!ModuleManager::instance->isRunning)
    {
        emit(calibCanceled());
        return false;
    }

    { //wait for capbits
        QElapsedTimer t; t.start();
        while (capbits_updated < 1 && t.elapsed() < 50) {
            qApp->processEvents();
            SleepExport::msleep2(2);
        }
        t.restart();
        while (capbits_updated < 3 && t.elapsed() < 50) {
            qApp->processEvents();
            SleepExport::msleep2(2);
        }
    }

    QWidgetList cwl;
    if (!use_direct_get_data) cwl << osc;
    else { cwl << osc->CalibWidget << osc->DacsWidget << osc->acp; }
    foreach (QWidget *wid, cwl)
        wid->setEnabled(false);

    QString name = osc->CalibWidget->Scanner_type->tabText(scanner);
    QString status = name + " " + tr("Sensors autocalibration");

    isInterrupted = false;

    barRecalc = new StatusBar(status, 6);
    connect(barRecalc, SIGNAL(canceled()), SLOT(cancelOperation()));
    barRecalc->add();

    //backup values for cancel
    DWORD Scanner_data_scale[3];
    long Scanner_data_Shift[3];
    if (scanner == scHead) {
        memcpy(Scanner_data_scale, osc->CalibWidget->Scanner_data_scale, sizeof(DWORD)*3);
        memcpy(Scanner_data_Shift, osc->CalibWidget->Scanner_data_Shift, sizeof(long)*3);
    } else {
        memcpy(Scanner_data_scale, osc->CalibWidget->Scanner2_data_scale, sizeof(DWORD)*3);
        memcpy(Scanner_data_Shift, osc->CalibWidget->Scanner2_data_Shift, sizeof(long)*3);
    }

    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    double default_scale = qFloor(settings.value("sensors_autocalib_gain", 100).toDouble());
    double default_reserve = settings.value("sensors_autocalib_reserve", 0.025).toDouble();
    double default_middle_x = 0.5;
    double default_middle_y = 0.5;
    double default_middle_z = 0.5;
    double default_scale_[3];
    for (double &defScale : default_scale_)
        defScale = default_scale;


    bool is_optical = is_optical_scanner[scanner];
    if (scanner == scHead) {
        default_middle_x = settings.value("sensors_autocalib_middle_x_h", is_optical ? 0.5 : 0.55).toDouble();
        default_middle_y = settings.value("sensors_autocalib_middle_y_h", is_optical ? 0.5 : 0.48).toDouble();
        default_middle_z = settings.value("sensors_autocalib_middle_z_h", is_optical ? 0.5 : 0.55).toDouble();

        for (auto &lineSensor : osc->CalibWidget->lineSensorScale0)
            lineSensor->setValue(0);
    } else {
        default_middle_x = settings.value("sensors_autocalib_middle_x_s", is_optical ? 0.5 : 0.55).toDouble();
        default_middle_y = settings.value("sensors_autocalib_middle_y_s", is_optical ? 0.5 : 0.55).toDouble();
        default_middle_z = settings.value("sensors_autocalib_middle_z_s", is_optical ? 0.5 : 0.55).toDouble();

        for (auto &lineSensor : osc->CalibWidget->lineSensorScale02)
            lineSensor->setValue(0);
    }

    osc->CalibWidget->Scanner_type->setCurrentIndex(scanner);
    bool old_fb = osc->CalibWidget->fb_on_off_xy->isOn;
    bool old_fbz = osc->CalibWidget->fb_on_off_z->isOn;
    bool old_nl = osc->CalibWidget->nl_on_off_ol->isChecked();
    if (old_fb) osc->CalibWidget->fb_on_off_xy->change_state_man_with_sig(false);
    if (old_fbz) osc->CalibWidget->fb_on_off_z->change_state_man_with_sig(false);
    if (old_nl) osc->CalibWidget->nl_on_off_ol->click();

    int old_ch[4];
    if (!use_direct_get_data)
    for (int i = 0; i < 4; i++) {
        old_ch[i] = osc->CBChannel[i]->currentIndex();
        osc->CBChannel[i]->setCurrentIndex(srcNone);
    }

    double x_range = osc->CalibWidget->Scanner_data_pid_cal_range[45];//mkm
    double y_range = osc->CalibWidget->Scanner_data_pid_cal_range[46];//mkm
    double z_range = osc->CalibWidget->Scanner_data_pid_cal_range[47];//mkm
    if (scanner == scStage) {
        x_range = osc->CalibWidget->Scanner2_data_pid_cal_range[45];//mkm
        y_range = osc->CalibWidget->Scanner2_data_pid_cal_range[46];//mkm
        z_range = osc->CalibWidget->Scanner2_data_pid_cal_range[47];//mkm
    }

    osc->CalibWidget->setSensorOverflowMode(scanner, 1);

    emit calibInProgress(true);

label_restart_calib:
    if (default_scale_[0] < 1 || default_scale_[1] < 1 || default_scale_[2] < 1) {
        auto opt = options::GetIt();
        if (!silent && !opt->optWin->chHideDidntFindedSensorsErrorMessages->isChecked()) {
            new Notification(tr("Sensor auto calibration") + ": " + name + ": " + tr("No data, or no sensors!"), Notification::Bad, 30);
            doBeep();
        }
        destructStatus(barRecalc);
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }

    barRecalc->setValue(0);
    qApp->processEvents();

    if (scanner == scHead) {
        for (int i = 0; i < 3; i++)
            osc->CalibWidget->lineSensorScale[i]->setValue(default_scale_[i]);
    } else {
        for (int i = 0; i < 3; i++)
            osc->CalibWidget->lineSensorScale2[i]->setValue(default_scale_[i]);
    }
    osc->CalibWidget->s_sendData();

    double X[2][2] = {{-1, -1}, {66000, 66000}};
    double Y[2][2] = {{-1, -1}, {66000, 66000}};
    double Z[2] = {-1, 66000};
    double nullx = -1, nully = -1, nullz = -1;

    read_osc_xyz_no_outer_drift(0, 0, 0, &nullx, &nully, &nullz, 1, 1, 1);
    if (isInterrupted)
    {
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }
    if (fixScaleAtBounds(nullx, nully, nullz, default_scale_))
        goto label_restart_calib;

    barRecalc->setValue(1);
    qApp->processEvents();

    read_osc_xyz_no_outer_drift(x_range, y_range, z_range, &X[1][1], &Y[1][1], &Z[1], -1, -1, -1);
    if (isInterrupted)
    {
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }
    if (fixScaleAtBounds(X[1][1], Y[1][1], Z[1], default_scale_))
        goto label_restart_calib;

    barRecalc->setValue(2);
    qApp->processEvents();

    read_osc_xyz_no_outer_drift(0, 0, z_range, &X[0][0], &Y[0][0], &Z[1], 1, 1, -1);
    if (isInterrupted)
    {
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }
    if (fixScaleAtBounds(X[0][0], Y[0][0], Z[1], default_scale_))
        goto label_restart_calib;

    barRecalc->setValue(3);
    qApp->processEvents();

    nullx = -1, nully = 66000, nullz = -1;
    read_osc_xyz_no_outer_drift(0, y_range, 0, &nullx, &nully, &nullz, 1, -1, 1);
    if (isInterrupted)
    {
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }
    if (fixScaleAtBounds(nullx, nully, nullz, default_scale_))
        goto label_restart_calib;

    barRecalc->setValue(4);
    qApp->processEvents();

    read_osc_xyz_no_outer_drift(x_range, 0, 0, &X[1][0], &Y[0][1], &Z[0], -1, 1, 1);
    if (isInterrupted)
    {
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }
    if (fixScaleAtBounds(X[1][0], Y[0][1], Z[0], default_scale_))
        goto label_restart_calib;

    barRecalc->setValue(5);
    qApp->processEvents();

    read_osc_xyz_no_outer_drift(0, y_range, 0, &X[0][1], &Y[1][0], &Z[0], 1, -1, 1);
    if (isInterrupted)
    {
        emit(calibCanceled());
        return restoreSensorReCalc(x_range, y_range, z_range,
                                   old_fb, old_fbz, old_nl,
                                   Scanner_data_scale, Scanner_data_Shift,
                                   old_ch, cwl);
    }
    if (fixScaleAtBounds(X[0][1], Y[1][0], Z[0], default_scale_))
        goto label_restart_calib;

    barRecalc->setValue(6);
    qApp->processEvents();


    //rescale
    X[0][0] = (X[0][0] - 32768) / default_scale_[0] + 32768;
    X[1][0] = (X[1][0] - 32768) / default_scale_[0] + 32768;
    X[0][1] = (X[0][1] - 32768) / default_scale_[0] + 32768;
    X[1][1] = (X[1][1] - 32768) / default_scale_[0] + 32768;

    Y[0][0] = (Y[0][0] - 32768) / default_scale_[1] + 32768;
    Y[1][0] = (Y[1][0] - 32768) / default_scale_[1] + 32768;
    Y[0][1] = (Y[0][1] - 32768) / default_scale_[1] + 32768;
    Y[1][1] = (Y[1][1] - 32768) / default_scale_[1] + 32768;

    Z[0] = (Z[0] - 32768) / default_scale_[2] + 32768;
    Z[1] = (Z[1] - 32768) / default_scale_[2] + 32768;

    //calc
    double shift[3];
    double mult[3];

    double _x[2];
    _x[0] = X[0][0];
    _x[1] = X[1][0];
    if (X[0][1] > X[0][0]) _x[0] = X[0][1];
    if (X[1][1] < X[1][0]) _x[1] = X[1][1];
    if (_x[1] < _x[0]) qSwap(_x[0], _x[1]);

    double _y[2];
    _y[0] = Y[0][0];
    _y[1] = Y[0][1];
    if (Y[1][0] > Y[0][0]) _y[0] = Y[1][0];
    if (Y[1][1] < Y[0][1]) _y[1] = Y[1][1];
    if (_y[1] < _y[0]) qSwap(_y[0], _y[1]);
    //QVector<double> srt;
    //srt << Y[0][0] << Y[0][1] << Y[1][0] << Y[1][1];
    //qSort(srt);
    //_y[0] = srt.at(2); _y[1] = srt.at(1);

    shift[0] = (1.0-default_middle_x)*_x[0] + default_middle_x*_x[1] - 32768;
    shift[1] = (1.0-default_middle_y)*_y[0] + default_middle_y*_y[1] - 32768;
    shift[2] = (1.0-default_middle_z)*Z[0] + default_middle_z*Z[1] - 32768;

    //round shift
    shift[0] = qRound(shift[0]);
    shift[1] = qRound(shift[1]);
    shift[2] = qRound(shift[2]);

    //apply shift
    X[0][0] -= shift[0]; X[1][0] -= shift[0]; X[0][1] -= shift[0]; X[1][1] -= shift[0];
    Y[0][0] -= shift[1]; Y[1][0] -= shift[1]; Y[0][1] -= shift[1]; Y[1][1] -= shift[1];
    Z[0] -= shift[2]; Z[1] -= shift[2];

    //fix shift sign
    shift[0] *= -1;
    shift[1] *= -1;
    shift[2] *= -1;

    mult[0] = qMin(fabs(X[0][0] - 32768), qMin(fabs(X[0][1] - 32768),
              qMin(fabs(X[1][0] - 32768), fabs(X[1][1] - 32768))));
    mult[1] = qMin(fabs(Y[0][0] - 32768), qMin(fabs(Y[0][1] - 32768),
              qMin(fabs(Y[1][0] - 32768), fabs(Y[1][1] - 32768))));
    mult[2] = qMin(fabs(Z[0] - 32768), fabs(Z[1] - 32768));

    mult[0] = (1+default_reserve)*32768/mult[0];
    mult[1] = (1+default_reserve)*32768/mult[1];
    mult[2] = (1+default_reserve)*32768/mult[2];

    //round mult
    mult[0] = qCeil(mult[0]);
    mult[1] = qCeil(mult[1]);
    mult[2] = qCeil(mult[2]);

    //errors checks
    QString err;
    bool isOk = true;
    if (X[0][0] == X[0][1] && X[0][1] == X[1][0] && X[1][0] == X[1][1]) {
        shift[0] = 0;
        mult[0] = 1;
        err += "X";
        isOk = false;
    }
    if (Y[0][0] == Y[0][1] && Y[0][1] == Y[1][0] && Y[1][0] == Y[1][1]) {
        shift[1] = 0;
        mult[1] = 1;
        err += "Y";
        isOk = false;
    }
    if (Z[0] == Z[1] && !sensorNoZ) {
        shift[2] = 0;
        mult[2] = 1;
        err += "Z";
        isOk = false;
    }
    auto opt = options::GetIt();
    if (!err.isEmpty() && !silent && !opt->optWin->chHideDidntFindedSensorsErrorMessages->isChecked()) {
        err = name + tr(" %1 not found").arg(err);
        new Notification(tr("Sensor auto calibration") + ": " + err, Notification::Bad, 30);
        doBeep();
    }
    if (sensorNoZ) {
        shift[2] = Scanner_data_Shift[2];
        mult[2] = Scanner_data_scale[2];
        qDebug() << name + " Z skip";
    }


    //set settings
    if (is_optical_scanner[scanner]) {
        double range[3];
        double optical_correction = qMax(0.1, settings.value("sensors_autocalib_optical_range_correction", 1.1).toDouble());
        for (int i = 0; i < 3; i++) {
            //scale_period_um, interp_factor, adc_range, mult_prescale, mult
            range[i] = (20. / 16384.) * (65536. * (256. / mult[i]));
            range[i] /= optical_correction;
        }

        if (scanner == scHead)
            for (int i = 0; i < 3; i++)
                osc->CalibWidget->lineScannerRangeCL[i]->setValue(range[i]);
        else
            for (int i = 0; i < 3; i++)
                osc->CalibWidget->lineScannerRangeCL2[i]->setValue(range[i]);
    }
    if (scanner == scHead) {
        for (int i = 0; i < 3; i++) {
            osc->CalibWidget->lineSensorScale[i]->setValue(mult[i]);
            osc->CalibWidget->lineSensorScale0[i]->setValue(shift[i]);
        }
        if (!is_optical_scanner[scanner])
            osc->CalibWidget->lineSensorScale0[1]->setValue(-shift[1]); //y shift is in another dir in old cap sensors heads
    } else {
        for (int i = 0; i < 3; i++) {
            osc->CalibWidget->lineSensorScale2[i]->setValue(mult[i]);
            osc->CalibWidget->lineSensorScale02[i]->setValue(shift[i]);
        }
    }
    if (old_fb) osc->CalibWidget->fb_on_off_xy->change_state_man_with_sig(true);
    if (old_fbz) osc->CalibWidget->fb_on_off_z->change_state_man_with_sig(true);
    if (old_nl) osc->CalibWidget->nl_on_off_ol->click();
    osc->CalibWidget->s_sendData();


    sensorNoZ = false;

    osc->CalibWidget->setSensorOverflowMode(scanner, 0);

    emit calibInProgress(false);

    go_osc_xyz(x_range/2, y_range/2, z_range/2);


    if (!use_direct_get_data)
    for (int i = 0; i < 4; i++)
        osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    destructStatus(barRecalc);

    foreach (QWidget *wid, cwl)
        wid->setEnabled(true);

    if (!silent)
        doBeep();

    emit(sensorsRecalibDone(isOk));
    return isOk;
}

//----------------------------------------------------------

void AutoCalib::restorePICalc(
        double x_range, double y_range, double z_range,
        bool old_fb, bool old_fbz, bool old_nl,
        int old_ch[4], QWidgetList cwl)
{
    if (old_fb != osc->CalibWidget->fb_on_off_xy->isOn)
        osc->CalibWidget->fb_on_off_xy->change_state_man_with_sig(old_fb);
    if (old_fbz != osc->CalibWidget->fb_on_off_z->isOn)
        osc->CalibWidget->fb_on_off_z->change_state_man_with_sig(old_fbz);
    if (old_nl != osc->CalibWidget->nl_on_off_ol->isChecked())
        osc->CalibWidget->nl_on_off_ol->click();
    osc->CalibWidget->s_sendData();


    emit calibInProgress(false);

    go_osc_xyz(x_range/2, y_range/2, z_range/2);


    if (!use_direct_get_data)
        for (int i = 0; i < 4; i++)
            osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    foreach (QWidget *wid, cwl)
        wid->setEnabled(true);
}

//V2008 Cyclomatic complexity:   79
void AutoCalib::PICalcAuto(BYTE _scanner, bool silent)
{
    scanner = _scanner;
    if (!ModuleManager::instance->isRunning)
        return;

    QWidgetList cwl;
    if (!use_direct_get_data) cwl << osc;
    else { cwl << osc->CalibWidget << osc->DacsWidget << osc->acp; }
    foreach (QWidget *wid, cwl)
        wid->setEnabled(false);

    QString name = osc->CalibWidget->Scanner_type->tabText(scanner);
    QString status = name + " " + tr("PI Calibration");

    isInterrupted = false;

    int count = use_direct_get_data ? 1000 : 100;

    QPointer<StatusBar> bar = new StatusBar(status, count);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    osc->CalibWidget->Scanner_type->setCurrentIndex(scanner);
    bool old_fb = osc->CalibWidget->fb_on_off_xy->isOn;
    bool old_fbz = osc->CalibWidget->fb_on_off_z->isOn;
    bool old_nl = osc->CalibWidget->nl_on_off_ol->isChecked();
    if (old_fb) osc->CalibWidget->fb_on_off_xy->change_state_man_with_sig(false);
    if (old_fbz) osc->CalibWidget->fb_on_off_z->change_state_man_with_sig(false);
    if (old_nl) osc->CalibWidget->nl_on_off_ol->click();
    osc->CalibWidget->s_sendData();

    emit calibInProgress(true);

    int old_ch[4];
    if (!use_direct_get_data)
    for (int i = 0; i < 4; i++) {
        old_ch[i] = osc->CBChannel[i]->currentIndex();
        osc->CBChannel[i]->setCurrentIndex(srcNone);
    }

    double x_range = osc->CalibWidget->Scanner_data_pid_cal_range[45];//mkm
    double y_range = osc->CalibWidget->Scanner_data_pid_cal_range[46];//mkm
    double z_range = osc->CalibWidget->Scanner_data_pid_cal_range[47];//mkm
    if (scanner == scStage) {
        x_range = osc->CalibWidget->Scanner2_data_pid_cal_range[45];//mkm
        y_range = osc->CalibWidget->Scanner2_data_pid_cal_range[46];//mkm
        z_range = osc->CalibWidget->Scanner2_data_pid_cal_range[47];//mkm
    }

    double c[3] = {0, 0, 0};
    double v[3] = {0, 0, 0};
    int c0[3] = {0, 0, 0};
    double pval[3] = {1e3, 1e3, 1e3};
    int i;

    //scan v[0->max] until we find c>0, measure N times, switch v-direction, measure till c==0
    //same for v[max->0], c<65535

    //c-low, v-up
    for (i = 0; i < count/2; i++) {
        double c1[3] = {66000, 66000, 66000};
        double v1[3];
        read_osc_xyz_no_outer_drift(x_range * i / count, y_range * i / count, z_range * i / count, &c1[0], &c1[1], &c1[2], -1, -1, -1);
        read_osc_xyz_v(&v1[0], &v1[1], &v1[2]);
        if (isInterrupted)
            return restorePICalc(x_range, y_range, z_range,
                                 old_fb, old_fbz, old_nl,
                                 old_ch, cwl);

        bool fin = true;
        for (int j = 0; j < 3; j++) {
            if (c[j] > 0) {
                v[j] -= v1[j];
                c[j] -= c1[j];

                if (c[j] < -5 && v[j] < -5) {
                    double dvdc = v[j] / c[j];
                    //qDebug() << j << dvdc;
                    pval[j] = qMin(pval[j], dvdc);
                    c0[j]++;
                } else
                    i++;
            } else
                i++;

            v[j] = v1[j];
            c[j] = c1[j];
            if (c0[j] < 6)
                fin = false;
        }
        if (fin)
            break;

        bar->setValue(i);
        qApp->processEvents();
    }
    //qDebug() << "_1";
    //c-low, v-down
    for (int &c : c0)
        c = 0;
    for (; i >= 0; i--) {
        double c1[3] = {0, 0, 0};
        double v1[3];
        read_osc_xyz_no_outer_drift(x_range * i / count, y_range * i / count, z_range * i / count, &c1[0], &c1[1], &c1[2], 1, 1, 1);
        read_osc_xyz_v(&v1[0], &v1[1], &v1[2]);
        if (isInterrupted)
            return restorePICalc(x_range, y_range, z_range,
                                 old_fb, old_fbz, old_nl,
                                 old_ch, cwl);

        bool fin = true;
        for (int j = 0; j < 3; j++) {
            if (c1[j] > 0) {
                v[j] -= v1[j];
                c[j] -= c1[j];

                if (c[j] > 5 && v[j] > 5) {
                    double dvdc = v[j] / c[j];
                    //qDebug() << j << dvdc;
                    pval[j] = qMin(pval[j], dvdc);
                    c0[j]++;
                }
            } else
                c0[j] = 6;

            v[j] = v1[j];
            c[j] = c1[j];
            if (c0[j] < 6)
                fin = false;
        }
        if (fin)
            break;

        bar->setValue(i);
        qApp->processEvents();
    }
    //qDebug() << "_2";
    //c-high, v-down
    int c0_[3] = {0, 0, 0};
    double c_[3] = {65535, 65535, 65535};
    double v_[3] = {65535, 65535, 65535};
    for (i = count; i > count/2; i--) {
        double c1[3] = {0, 0, 0};
        double v1[3];
        read_osc_xyz_no_outer_drift(x_range * i / count, y_range * i / count, z_range * i / count, &c1[0], &c1[1], &c1[2], 1, 1, 1);
        read_osc_xyz_v(&v1[0], &v1[1], &v1[2]);
        if (isInterrupted)
            return restorePICalc(x_range, y_range, z_range,
                                 old_fb, old_fbz, old_nl,
                                 old_ch, cwl);

        bool fin = true;
        for (int j = 0; j < 3; j++) {
            if (c_[j] < 65535) {
                v_[j] -= v1[j];
                c_[j] -= c1[j];

                if (c_[j] > 5 && v_[j] > 5) {
                    double dvdc = v_[j] / c_[j];
                    //qDebug() << j << dvdc;
                    pval[j] = qMin(pval[j], dvdc);
                    c0_[j]++;
                } else
                    i--;
            } else
                i--;

            v_[j] = v1[j];
            c_[j] = c1[j];
            if (c0_[j] < 6)
                fin = false;
        }
        if (fin)
            break;

        bar->setValue(count*1.5 - i);
        qApp->processEvents();
    }
    //qDebug() << "_3";
    //c-high, v-up
    for (int &c : c0_)
        c = 0;
    for (; i <= count; i++) {
        double c1[3] = {66000, 66000, 66000};
        double v1[3];
        read_osc_xyz_no_outer_drift(x_range * i / count, y_range * i / count, z_range * i / count, &c1[0], &c1[1], &c1[2], -1, -1, -1);
        read_osc_xyz_v(&v1[0], &v1[1], &v1[2]);
        if (isInterrupted)
            return restorePICalc(x_range, y_range, z_range,
                                 old_fb, old_fbz, old_nl,
                                 old_ch, cwl);

        bool fin = true;
        for (int j = 0; j < 3; j++) {
            if (c_[j] < 65535) {
                v_[j] -= v1[j];
                c_[j] -= c1[j];

                if (c_[j] < -5 && v_[j] < -5) {
                    double dvdc = v_[j] / c_[j];
                    //qDebug() << j << dvdc;
                    pval[j] = qMin(pval[j], dvdc);
                    c0_[j]++;
                }
            } else
                c0_[j] = 6;

            v_[j] = v1[j];
            c_[j] = c1[j];
            if (c0_[j] < 6)
                fin = false;
        }
        if (fin)
            break;

        bar->setValue(count*1.5 - i);
        qApp->processEvents();
    }

    qDebug() << pval[0] << pval[1] << pval[2];

    //guard
    const double prec = 0.01;
    for (int j = 0; j < 3; j++) {
        if (pval[j] <= prec || pval[j] > 10)
            pval[j] = prec;

        // /10, /2 is in p_scale
        pval[j] /= 5;
    }

    //calc I=P/(coeff*osc_period*csb_rate), P
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    double ifactor = settings.value("pi_autocalib_i_scale", 10).toDouble();
    int csb_rate = osc->CalibWidget->get_csb_speed();

    double ci[3];
    const double oscT = 0.1;
    double scale = settings.value("pi_autocalib_p_scale", 0.45).toDouble();
    for (int j = 0; j < 3; j++) {
        ci[j] = pval[j] / (ifactor * oscT * csb_rate);
        pval[j] *= scale;

        //round
        double step = qPow(10, qFloor(log10(pval[j]) - 1));
        pval[j] = qRound(pval[j] / step) * step;

        double step2 = qPow(10, qFloor(log10(ci[j]) - 1));
        ci[j] = qRound(ci[j] / step2) * step2;
    }


    //set val
    if (scanner == scHead) {
        for (int j = 0; j < 3; j++) {
            osc->CalibWidget->linePID[0][j]->setValue(pval[j]);
            osc->CalibWidget->linePID[1][j]->setValue(ci[j]);
            osc->CalibWidget->linePID[2][j]->setValue(0);
        }
    } else {
        for (int j = 0; j < 3; j++) {
            osc->CalibWidget->linePID2[0][j]->setValue(pval[j]);
            osc->CalibWidget->linePID2[1][j]->setValue(ci[j]);
            osc->CalibWidget->linePID2[2][j]->setValue(0);
        }
    }
    if (old_fb) osc->CalibWidget->fb_on_off_xy->change_state_man_with_sig(true);
    if (old_fbz) osc->CalibWidget->fb_on_off_z->change_state_man_with_sig(true);
    if (old_nl) osc->CalibWidget->nl_on_off_ol->click();
    osc->CalibWidget->s_sendData();


    emit calibInProgress(false);

    go_osc_xyz(x_range/2, y_range/2, z_range/2);


    if (!use_direct_get_data)
        for (int i = 0; i < 4; i++)
            osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    destructStatus(bar);

    foreach (QWidget *wid, cwl)
        wid->setEnabled(true);

    if (!silent)
        doBeep();
}
