﻿#include "ccdmatrix.h"
#include "gui/status.hpp"

#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include "windows.h"
#include <math.h>
#include <QElapsedTimer>
#include <QSettings>
#include <QApplication>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtMath>
#else
#include <qmath.h>
#include <qnumeric.h>
#endif

#include "i_ccd.hpp"
#include "i_spectr.hpp"
#include "hardware/andor.hpp"
#include "hardware/ormins.hpp"
#include "hardware/sol_ccd.hpp"
#include "modules.hpp"
#include "mobile/remote_mobile.hpp"
#include "remote_nst_dataid.hpp"

void Fill_T2Darr_from_long_zoom(long * data,  int Xmin, int Xmax, int Ymin, int Ymax, int Nx, int /*Ny*/, plot::T2dArray & array)
{
    for (int Yind = Ymin; Yind <= Ymax; Yind++) {
        for (int Xind = Xmin; Xind <= Xmax; Xind++)
            array[Yind-Ymin][Xind-Xmin] = data[Nx*Yind+Xind];
    }
}


double wavelength(double central_wl, double disp, double pixelsize, int pixel_idx, int cp)
{
    return central_wl + disp * pixelsize * (pixel_idx - cp);
}


double CCDMatrix::wavelengthNL(double central_wl, double disp, double pixelsize, int pixel_idx, int cp)
{
    double pxlWL = 0;
    bool status = false;

    if (nonlinearMode == 1 && m_spectrInterface) {
        pxlWL = m_spectrInterface->getPixelWl(pixel_idx);
        status = pxlWL > 0;
    } else if (nonlinearMode == 2) {
        QScriptValue res = nlcFun.call(QScriptValue(),
                                    QScriptValueList() << pixel_idx << cp << central_wl << disp << pixelsize);

        pxlWL = res.toNumber();
        status = res.isValid() && !res.isError();
    }

    if (!status)
        pxlWL = wavelength(central_wl, disp, pixelsize, pixel_idx, cp);
    return pxlWL;
}
void CCDMatrix::nlcPrepare(double pixelsize, int nx, int central_pixel)
{
    if (nonlinearMode == 1) {
        bool ok = false;
        if (m_spectrInterface)
            ok = m_spectrInterface->preparePixelWl(pixelsize, nx, central_pixel);
        if (!ok)
            nonlinearMode = 0;
    }

    if (nonlinearMode != 2)
        return;

    try {
        nlcEngine.clearExceptions();
        nlcEngine.evaluate(nonlinearCalibFormula);
        nlcFun = nlcEngine.globalObject().property("calibration");
        QScriptValue res = nlcFun.call(QScriptValue(), QScriptValueList() << 1 << 1 << 1 << 1 << 1);

        if (!nlcFun.isFunction() || nlcFun.isError() || res.isError() || !res.isValid())
            nonlinearMode = 0;
    } catch (...) {
        qDebug() << "nlcfail";
        nonlinearMode = 0;
    }
}

int wavelength_Inverse(double wavelength, double central_wl, double disp, double pixelsize, int cp)
{
    return qRound((wavelength - central_wl) / disp / pixelsize + cp);
}
double raman_shift(double laser_wl, double wl)
{
    if (qIsNull(wl) || !qIsFinite(wl)) return 0;
    return 1e7 * (1 / laser_wl - 1 / wl);
}
double raman_shift_Inverse(double raman_shift, double laser_wl)
{
    return 1 / (1 / laser_wl - raman_shift / 1e7);
}
void cut_pikes(long* data,int datasizeX, int datasizeY, long RMS)
{
    for (int j = 0; j < datasizeY; j++) {
        for (int i = 2; i < (datasizeX-2); i++) {
            bool isSpike = false;
            int idx = i+datasizeX*j;
            int idx_prev = i+datasizeX*j-1;
            int idx_next = i+datasizeX*j+1;

            if (((data[idx_next] + RMS) < data[idx_prev]) && ((data[idx] + RMS) < data[idx_next]))
                isSpike = true;
            else if (data[idx] > (data[idx_prev] + RMS) &&
                     data[idx] > (data[idx_next] + RMS) &&
                     abs(data[idx_prev]- data[idx_prev-1] < RMS) &&
                     abs(data[idx_next]- data[idx_next+1] < RMS) ) //fixme strange place, why abs(bool) ?
                isSpike = true;

            if (isSpike)
                data[idx] = (data[idx_prev] + data[idx_next])/2;
        }
    }
}

void set_spectra_calib(plot::TCurveData &curve, plot::TMetaStorage &attrib, int scale_idx)
{
    for (int i = 0; i < curve.XSeq.size(); i++) {
        switch (scale_idx) {
        case 0:
            curve.XSeq[i] = i;
            break;
        case 1:
            curve.XSeq[i] =
                    1e-9*wavelength(attrib.value("Central Wavelength, nm").toDouble(),
                                    attrib.value("Dispersion").toDouble(),
                                    attrib.value("pixelsize").toDouble(),
                                    i,
                                    attrib.value("Central Pixel").toInt());
            break;
        case 2:
            curve.XSeq[i] =
                    raman_shift(attrib.value("Laser Wavelength, nm").toDouble(),
                                wavelength(attrib.value("Central Wavelength, nm").toDouble(),
                                           attrib.value("Dispersion").toDouble(),
                                           attrib.value("pixelsize").toDouble(),
                                           i,
                                           attrib.value("Central Pixel").toInt()));
            break;
        }
    }
}

class IDummy : public I_CCD
{
public:
    IDummy()
    {
        xPixelsCount = 1024;
        yPixelsCount = 256;
        xPixelSize = 0.026;
        yPixelSize = 0.026;
    }
    bool initialize(int) override { initialized = false; return true; }
    bool shutdown() override { initialized = false; return true; }
    bool setAcquisitionMode(int mode) override { acqMode = mode; return true; }
    bool setResultsMode(int mode) override { resMode = mode; return true; }
    bool setTriggerMode(int mode) override { triggerMode = mode; return true; }
    bool setTemperature(int temp) override { temperature = temp; return true; }

    bool startAcquisition() override { return false; }
    bool stopAcquisition() override { return true; }
    bool isReady() override { return true; }
    bool getEffectiveTimings(double *, double *, double *) override { return true; }
};

#include <QComboBox>
#include <QCheckBox>
#include <QFormLayout>
#include <edit.hpp>
#include <controls.h>
class IEmulator : public IDummy
{
    QElapsedTimer timer;

public:
    bool initialize(int) override
    {
        p_data = new long[xPixelsCount*yPixelsCount];
        initialized = true;
        hasTemperature = true;
        return true;
    }
    bool shutdown() override
    {
        delete [] p_data;
        initialized = false;
        return true;
    }
    bool startAcquisition() override
    {
        timer.start();
        return true;
    }
    bool isReady() override
    {
        return timer.hasExpired(exposureTime * 1000);
    }
    bool stopAcquisition() override
    {
        timer.invalidate();
        return true;
    }
    bool getData() override
    {
        if (resMode == RESMODE_IMAGE) {
            for (int i = 0; i < yPixelsCount; i++)
                for (int j = 0; j < xPixelsCount; j++)
                    p_data[i*xPixelsCount + j] = (300*sin(0.1*i)*sin(0.1*i)*sin(0.1*j)*sin(0.1*j)+(rand()% 100));
        } else {
            int shiftus = rand()% 5000;
            for (int i = 0; i < xPixelsCount; i++)
                p_data[i] = shiftus+(long)(2000*exp(-0.05*(i-100)*(i-100))+
                                           1000*exp(-0.05*(i-500)*(i-512))+
                                           4000*exp(-0.05*(i-800)*(i-800))+(rand()% 30));
        }

        return true;
    }
    QWidget *createInterfaceWidget() override {

        QWidget *wid = new QWidget;
        QHBoxLayout *widLay = new QHBoxLayout(wid);
        widLay->setContentsMargins(0,0,0,0);

        QGroupBox *gb1 = new QGroupBox();
        widLay->addWidget(gb1);
        QHBoxLayout *gb1HLay = new QHBoxLayout(gb1);

        QVBoxLayout **gb1VLays = new QVBoxLayout*[3];
        for (int i=0; i<3; i++)
        {
            gb1VLays[i] = new QVBoxLayout();
            gb1HLay->addLayout(gb1VLays[i]);
        }

        QGroupBox *gb2 = new QGroupBox();
        widLay->addWidget(gb2);
        QVBoxLayout *gb2Lay = new QVBoxLayout(gb2);

        {
            QCheckBox *cooler = new QCheckBox(QWidget::tr("Cooler Level"));
            cooler->setTristate();
            cooler->setChecked(true);
            gb2Lay->addWidget(cooler);
        }

        {
            QCheckBox *shutter = new QCheckBox(QWidget::tr("Shutter") + " " + QWidget::tr("Open/Auto/Closed"));
            shutter->setTristate();
            shutter->setCheckState(Qt::PartiallyChecked);
            gb2Lay->addWidget(shutter);
        }

        {
            QCheckBox *emccd = new QCheckBox(QWidget::tr("EMCCD"));
            gb2Lay->addWidget(emccd);

            Edit *Eemgain = new Edit(0, 1000);
            Eemgain->setValue(1);
            Eemgain->setMaximumWidth(60);
            gb1VLays[0]->addWidget(new QLabel(QWidget::tr("EM gain: ")));
            gb1VLays[1]->addWidget(Eemgain);
        }

        QComboBox* CBhorSpeed = new QComboBox;
        CBhorSpeed->setMaximumWidth(60);
        CBhorSpeed->addItem("1.1");
        gb1VLays[0]->addWidget(new QLabel(QWidget::tr("Shift speed: ")));
        gb1VLays[1]->addWidget(CBhorSpeed);

        QComboBox* CBpreamp = new QComboBox;
        CBpreamp->setMaximumWidth(60);
        CBpreamp->addItem("x1");
        gb1VLays[0]->addWidget(new QLabel(QWidget::tr("Preamp: ")));
        gb1VLays[1]->addWidget(CBpreamp);

        return wid;
    }
};

CCDMatrix::~CCDMatrix()
{
    delete m_ccdInterface;
    delete m_spectrInterface;
}
CCDMatrix::CCDMatrix()
{
    idx_accum = 0;
    idx_kin_ser = 0;
    acquisitionMode_n = 0;
    dataSize = 0;
    imageX_min_n = 0;
    imageX_max_n = 0;
    imageY_min_n = 0;
    imageY_max_n = 0;
    Scale_n = 0;
    Laser_WL_n = 0.;
    Cent_Pixel_n = 0;
    Cent_WL_n = 0.;
    Disp_n = 0.;
    Disp_backup = 0.;
    last_wl = 0.;
    last_idx = 0;
    temp = 0;
    m_ccdInterface = nullptr;
    m_spectrInterface = nullptr;
    m_update_ccd = true;
    m_update_exposure = true;
    m_update_singleTrack = true;
    hasTemperature = false;

    panoram = false;
    panoram_min = 200;
    panoram_max = 800;

    m_readMode = RESMODE_FVB;

    QSettings settings("Param/ccd.conf", QSettings::IniFormat);

    base_line_first = settings.value("baseline_before_intensity", false).toBool();
    nonlinearMode = 1;

    m_closeThread = false;
    m_waitMainInit = true;

    Action1 = false;
    Action6 = false;
    Action8 = false;
    Action12 = false;
    Action19 = false;
    Action23 = false;
    Action24 = false;
    Action26 = false;

    m_ccdStart = false;
    m_runTillCancel = false;

    Nx = 1024;
    Ny = 256;

    imageX_min = 0;
    imageX_max = 1023;
    imageY_min = 0;
    imageY_max = 255;

    Cent_Pixel = 512;
    pixel_Xsize = 0.026f;
    Laser_WL = 532.1;
    Cent_WL = 532.1;
    Disp = 10;

    Cent_Pixel_real = Cent_Pixel;
    Cent_WL_real = Cent_WL;
    Disp_real = Disp;
    Nx_real = Nx;

    m_trackHeight = 10;

    m_exposureTime = 0.1;
    m_centralRow = 128;
    numOfAccumulations = 1;
    accumulationCycleTime = 0;
    Kin_ser_time = 0;
    Kin_ser_number= 1;
    acquisitionMode = ACQMODE_SINGLE;

    Auto_Calib = false;
    cut_pikes_on_off = false;
    base_line_on_off = false;
    intensity_calib_on_off = intensity_calib_pixel_on_off = false;

    Scale = 0;
    Scale_Dir = false;

    m_triggerMode = TRGMODE_AUTO;
    ManBaseLine = new long[80000];
    load_data();

    currentCameraTemperature = 20;

    start();
}

void CCDMatrix::initialize()
{
    if (!m_ccdInterface)
        return;

    m_ccdInterface->initialize();

    Nx = m_ccdInterface->xPixelsCount;
    Nx_real = Nx;
    Ny = m_ccdInterface->yPixelsCount;
    pixel_Xsize = m_ccdInterface->xPixelSize;

    imageX_min = 0;
    imageX_max = Nx-1;
    imageY_min = 0;
    imageY_max = Ny-1;
    dataSize = Nx;
    m_centralRow = (int)(Ny/2);
    m_trackHeight = (int)(Ny/20);
}
void CCDMatrix::load_data()
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);

    Scale = settings.value("scale_ind", 0).toInt();
    Laser_WL = settings.value("laser_wl", 532.1).toDouble();
    Cent_WL = settings.value("central_wl", 532.1).toDouble();
    Cent_Pixel = settings.value("central_px", 512).toInt();
    Disp = settings.value("dispersion", 10).toDouble();

    Cent_Pixel_real = Cent_Pixel;
    Cent_WL_real = Cent_WL;
    Disp_real = Disp;
}
int CCDMatrix::getCameraTemperature()
{
    return currentCameraTemperature;
}
QWidget *CCDMatrix::getCCDInterfaceWidget()
{
    if (m_ccdInterface)
        return m_ccdInterface->createInterfaceWidget();
    return nullptr;
}

void CCDMatrix::thread_close()
{
    m_closeThread = true;

    while (isRunning()) {
        qApp->processEvents();
        msleep(3);
    }
}
void CCDMatrix::startAcquisition()
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    Scale_Dir = !settings.value("scale_reverse", true).toBool();
    Action1 = true;
}
void CCDMatrix::s_abortAcq()
{
    m_ccdStart = false;
}
void CCDMatrix::sl_set_readMode(int mode)
{
    m_readMode = mode;
    m_update_ccd = true;
}
void CCDMatrix::sl_set_Triger(int trigger)
{
    m_triggerMode = trigger;
    m_update_ccd = true;
}
void CCDMatrix::sl_set_Acc_num(int number)
{
    numOfAccumulations = number;
}
void CCDMatrix::s_setTemp(int temp_new)
{
    temp = temp_new;
    Action6=true;
    #ifndef _ADROID
    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::setCCDTemperature, temp_new);
#endif
}
void CCDMatrix::sl_set_Exp_time(double time)
{
    m_exposureTime = time;
    m_update_exposure = true;

    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    settings.setValue("exposure_time", time);
    #ifndef _ADROID
    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::setCCDExposure, time);
#endif
}
void CCDMatrix::sl_set_acqMode(int mode)
{
    acquisitionMode_n=mode;
    Action8=true;
}
void CCDMatrix::sl_set_Acc_time(double time)
{
    accumulationCycleTime = (float)time;
}
void CCDMatrix::sl_set_Center_row(int row)
{
    m_centralRow = row;
    m_update_singleTrack = true;
}
void CCDMatrix::sl_set_Height(int height)
{
    m_trackHeight = height;
    m_update_singleTrack = true;
}
void CCDMatrix::s_setXAxisScale(int scale, double laser, int cent_pixel, double wl, double disp)
{
    Scale_n=scale;
    Cent_WL_n=wl;
    Disp_n = disp;
    Cent_Pixel_n=cent_pixel;
    Laser_WL_n=laser;
    Action12=true;
}
void CCDMatrix::s_isContScan(bool checked)
{
    m_runTillCancel = checked;
}
void CCDMatrix::sl_set_Kin_ser_num(int number)
{
    Kin_ser_number = number;
}
void CCDMatrix::sl_set_Kin_ser_time(double time)
{
    Kin_ser_time = time;
}
void CCDMatrix::sl_set_Auto_calib(bool value)
{
    Auto_Calib = value;
    Action19 = true;
}
void CCDMatrix::sl_cut_pikes(bool value)
{
    cut_pikes_on_off = value;
#ifndef _ADROID
    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::spectrSetCutSpikes, value);
#endif
}
void CCDMatrix::sl_set_new_3D_Limits(const Range &XLimits, const Range &YLimits)
{
    imageX_min_n = (int)XLimits[0];
    imageX_max_n = (int)XLimits[1];
    imageY_min_n = (int)YLimits[0];
    imageY_max_n = (int)YLimits[1];
    Action23 = true;
}
void CCDMatrix::sl_save_base_line(void)
{
    Action24 = true;
}
void CCDMatrix::sl_set_baseline_on_off(bool value)
{
    base_line_on_off = value;
#ifndef _ADROID
    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::spectrSetBaselineEnable, value);
#endif
}
void CCDMatrix::setOutput(plot::T2dArray* image_array_out)
{
    output2dArray = image_array_out;
}
void CCDMatrix::SetManualBaseLine(long *bl)
{
    for (int i = 0; i < qMin(80000, 2*Nx_real); i++)
        ManBaseLine[i] = bl[i];

    Action26 = true;
}

void CCDMatrix::run()
{
    long *dataArray = new long[1100000];
    long *aver = new long[1100000];
    long *Base_Line = new long[80000];

    memset(aver, 0, sizeof(long)*20000);
    memset(Base_Line, 0, sizeof(long)*40000);

    m_closeThread = false;
    m_waitMainInit = true;

#ifdef USE_CCD_EMULATOR
    {
        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        if (settings.value("ccd", false).toBool()  && !m_ccdInterface)
            m_ccdInterface = new IEmulator();
    }
#endif
#ifdef USE_SOL_CCD
    {
        QSettings settings("Param/ccd.conf", QSettings::IniFormat);
        if (settings.value("use_sol_ccd", false).toBool() && !m_ccdInterface)
            m_ccdInterface = new ISolCCD();
    }
#endif
#ifdef USE_ORMINS
    {
        QSettings settings("Param/ccd.conf", QSettings::IniFormat);
        if (settings.value("use_ormins", false).toBool() && !m_ccdInterface)
            m_ccdInterface = new IOrmins();
    }
#endif
#ifdef USE_ANDOR
    if (!m_ccdInterface)
        m_ccdInterface = new IAndor();
#endif


    initialize();

    if (!m_ccdInterface || !m_ccdInterface->initialized) {
        delete m_ccdInterface;
        m_ccdInterface = new IDummy;
    }

    while (m_waitMainInit && !m_closeThread) {
        msleep(50);
    }

    hasTemperature = m_ccdInterface->hasTemperature;
    emit sigUpdateNxNy();
    emit updateCCDStatus(m_ccdInterface->initialized ? EDok : EDno);

    while (!m_closeThread) {
        if (Action1) {
            Action1 = false;
            m_ccdStart = true;
        }
        if (Action6) {
            Action6 = false;
            if (m_ccdInterface->hasTemperature)
                m_ccdInterface->setTemperature(temp);
        }
        if (m_update_exposure) {
            m_update_exposure = false;
            m_ccdInterface->exposureTime = m_exposureTime;
            m_update_ccd = true;
        }
        if (Action8) {
            Action8 = false;
            acquisitionMode = acquisitionMode_n;
        }
        if (m_update_singleTrack) {
            m_update_singleTrack = false;
            m_ccdInterface->centralRow = m_centralRow;
            m_ccdInterface->trackHeight = m_trackHeight;
            m_update_ccd = true;
        }
        if (Action12) {
            Action12 = false;
            if (!Auto_Calib) {
                Disp = Disp_n;
                Cent_WL = Cent_WL_n;
                Cent_WL_real = Cent_WL;
                Disp_real = Disp;
            }

            Cent_Pixel = Cent_Pixel_n;
            Laser_WL = Laser_WL_n;
            Cent_Pixel_real = Cent_Pixel;
            Scale = Scale_n;
        }
        if (Action19) {
            Action19 = false;

            if (Auto_Calib && m_spectrInterface) {
                Cent_WL = m_spectrInterface->getWl();
                Disp = m_spectrInterface->getDispersion();
                Cent_WL_real = Cent_WL;
                Disp_real = Disp;
            }
        }
        if (Action23) {
            imageX_min = imageX_min_n;
            imageX_max = imageX_max_n;
            imageY_min = imageY_min_n;
            imageY_max = imageY_max_n;
            Action23 = false;
        }
        if (Action24) {
            for (int i = 0; i < qMin(80000, Nx_real); i++)
                Base_Line[i] = dataArray[i];

            for (int i = qMin(80000, Nx_real); i < qMin(80000, 2*Nx_real); i++)
                Base_Line[i] = Base_Line[Nx_real-1];

            Action24 = false;
        }
        if (Action26) {
            for (int i = 0; i < qMin(80000, Nx_real); i++)
                Base_Line[i] = ManBaseLine[!Scale_Dir ? i : Nx_real-i-1];

            for (int i = qMin(80000, Nx_real); i < qMin(80000, 2*Nx_real); i++)
                Base_Line[i] = Base_Line[Nx_real-1];

            Action26 = false;
        }

        static bool prevSpect = false;
        static QElapsedTimer stamp; //TODO invalidate on init
        static int st_mul = 0;
        if (!stamp.isValid() || stamp.hasExpired(100)) {
            stamp.start();

            st_mul++;
            if (m_ccdInterface->hasTemperature && st_mul >= 5) {
                st_mul = 0;
                currentCameraTemperature = m_ccdInterface->getTemperature();
            }

            if (m_spectrInterface) {
                bool val = m_spectrInterface->isReady();
                if (prevSpect != val) {
                    emit monoch_status(val);
                    prevSpect = val;
                }
            }
        }

        static bool prevStart = false;
        if (prevStart != m_ccdStart) {
            emit updateCCDStatus(m_ccdInterface->initialized ? (m_ccdStart ? EDplay : EDok) : EDno);
            prevStart = m_ccdStart;
        }

        if (m_ccdStart) {
            if (m_update_ccd) {
                m_update_ccd = false;

                m_ccdInterface->setAcquisitionMode(ACQMODE_SINGLE);
                m_ccdInterface->setResultsMode(m_readMode);
                m_ccdInterface->setTriggerMode(m_triggerMode);

                dataSize = Nx;
                if (m_readMode == RESMODE_IMAGE)
                    dataSize = Nx*Ny;
            }
            if (Auto_Calib && m_spectrInterface) {
                Cent_WL = m_spectrInterface->getWl();
                Disp = m_spectrInterface->getDispersion();
                Cent_WL_real = Cent_WL;
                Disp_real = Disp;
            }

            memset(aver, 0, sizeof(long)*20000);

            ObtainProccessedData(dataArray, aver, Base_Line);

            if (!m_runTillCancel) {
                m_ccdStart = false;
                emit ccd_finished();
            }
        }
        msleep(1);
    }

    delete[] dataArray;
    delete[] aver;
    delete[] Base_Line;

    if (m_ccdInterface)
        m_ccdInterface->shutdown();

    /*  //not owned by ccdmatrix thread for now
    if (m_spectrInterface)
        m_spectrInterface->shutdown(); //<- crashes
    */
}

int CCDMatrix::gen_panoram_cycles()
{
    if (!panoram)
        return 1;

    double range = 0.95 * Nx_real*Disp_real*pixel_Xsize;

    if (nonlinearMode && (m_spectrInterface || nonlinearMode != 1)) {
        nlcPrepare(pixel_Xsize, Nx, Cent_Pixel);
        if (nonlinearMode) {
            double wl1 = wavelengthNL(Cent_WL, Disp, pixel_Xsize, 0, Cent_Pixel);
            double wl2 = wavelengthNL(Cent_WL, Disp, pixel_Xsize, Nx-1, Cent_Pixel);

            range = 0.7 * (wl2 - wl1);
        }
    }

    int count = ceil( (panoram_max - panoram_min) / range );

    return count;
}
void CCDMatrix::gen_panoram_param(double *arr, int num)
{
    if (!panoram)
        return;

    for (int i = 0; i < num; i++) {
        arr[i] = panoram_min + (panoram_max - panoram_min)*(i+0.5)/num;
    }
}
void CCDMatrix::sew_panoram(long *result, long *add)
{
    if (!panoram || !result)
        return;

    int start = qMax(0, wavelength_Inverse(last_wl, Cent_WL_real, Disp_real, pixel_Xsize, Cent_Pixel_real));
    int end = qMin(Nx_real - 1, wavelength_Inverse(panoram_max, Cent_WL_real, Disp_real, pixel_Xsize, Cent_Pixel_real));

    int _nx = ceil( (panoram_max - panoram_min) / (Disp_backup * pixel_Xsize) );
    double _cwl = (panoram_min + (panoram_max - panoram_min) / 2.0);

    //linear interpolation of intersected regions
    {
        long last_idx2 = 0;
        for (int i = 0; i < start; i++) {
            double cwl = wavelength(Cent_WL_real, Disp_real, pixel_Xsize, i, Cent_Pixel_real);
            long pxl = wavelength_Inverse(cwl, _cwl, Disp_backup, pixel_Xsize, _nx / 2.0);

            long idx = i;
            if (Scale_Dir)
                idx = (Nx_real - 1) - i;

            if (pxl < 0 || pxl >= _nx)
                continue;

            result[pxl] = (result[pxl] * float(start - i) / start) + (add[idx] * float(i) / start);

            if (last_idx2 != 0)
                for (int j = last_idx2+1; j < pxl; j++) {
                    double scale = float(i) / start - float(pxl - j) / (pxl - last_idx2) / start;
                    result[j] = result[j] * (1 - scale) +
                            ( add[idx] * float(j - last_idx2) / (pxl - last_idx2) +
                              add[idx-1] * float(pxl - j) / (pxl - last_idx2) ) * scale;
                }

            if (last_idx2 < pxl)
                last_idx2 = pxl;
        }
        last_idx = last_idx2; //cut off tail of last part, that wasn't interpolated
    }

    //add new points
    for (int i = start; i <= end; i++) {
        double cwl = wavelength(Cent_WL_real, Disp_real, pixel_Xsize, i, Cent_Pixel_real);
        long pxl = wavelength_Inverse(cwl, _cwl, Disp_backup, pixel_Xsize, _nx / 2.0);
        if (pxl < last_idx || pxl >= _nx)
            continue;

        long idx = i;
        if (Scale_Dir)
            idx = (Nx_real - 1) - i;

        if (pxl == last_idx && last_idx != 0) {
            result[pxl] += add[idx];
            result[pxl] /= 2;
        } else {
            if (last_idx == 0)
                for (int i = 0; i < pxl; i++)
                    result[i] = add[idx];

            result[pxl] = add[idx];
        }

        for (int j = last_idx + 1; j < pxl; j++) {
            result[j] = result[last_idx] * float(pxl - j) / (pxl - last_idx) +
                    result[pxl] * float(j - last_idx) / (pxl - last_idx);
        }

        if (last_idx < pxl)
            last_idx = pxl;
    }

    last_wl = wavelength(Cent_WL_real, Disp_real, pixel_Xsize, end, Cent_Pixel_real);

    {
        int _nx = last_idx;//ceil( (last_wl - panoram_min) / (Disp_backup * pixel_Xsize) );

        //fill data
        Fillspectrum(result, Laser_WL,
                     (panoram_min + (last_wl - panoram_min) / 2.0),
                     _nx, Disp_backup, pixel_Xsize, _nx / 2.0, false);
    }
}

void CCDMatrix::ObtainProccessedData(long *dataArray, long *aver, long *base_line)
{
    int panoram_cycles = gen_panoram_cycles();
    double *pan_central = new double[panoram_cycles];
    gen_panoram_param(pan_central, panoram_cycles);

    int _nx = ceil( (panoram_max - panoram_min)/(Disp_real*pixel_Xsize) );
    long *full_pan = nullptr;
    if (panoram) {
        try {
            full_pan = new long[_nx + 1];
        } catch (...) {
            full_pan = nullptr;
            qDebug() << "panoram alloc fail" << _nx;
        }
    }

    Disp_backup = Disp_real;

    idx_kin_ser = 0; //kinetic series cycles
    do {
        if (panoram && full_pan)
            memset(full_pan, 0, _nx*sizeof(long));

        last_idx = 0;
        last_wl = panoram_min;

        for (int i = 0; i < panoram_cycles; i++) {
            if (panoram) {
                if (m_spectrInterface) {
                    while (!m_spectrInterface->isReady()) msleep(1);//wait
                    m_spectrInterface->setWl(pan_central[i], false);
                    while (!m_spectrInterface->isReady()) msleep(1);//wait
                }

                if (Auto_Calib && m_spectrInterface) {
                    Cent_WL = m_spectrInterface->getWl();
                    Disp = m_spectrInterface->getDispersion();
                    Cent_WL_real = Cent_WL;
                    Disp_real = Disp;
                } else {
                    Cent_WL = pan_central[i];
                    Cent_WL_real = Cent_WL;
                }
            }

            idx_accum = 0; //accumulation cycles
            do {
                if (!m_ccdInterface->startAcquisition())
                    break;

                while (m_ccdStart && !m_ccdInterface->isReady())
                    msleep(1);

                if (!m_ccdStart) {
                    m_ccdInterface->stopAcquisition();
                    break;
                }

                if (!m_ccdInterface->getData())
                    break;

                memcpy(dataArray, m_ccdInterface->p_data, sizeof(long)*dataSize);

                processData(dataArray, aver, base_line);

                if (acquisitionMode != ACQMODE_SINGLE && m_triggerMode == TRGMODE_AUTO && (int)(accumulationCycleTime*1000.0) >= 1)
                    msleep((int)(accumulationCycleTime*1000.0));

                idx_accum++;
            } while (acquisitionMode != ACQMODE_SINGLE && idx_accum < numOfAccumulations);
            if (!m_ccdStart)
                break;


            sew_panoram(full_pan, dataArray);
        }
        if (!m_ccdStart)
            break;


        if (panoram && full_pan) {
            int _nx = ceil( (panoram_max - panoram_min) / (Disp_backup * pixel_Xsize) );

            for (int j = last_idx+1; j <= _nx; j++)
                full_pan[j] = full_pan[last_idx];

            memcpy(dataArray, full_pan, _nx*sizeof(long));

            //fill data
            Fillspectrum(dataArray, Laser_WL,
                         (panoram_max + panoram_min) / 2.0,
                         _nx, Disp_backup, pixel_Xsize, _nx / 2.0, false);

            if (acquisitionMode == ACQMODE_KINETIC)
                emit sig_cur_kin_ser_num(idx_kin_ser); //add curve
        }

        if (acquisitionMode == ACQMODE_KINETIC && m_triggerMode == TRGMODE_AUTO && (int)(Kin_ser_time*1000.0) >= 1)
            msleep((int)(Kin_ser_time*1000.0));

        idx_kin_ser++;
    } while (acquisitionMode == ACQMODE_KINETIC && idx_kin_ser < Kin_ser_number);

    delete [] pan_central;
    delete [] full_pan;
}
void CCDMatrix::nonlinear_correction(long *dataArray)
{
    Nx_real = Nx;
    if (!nonlinearMode || m_readMode == RESMODE_IMAGE || (!m_spectrInterface && nonlinearMode == 1))
        return;

    nlcPrepare(pixel_Xsize, Nx, Cent_Pixel);
    if (!nonlinearMode)
        return;

    double old = -1e10;
    double min = 1e10;
    for (int i = 0; i < Nx; i++) {
        double wl = wavelengthNL(Cent_WL, Disp, pixel_Xsize, i, Cent_Pixel);
        double diff = fabs(wl - old);
        if (diff < min)
            min = diff;

        old = wl;
    }

    double wmin = wavelengthNL(Cent_WL, Disp, pixel_Xsize, 0, Cent_Pixel);
    double wmax = wavelengthNL(Cent_WL, Disp, pixel_Xsize, Nx-1, Cent_Pixel);
    if (wmax < wmin) qSwap(wmin, wmax);

    Disp_real = Disp;
    if (min > 0)
        Disp_real = min / pixel_Xsize;
    Nx_real = (wmax - wmin) / (Disp_real * pixel_Xsize) + 1;
    //qDebug() << "disp nx min" << Disp << Nx << min;
    if (Nx_real > 4 * Nx || Nx_real < 0.1 * Nx) {
        Nx_real = Nx * 3.5;
        Disp_real = (wmax - wmin) / (Nx_real-1) / pixel_Xsize;
    }

    long *calibArr;
    try {
        calibArr = new long[Nx_real+1];
    } catch (...) {
        qDebug() << "nlc calib fail" << Nx_real;
        Disp_real = Disp;
        Nx_real = Nx;
        return;
    }

    int lasti = 0;

    memset(calibArr, 0, Nx_real * sizeof(long));

    Cent_Pixel_real = Nx_real / 2;
    double wc1 = wmin + (Disp_real * Nx_real * pixel_Xsize) / 2;
    double wc2 = wmax - (Disp_real * Nx_real * pixel_Xsize) / 2;
    Cent_WL_real = wc1/2 + wc2/2;
    //qDebug() << "central" << wc1 << wc2 << Cent_WL;

    for (int i = 0; i < Nx; i++) {
        int idx0 = i;
        if (Scale_Dir)
            idx0 = (Nx - 1) - i;

        double wl = wavelengthNL(Cent_WL, Disp, pixel_Xsize, i, Cent_Pixel);

        int idx = wavelength_Inverse(wl, Cent_WL_real, Disp_real, pixel_Xsize, Cent_Pixel_real);

        if (idx < lasti || idx >= Nx_real)
            continue;

        if (lasti == 0)
            for (int i = 0; i <= idx; i++)
                calibArr[i] = dataArray[idx0];

        if (idx == lasti) {
            calibArr[idx] += dataArray[idx0];
            calibArr[idx] /= 2;
        } else {
            calibArr[idx] = dataArray[idx0];
        }

        for (int i = lasti+1; i < idx; i++)
            calibArr[i] = (calibArr[lasti] * (idx - i)  +
                           calibArr[idx] * (i - lasti) ) / (idx - lasti);

        lasti = idx;
    }

    for (int i = lasti + 1; i < Nx_real; i++)
        calibArr[i] = calibArr[lasti];

    for (int i = 0; i < Nx_real; i++)
        dataArray[i] = calibArr[Scale_Dir ? Nx_real - 1 - i : i];

    delete [] calibArr;
}

void CCDMatrix::processData(long *dataArray, long *aver, long *base_line)
{
    if (cut_pikes_on_off) {
        if (m_readMode == RESMODE_IMAGE)
            cut_pikes(dataArray, Nx, Ny, 3);
        else
            cut_pikes(dataArray, Nx_real, 1, 3);
    }

    if (intensity_calib_pixel_on_off && m_readMode != RESMODE_IMAGE && intensity_calib_pixel_data.size()) {
        for (int i = 0; i < intensity_calib_pixel_data.first().first; i++)
            dataArray[i] *= intensity_calib_pixel_data.first().second;

        for (int i = 0; i < intensity_calib_pixel_data.size()-1; i++) 
		{
			auto pixDataI = intensity_calib_pixel_data.at(i);
			auto pixDataNext = intensity_calib_pixel_data.at(i + 1);
            for (int j = qMax(0, pixDataI.first); j < qMin(Nx, pixDataNext.first); j++)
                dataArray[j] *= pixDataI.second + (pixDataNext.second - pixDataI.second) *
                        (double(j - pixDataI.first) / (pixDataNext.first - pixDataI.first));
        }

        for (int i = qMax(0, intensity_calib_pixel_data.last().first); i < Nx; i++)
            dataArray[i] *= intensity_calib_pixel_data.last().second;
    }

    nonlinear_correction(dataArray);

    if (base_line_on_off && base_line_first)
        for (int i = 0; i < qMin(80000, 2*Nx_real); i++)
            dataArray[i] -= base_line[i];

    if (intensity_calib_on_off && m_readMode != RESMODE_IMAGE && intensity_calib_data.size()) {
        QVector<int> calib(intensity_calib_data.size());
        for (int i = 0; i < intensity_calib_data.size(); i++) {
            int idx = i;
            if (Scale_Dir) idx = intensity_calib_data.size() - 1 - i;
            calib[idx] = wavelength_Inverse(intensity_calib_data.at(i).first, Cent_WL_real, Disp_real, pixel_Xsize, Cent_Pixel_real);
            if (Scale_Dir) calib[idx] = Nx_real - 1 - calib[idx];
        }

        for (int i = 0; i < calib.first(); i++)
            dataArray[i] *= (!Scale_Dir ? intensity_calib_data.first() : intensity_calib_data.last()).second;

        for (int i = 0; i < calib.size()-1; i++) {
            int idx = i;
            if (Scale_Dir) idx = intensity_calib_data.size() - 1 - i;
            for (int j = qMax(0, calib.at(i)); j < qMin(Nx_real, calib.at(i+1)); j++)
                dataArray[j] *= intensity_calib_data.at(idx).second + (intensity_calib_data.at(idx+(!Scale_Dir ? 1 : -1)).second - intensity_calib_data.at(idx).second) *
                        (double(j - calib.at(i)) / (calib.at(i+1) - calib.at(i)));
        }

        for (int i = qMax(0, calib.last()); i < Nx_real; i++)
            dataArray[i] *= (Scale_Dir ? intensity_calib_data.first() : intensity_calib_data.last()).second;
    }

    if (base_line_on_off && !base_line_first)
        for (int i = 0; i < qMin(80000, 2*Nx_real); i++)
            dataArray[i] -= base_line[i];

    if (m_readMode == RESMODE_IMAGE) {
        Fill_T2Darr_from_long_zoom(dataArray, imageX_min, imageX_max, imageY_min, imageY_max, Nx, Ny, *output2dArray);
        emit sig_Update();
    } else {
        if (!panoram && (acquisitionMode == ACQMODE_SINGLE || idx_accum < numOfAccumulations))
            Fillspectrum(dataArray, Laser_WL, Cent_WL_real, Nx_real, Disp_real, pixel_Xsize, Cent_Pixel_real, Scale_Dir);
    }

    if (acquisitionMode != ACQMODE_SINGLE) {
        emit sig_cur_ac_num(idx_accum + 1);

        int datasz = Nx_real;
        if (m_readMode == RESMODE_IMAGE)
            datasz = dataSize;

        for (int i = 0; i < datasz; i++)
            aver[i] += dataArray[i];

        if (idx_accum + 1 >= numOfAccumulations) {
            if (acquisitionMode == ACQMODE_KINETIC)
                emit sig_cur_kin_ser_num(idx_kin_ser + 1);

            for (int i = 0; i < datasz; i++)
                dataArray[i] = aver[i] / numOfAccumulations;
            memset(aver, 0, sizeof(long)*datasz);

            if (m_readMode == RESMODE_IMAGE) {
                Fill_T2Darr_from_long_zoom(dataArray, imageX_min, imageX_max, imageY_min, imageY_max, Nx, Ny, *output2dArray);
                emit sig_Update();
            } else {
                if (!panoram //) {
                        || acquisitionMode == ACQMODE_KINETIC) { //this clause part adds source curves of panoram spectra as separate curves, when kinetic mode enabled
                    Fillspectrum(dataArray, Laser_WL, Cent_WL_real, Nx_real, Disp_real, pixel_Xsize, Cent_Pixel_real, Scale_Dir);
                }
            }
        }
    }
}

void CCDMatrix::Fillspectrum(long *dataArray, double laser_wl, double central_wl, double nx, double disp, double px_size, double central_px, bool scale_reverse)
{
    curveCalib->insert("Laser Wavelength, nm", laser_wl);
    curveCalib->insert("Central Wavelength, nm", central_wl);
    curveCalib->insert("Dispersion", disp);
    curveCalib->insert("pixelsize", px_size);
    curveCalib->insert("Nx", nx);
    curveCalib->insert("Central Pixel", central_px);

    QVector<double> xdata, ydata;
    try {
        xdata.reserve(nx);
        ydata.reserve(nx);
    } catch (...) {
        qDebug() << "ccd oom";
        xdata.clear();
        ydata.clear();
        emit new2dData(ydata, xdata);
        return;
    }
    for (int i = 0; i < nx; i++) {
        int idx = i;
        if (scale_reverse)
            idx = nx - i - 1;

        double idx2 = i;
        switch (Scale) {
        case 0:
            idx2 = i;
            break;
        case 1:
            idx2 = 1e-9 * wavelength(central_wl, disp, px_size, i, central_px);
            break;
        case 2:
            idx2 = raman_shift(laser_wl, wavelength(central_wl, disp, px_size, i, central_px));
            break;
        }

        xdata.append(idx2);
        ydata.append(dataArray[idx]);
    }

    emit new2dData(ydata, xdata);
}

CCDMatrix* CCDMatrix::GetIt() noexcept
{
    #ifndef _ADROID
    auto ccdMatrix = ModuleManager::instance->find1VoidPtr("ccdmatrix");
    return static_cast<CCDMatrix*>(ccdMatrix);
#else
    return nullptr;
#endif
}
