#include "controls.h"
#include "nstslider.h"
#include <QPixmapCache>
#include <QSettings>
#include <QFormLayout>
#include <QDoubleValidator>
#include <QShortcut>
#include <math.h>
#include <QDebug>
#include <QPainter>
#include <QPainterPath>
#include <QEvent>

QIcon rotareIcon(const QIcon icon, QSize size, int gradus)
{
    QPixmap origPix(icon.pixmap(size));
    
    QMatrix rm;
    rm.rotate(gradus);
    return origPix.transformed(rm);
}

void rotareButtonIcon(QPushButton *btn, int gradus)
{
    btn->setIcon(rotareIcon(btn->icon(), btn->size(), gradus));
}

QIcon makeHugeIcon(QString path)
{
    static int ICO_SZ = -1;
    if (ICO_SZ == -1) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        ICO_SZ = qMin(settings.value("buttonsize", 25).toInt(), 512);
    }

    if (ICO_SZ <= 100)
        return QIcon(path); //return delay-loaded icon as we know that all images are >= 100px. saves some startup time

    QPixmap px;
    if (!QPixmapCache::find(path, &px)) {
        px = QPixmap(path);
        QSize sz = px.size();
        while (sz.width() < ICO_SZ || sz.height() < ICO_SZ)
            sz *= 2;

        if (sz != px.size())
            px = px.scaled(sz);
        QPixmapCache::insert(path, px);
    }

    return QIcon(px);
}

QIcon makeHugeIconBounded(QString path, QColor color, qreal bWidth)
{
    QPixmap px;
    if (!QPixmapCache::find("bounded_" + path, &px)) {
        px = QPixmap(path);
        QSize sz(100, 100);
        px = px.scaled(sz);

        QPainter paint(&px);
        paint.setPen(QPen(QBrush(color), bWidth));
        paint.drawRect(0, 0, px.size().width() - bWidth*0.2, px.size().height() - bWidth*0.2);

        QPixmapCache::insert("bounded_" + path, px);
    }

    return QIcon(px);
}

QString *makeVerticalText(const QString &text)
{
    if (!text.size())
        throw "empty text in makeVerticalText!";
    QString *str = new QString();
    for (int i=0; i<text.length()-1; i++) {
        str->append(text[i] + '\n');
    }
    return str;
}

QFormLayout *makeFormLayout(QWidget *parent)
{
    QFormLayout *lay = new QFormLayout(parent);
    //lay->setFieldGrowthPolicy(QFormLayout::AllNonFixedFieldsGrow);
    lay->setLabelAlignment(Qt::AlignLeft);
    lay->setFieldGrowthPolicy(QFormLayout::ExpandingFieldsGrow);
    return lay;
}

static QPalette changedPalette_glbl;
DoubleEdit::DoubleEdit(double min, double max, int acc)
{
    range_validator = new QDoubleValidator(min, max, acc, this);
    setValidator(range_validator);

    prev_value = 0;
    setText("0");
    setAlignment(Qt::AlignRight);

    connect(this, SIGNAL(returnPressed()), SLOT(setNewValue()));
    new QShortcut(QKeySequence(Qt::Key_Escape), this, SLOT(resetValue()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_Up), this, SLOT(stepUp()), nullptr, Qt::WidgetWithChildrenShortcut);
    new QShortcut(QKeySequence(Qt::Key_Down), this, SLOT(stepDown()), nullptr, Qt::WidgetWithChildrenShortcut);

    changedPalette_glbl.setBrush(QPalette::Active, QPalette::Text, QColor(Qt::red));
    setPalette(QPalette());

    setMaximumWidth(70);
    setMaxLength(15);
}

void DoubleEdit::stepUp()
{
    setValue(QLocale::system().toDouble(text()) + pow(10., -range_validator->decimals()));
    setPalette(changedPalette_glbl);
}
void DoubleEdit::stepDown()
{
    setValue(QLocale::system().toDouble(text()) - pow(10., -range_validator->decimals()));
    setPalette(changedPalette_glbl);
}

void DoubleEdit::setMin(double value_min)
{
   range_validator->setBottom(value_min);

   if (value() < value_min)
       setValue(value_min);
}
void DoubleEdit::setMax(double value_max)
{
   range_validator->setTop(value_max);

   if (value() > value_max)
       setValue(value_max);
}

void DoubleEdit::resetValue()
{
    setValue(prev_value);
    clearFocus();
}

void DoubleEdit::setNewValue()
{
    prev_value = QLocale::system().toDouble(text());

    emit newValue(int(prev_value));
    emit newValue(prev_value);
    emit newValue();

    setPalette(QPalette());
    clearFocus();
}

void DoubleEdit::setValue(double value)
{
    if (value > range_validator->top())
        value = range_validator->top();
    if (value < range_validator->bottom())
        value = range_validator->bottom();

    prev_value = value;
    setText(QLocale::system().toString(value, 'g', 4)); //prec 8 - wrong string
    setPalette(QPalette());
}

double DoubleEdit::value() const
{
    return prev_value;
}

void DoubleEdit::focusInEvent(QFocusEvent *event)
{
    emit editingStarted();
    setPalette(changedPalette_glbl);

    QLineEdit::focusInEvent(event);
}
void DoubleEdit::focusOutEvent(QFocusEvent *event)
{
    setValue(prev_value);

    QLineEdit::focusOutEvent(event);
}

//--------
#include "edit.hpp"
void linkDoubleChangers(DoubleChanger *master, DoubleChanger *slave, bool linkRange)
{
    if (linkRange) {
        QObject::connect(master, SIGNAL(newDecimals(int)), slave, SLOT(setDecimals(int)));
        QObject::connect(master, SIGNAL(newMinimum(double)), slave, SLOT(setMinimum(double)));
        QObject::connect(master, SIGNAL(newMaximum(double)), slave, SLOT(setMaximum(double)));
        slave->setMinimum(master->edit->minimum());
        slave->setMaximum(master->edit->maximum());
        slave->setDecimals(master->edit->decimals());
    }

    QObject::connect(master, SIGNAL(newValue(double)), slave, SLOT(setValue(double)));
    QObject::connect(slave, SIGNAL(newValue(double)), master, SLOT(setValue(double)));
    QObject::connect(slave, SIGNAL(newValue()), master, SLOT(setNewValue()));
    slave->setValue(master->value());
}

DoubleChanger::DoubleChanger(double min, double max, int acc, bool vertical)
{
    this->vertical = vertical;
    QBoxLayout *mainLayout;
    if (vertical)
        mainLayout = new QVBoxLayout(this);
    else
        mainLayout = new QHBoxLayout(this);
    mainLayout->setContentsMargins(0, 0, 0, 0);

    edit = new Edit(min, max, acc);
    edit->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    slider = new NSTSlider(min, max, acc, this, vertical);
    slider->setFocusPolicy(Qt::NoFocus);

    mainLayout->addWidget(slider, 1);
    mainLayout->addWidget(edit);

    //
    connect(edit, SIGNAL(editingStarted()), SIGNAL(editingStarted()));
    connect(edit, SIGNAL(editingFinished()), SIGNAL(editingFinished()));

    connect(slider, SIGNAL(sliderPressed()), SIGNAL(editingStarted()));
    connect(slider, SIGNAL(sliderReleased()), SIGNAL(editingFinished()));

    //
    connect(edit, SIGNAL(newValue(double)), slider, SLOT(setValue(double)));
    connect(slider, SIGNAL(newValue(double)), edit, SLOT(setValue(double)));
    connect(slider, SIGNAL(sliderMoved(double)), edit, SLOT(setValue(double)));

    //
    connect(edit, SIGNAL(newValue()), SIGNAL(newValue()));
    connect(edit, SIGNAL(newValue(int)), SIGNAL(newValue(int)));
    connect(edit, SIGNAL(newValue(double)), SIGNAL(newValue(double)));

    connect(slider, SIGNAL(newValue()), SIGNAL(newValue()));
    connect(slider, SIGNAL(newValue(int)), SIGNAL(newValue(int)));
    connect(slider, SIGNAL(newValue(double)), SIGNAL(newValue(double)));
}
void DoubleChanger::setDecimals(int decimals)
{
    edit->setDecimals(decimals);
    edit->setSingleStep(pow(10., -decimals));
    slider->setMainRangeAndStep(slider->minV, slider->maxV, pow(10., -decimals));
    emit newDecimals(decimals);
}
void DoubleChanger::setMinimum(double value_min)
{
    edit->setMinimum(value_min);
    slider->setMainRangeAndStep(value_min, slider->maxV, slider->stepV);
    emit newMinimum(value_min);
}
void DoubleChanger::setMaximum(double value_max)
{
    edit->setMaximum(value_max);
    slider->setMainRangeAndStep(slider->minV, value_max, slider->stepV);
    edit->resetValue();
    emit newMaximum(value_max);
}
void DoubleChanger::setNewValue(bool emitNewValue)
{
    edit->setNewValue(emitNewValue);
}
void DoubleChanger::setValue(double value)
{
    edit->setValue(value);
    slider->setValue(value);
}
double DoubleChanger::value() const
{
    return edit->value();
}
double DoubleChanger::maximum() const
{
    return slider->maxValue;
}
double DoubleChanger::minimum() const
{
    return slider->minValue;
}
QSize DoubleChanger::minimumSizeHint() const
{
    QSize size;
    if (vertical)
    {
        int maxWidth;
        if (slider->minimumSizeHint().width() > edit->minimumSizeHint().width())
            maxWidth = slider->minimumSizeHint().width();
        else
            maxWidth = edit->minimumSizeHint().width();
        size.setWidth(maxWidth);
        
        size.setHeight(slider->minimumSizeHint().height() + edit->minimumSizeHint().height());
    }
    else
    {
        int maxHeight;
        if (slider->minimumSizeHint().height() > edit->minimumSizeHint().height())
            maxHeight = slider->minimumSizeHint().height();
        else
            maxHeight = edit->minimumSizeHint().height();
        size.setHeight(maxHeight);
        
        size.setWidth(slider->minimumSizeHint().width() + edit->minimumSizeHint().width());
    }
    return size;
}
//--------

TwoStateButton::TwoStateButton(QString off, QString on)
{
    offstring = off;
    onstring = on;
    isOn = false;

    setIcon(makeHugeIcon(offstring));
    //btnSize = iconSize();
    connect(this, SIGNAL(clicked()), SLOT(change_state()));
}
TwoStateButton::TwoStateButton(const QString & text, QString off, QString on) : QPushButton (QIcon(off), text)
{
    offstring = off;
    onstring = on;
    isOn = false;

    setIcon(makeHugeIcon(offstring));
    //btnSize = iconSize();

    connect(this, SIGNAL(clicked()), SLOT(change_state()));
}
void TwoStateButton::rotare(int gradus)
{
    rotareGradus = gradus;
    setIcon(rotareIcon(icon(), size(), rotareGradus));
}
/*void TwoStateButton::setIconSize2(const QSize& size)
{
    btnSize = size;
    setIconSize(btnSize);
}*/
void TwoStateButton::change_state(void)
{
    change_state_man_with_sig(!isOn);
}
void TwoStateButton::change_state_noEmit(void)
{
    change_state_man(!isOn);
}
void TwoStateButton::change_state_man(bool st)
{
    setChecked(st);
    setIcon(makeHugeIcon(st ? onstring : offstring));
    if (rotareGradus)
    {
        setIcon(rotareIcon(icon(), size(), rotareGradus));
    }
    //setIconSize(btnSize);
    isOn = st;
    emit state_changed_man(st);
}
void TwoStateButton::change_state_man_with_sig(bool st)
{
    change_state_man(st);
    emit state_changed(st);
    emit state_changed_man(st);
}

TwoStateButton2::TwoStateButton2(QString off, QString on, int iconSize)
{
    offstring = off;
    onstring = on;
    isOn = false;

    iconWidth = iconSize;
    setIcon(makeHugeIcon(offstring));
    setIconSize(QSize(iconSize, iconSize));

    connect(this, SIGNAL(clicked()), SLOT(change_state()));
}
void TwoStateButton2::change_state(void)
{
    change_state_man_with_sig(!isOn);
}
void TwoStateButton2::change_state_noEmit(void)
{
    change_state_man(!isOn);
}
void TwoStateButton2::change_state_man(bool st)
{
    setChecked(st);
    setIcon(makeHugeIcon(st ? onstring : offstring));
    setIconSize(QSize(iconWidth, iconWidth));
    isOn = st;
}
void TwoStateButton2::change_state_man_with_sig(bool st)
{
    change_state_man(st);
    emit state_changed(st);
}

bool TwoStateButton2::event(QEvent *e)
{
    if (e->type() == QEvent::Resize)
        setIconSize(QSize(iconWidth, iconWidth));

    return QPushButton::event(e);
}
