#include "edit.hpp"

#include <QShortcut>
#include <QLineEdit>
#include <QSettings>
#include <math.h>

static QPalette changedPalette;
Edit::Edit(double min, double max, int decimals) : prevValue(min)
{
    if (min <= 0 && max >= 0)
        prevValue = 0;

#ifndef _ADROID
    static int has_buttons = 1;
    if (has_buttons == 0)
#endif
        setButtonSymbols(NoButtons);

    setRange(min, max);
    setDecimals(decimals);
    setSingleStep(pow(10., -decimals));
    setValue(prevValue);

    changedPalette.setBrush(QPalette::Active, QPalette::Text, QColor(Qt::red));

    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);

    connect(this, SIGNAL(valueChanged(double)), this, SLOT(updated()));

    new QShortcut(QKeySequence(Qt::Key_Enter), this, SLOT(setNewValue()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_Return), this, SLOT(setNewValue()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_Escape), this, SLOT(resetValue()), nullptr, Qt::WidgetShortcut);
}

void Edit::setImmediateMode(bool on)
{
    if (immediateMode == on)
        return;

    foreach (QShortcut *sc, findChildren<QShortcut *>())
        if (sc->parent() == this &&
                (sc->key() == QKeySequence(Qt::Key_Enter) ||
                 sc->key() == QKeySequence(Qt::Key_Return)))
            delete sc;

    if ((immediateMode = on)) {
        new QShortcut(QKeySequence(Qt::Key_Enter), this, SLOT(resetValue()), nullptr, Qt::WidgetShortcut);
        new QShortcut(QKeySequence(Qt::Key_Return), this, SLOT(resetValue()), nullptr, Qt::WidgetShortcut);
    } else {
        new QShortcut(QKeySequence(Qt::Key_Enter), this, SLOT(setNewValue()), nullptr, Qt::WidgetShortcut);
        new QShortcut(QKeySequence(Qt::Key_Return), this, SLOT(setNewValue()), nullptr, Qt::WidgetShortcut);
    }
}

double Edit::value() const
{
    return prevValue;
}

QString Edit::textFromValue(double value) const
{
    //TODO fix for decimals >= 8
    QString str = QDoubleSpinBox::textFromValue(value);
    if (decimals() > 0) {
        str.remove(QRegExp("[0]*$")); //trailing zeroes
        str.remove(QRegExp("[.]$")); //trailing point
        str.remove(QRegExp("[,]$")); //trailing comma
    }
    return str;
}

void Edit::setNewValue(bool emitNewValue)
{
    prevValue = QDoubleSpinBox::value();

    if (emitNewValue)
    {
        emit newValue(int(prevValue));
        emit newValue(double(prevValue));
        emit newValue();
    }

    if (!immediateMode) {
        clearFocus();
        lineEdit()->setPalette(QPalette());
    }
}

double Edit::fixedValue(double value) const
{
    double step = pow(10., -decimals());
    long base = value;
    double delta = value - base;
    return base + qRound(delta / step) * step;
}

void Edit::setValue(double value)
{
    prevValue = qBound(minimum(), fixedValue(value), maximum());
    QDoubleSpinBox::setValue(prevValue);
}

void Edit::resetValue()
{
    setValue(prevValue);
    clearFocus();
    lineEdit()->setPalette(QPalette());
}

void Edit::stepBy(int steps)
{
    QDoubleSpinBox::stepBy(steps);

    lineEdit()->deselect();
    lineEdit()->setCursorPosition(lineEdit()->text().length());
}

void Edit::updated()
{
    if (immediateMode) {
        setNewValue(true);
    } else {
        //TODO fix for values == 0
        lineEdit()->setPalette( !qFuzzyCompare(QDoubleSpinBox::value(), prevValue) ? changedPalette : QPalette());
    }
}

void Edit::focusInEvent(QFocusEvent *event)
{
    emit editingStarted();

    QDoubleSpinBox::focusInEvent(event);
}

void Edit::focusOutEvent(QFocusEvent *event)
{
    setValue(prevValue);
    lineEdit()->setPalette(QPalette());

    QDoubleSpinBox::focusOutEvent(event);
}

QSize Edit::sizeHint() const
{
    QSize sz = QDoubleSpinBox::sizeHint();

    int pad = decimals() - 1;
    while (pad >= 0) { //count decimals already in QDSB hint
        double step = pow(10., -pad);
        bool round0 = fabs(minimum() / step - qRound(minimum() / step)) < 0.1;
        bool round1 = fabs(maximum() / step - qRound(maximum() / step)) < 0.1;
        if (!round1 || !round0)
            break;
        pad--;
    }
    pad = decimals() - pad - 1;
    if (pad > 0)
        sz.rwidth() += pad * fontMetrics().width('0') + fontMetrics().width('.');
    return sz;
}
