#include "nstslider.h"

#include <QtEvents>
#include <QShortcut>
#include <QApplication>
#include <QDesktopWidget>
#include <QGesture>
#include <QGestureEvent>
#include <QPainter>
#include <QPainterPath>
#include <qdrawutil.h>
#include "math.h"
#include "core/plot_core.h"

NSTSlider::NSTSlider(double min, double max, int decimals, QWidget *parent, bool vertical)
    : QwtSlider(parent,
                (vertical ? Qt::Vertical : Qt::Horizontal),
                (vertical ? QwtSlider::RightScale : QwtSlider::TopScale))
{
    prevZoom = 0;
    plot::TScale *s;
    setScaleDraw(s = new plot::TScale);
    setScaleEngine(new plot::TScaleDiv(s));

    stepV = pow(10., -decimals);
    minV = fixedValue(min);
    maxV = fixedValue(max);
    setRange(minV, maxV, stepV);

    pressed = 0;

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
    double scale = qApp->desktop()->logicalDpiX() / 101.;
#else
    double scale = qApp->desktop()->physicalDpiX() / 101.;
#endif

    double x = 12. * scale;
    double y = 2. * x;
#ifdef _ADROID
    x *= 1.5;
    y *= 1.2;
#endif

    setBorderWidth(1);
    if (vertical) {
        setHandleSize(y, x);
        setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);
    } else {
        setHandleSize(x, y);
        setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    }

    connect(this, SIGNAL(sliderPressed()), this, SLOT(slPressed()));
    connect(this, SIGNAL(sliderReleased()), this, SLOT(slUnPressed()));

    connect(this, SIGNAL(sliderReleased()), this, SLOT(setNewValue()));
    new QShortcut(QKeySequence(Qt::Key_Up), this, SLOT(stepUp()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_Down), this, SLOT(stepDown()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_Right), this, SLOT(stepUp()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_Left), this, SLOT(stepDown()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_PageUp), this, SLOT(pageUp()), nullptr, Qt::WidgetShortcut);
    new QShortcut(QKeySequence(Qt::Key_PageDown), this, SLOT(pageDown()), nullptr, Qt::WidgetShortcut);

    grabGesture(Qt::PinchGesture);
    grabGesture(Qt::TapAndHoldGesture);

    setFocusPolicy(Qt::StrongFocus);
}

double NSTSlider::fixedValue(double value) const
{
    long base = value;
    double delta = value - base;
    return base + qRound(delta / stepV) * stepV;
}

void NSTSlider::setValue(double new_value)
{
    QwtSlider::setValue(fixedValue(new_value));
}

void NSTSlider::setNewValue()
{
    emit newValue(int(value));
    emit newValue(double(value));
    emit newValue();
}

bool NSTSlider::event(QEvent *event)
{
    if (event->type() == QEvent::Gesture) {
        return gestureEvent(static_cast<QGestureEvent*>(event));
    }
    return QwtSlider::event(event);
}

void NSTSlider::mousePressEvent(QMouseEvent *event)
{
    pos = event->pos();

    if (getScrollMode(event->pos()))
        QwtSlider::mousePressEvent(event);
}

void NSTSlider::mouseMoveEvent(QMouseEvent *event)
{
    if (pressed == 1) {
        QwtSlider::mouseMoveEvent(event);
        return;
    }
    if (pressed)
        return;

    double delta = maxValue - minValue;
    if (orientation == Qt::Horizontal)
        delta *= 1.0 * (event->pos() - pos).x() / width();
    else
        delta *= -1.0 * (event->pos() - pos).y() / height();

    if (fabs(delta) < stepV)
        return;
    delta = fixedValue(delta);
    pos = event->pos();

    double min = minV, max = maxV;
    double newMin = minValue - delta;
    double newMax = maxValue - delta;
    if (minV > maxV) {
        qSwap(newMin, newMax);
        min = maxV;
        max = minV;
    }

    if (newMin < min) {
        newMax += (min - newMin);
        newMin = min;
    }
    if (newMax > max) {
        newMin -= (newMax - max);
        newMax = max;
        if (newMin < min)
            newMin = min;
    }

    newMin = fixedValue(newMin);
    newMax = fixedValue(newMax);

    if ((value < newMin) || (value > newMax))
        return;

    if (minV > maxV)
        qSwap(newMin, newMax);

    setRange(newMin, newMax, stepV);
}

void NSTSlider::setMainRangeAndStep(double min, double max, double step)
{
    if ((minV != min) || (maxV != max) || (stepV != step)) {
        stepV = step;
        minV = fixedValue(min);
        maxV = fixedValue(max);

        setRange(minV, maxV, stepV);
        emit rangeStepChanged(minV, maxV, stepV);
    }
}

#if (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
#include <QInputContext>
#endif
void NSTSlider::mouseDoubleClickEvent(QMouseEvent *)
{
    if (pressed == 1)
        return;

#if (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
    if (qApp->inputContext()) {
        QEvent ev(QEvent::RequestSoftwareInputPanel);
        qApp->inputContext()->filterEvent(&ev);
    }
#endif
    setRange(minV, maxV, stepV);
}

void NSTSlider::wheelEvent(QWheelEvent *event)
{
    event->accept();

    if (pressed == 1)
        return;

    int steps = qCeil(event->delta() / 8. / 15.);

    if (event->modifiers() & Qt::ControlModifier && (event->modifiers() & (Qt::ShiftModifier | Qt::AltModifier)))
        steps *= 10;
    if (event->modifiers() & (Qt::ShiftModifier | Qt::ControlModifier | Qt::AltModifier)) {
        stepBy(steps);
        return;
    }

    double toMin = fabs(value - minValue);
    double toMax = fabs(maxValue - value);
    double min = minV, max = maxV;
    if (minV > maxV) {
        qSwap(toMin, toMax);
        min = maxV;
        max = minV;
    }

    double newMin, newMax;

    //zoom in
    if (steps > 0) {
        newMin = value - toMin / 2;
        newMax = value + toMax / 2;

        if (newMax - newMin < 5*stepV)
            return;
    } else {
        newMin = value - toMin * 2;
        newMax = value + toMax * 2;

        if (newMin < min) {
            newMax += (min - newMin);
            newMin = min;
        }
        if (newMax > max) {
            newMin -= (newMax - max);
            newMax = max;
            if (newMin < min)
                newMin = min;
        }
    }

    newMin = fixedValue(newMin);
    newMax = fixedValue(newMax);

    if (minV > maxV)
        qSwap(newMin, newMax);

    setRange(newMin, newMax, stepV);
}

bool NSTSlider::gestureEvent(QGestureEvent *event)
{
    if (pressed == 1)
        return false;

    if (QGesture *pinch = event->gesture(Qt::PinchGesture)) {
        QPinchGesture *p = (QPinchGesture *)pinch;

        if (p->state() == Qt::GestureStarted)
            prevZoom = 0;

        double scale = log(p->totalScaleFactor()) / log(2.) * 2. - prevZoom;
        if (fabs(scale) > 1.)
            prevZoom += (scale > 0 ? floor(scale) : ceil(scale));

        for (int i = 0; i < floor(fabs(scale)); i++) {
            QWheelEvent wev(QPoint(), scale * 8.*15., Qt::MiddleButton, Qt::NoModifier);
            wheelEvent(&wev);
        }
        pressed = 2;
    }
    bool fin = true;
    foreach (QGesture *g, event->gestures())
        if (g->state() != Qt::GestureFinished && g->state() != Qt::GestureCanceled) {
            fin = false;
            break;
        }
    if (fin)
        pressed = 0;

    return true;
}

void NSTSlider::resizeEvent(QResizeEvent *e)
{
    QwtSlider::resizeEvent(e);
    rangeChange();
}

/*void NSTSlider::drawSlider (QPainter *p, const QRect &r) const
{
    QwtSlider::drawSlider(p, r);
    //color drawing is in drawHandle, as this line will draw over it
}*/
#define SLIDER_COLOR QColor(0, 0, 255)
void NSTSlider::drawHandle(QPainter *p, const QRect &r, int pos) const
{
    int ws = 4;
    int ds = handleSize().width() / 2 - 4;
    if ( ds < 1 )
        ds = 1;

    QRect rSlot;
    if ( orientation == Qt::Horizontal )
    {
        if ( r.height() & 1 )
            ws++;

        rSlot = QRect( r.x() + ds,
                r.y() + ( r.height() - ws ) / 2,
                r.width() - 2 * ds, ws );
        if (minV < maxV)
            rSlot.setRight(pos);
        else
            rSlot.setLeft(pos);
    }
    else
    {
        if ( r.width() & 1 )
            ws++;

        rSlot = QRect( r.x() + ( r.width() - ws ) / 2,
                       r.y() + ds,
                       ws, r.height() - 2 * ds );
        if (minV < maxV)
            rSlot.setTop(pos);
        else
            rSlot.setBottom(pos);
    }

    QColor clr = isEnabled() ? SLIDER_COLOR : SLIDER_COLOR.lighter();
    QBrush brush = QBrush(clr);
    if (rSlot.isValid())
    qDrawShadePanel( p, rSlot, palette(), true, 1 , &brush );


    const int bw = borderWidth();

    //additional marks
    for (int i = 0; i < marks.size(); i++) {
        QPair<double, QColor> v = marks.at(i);
        p->setPen(v.second);
        int pos = transform(v.first);
        if ( orientation == Qt::Horizontal )
            p->drawLine(pos, r.top() + bw, pos, r.bottom() - bw);
        else // Vertical
            p->drawLine(r.left() + bw, pos, r.right() - bw, pos);
    }


    QwtSlider::drawHandle(p, r, pos);


    p->setPen(clr);

    //pos++; // shade line points one pixel below
    if ( orientation == Qt::Horizontal )
        p->drawLine(pos, r.top() + bw, pos, r.bottom() - bw);
    else // Vertical
        p->drawLine(r.left() + bw, pos, r.right() - bw, pos);
}

QSize NSTSlider::minimumSizeHint() const
{
    return QwtSlider::minimumSizeHint();
}
