#include "status.hpp"

#include <QApplication>
#include <QDesktopWidget>
#include <QSettings>
#include <QHBoxLayout>
#include <QStatusBar>
#include <QPainter>
#include <QPainterPath>

void StatusWidget::set(Shape shape, QBrush brush, QPixmap custom)
{
    if (shape == Custom) {
        setPixmap(custom);
        return;
    }

    QPixmap pix(size());
    QPainter p(&pix);
    if (!p.isActive())
        return;

    p.fillRect(pix.rect(), palette().background());

    int width = pix.width() / 5;
    QRect rect = pix.rect().adjusted(width, width, -width, -width);

    p.setBrush(brush);
    p.setPen(QPen(brush, 0));

    QPointF pts[3];

    switch (shape) {
    case Square:
        p.drawRect(rect);
        break;
    case Circle:
        p.setPen(QPen(brush, 3)); //width < 3 = polygon instead of circle
        p.drawEllipse(rect);
        break;
    case Cross:
        p.setPen(QPen(brush, width));
        p.drawLine(rect.topLeft(), rect.bottomRight());
        p.drawLine(rect.topRight(), rect.bottomLeft());
        break;
    case Play:
        pts[0] = rect.topLeft();
        pts[2] = rect.bottomLeft();
        pts[1].rx() = rect.right();
        pts[1].ry() = rect.center().y();
        p.drawPolygon(pts, 3);
        break;
    case Up:
        pts[0] = rect.bottomLeft();
        pts[2] = rect.bottomRight();
        pts[1].rx() = rect.center().x();
        pts[1].ry() = rect.top();
        p.drawPolygon(pts, 3);
        break;
    case Down:
        pts[0] = rect.topLeft();
        pts[2] = rect.topRight();
        pts[1].rx() = rect.center().x();
        pts[1].ry() = rect.bottom();
        p.drawPolygon(pts, 3);
        break;
    }

    setPixmap(pix);
}

StatusWidget1::StatusWidget1(QString text)
{
    label = new QLabel(text);
    label->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

    //TODO global get_dpi, get_buttonsize funcs
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
    double scale = qApp->desktop()->logicalDpiX() / 101.;
#else
    double scale = qApp->desktop()->physicalDpiX() / 101.;
#endif
    static int ICO_SZ = -1;
    if (ICO_SZ == -1) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        ICO_SZ = qMin(settings.value("buttonsize", 25).toInt(), 512);
    }

    int sz = qMax(ICO_SZ * 2 / 3., scale * 20);
    status = new StatusWidget(sz);

    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->setSpacing(3);
    layout->addWidget(label);
    layout->addWidget(status);

    layout->setContentsMargins(3, 0, 3, 0);
}

//----------------------------------------------


QList<StatusBar *> statusList; //use only in gui thread
QStatusBar *stbar = nullptr;
void setMainStatusBar(QStatusBar *bar)
{
    stbar = bar;
}
int statusBarCount()
{
    return statusList.count();
}
void cancelAllStatusBars()
{
    for (int i = 0; i < statusList.size(); i++)
        statusList[i]->cancel->click();
}

StatusBar::StatusBar(QString status, int max)
{
    statusList.push_back(this);

    setStyleSheet("QToolBar { spacing: 6px; padding: 3px; }");

    statusLabel = new QLabel(status);

    cancel = new QPushButton(tr("Cancel"));
    connect(cancel, SIGNAL(clicked()), SIGNAL(canceled()));
    connect(cancel, SIGNAL(clicked()), SLOT(deleteLater()));

    bar = new QProgressBar;
    bar->setUpdatesEnabled(false);
    bar->setValue(0);
    bar->setMaximum(max);
    bar->setFormat("%p% : %v / %m");
    bar->setUpdatesEnabled(true);

    addWidget(statusLabel);
    addWidget(cancel);
    addWidget(bar);
}

StatusBar::~StatusBar()
{
    stbar->removeWidget(this);
    statusList.removeAll(this);
}

void StatusBar::add()
{
    stbar->addWidget(this, this->sizeHint().width());
}

void StatusBar::setValue(int val)
{
    bar->setUpdatesEnabled(false);
    bar->setValue(val);
    bar->setUpdatesEnabled(true);
}


#include <QGraphicsOpacityEffect>
#include <QPropertyAnimation>
#include <QTimer>
Notification::Notification(QString status, Tint tint, int timeout) : StatusBar(QString())
{
    statusList.removeAll(this);

    clear();
    bar = nullptr;
    statusLabel = nullptr;
    cancel = new QPushButton(status);
    cancel->setToolTip(status + "\n" + tr("Click to dismiss"));
    cancel->setFlat(true);
    cancel->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Minimum);
    connect(cancel, SIGNAL(clicked()), SLOT(hide()));
    connect(cancel, SIGNAL(clicked()), SLOT(deleteLater()));
    addWidget(cancel);

    if (timeout) {
        QGraphicsOpacityEffect *g = new QGraphicsOpacityEffect(this);
        g->setOpacity(1);
        setGraphicsEffect(g);
        QPropertyAnimation *anim = new QPropertyAnimation(g, "opacity", this);
        anim->setDuration(2000);
        anim->setEndValue(0);

        QTimer *t = new QTimer(this);
        t->setSingleShot(true);
        //connect(t, SIGNAL(timeout()), cancel, SLOT(click()));
        connect(t, SIGNAL(timeout()), anim, SLOT(start()));
        connect(anim, SIGNAL(finished()), cancel, SLOT(click()));
        t->start(timeout * 1000);
    }

    setTint(tint);

    add();
}
void Notification::setTint(Tint tint)
{
    if (tint == None)
        return;

    setStyleSheet(styleSheet() +
                  "QToolBar { background-color: " +
                  (tint == Warn ? "#FFA500" : tint == Bad ? "#FF6347" : "#ADFF2F") +
                  "; }");
}
