#include "viewer3d.hpp"
#include <QVector3D>
#include <QtEvents>

#include <math.h>
#if defined(Q_WS_MAC)
#include <OpenGL/glu.h>
#else
#ifndef _ADROID
#include <GL/glu.h>
#endif
#endif

T3dViewer::T3dViewer(QWidget *parent) : QGLWidget(QGLFormat(QGL::SampleBuffers),
                                                  parent)
#ifndef NO_GLES_3D
    , program(context()), textprogram(context())
#endif
{
    _x_rot = _y_rot = _z_rot = 0.;
    _x_pos = _y_pos = _z_pos = 0.;
    setAttribute(Qt::WA_DeleteOnClose, true);
    setWindowTitle("3D View");

    itexture = -1;
    ititletexture = -1;
    scale = 1;
    z_scale = 1;
    xyaspect = 1;
    bgColor = QColor(227, 227, 236);//Qt::black;
    fgColor = QColor(227, 227, 236);//Qt::white;
    showAxes = false;
    invertY = false;
    for (int i = 0; i < 3; i++)
        iaxistexture[i] = -1;

    mousemode = 0;
    pressed = false;
    x_rot = y_rot = z_rot = 0;
    x_pos = y_pos = z_pos = 0;
#ifdef NO_GLES_3D
    m_glListId = -1;
    m_axesLists = -1;
#else
    m_glVboId = -1;
    m_glIndicesId = -1;
    m_glBoxIndices = -1;
    m_glFSQuadIndices = -1;
    vboSize = 0;
#endif
}

T3dViewer::~T3dViewer()
{
    makeCurrent();
    if (itexture != -1)
        deleteTexture(itexture);
    if (ititletexture != -1)
        deleteTexture(ititletexture);
    for (int i = 0; i < 3; i++)
        if (iaxistexture[i] != -1)
            deleteTexture(iaxistexture[i]);
#ifdef NO_GLES_3D
    if (glIsList(m_glListId))
        glDeleteLists(m_glListId, 1);
    if (glIsList(m_axesLists))
        glDeleteLists(m_axesLists, 1);
#else
    if (m_glVboId != -1) {
        GLuint buffers[4];
        buffers[0] = m_glVboId;
        buffers[1] = m_glIndicesId;
        buffers[2] = m_glBoxIndices;
        buffers[3] = m_glFSQuadIndices;
        glDeleteBuffers(4, buffers);
    }
    program.removeAllShaders();
    textprogram.removeAllShaders();
#endif
}

void T3dViewer::initializeGL()
{
    glClearColor(fgColor.redF(), fgColor.greenF(), fgColor.blueF(), 0.0);
    glEnable(GL_DEPTH_TEST);
    //glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
    //glPolygonMode( GL_FRONT_AND_BACK, GL_POLYGON );

#ifdef NO_GLES_3D
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    GLfloat global_ambient[] = { 0.5f, 0.5f, 0.5f, 1.0f };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

    GLfloat specular[] = {1.0f, 1.0f, 1.0f , 1.0f};
    glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

    GLfloat diffuse[] = {1.0f, 1.0f, 1.0f , 1.0f};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);

    GLfloat ambient[] = {0.0f, 0.0f, 0.0f, 0.0f};
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);

    GLfloat position[] = {3.f, -1.f, -2.f, 0.0f};
    glLightfv(GL_LIGHT0, GL_POSITION, position);

    float mcolor[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mcolor);
    //glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    //glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.0f);

    glFrontFace(GL_CCW);
    //glEnable(GL_COLOR_MATERIAL);
    //glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
#else
    initializeGLFunctions();

    QString vs = //"precision highp float;"
            "uniform mat4 mvp_matrix;"
            //"uniform mat4 mv_matrix;"
            "uniform mat4 n_matrix;"
            "attribute vec3 a_position;"
            "attribute vec3 a_normal;"
            "varying vec3 v_ecNormal;"
            //"varying vec3 v_ecVertex;"
            "varying vec2 v_texcoord;"
            "void main() {"
            "gl_Position = mvp_matrix * vec4(a_position, 1.0);"
            "v_ecNormal = normalize(vec3(n_matrix * vec4(a_normal, 0.0)));"
            //"v_ecVertex = vec3(mv_matrix * vec4(a_position, 0.0));"
            "v_texcoord = a_position.xy/2. + .5;"
            "}";
    QString ps = //"precision highp float;"
            "struct Material { float cAmb; float cDiff; float cSpec; float shininess; };"
            "uniform vec4 u_lightPosDir;"
            "uniform Material u_material;"
            "varying vec3 v_ecNormal;"
            //"varying vec3 v_ecVertex;"
            "varying vec2 v_texcoord;"
            "uniform sampler2D texture;"
            "uniform vec4 u_colorOverride;"
            "void main() {"
            "if (u_colorOverride.w > 0.0) {"
            "gl_FragColor = vec4(u_colorOverride.rgb, 1.0);"
            "return; }"
            "vec3 lightdir = u_lightPosDir.xyz;"
            //"vec3 lightdir = normalize(u_lightPosDir.w > 0 ? u_lightPosDir.xyz - v_ecVertex : u_lightPosDir.xyz);"
            "float ecNormalDotLightDirection = max(0.0, dot(v_ecNormal, lightdir));"
            //"float ecNormalDotLightHalfplane = max(0.0, reflect(lightdir, v_ecNormal).z);"
            "float ecNormalDotLightHalfplane = max(0.0, dot(v_ecNormal, normalize(vec3(0.0, 0.0, 1.0) + lightdir)));"
            "gl_FragColor = texture2D(texture, v_texcoord) * (u_material.cAmb + ecNormalDotLightDirection * u_material.cDiff);"
            "if (ecNormalDotLightDirection != 0.)"
            "gl_FragColor.rgb += vec3(1, 1, 1) * u_material.cSpec * pow(ecNormalDotLightHalfplane, u_material.shininess);"
            //"gl_FragColor.rgb = v_ecNormal;"
            "}";
    program.addShaderFromSourceCode(QGLShader::Vertex, vs);
    program.addShaderFromSourceCode(QGLShader::Fragment, ps);
    program.link();

    QString vst = //"precision highp float;"
            "attribute vec3 a_position;"
            "uniform vec4 u_rect;"
            "varying vec2 v_texcoord;"
            "void main() {"
            "gl_Position = vec4(u_rect.x + a_position.x * u_rect.z, u_rect.y + a_position.y * u_rect.w, 0.0, 1.0);"
            "v_texcoord = a_position.xy/2. + .5;"
            "}";
    QString pst = //"precision highp float;"
            "varying vec2 v_texcoord;"
            "uniform sampler2D texture;"
            "uniform vec3 u_color;"
            "void main() {"
            "float alpha = texture2D(texture, v_texcoord).a;"
            //"if (alpha < 0.01) discard;"
            "gl_FragColor = vec4(u_color, alpha);"
            "}";
    textprogram.addShaderFromSourceCode(QGLShader::Vertex, vst);
    textprogram.addShaderFromSourceCode(QGLShader::Fragment, pst);
    textprogram.link();

    if (m_glVboId == -1) {
        GLuint buffers[4];
        glGenBuffers(4, buffers);
        m_glVboId = buffers[0];
        m_glIndicesId = buffers[1];
        m_glBoxIndices = buffers[2];
        m_glFSQuadIndices = buffers[3];
    }
#endif
}

void T3dViewer::resizeGL(int w, int h)
{
    makeCurrent();

    glViewport(0, 0, (GLint)w, (GLint)h);

    qreal aspect = (qreal)w / (h ? h : 1);
    const qreal zNear = 1.0, zFar = 100.0, fov = 40.0 / scale;

#ifdef NO_GLES_3D
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //glOrtho(-1/scale, 1/scale, -1/scale, 1/scale, zNear, zFar);
    gluPerspective(fov, aspect, zNear, zFar);
    //glFrustum(-10/scale, 10/scale, -1/scale, 1/scale, zNear, zFar);
    //glScalef(1.*h/w, 1, 1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
#endif

    projection.setToIdentity();
    //projection.ortho(-1/scale, 1/scale, -1/scale, 1/scale, zNear, zFar);
    projection.perspective(fov, aspect, zNear, zFar);
    projection.translate(0, 0, -2);
}

struct VertexData
{
    short position[3];
    short normal[3];
};
int zcmp(const void *p1, const void *p2)
{
    //if (((QVector3D *)p1)->z() == ((QVector3D *)p2)->z())
    //    return 0;
    return (((QVector3D *)p1)->z() >= ((QVector3D *)p2)->z()) ? 1 : -1;
}
#include <QVector2D>
void T3dViewer::paintGL()
{
    glClearColor(bgColor.redF(), bgColor.greenF(), bgColor.blueF(), 1.f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /*
    QPainter p(this);
    p.setPen(Qt::white);
    p.drawText(0.3*width(), 0.1*height(), QString("Title"));
    p.end();
    */

    if (itexture != -1) {
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, itexture);
    } else {
        glDisable(GL_TEXTURE_2D);
    }

    QMatrix4x4 matrix;
    matrix.lookAt(QVector3D(0, -2, 0), QVector3D(0, 0, 0), QVector3D(0, 0, 1));
#ifdef NO_GLES_3D
    matrix.scale(2);
#endif
    matrix.translate(x_pos, z_pos, y_pos);
#ifdef NO_GLES_3D
    matrix.scale(0.5);
#endif
    matrix.rotate(x_rot, 1, 0, 0);
    matrix.rotate(y_rot, 0, 1, 0);
    matrix.rotate(z_rot, 0, 0, 1);
    matrix.scale(xyaspect, 1/xyaspect * (invertY ? -1 : 1), z_scale);

#ifdef NO_GLES_3D
    glLoadIdentity();
    gluLookAt(0, -2, 0, 0, 0, 0, 0, 0, 1);

    glTranslatef(x_pos, z_pos, y_pos);
    glRotatef(x_rot, 1, 0, 0);
    glRotatef(y_rot, 0, 1, 0);
    glRotatef(z_rot, 0, 0, 1);

    glScalef(xyaspect, 1/xyaspect * (invertY ? -1 : 1), z_scale);

    glEnable(GL_LIGHTING);

    if (m_glListId != -1)
        glCallList(m_glListId);

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glColor3f(fgColor.redF(), fgColor.greenF(), fgColor.blueF());
    glLineWidth(2);

     if (showAxes && m_axesLists != -1)
        glCallList(m_axesLists);
#else
    program.bind();
    program.setUniformValue("texture", 0);

    QMatrix4x4 matrix2 = matrix.inverted().transposed(); //normal matrix (inverse transpose of the modelview matrix)

    //program.setUniformValue("mv_matrix", matrix);
    program.setUniformValue("n_matrix", matrix2);
    program.setUniformValue("mvp_matrix", projection * matrix);

    program.setUniformValue("u_colorOverride", QVector4D(fgColor.redF(), fgColor.greenF(), fgColor.blueF(), 0));
    program.setUniformValue("u_lightPosDir", QVector4D(3, -1, -2, 0));
    program.setUniformValue("u_material.cAmb", (GLfloat).5);
    program.setUniformValue("u_material.cDiff", (GLfloat).25);
    program.setUniformValue("u_material.cSpec", (GLfloat).5);
    program.setUniformValue("u_material.shininess", (GLfloat)5);

    glBindBuffer(GL_ARRAY_BUFFER, m_glVboId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_glIndicesId);

    int vertexLocation = program.attributeLocation("a_position");
    program.enableAttributeArray(vertexLocation);
    program.setAttributeBuffer(vertexLocation, GL_SHORT, 0, 3, sizeof(VertexData));

    int normalLocation = program.attributeLocation("a_normal");
    program.enableAttributeArray(normalLocation);
    program.setAttributeBuffer(normalLocation, GL_SHORT, 3*sizeof(short), 3, sizeof(VertexData));

    glDrawElements(GL_TRIANGLE_STRIP, vboSize, GL_UNSIGNED_INT, nullptr);

    if (showAxes) {
        glDisable(GL_TEXTURE_2D);
        program.setUniformValue("u_colorOverride", QVector4D(fgColor.redF(), fgColor.greenF(), fgColor.blueF(), 1));
        glLineWidth(2);

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_glBoxIndices);
        glDrawElements(GL_LINES, 12*2, GL_UNSIGNED_INT, nullptr);
    }

    program.release();
#endif

    if (ititletexture != -1) {
        titlerect.moveTo(0.5 - titlerect.width() / 2. / width(),
                         0.95);
        drawGuiTexture(ititletexture, titlerect);
    }

    if (showAxes) {
        QVector3D pos[3][4] = {
            {QVector3D(0, 1, 1), QVector3D(0, -1, 1), QVector3D(0, 1, -1), QVector3D(0, -1, -1)},
            {QVector3D(1, 0, 1), QVector3D(-1, 0, 1), QVector3D(1, 0, -1), QVector3D(-1, 0, -1)},
            {QVector3D(1, 1, 0), QVector3D(-1, 1, 0), QVector3D(1, -1, 0), QVector3D(-1, -1, 0)}
        };
        QMatrix4x4 mvp = projection * matrix;
        for (int i = 0; i < 3; i++)
            if (iaxistexture[i] != -1) {
                for (int j = 0; j < 4; j++)
                    pos[i][j] = (mvp * QVector4D(pos[i][j], 1)).toVector3DAffine();
                qsort(pos[i], 4, sizeof(QVector3D), zcmp);

                QPointF halfrect(axisrect[i].width() / 2. / width(), axisrect[i].height() / 2. / height());
                QPointF extrude_ajust = (QVector2D(halfrect) *
                                         (pos[i][1].toVector2D() - (mvp * QVector4D(0, 0, 0, 1))
                                         .toVector3DAffine().toVector2D()).normalized() * 1.4).toPointF();
                axisrect[i].moveTo((pos[i][1].toPointF() + QPointF(1, 1)) / 2 - halfrect + extrude_ajust);

                drawGuiTexture(iaxistexture[i], axisrect[i]);
            }
    }

}
void T3dViewer::drawGuiTexture(int texture, QRectF rect0)
{
    //pixel rounding prevents ugly half-pixel aliasing/flickering
    QRectF rect(qRound(rect0.left() * width()) * 1. / width(),
                qRound(rect0.top() * height()) * 1. / height(),
                rect0.width() * 1.  / width(),
                rect0.height() * 1. / height());
    float x0 = -1. + 2. * rect.left() + rect.width(); //map to -1..1 range
    float y0 = -1. + 2. * rect.top()  + rect.height();

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texture);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glDisable(GL_DEPTH_TEST);

#ifdef NO_GLES_3D
    //glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    glTranslatef(x0, y0, 0.0f);
    glScalef(rect.width(), rect.height(), 1.0f);

    glBegin(GL_TRIANGLE_STRIP);

    glTexCoord2d(0, 0);
    glVertex3f(-1, -1, -1);

    glTexCoord2d(0, 1);
    glVertex3f(-1, 1, -1);

    glTexCoord2d(1, 0);
    glVertex3f(1, -1, -1);

    glTexCoord2d(1, 1);
    glVertex3f(1, 1, -1);

    glEnd();

    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
#else
    textprogram.bind();
    textprogram.setUniformValue("texture", 0);
    textprogram.setUniformValue("u_color", QVector3D(fgColor.redF(), fgColor.greenF(), fgColor.blueF()));
    textprogram.setUniformValue("u_rect", QVector4D(x0, y0, rect.width(), rect.height()));

    glBindBuffer(GL_ARRAY_BUFFER, m_glVboId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_glFSQuadIndices);

    int vertexLocation = textprogram.attributeLocation("a_position");
    textprogram.enableAttributeArray(vertexLocation);
    textprogram.setAttributeBuffer(vertexLocation, GL_SHORT, 0, 3, sizeof(VertexData));

    glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_INT, nullptr);

    textprogram.release();
#endif

    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
}

void T3dViewer::updateData(float *z_values, int nx, int ny, float k)
{
    makeCurrent();

    xyaspect = sqrt(k);

    //TODO 2x resolution??
    QVector3D *temp = new QVector3D[nx*ny];
    QVector3D *normals = new QVector3D[nx*ny];

    for (int i = 0; i < nx; i++) {
        for (int j = 0; j < ny; j++) {
            temp[i*ny + j].setX((-0.5 + 1.*(i)/nx));
            temp[i*ny + j].setY((0.5 - 1.*(j)/ny));
            temp[i*ny + j].setZ(z_values[(i)*ny + (j)]);
        }
    }

    for (int i = 1; i < nx-1; i++) {
        for (int j = 1; j < ny-1; j++) {
            normals[i*ny + j] += QVector3D::normal(temp[i*ny + j], temp[(i+1)*ny + (j)], temp[(i)*ny + (j+1)]) / 4;
            normals[i*ny + j] += QVector3D::normal(temp[i*ny + j], temp[(i)*ny + (j+1)], temp[(i-1)*ny + (j)]) / 4;
            normals[i*ny + j] += QVector3D::normal(temp[i*ny + j], temp[(i-1)*ny + (j)], temp[(i)*ny + (j-1)]) / 4;
            normals[i*ny + j] += QVector3D::normal(temp[i*ny + j], temp[(i)*ny + (j-1)], temp[(i+1)*ny + (j)]) / 4;
        }
    }
    for (int i = 0; i < nx; i++) { int j = 0;
        normals[i*ny + j] = normals[i*ny + j+1]; j = ny-1;
        normals[i*ny + j] = normals[i*ny + j-1]; }
    for (int j = 0; j < ny; j++) { int i = 0;
        normals[i*ny + j] = normals[(i+1)*ny + j]; i = nx-1;
        normals[i*ny + j] = normals[(i-1)*ny + j]; }

#ifdef NO_GLES_3D
    if (glIsList(m_glListId))
        glDeleteLists(m_glListId, 1);

    m_glListId = glGenLists(1);
    glNewList(m_glListId, GL_COMPILE);
    glBegin(GL_TRIANGLES);

    int idx;
    for (int i = 0; i < nx-1; i++) {
        for (int j = 0; j < ny-1; j++) {
            idx = i*ny + j;
            glTexCoord2d(1.*(i)/nx, 1 - 1.*(j)/ny);
            glNormal3f(normals[idx].x(), normals[idx].y(), normals[idx].z());
            glVertex3f(temp[idx].x(), temp[idx].y(), temp[idx].z());

            idx = (i+1)*ny + j;
            glTexCoord2d(1.*(i+1)/nx, 1 - 1.*(j)/ny);
            glNormal3f(normals[idx].x(), normals[idx].y(), normals[idx].z());
            glVertex3f(temp[idx].x(), temp[idx].y(), temp[idx].z());

            idx = (i+1)*ny + (j+1);
            glTexCoord2d(1.*(i+1)/nx, 1 - 1.*(j+1)/ny);
            glNormal3f(normals[idx].x(), normals[idx].y(), normals[idx].z());
            glVertex3f(temp[idx].x(), temp[idx].y(), temp[idx].z());

            //

            idx = i*ny + j;
            glTexCoord2d(1.*(i)/nx, 1 - 1.*(j)/ny);
            glNormal3f(normals[idx].x(), normals[idx].y(), normals[idx].z());
            glVertex3f(temp[idx].x(), temp[idx].y(), temp[idx].z());

            idx = (i+1)*ny + (j+1);
            glTexCoord2d(1.*(i+1)/nx, 1 - 1.*(j+1)/ny);
            glNormal3f(normals[idx].x(), normals[idx].y(), normals[idx].z());
            glVertex3f(temp[idx].x(), temp[idx].y(), temp[idx].z());

            idx = i*ny + (j+1);
            glTexCoord2d(1.*(i)/nx, 1 - 1.*(j+1)/ny);
            glNormal3f(normals[idx].x(), normals[idx].y(), normals[idx].z());
            glVertex3f(temp[idx].x(), temp[idx].y(), temp[idx].z());
        }
    }

    glEnd();
    glEndList();

    if (!glIsList(m_axesLists)) {
        m_axesLists = glGenLists(1);
        glNewList(m_axesLists, GL_COMPILE);
        glBegin(GL_LINES);

        for (int i = 0; i < 3; i++)
            for (int j = 0; j < 4; j++) {
                glVertex3f((i == 0 ? -.5 : j % 2 == 0 ? -.5 : .5),
                           (i == 1 ? -.5 : j / 2 == 0 ? -.5 : .5),
                           (i == 2 ? -.5 : (i == 1 ? j / 2 : j % 2) == 0 ? -.5 : .5));
                glVertex3f((i == 0 ? .5 : j % 2 == 0 ? -.5 : .5),
                           (i == 1 ? .5 : j / 2 == 0 ? -.5 : .5),
                           (i == 2 ? .5 : (i == 1 ? j / 2 : j % 2) == 0 ? -.5 : .5));
            }

        glEnd();
        glEndList();
    }
#else
    vboSize = (nx*2 + 2)*(ny - 1);
    VertexData *vertices = new VertexData[nx*ny+8];
    GLuint *indices = new GLuint[vboSize];
    int ind_i = 0;
    for (int j = 0; j < ny; j++) {
        for (int i = 0; i < nx; i++) {
            vertices[j*nx + i].position[0] = 65535. * temp[i*ny + j].x();
            vertices[j*nx + i].position[1] = 65535. * temp[i*ny + j].y();
            vertices[j*nx + i].position[2] = 65535. * temp[i*ny + j].z();
            vertices[j*nx + i].normal[0] = 32767. * normals[i*ny + j].x();
            vertices[j*nx + i].normal[1] = 32767. * normals[i*ny + j].y();
            vertices[j*nx + i].normal[2] = 32767. * normals[i*ny + j].z();
        }
    }
    for (int i = 0; i < 2; i++) //box
        for (int j = 0; j < 2; j++)
            for (int k = 0; k < 2; k++) {
                vertices[k*4 + j*2 + i + nx*ny].position[0] = i ? 32767 : -32766;
                vertices[k*4 + j*2 + i + nx*ny].position[1] = j ? 32767 : -32766;
                vertices[k*4 + j*2 + i + nx*ny].position[2] = k ? 32767 : -32766;
            }
    for (int j = 0; j < ny-1; j++) {
        for (int i = 0; i < nx; i++) {
            indices[ind_i++] = j*nx + i;
            indices[ind_i++] = (j+1)*nx + i;
        }
        indices[ind_i++] = (j+2)*nx - 1;
        indices[ind_i++] = (j+1)*nx;
    }

    initializeGLFunctions();
    if (m_glVboId == -1) {
        GLuint buffers[4];
        glGenBuffers(4, buffers);
        m_glVboId = buffers[0];
        m_glIndicesId = buffers[1];
        m_glBoxIndices = buffers[2];
        m_glFSQuadIndices = buffers[3];
    }

    glBindBuffer(GL_ARRAY_BUFFER, m_glVboId);
    glBufferData(GL_ARRAY_BUFFER, (nx * ny +8) * sizeof(VertexData), vertices, GL_STATIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_glIndicesId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, vboSize * sizeof(GLuint), indices, GL_STATIC_DRAW);

    GLuint box[12*2] = {
        0, 4,
        1, 5,
        2, 6,
        3, 7,

        0, 2,
        1, 3,
        4, 6,
        5, 7,

        0, 1,
        2, 3,
        4, 5,
        6, 7
    };
    for (int i = 0; i < 12*2; i++)
        box[i] += nx*ny;
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_glBoxIndices);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 12*2 * sizeof(GLuint), box, GL_STATIC_DRAW);

    GLuint quad[4] = {0, 1, 2, 3};
    for (int i = 0; i < 4; i++)
        quad[i] += nx*ny;
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_glFSQuadIndices);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * sizeof(GLuint), quad, GL_STATIC_DRAW);

    delete [] vertices;
    delete [] indices;
#endif

    delete [] temp;
    delete [] normals;
}

void T3dViewer::updateTexture(QImage texture)
{
    makeCurrent();

    if (itexture != -1)
        deleteTexture(itexture);

    itexture = bindTexture(texture);
}


void T3dViewer::setTitle(QString title, QFont font)
{
    makeCurrent();

    if (ititletexture != -1)
        deleteTexture(ititletexture);
    ititletexture = -1;

    if (title.isEmpty())
        return;

    QImage tex = textureFromString(title, font);
    titlerect.setSize(tex.size());
    ititletexture = bindTexture(tex);
}
void T3dViewer::setAxisTitle(int idx, QString title, QFont font)
{
    if (idx < 0 || idx >= 3)
        return;

    makeCurrent();

    if (iaxistexture[idx] != -1)
        deleteTexture(iaxistexture[idx]);
    iaxistexture[idx] = -1;

    if (title.isEmpty())
        return;

    QImage tex = textureFromString(title, font);
    axisrect[idx].setSize(tex.size());
    iaxistexture[idx] = bindTexture(tex);
}
QImage T3dViewer::textureFromString(QString string, QFont font)
{
    QFontMetrics fm(font);
    QSize sz = fm.size(0, string);
    QImage tex(sz.width() + 2, sz.height() + fm.descent() + 2, QImage::Format_ARGB32);
    tex.fill(Qt::transparent);

    font.setStyleStrategy(QFont::PreferAntialias); //might force grayscale antialiasing, better than nothing

    QPainter p(&tex);
    p.setFont(font);
    p.setPen(Qt::white); //for modulate
    p.drawText(1, sz.height()+1, string);
    p.end();
    return tex;
}


void T3dViewer::mousePressEvent(QMouseEvent *e)
{
    pressed = true;
    old = e->pos();
    _x_rot = x_rot;
    _y_rot = y_rot;
    _z_rot = z_rot;
    _x_pos = x_pos;
    _y_pos = y_pos;
    _z_pos = z_pos;
}
void T3dViewer::mouseReleaseEvent(QMouseEvent *)
{
    pressed = false;
}
bool T3dViewer::event(QEvent *event)
{
    emit eventSig(event);
    if (event->type() == QEvent::Resize)
        emit resized(size());

    return QGLWidget::event(event);
}
void T3dViewer::mouseMoveEvent(QMouseEvent *e)
{
    if (pressed) {
        if (mousemode == 0 && !(e->modifiers() & Qt::ControlModifier)) { //rotate
            auto pos = e->pos();
            double relx = 360*(old - pos).x() / width();
            double relyz = 360*(old - pos).y() / height();

            x_rot = qRound(_x_rot - relyz) % 360;
            //y_rot = qRound(_y_rot + relx) % 360;
            z_rot = qRound(_z_rot - relx) % 360;

            updateGL();
        }
        if (mousemode == 1 || (e->modifiers() & Qt::ControlModifier)) {
            auto pos = e->pos();
            double relx = 1*(old - pos).x() / width();
            double relyz = 1*(old - pos).y() / height();

            x_pos = _x_pos - relx;
            y_pos = _y_pos + relyz;
            //z_pos = _z_pos + relyz;

            updateGL();
        }
    }
}
void T3dViewer::wheelEvent(QWheelEvent *e)
{
    int numDegrees = e->delta() / 8;
    int numSteps = numDegrees / 15;

    if (e->modifiers() == Qt::ShiftModifier) {
        z_scale *= pow(1.1, numSteps);
    } else {
        scale *= pow(1.1, numSteps);
        resizeGL(width(), height());
    }

    updateGL();
}
