#include "EG3000.h"
#include <i_eg3000.h>
#include "gui/status.hpp"
#include <QSettings>
#include <QTimer>
#include <QApplication>
#include <QFile>
#include <QDebug>
#include <QAction>
#include "scanplot.h"
#include "modules.hpp"
#include "motors.hpp"

inline
double linear_calib(double value, double from_min, double from_max, double to_min, double to_max)
{
    return (value - from_min) / (from_max - from_min) * (to_max - to_min) + to_min;
}

static
QHash<BYTE, double> analog_ch_gain;
double input_gain(BYTE ch)
{
    return analog_ch_gain.value(ch, 1);
}
void input_gain_set(BYTE ch, double gain)
{
    analog_ch_gain.insert(ch, gain);
}
void input_gain_clear(BYTE ch)
{
    analog_ch_gain.remove(ch);
}

static
QHash<BYTE, double> analog_ch_shift;
double input_shift(BYTE ch)
{
    return analog_ch_shift.value(ch, 0);
}
void input_shift_set(BYTE ch, double shift)
{
    analog_ch_shift.insert(ch, shift);
}
void input_shift_clear(BYTE ch)
{
    analog_ch_shift.remove(ch);
}

//from [a..b] to [a0..b0]
double analog_calib(double value, const char *name, double a0, double b0)
{
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    double a = settings_double(settings, QString() + name + "_min", a0);
    double b = settings_double(settings, QString() + name + "_max", b0);
    return linear_calib(value, a, b, a0, b0);
}

double analog_calib_inverse(double value, const char *name, double a0 = -10., double b0 = 10., double a = -10., double b = 10.)
{
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    a = settings_double(settings, QString() + name + "_min", a);
    b = settings_double(settings, QString() + name + "_max", b);
    return linear_calib(value, a0, b0, a, b);
}

//

static
QMap<int, QPair<double, double> > analog_map_cache;
static
void init_analog_calib_cache()
{
    analog_map_cache.clear();

    QSettings settings("Param/analog.conf", QSettings::IniFormat);

    QList<const char *> lst;
    QList<int> key;
    lst << "adc/v1" << "adc/v1mag" << "adc/v2" << "adc/v2mag" << "adc/v3" << "adc/v4" << "adc/v5" << "adc/v1hr" << "adc/v2hr" << "dac/vtip";
    key << 1 << 11 << 2 << 12 << 3 << 4 << 5 << 21 << 22 << 101;
    for (int i = 0; i < lst.size() && i < key.size(); i++) {
        double a = settings_double(settings, QString() + lst.at(i) + "_min", -10);
        double b = settings_double(settings, QString() + lst.at(i) + "_max", 10);

        analog_map_cache.insert(key.at(i), qMakePair(a, b));
    }
}

QPair<double, double> GetPairAnalogMapCache(int key)
{
    return analog_map_cache.value(key, qMakePair(-10., 10.));
}

double input_calib(BYTE ch, double val)
{
    double gain = input_gain(ch);
    double shift = input_shift(ch);

    //bool unwrap_unsigned = false;
    int idx = 0;
    switch (ch) {
    default:
        return (linear_calib(val, 0, 65535, -10, 10) + shift) * gain;
        break;
    case srcADC1:
    case srcDFL:
        idx = 1;
        break;
    case srcMag1:
        //unwrap_unsigned = true;
    case srcMagCos1:
    case srcMagSin1:
        idx = 11;
        break;
    case srcADC2:
    case srcLFL:
        idx = 2;
        break;
    case srcMag2:
        //unwrap_unsigned = true;
        idx = 12;
        break;
    case srcADC3:
    case srcIntensity:
        idx = 3;
        break;
    case srcADC4:
        idx = 4;
        break;
    case srcADC5:
        idx = 5;
        break;
    case srcADCHR1:
        idx = 21;
        break;
    case srcADCHR2:
        idx = 22;
        break;
    case srcVTip:
        idx = 101;
        break;
    }

    //fix negative part to be at top half for sign-scrambled unsigned values
    //if (unwrap_unsigned && val < 32768)
    //    val += 65536;

    QPair<double, double> p = analog_map_cache.value(idx, qMakePair(-10., 10.));
    return (linear_calib(val, 0., 65535., p.first, p.second) + shift) * gain;
}


QHash<BYTE, QString> analog_ch_unit;
QString input_units(BYTE ch)
{
    return analog_ch_unit.value(ch, QObject::tr("V"));
}
void input_unit_set(BYTE ch, QString unit)
{
    analog_ch_unit.insert(ch, unit);
}
void input_unit_clear(BYTE ch)
{
    analog_ch_unit.remove(ch);
}


//------------------------------------------------------------------------
static
bool validateHex(QString filename)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "validateHex can't open file";
        return false;
    }

    bool end = false;
    int c = 0;
    while (!file.atEnd()) {
        c++;
        QByteArray line = file.readLine();
        line.chop(1); // windows, "\cr\lf\", \cr remove

        if (line.size() < 11) {
            qDebug() << "validateHex short line" << c;
            file.close();
            return false;
        }
        if (line.at(0) != ':') {
            qDebug() << "validateHex no ':'" << c;
            file.close();
            return false;
        }
        if ((line.size() - 11) != 2*((QString(line.at(1)).toInt(nullptr, 16) << 4) + QString(line.at(2)).toInt(nullptr, 16))) {
            qDebug() << "validateHex wrong data len" << c;
            file.close();
            return false;
        }
        if (line.at(7) != '0') {
            qDebug() << "validateHex wrong record type1" << c;
            file.close();
            return false;
        }
        if (line.at(8) != '0' && line.at(8) != '1' && line.at(8) != '2' &&
                line.at(8) != '3' && line.at(8) != '4' && line.at(8) != '5') {
            qDebug() << "validateHex wrong record type2" << c;
            file.close();
            return false;
        }
        if (line.at(8) != '0' &&
                (line.at(3) != '0' ||
                 line.at(4) != '0' ||
                 line.at(5) != '0' ||
                 line.at(6) != '0')) {
            qDebug() << "validateHex address not 0" << c;
            file.close();
            return false;
        }

        if (line.at(8) == '1')
            end = true;
        if (end && !file.atEnd()) {
            qDebug() << "validateHex end record not last" << c;
            file.close();
            return false;
        }

        unsigned long checksum = 0;
        for (int i = 0; i < (line.size() - 1)/2; i++) {
            checksum += (QString(line.at(2*i+1)).toInt(nullptr, 16) << 4) + QString(line.at(2*i+2)).toInt(nullptr, 16);
        }
        if ((0xFF & checksum) != 0) {
            qDebug() << "validateHex wrong checksum" << c;
            file.close();
            return false;
        }
    }

    if (!end)
        qDebug() << "validateHex no eof record";

    return end;
}

static
unsigned long bufferHex(QString filename, unsigned char *buffer, unsigned int max_size)
{
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "bufferHex can't open file";
        return 0;
    }

    memset(buffer, 0xFF, max_size);

    unsigned long max_addr = 0;
    unsigned long high_addr = 0;
    while (!file.atEnd()) {
        QByteArray line = file.readLine();
        line.chop(1); // windows, "\cr\lf\", \cr remove

        if (line.at(8) == '0') {
            if (high_addr < 0x1D00 || high_addr > 0x1D08)
                continue;

            unsigned long size = (QString(line.at(1)).toInt(nullptr, 16) << 4) + QString(line.at(2)).toInt(nullptr, 16);
            unsigned long addr = (((QString(line.at(3)).toInt(nullptr, 16) << 4) + QString(line.at(4)).toInt(nullptr, 16)) << 8) +
                    (QString(line.at(5)).toInt(nullptr, 16) << 4) + QString(line.at(6)).toInt(nullptr, 16);
            addr += (high_addr - 0x1D00) << 16;

            for (unsigned int i = 0; i < size && addr + i < max_size; i++) {
                buffer[addr + i] = (QString(line.at(i*2+9)).toInt(nullptr, 16) << 4) + QString(line.at(i*2+10)).toInt(nullptr, 16);
            }

            if (addr + size >= max_size)
                qDebug() << "bufferHex overflow!" << addr + size << max_size;

            if (addr + size > max_addr)
                max_addr = addr + size;
        } else if (line.at(8) == '1') {
            break;
        } else if (line.at(8) == '4') {
            high_addr = (((QString(line.at(9)).toInt(nullptr, 16) << 4) + QString(line.at(10)).toInt(nullptr, 16)) << 8) +
                    (QString(line.at(11)).toInt(nullptr, 16) << 4) + QString(line.at(12)).toInt(nullptr, 16);
        }
    }

    return max_addr;
}

static
bool program_hex(QString filename, struct eg3000_read_handle *read_handle, void *devHandle)
{
    if (!validateHex(filename))
        return false;

    unsigned char *buf = new unsigned char[520000];
    unsigned size = 4 + ((~3) & bufferHex(filename, buf, 520000));

    bool ok = RET_TRUE == eg3000_bl_program(devHandle, buf, size, read_handle, 0);
    eg3000_bl_stats(devHandle);
    delete[] buf;

    return ok;
}
static inline
BYTE invert_byte(BYTE b)
{ //from SO sth
    b = (b & 0xF0) >> 4 | (b & 0x0F) << 4; //swap every 4 bits
    b = (b & 0xCC) >> 2 | (b & 0x33) << 2; //swap every 2 bits
    b = (b & 0xAA) >> 1 | (b & 0x55) << 1; //swap every 1 bits
    return b;
}
static
bool program_bin(QString filename, struct eg3000_read_handle *read_handle, void *devHandle)
{
    QFile bin(filename);
    bin.open(QIODevice::ReadOnly);
    QByteArray buf = bin.readAll();
    bin.close();

    if (filename.endsWith(".rpd")) //flash lsb first
        for (int i = 0; i < buf.size(); i++)
            buf[i] = invert_byte(buf.at(i));

    bool ok = RET_TRUE == eg3000_bl_program(devHandle, (BYTE *)buf.data(), buf.size(), read_handle, 1);
    //need?? eg3000_bl_stats(devHandle);
    return ok;
}

//------------------------------------------------------------------------
EG3000_CONTROL::EG3000_CONTROL()
{
    need_program = false;

    init_analog_calib_cache();

    prev_usb_stat = -1;
    SWusb1 = new StatusWidget1(tr("USB"));
    SWusb1->label->setToolTip(tr("EG-3000"));
    s_usbStat(EDno);

    QAction *a;
    SWusb1->setContextMenuPolicy(Qt::ActionsContextMenu);
    SWusb1->addAction(a = new QAction(tr("Reconnect"), SWusb1));
    connect(a, SIGNAL(triggered()), this, SLOT(drop_connect()));

    connect(this, SIGNAL(usbStatus(int)), this, SLOT(s_usbStat(int)));
    connect(&rThread, SIGNAL(status(int)), this, SLOT(s_usbStat(int)));

    QTimer *boot_timeout = new QTimer(this);
    boot_timeout->setInterval(2500);
    boot_timeout->setSingleShot(true);
    connect(boot_timeout, SIGNAL(timeout()), this, SLOT(s_boot_timeout()));
    connect(this, SIGNAL(restart_boot_timer()), boot_timeout, SLOT(start()));


    if (sizeof(BYTE) != 1 || sizeof(WORD) != 2 || sizeof(DWORD) != 4 || sizeof(float) != 4)
        qDebug() << "wrong data size" <<
                    sizeof(BYTE) << sizeof(WORD) << sizeof(DWORD) << sizeof(float);
    if (sizeof(EG_CalData) != 228 || sizeof(EG_ForceCurve) != 20 ||
        sizeof(EG_ScanParams) != 108 || sizeof(EG_StepMotorData) != 16)
        qDebug() << "wrong eg_data size" <<
                    sizeof(EG_CalData) << sizeof(EG_ForceCurve) <<
                    sizeof(EG_ScanParams) << sizeof(EG_StepMotorData);
}

void EG3000_CONTROL::program(QString val)
{
    need_program = true;
    path = val;
    emit device_status(2);
}

void EG3000_CONTROL::s_boot_timeout()
{
    boot_timeout = true;
}

void EG3000_CONTROL::run()
{
    m_closeThread = false;
    m_waitMainInit = true;
    need_reconnect = false;

    struct eg3000_read_handle read_handle;
    eg3000_init_read_handle(&read_handle);
    rThread.read_handle = &read_handle;
    void *devHandle = nullptr;
    bool st = false;

    bool need_boot = true;
    bool is_flash_good = true;
    boot_timeout = false;
    int prevStatus = -1;
    bool use_4000 = false;
    bool slept = true;

    while (!m_closeThread) {
        if (st)
            st = eg3000_is_ok(devHandle);
        if (need_reconnect) {
            st = false;
            need_reconnect = false;
        }

        //prepare for reprogramming
        if (st && need_program && !need_boot) {
            wThread.start_stop(false);
            rThread.start_stop(false);
            emit usbStatus(prevStatus = EDconn);

            msleep(50);

            while (!wThread.write_lock.tryLock(1))
                msleep(10);

            //
            eg3000_fb_control(devHandle, fbOff); //disable feedback
            eg3000_osc_startstop(devHandle, 0); //stop oscilloscope
            eg3000_status_onoff(devHandle, 0); //disable spm info send
            eg3000_set_head_dac(devHandle, 32767); //set tip voltage ~0
            eg3000_hv_zero(devHandle); //hv outs to 0
            qDebug() << "eg3000 stop";

            msleep(500);

            eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_GO_REBOOT);
            qDebug() << "eg3000 reset";

            wThread.write_lock.unlock();

            need_boot = true;
            is_flash_good = true;

            msleep(500);

            eg3000_clear(devHandle);
            eg3000_clear_read_handle(&read_handle);

            st = eg3000_is_ok(devHandle);
        }

        //boot start ---------------------------------------
        if (st && need_boot && !m_closeThread) {
            rThread.read_lock.lock();
            wThread.write_lock.lock();

            if (!need_program) {
                //query
                eg3000_bl_stats(devHandle); //ping bootloader
                eg3000_cpu_loopback(devHandle); //ping cpu
                st = eg3000_is_ok(devHandle);
            }

            emit restart_boot_timer();
            boot_timeout = false;

            if (!(st && need_boot && !m_closeThread)) { //need_boot is always true here
                rThread.read_lock.unlock();
                wThread.write_lock.unlock();
            }
        }
        bool boot_retried = false;
        BYTE boot_timeout_count = 0;
        while (st && need_boot && !m_closeThread) {
            st = eg3000_update_parse(devHandle, &read_handle);
            BYTE Address = read_handle.address;
            BYTE SubAddress = read_handle.sub_address;

            //packet parsing
            if (read_handle.packet_recieved) {
                if (Address == ADR_BOOTLOADER) {
                    if (SubAddress == BL_ERROR) {
                        qDebug() << "bl error" << (void*)((BYTE *)read_handle.buffer)[0];
                    }
                    if (SubAddress == BL_STATS) {
                        //unsigned int flash_size = ((DWORD*)read_handle.buffer)[0];
                        is_flash_good = ((BYTE *)read_handle.buffer)[4];

                        if (!need_program) {
                            if (!is_flash_good) {
                                if (boot_timeout_count == 0) {
                                    if (!m_waitMainInit)
                                        boot_timeout_count = 4; //skip second warning
                                    qDebug() << "no flash in controller!";
                                    emit device_status(0);
                                }
                            } else { //auto_boot
                                st = eg3000_bl_run(devHandle);
                                boot_retried = false;
                            }
                        }
                    }
                    if (SubAddress == BL_NO_FLASH) {
                        if (!need_program) {
                            if (boot_retried) {
                                if (boot_timeout_count == 0) {
                                    if (!m_waitMainInit)
                                        boot_timeout_count = 4; //skip second warning
                                    qDebug() << "no flash in controller!";
                                    emit device_status(0);
                                }
                            } else {
                                boot_retried = true;
                                msleep(100);
                                st = eg3000_bl_run(devHandle);
                            }
                        }
                    }

                    if (SubAddress == BL_BOOT_OK) { //booted
                        st = eg3000_cpu_loopback(devHandle); //ping
                    }
                }
                if (Address == ADR_CPU) {
                    need_boot = false;
                }

                eg3000_discard_packet(&read_handle);
            } else
                msleep(10);

            //program
            if (need_program && need_boot) {
                emit device_status(3);
                bool ok = (path.endsWith(".hex") ? program_hex : program_bin)(path, &read_handle, devHandle);
                st = eg3000_is_ok(devHandle);
                if (ok && st) {
                    need_program = false;
                    emit device_status(4);
                } else {
                    qDebug() << "retry program";
                    emit device_status(2);
                }

                emit restart_boot_timer();
                boot_timeout = false;
                boot_timeout_count = 0;

                msleep(100);
            }

            //fail recovery
            if (st && (boot_timeout || read_handle.packet_malformed) && need_boot && !m_closeThread && is_flash_good) {
                emit restart_boot_timer();
                boot_timeout = false;

                if (read_handle.packet_malformed) {
                    eg3000_clear_read_handle(&read_handle);
                    eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_USB_BUF_RST);
                    msleep(100);
                } else if (!m_waitMainInit)
                    boot_timeout_count++;

                //resend query
                eg3000_bl_stats(devHandle); //ping bootloader
                eg3000_cpu_loopback(devHandle); //ping cpu
                st = eg3000_is_ok(devHandle);

                if (boot_timeout_count == 3) {
                    qDebug() << "boot timeout";
                    emit device_status(1);
                }
            }

            if (!(st && need_boot && !m_closeThread)) {
                rThread.read_lock.unlock();
                wThread.write_lock.unlock();
            }
        }
        //boot end------------------------------------

        {
            int stat = st ? (need_boot ? EDconn : EDok) : EDno;
            if (!m_waitMainInit && prevStatus != stat) {
                wThread.start_stop(st);
                rThread.start_stop(st);
                emit usbStatus(prevStatus = stat);
            }
        }

        //close
        if (!st && devHandle != nullptr) {
            rThread.read_lock.lock();
            wThread.write_lock.lock();

            if (eg3000_is_ok(devHandle))
                eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_USB_BUF_RST); //l2k soft reset = need cleanup

            eg3000_close(devHandle);
            devHandle = nullptr;

            rThread.read_lock.unlock();
            wThread.write_lock.unlock();
        }

        //open
        if (!st && !m_closeThread) {
            QSettings settings("Param/controller.conf", QSettings::IniFormat);
            bool use = settings.value("use_eg4000", true).toBool();
            if (use_4000 != use)
                eg3000_set_backend(use ? 1 : 0); //libusb:ftdi
            use_4000 = use;

            need_boot = true;
            is_flash_good = true;

            const WORD pid = PID+(use_4000 ? 5 : 0);
            struct s_info array;
            WORD size = 0;
            st = eg3000_list(VID, pid, &size, &array, 1);
            if (st && size != 0) {
                st = eg3000_open(array.handle, VID, pid, &devHandle);
                rThread.devHandle = devHandle;
                wThread.devHandle = devHandle;
            }

            if (st) {
                st = eg3000_clear(devHandle);
                eg3000_clear_read_handle(&read_handle);
                if (slept) {
                    slept = false;
                    continue;
                }
            }
        }

        slept = true;
        for (int i = 0; i < (m_waitMainInit ? 1 : 10) && !m_closeThread; i++)
            msleep(50);
    }

    if (st && !need_boot) {
        eg3000_fb_control(devHandle, fbOff); //disable feedback
        eg3000_osc_startstop(devHandle, 0); //stop oscilloscope
        eg3000_status_onoff(devHandle, 0); //disable spm info send
        eg3000_set_head_dac(devHandle, 32767); //set tip voltage ~0
        eg3000_hv_zero(devHandle); //hv outs to 0

        eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_USB_BUF_RST);
    }

    if (devHandle != nullptr)
        eg3000_close(devHandle);
    eg3000_close_read_handle(&read_handle);
}

void EG3000_CONTROL::thread_close()
{
    wThread.start_stop(false);
    rThread.start_stop(false);
    while (wThread.isRunning() && rThread.isRunning()) {
        qApp->processEvents();
        msleep(3);
    }

    m_closeThread = true;
    while (isRunning()) {
        qApp->processEvents();
        msleep(3);
    }
}
void EG3000_CONTROL::drop_connect()
{
    need_reconnect = true;
}

void EG3000_CONTROL::s_usbStat(int idx)
{
    if (prev_usb_stat == idx)
        return;
    prev_usb_stat = idx;

    if (!SWusb1)
        return;

	auto status = SWusb1->status;

    switch (prev_usb_stat) {
    case EDno:
        status->set(StatusWidget::Cross, Qt::red);
        status->setToolTip(tr("No device"));
        break;
    case EDconn:
        status->set(StatusWidget::Circle, QColor(255, 128, 0));
        status->setToolTip(tr("Loading"));
        break;
    case EDok:
        status->set(StatusWidget::Circle, Qt::green);
        status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        status->set(StatusWidget::Play, Qt::green);
        status->setToolTip(tr("Working"));
        break;
    }

#ifdef _NEW_MOTORS_CHANGES
    if (prev_usb_stat == EDok || prev_usb_stat == EDno)
    {
        auto m = Motors::GetIt();
        if (m)
        {
            m->motorsDebug[0]->setText(tr("UART1 = none"));
            m->motorsDebug[1]->setText(tr("UART2 = none"));
            m->motorsDebug[2]->setText(tr("UART3 = none"));
            m->motorsDebugNumMot[0] = -1;
            m->motorsDebugNumMot[1] = -1;
            m->motorsDebugNumMot[2] = -1;
        }
    }
#endif
}

//------------------------------------------------------------------------
USB_WRITE_THREAD::USB_WRITE_THREAD()
{
    want_running = false;
    devHandle = nullptr;
}
void USB_WRITE_THREAD::start_stop(bool st)
{
    if (want_running == st)
        return;

    if (want_running = st)
        start(QThread::HighPriority);
}
void USB_WRITE_THREAD::run()
{
    for (int i = 0; i < writers.count(); i++)
        writers.at(i)->resetWriter();

    qDebug() << "eg3000 started";

    while (want_running) {
        if (!write_lock.tryLock(1)) {
            msleep(1);
            continue;
        }

        bool sent = false;
        for (int i = 0; i < writers.count(); i++) {
            if (writers.at(i)->sendData(devHandle))
                sent = true;
            if (writers.at(i)->sendDataIndirect(this))
                sent = true;
        }

        if (!eg3000_is_ok(devHandle))
            want_running = false;

        write_lock.unlock();
        if (!sent)
            msleep(1);
    }
}

BYTE USB_WRITE_THREAD::raw_send(const void *buffer, WORD data_size)
{
    return eg3000_raw_send(devHandle, buffer, data_size);
}
BYTE USB_WRITE_THREAD::send_packet(BYTE Address, BYTE SubAddress, const void *buffer, WORD data_size)
{
    return eg3000_send_packet(devHandle, Address, SubAddress, buffer, data_size);
}

//------------------------------------------------------------------------
USB_READ_THREAD::USB_READ_THREAD()
{
    want_running = false;
    devHandle = nullptr;
    read_handle = nullptr;
}
void USB_READ_THREAD::start_stop(bool st)
{
    if (want_running == st)
        return;

    if (want_running = st) {
        emit capBits(0, 0); //eg3k
        emit capBits(2, 0); //head
        emit capBits(3, 0); //stage
        emit capBits(4, 0); //z3
        emit capBits(6, 0); //motors

        start(QThread::HighestPriority);
    }
}
void USB_READ_THREAD::run()
{
    DWORD last_bytes_cnt = 0;
    int prevStatus = -1;

    while (want_running) {
        if (!read_lock.tryLock(1)) {
            msleep(1);
            continue;
        }

        if (!eg3000_update_parse(devHandle, read_handle))
            want_running = false;

        BYTE Address = read_handle->address;
        BYTE SubAddress = read_handle->sub_address;
        bool maybe_have_more_data = (last_bytes_cnt != read_handle->buf_data_length) || read_handle->header_read;
        last_bytes_cnt = read_handle->buf_data_length;

        //packet parsing
        if (read_handle->packet_recieved) {
            for (int i = 0; i < readers.count(); i++)
                if (readers.at(i)->processPacket(Address, SubAddress, read_handle->buffer, read_handle->packet_size))
                    break;

            if (SubAddress == CPU_DEVICE_CAP_BITS &&
                    (Address - ADR_HEAD) >= 0 &&
                    (Address - ADR_HEAD) <= 8) {
                emit capBits(2 + (Address - ADR_HEAD), *((DWORD *)read_handle->buffer));
            } else if (Address == ADR_CPU) {
                if (SubAddress == CPU_CAP_BITS)
                    emit capBits(0, *((DWORD *)read_handle->buffer));

                if (SubAddress == CPU_STATUS) {
                    EG_Status *Deflectomter = (EG_Status *)read_handle->buffer;

                    int stat = Deflectomter->Scanning != mdIdle ? EDplay : EDok;
                    if (prevStatus != stat)
                        emit status(prevStatus = stat);
                }
            }

            eg3000_discard_packet(read_handle);
        } else if (!maybe_have_more_data)
            msleep(1);

        read_lock.unlock();
    }
}
