#include "mainwindow.h"

#include "hardware/EG3000.h"
#include "container/datalist.hpp"
#include "spmtunning.h"
#include "options_window.h"
#include "gui/status.hpp"
#include "scanplot.h"

#include <QCheckBox>
#include <QHBoxLayout>
#include <QKeyEvent>
#include <QApplication>
#include <QSplashScreen>
#include <QStyleFactory>
#include <QProgressBar>
#include <QStatusBar>
#include <QMessageBox>
#include <QShortcut>
#include "3d/plot_3dplot.h"
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtConcurrent>
#else
#include <QtCore>
#endif


//----------------------------------------------
void send2dPlot(int id, const plot::T2dPlot *plot2d, bool can_drop)
{
    auto mob = MobileServer::GetIt();
    if (!mob)
        return;
    if (!mob->need_send())
        return;

    plot::TCurvesDataOwner data = plot::getLock(plot2d->Pure2dPlot->SourceData);

    int idx = -1;
    for (int i = 0; i < data.size(); i++)
        if (data.at(i) && data.at(i)->size()) { idx = i; break; }
    if (idx < 0)
        return;

    if (mob->any_need_data("need_2dPlot"))
    {
        mob->setPropIdVar(DataID::PlotDataID, id);
        mob->setPropIdVar(DataID::PlotData2dInfo, QVariant::fromValue(*((plot::TInfoFor2dPlot *)plot2d->Pure2dPlot)));
        //mob->setPropIdVar(DataID::PlotData2d, QVariant::fromValue(*data.at(idx).data()), can_drop ? id+2 : 0);

        for (int i = 0; i < data.size(); i++)
            if (data.at(i) && data.at(i)->size())
            {
                auto dt = *data.at(i).data();
                mob->setPropIdVar(DataID::PlotData2d_A+i, QVariant::fromValue(dt), can_drop ? id+2 : 0);
            }
    }
}

void send3dPlot(int id, const plot::T3dPlot *plot3d, int plotid, bool can_drop, bool forceSend, bool isCCD)
{
    auto mob = MobileServer::GetIt();
    if (!mob)
        return;
    if (!mob->need_send())
        return;

    plot::T3dDataOwner data = plot::getLock(plot3d->ThreeDPlots.at(plotid)->PaintData);

    int idx = -1;
    for (int i = 0; i < data.size(); i++)
        if (data.at(i))
            { idx = i; break; }
    if (idx < 0)
        return;

    if (id != 0 && (mob->any_need_data("scan_plot_" + QString::number(plotid)) || forceSend))
    {
        if (plotid == 1 && isCCD && !forceSend)
        {
            QByteArray vdata;
            QBuffer buf(&vdata);
            buf.open(QBuffer::WriteOnly);
            int size = 500;
            plot3d->ThreeDPlots[1]->MakePreview(plot3d->PlotControl, size, size).save(&buf, "bmp");
            buf.close();

            mob->setPropIdVar(DataID::plot3DCCDImageData, vdata, 1);
            return;
        }
        plot::TInfoFor3dPlot info = *((plot::TInfoFor3dPlot *)plot3d->ThreeDPlots.at(plotid));
        //info.sliceBounds = false;

        mob->setPropIdVar(DataID::next3DPlotDataForPlot, plotid);
        mob->setPropIdVar(DataID::PlotDataID, id);
        mob->setPropIdVar(DataID::PlotData3dInfo, QVariant::fromValue(info));
        mob->setPropIdVar(DataID::PlotData3d, QVariant::fromValue(*data.at(idx).data()), can_drop ? id+2 : 0);
    }
}

//----------------------------------------------
void MainWindow::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        connect(&m->wThread, SIGNAL(finished()), this, SLOT(cancelOperation()));
    }

    if (I_GUI_MODULE_BASE *m = (I_GUI_MODULE_BASE *)module->basecast("gui")) {
        QPushButton *pb = m->createPanelButton();
        if (pb) {
            guiModuleButtons.insert(m, pb);
            ContentStackedWidget->addWidget((QWidget *)m->basecast());
            connect(pb, SIGNAL(clicked()), this, SLOT(moduleButtonClicked()));

            pb->setFlat(true);
            pb->setCheckable(true);
            if (guiModuleButtons.size() == 1) {
                pb->setChecked(true);
                prev_button = pb;
            }
            pb->setShortcut(QKeySequence(Qt::ALT + Qt::Key_1 + guiModuleButtons.size()-1));
            pb->setAcceptDrops(true);
            pb->installEventFilter(this);

            QToolBar *TBMain = findChild<QToolBar *>("MainTB");
            QWidget *wid = new QWidget();
            TBMain->addWidget(wid);
            QHBoxLayout *lay = new QHBoxLayout(wid);
            lay->setContentsMargins(0, 0, 0, 0);
            lay->addWidget(pb);
        }

        int group = 0;
        QToolBar *TBStatus = findChild<QToolBar *>("StatusTB");
        foreach (QWidget *w, m->getStatusWidgets()) {
            if ((long)w < 100) {
                group = (long)w;
                continue;
            }

            QAction *next = nullptr;
            foreach (int gr, statusGroups.keys()) //sorted
                if (gr > group) {
                    next = statusGroups.value(gr);
                    break;
                }
            if (group > 0 && !statusGroups.contains(group))
                statusGroups.insert(group, TBStatus->insertSeparator(next));

            TBStatus->insertWidget(next, w);
        }
    }
    if (name == QLatin1String("net")) {
        MobileServer *m = (MobileServer *)module->basecast();

        QToolBar *TBStatus = findChild<QToolBar *>("StatusTB");
        TBStatus->addSeparator(); //6
        TBStatus->addWidget(m->networkStatus);

        //add<Edit>(this, SIGNAL(newValue(double)), SLOT(setPropDbl(double)));
        add<DoubleChanger>(this, SIGNAL(newValue(double)), SLOT(setPropDbl(double)));
        //add<NSTSlider>(this, SIGNAL(newValue(double)), SLOT(setPropDbl(double)));
        //add<QPushButton>(this->PBname, SIGNAL(clicked()), SLOT(setPropVar())); //requres manual direct call
        add<TwoStateButton>(this, SIGNAL(state_changed(bool)), SLOT(setPropBool(bool)));
        add<QCheckBox>(this, SIGNAL(toggled(bool)), SLOT(setPropBool(bool)));
        add<QComboBox>(this, SIGNAL(currentIndexChanged(int)), SLOT(setPropInt(int)));
    }
}
void MainWindow::disconnected(const QLatin1String , I_MODULE *module)
{
    if (guiModuleButtons.contains(module)) {
        delete guiModuleButtons.value(module)->parentWidget();//delete guiModuleButtons.value(module);
        guiModuleButtons.remove(module);
    }
}


void MainWindow::initialSend(MobileInstance *client)
{
    foreach (int id, netidmap.uniqueKeys()) {
        QObject *obj = netidmap.value(id);
        //if (Edit *v = qobject_cast<Edit *>(obj))
        //    client->serv->setPropIdVarS(client->socket, id, v->value());
        if (DoubleChanger *v = qobject_cast<DoubleChanger *>(obj))
            client->serv->setPropIdVarS(client->socket, id, v->value());
        if (TwoStateButton *v = qobject_cast<TwoStateButton *>(obj))
            client->serv->setPropIdVarS(client->socket, id, v->isOn);
        if (QComboBox *v = qobject_cast<QComboBox *>(obj))
            client->serv->setPropIdVarS(client->socket, id, v->currentIndex());
    }
}
void MainWindow::setValue(int id, QVariant val, QObject *)
{
    foreach (QObject *obj, netidmap.values(id)) {
        //if (Edit *v = qobject_cast<Edit *>(obj)) {
        //    v->setValue(val.toDouble());
        //    v->setNewValue();
        //}
        if (DoubleChanger *v = qobject_cast<DoubleChanger *>(obj)) {
            v->setValue(val.toDouble());
            v->setNewValue();
        }
        //if (NSTSlider *v = qobject_cast<NSTSlider *>(obj)) {
        //    v->setValue(val.toDouble());
        //    v->setNewValue();
        //}
        if (TwoStateButton *v = qobject_cast<TwoStateButton *>(obj))
            v->change_state_man_with_sig(val.toBool());
        //else if (QPushButton *v = qobject_cast<QPushButton *>(obj))
            //v->click();
        if (QCheckBox *v = qobject_cast<QCheckBox *>(obj))
            v->setChecked(val.toBool());
        if (QComboBox *v = qobject_cast<QComboBox *>(obj))
            v->setCurrentIndex(val.toInt());
    }
}
void MainWindow::setPropBool(bool val)
{
    setPropVar(val);
}
void MainWindow::setPropInt(int val)
{
    setPropVar(val);
}
void MainWindow::setPropDbl(double val)
{
    setPropVar(val);
}
void MainWindow::setPropVar(QVariant val)
{
    QObject *p = sender();
    if (!p || !p->property("id").isValid())
        return;

    bool ok;
    int id = p->property("id").toInt(&ok);
    if (ok)
        MobileServer::sendData(id, val);
}


//fast widget creation guide:
//1: construct widget-layout tree top-down, with parents. prevents VERY expensive reparenting
//2: avoid setFixed*(sizeHint()), it calls expensive layouting.
//setSizePolicy(QSizePolicy::Fixed); in layouts
//layout()->setSizeConstraint(QLayout::SetFixedSize); for windows
//setFixed() for qsplitter :/

MainWindow::MainWindow()
{
    prev_button = nullptr;

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
    QApplication::setStyle(QStyleFactory::create("Fusion"));
#else
    QApplication::setStyle(QStyleFactory::create("Plastique"));
#endif

    ContentStackedWidget = new QStackedWidget(this);
    ContentStackedWidget->setMinimumSize(1024, 600);
    setCentralWidget(ContentStackedWidget);

    QToolBar *TBMain = new QToolBar(this);
    addToolBar(TBMain);
    TBMain->setObjectName("MainTB");
    TBMain->toggleViewAction()->setVisible(false);

    QToolBar *TBStatus = new QToolBar(this);
    addToolBar(TBStatus);
    TBStatus->setObjectName("StatusTB");
    TBStatus->toggleViewAction()->setVisible(false);
    TBStatus->addSeparator();

    QToolBar *TBProgress = new QToolBar(this);
    addToolBar(TBProgress);
    TBProgress->setObjectName("ProgressTB");
    TBProgress->setAllowedAreas(Qt::TopToolBarArea | Qt::BottomToolBarArea);
    TBProgress->toggleViewAction()->setVisible(false);

    QStatusBar *stbar = new QStatusBar;
    stbar->setSizeGripEnabled(false);
    TBProgress->addWidget(stbar);
    setMainStatusBar(stbar);
    {
        StatusBar *bar = new StatusBar("0", 1);
        bar->add();
        stbar->setMinimumHeight(stbar->sizeHint().height());
        bar->hide();
        bar->deleteLater();
    }


    QTimer *timer0 = new QTimer(this);
    connect(timer0, SIGNAL(timeout()), this, SLOT(s_update_singleapp()));
    timer0->start(5*1000);


    setWindowTitle(tr("Nspec Universal Software"));

    qApp->installEventFilter(this);

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    foreach (QToolBar *t, findChildren<QToolBar *>())
        t->setMovable(settings.value("gui_unlock", false).toBool());

    if (settings.contains("geometry")) {
        restoreGeometry(settings.value("geometry").toByteArray());
        restoreState(settings.value("state/main").toByteArray());
    } else
        setWindowState(Qt::WindowMaximized);
}
void MainWindow::setRunning(bool on) {
    if (!on) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        settings.setValue("geometry", saveGeometry());
        settings.setValue("state/main", saveState());
    }
}
void MainWindow::cancelOperation()
{
    cancelAllStatusBars();
}


QByteArray singleappGet();
bool singleappSet(const QByteArray ba);
void MainWindow::s_update_singleapp()
{
    QByteArray byteArray = singleappGet();
    if (!byteArray.size() || byteArray.left(1) == "0")
        return;

    bool openfile = (byteArray.left(1) == "2");
    byteArray.remove(0, 1);
    QString message = QString::fromUtf8(byteArray.constData());

    byteArray = "0\0";
    singleappSet(byteArray);

    raise();
    activateWindow();

    if (openfile && !message.isEmpty()) {
        qDebug() << "singleapp open message" << message;
        if (I_MODULE *datam = ModuleManager::instance->find1("data")) {
            DataPlot *data = (DataPlot *)datam->basecast();
            QPushButton *pb = guiModuleButtons.value(datam);
            if (pb && pb->isVisible())
                pb->click();

            data->Container->OpenFile(message, true);
        }
    }
}


void MainWindow::AttachDetachModuleWindow()
{
    I_GUI_MODULE_BASE *m = nullptr;
    foreach (I_MODULE *m0, ModuleManager::instance->modules)
        if (m = (I_GUI_MODULE_BASE *)m0->basecast("gui")) {
            if (m->detachButton() == sender())
                break;
            else
                m = nullptr;
        }
    if (m == nullptr)
        return;

    QWidget *w = (QWidget *)m->basecast();
    QWidget *button = guiModuleButtons.value(m);
    bool onoff = w->parentWidget();

    button->setVisible(!onoff);
    if (onoff) {
        QPoint wpos = w->mapToGlobal(pos() - geometry().topLeft()); //-decoration size
        ContentStackedWidget->removeWidget(w);
        w->setParent(nullptr);
        w->setWindowFlags(Qt::Window);
        w->show();
        w->move(wpos);

        foreach (I_MODULE *m0, ModuleManager::instance->modules)
            if (I_GUI_MODULE_BASE *m1 = (I_GUI_MODULE_BASE *)m0->basecast("gui"))
                if (m1 != m && ((QWidget *)m0->basecast())->parentWidget()) {
                    ShowModuleWidget(m0);
                    break;
                }
    } else {
        ContentStackedWidget->addWidget(w);
        w->setWindowFlags(Qt::Widget);
        ShowModuleWidget(m);
        raise();
    }
}
void MainWindow::moduleButtonClicked()
{
    QHash<I_MODULE *, QPushButton *>::const_iterator i = guiModuleButtons.constBegin();
    while (i != guiModuleButtons.constEnd()) {
        if (i.value() == sender()) {
            ShowModuleWidget(i.key());
            break;
        }
        ++i;
    }
}
void MainWindow::ShowModuleWidget(I_MODULE *m)
{
    if (prev_button) {
        prev_button->setChecked(false);
        prev_button->setEnabled(true);
    }
    prev_button = guiModuleButtons.value(m);

    if (prev_button) {
        prev_button->setChecked(true);
        prev_button->setEnabled(false);
    }
    ContentStackedWidget->setCurrentWidget((QWidget *)m->basecast());
}

void ManagedThreadedCleaner::registerManagedObject(QObject *obj)
{
    auto mtc = ManagedThreadedCleaner::GetIt();
    QMutexLocker lck(&mtc->shutdownDataMutex);
    mtc->shutdownObjects << obj;
}
void ManagedThreadedCleaner::registerManagedFuture(QFuture<void> obj)
{
    auto mtc = ManagedThreadedCleaner::GetIt();
    QMutexLocker lck(&mtc->shutdownDataMutex);
    mtc->shutdownActions << obj;
}
void ManagedThreadedCleaner::unregisterManagedObject(QObject *obj)
{
    auto mtc = ManagedThreadedCleaner::GetIt();
    QMutexLocker lck(&mtc->shutdownDataMutex);
    int i = mtc->shutdownObjects.indexOf(obj);
    if (i != -1)
        mtc->shutdownObjects.remove(i);
}
void ManagedThreadedCleaner::unregisterManagedFuture(QFuture<void> obj)
{
    auto mtc = ManagedThreadedCleaner::GetIt();
    QMutexLocker lck(&mtc->shutdownDataMutex);
    int i = mtc->shutdownActions.indexOf(obj);
    if (i != -1)
        mtc->shutdownActions.remove(i);
}

ManagedThreadedCleaner* ManagedThreadedCleaner::GetIt() noexcept
{
    auto mtc = ModuleManager::instance->find1VoidPtr("managedthreadedcleaner");
    return static_cast<ManagedThreadedCleaner*>(mtc);
}

static void deleteObject(QObject *obj)
{
    delete obj;
}
void ManagedThreadedCleaner::setRunning(bool on)
{
    if (!on) {
        QElapsedTimer timeout;
        QWidget *parent = nullptr;
        foreach (QWidget *w, qApp->topLevelWidgets())
            if (w->isVisible() && !w->windowTitle().isNull()) {
                parent = w;
                break;
            }

        //TODO meaningful operation description
        for (int i = 0; i < shutdownActions.size(); i++) {
            timeout.start();
            while (shutdownActions[i].isRunning() && timeout.elapsed() < 10000) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }

            if (timeout.elapsed() >= 10000 && QMessageBox::Yes ==
                    QMessageBox::question(parent, QObject::tr("Nspec"),
                                          QObject::tr("Some operation hasn't finished yet.") + "\n" +
                                          QObject::tr("Wait?"), QMessageBox::Yes, QMessageBox::No))
                i--;
        }
        shutdownActions.clear();

        QMutexLocker lck(&shutdownDataMutex);
        for (int i = 0; i < shutdownObjects.size(); i++) {
            shutdownObjects.at(i)->moveToThread(nullptr);
            shutdownActions << QtConcurrent::run(deleteObject, shutdownObjects.at(i));
        }
        //shutdownObjects.clear();
        lck.unlock();

        for (int i = 0; i < shutdownActions.size(); i++) {
            timeout.start();
            while (shutdownActions[i].isRunning() && timeout.elapsed() < 10000) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }

            if (timeout.elapsed() >= 10000 && QMessageBox::Yes ==
                    QMessageBox::question(parent, QObject::tr("Nspec"),
                                          QObject::tr("%1 Object cleanup hasn't finished yet.").arg(shutdownObjects.at(i)->metaObject()->className()) + "\n" +
                                          QObject::tr("Wait?"), QMessageBox::Yes, QMessageBox::No))
                i--;
        }
        shutdownObjects.clear();
        shutdownActions.clear();
    }
}

void MainWindow::resizeEvent(QResizeEvent *)
{
    QToolBar *main = findChild<QToolBar *>("MainTB");
    QToolBar *status = findChild<QToolBar *>("StatusTB");
    QToolBar *progress = findChild<QToolBar *>("ProgressTB");
    if (!main || !status || !progress)
        return;

    int mainw = 0, statw = 0;
    if (toolBarArea(progress) == toolBarArea(main))
        mainw = qMin(main->width(), main->sizeHint().width());
    if (toolBarArea(progress) == toolBarArea(status))
        statw = qMin(status->width(), status->sizeHint().width());
    if (toolBarBreak(status) || toolBarBreak(main)) {
        mainw = qMax(mainw, statw);
        statw = 0;
    }

    if (toolBarBreak(progress) && (width() - (mainw+statw)) > 0.35*width()) {
        removeToolBarBreak(progress);
        layout()->activate(); //update for re-break check
    }
    //break, re-break if not enough space
    if (!toolBarBreak(progress) && progress->width() < 0.3*width())
        insertToolBarBreak(progress);
}

bool MainWindow::eventFilter(QObject *obj, QEvent *event)
{
    if (event->type() == QEvent::DragEnter) {
        QPushButton *button = qobject_cast<QPushButton *>(obj);
        if (button)
            button->click();
    }
    if (event->type() == QEvent::KeyPress) {
        QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);

        auto spm = SPMTunning::GetIt();
        if (!spm || obj == this)
            return QMainWindow::eventFilter(obj, event);

        if (keyEvent->key() == Qt::Key_Escape)
        {
            spm->motors_stop(); //TODO event filter in spm on qapp?
            return true;
        }

        auto scan = ScanPlot::GetIt();
        if (!scan)
            return QMainWindow::eventFilter(obj, event);

        scan->eventFilter(obj, event);
    }

    if (obj == this) {
        if (event->type() == QEvent::Close) {
            if (statusBarCount()) {
                QMessageBox::information(this, tr("Nspec"),
                                         tr("You should finish or cancel all active operations before you can quit"));
                event->ignore();
                return true;
            }
            if (auto data = DataPlot::GetIt()) {
                bool notSaved = false;
                for (int i = 0; i < data->Container->CBDataLists->count(); i++) {
                    if (((TListWidget *)data->Container->CBDataLists->widget(i))->isUpdated &&
                            ((TListWidget *)data->Container->CBDataLists->widget(i))->DataList.size()) {
                        notSaved = true;
                        break;
                    }
                }
                if (notSaved && QMessageBox::No ==
                        QMessageBox::question(this, tr("Nspec"),
                                              tr("You have unsaved data") + "\n" +
                                              tr("Still want to quit?"), QMessageBox::Yes, QMessageBox::No)) {
                    event->ignore();
                    return true;
                }
            }

            auto setts = options::GetIt();
            auto osc = Oscilloscope::GetIt();
            if (osc && setts && setts->optWin->chMoveXYZToZeroOff->isChecked())
            {
                auto dacs = osc->DacsWidget;
                DoubleChanger **changers = new DoubleChanger*[3];
                changers[0] = dacs->X2;
                changers[1] = dacs->Y2;
                changers[2] = dacs->Z2;
                QVariant datas[3];
                datas[0] = QVariant::fromValue((void *)changers[0]);
                datas[1] = QVariant::fromValue((void *)changers[1]);
                datas[2] = QVariant::fromValue((void *)changers[2]);
                for (int i=0; i<3; i++)
                {
                    int index = dacs->DacsChannel->findData(datas[i]);
                    if (index == -1)
                        continue;
                    dacs->DacsChannel->setCurrentIndex(index);
                    changers[i]->setValue(0);
                    changers[i]->setNewValue(true);

                    int b=0;
                    while (b<(setts->optWin->edZeroWaitPerChannel->value()*0.1)) {
                        qApp->processEvents();
                        QThread::msleep(10);
                        b++;
                    }
                }
            }

            foreach (QWidget *widget, QApplication::topLevelWidgets())
                widget->hide();

            QSplashScreen *close_widget = new QSplashScreen;
            close_widget->setWindowFlags(close_widget->windowFlags() | Qt::WindowStaysOnTopHint);
            close_widget->setPixmap(QPixmap(":/MainToolbar/close_app.png"));
            close_widget->resize(200,100);
            close_widget->setEnabled(false);
            close_widget->show();
            qApp->processEvents();


            ModuleManager::instance->setRunning(false);


            close_widget->hide();
            close_widget->close();
            close_widget->deleteLater();

#if (QT_VERSION >= QT_VERSION_CHECK(4, 8, 0))
            QMessageBox *pool_warn = nullptr;
            while (!QThreadPool::globalInstance()->waitForDone(0)) {
                //*((int *)(-1)) = 0; //abort(); //exit(0); //sepukku //not works! :(
                if (!pool_warn) {
                    pool_warn = new QMessageBox(QMessageBox::Critical, tr("Nspec"),
                                                tr("There are %n thread(s) left that I can't clean up", nullptr, QThreadPool::globalInstance()->activeThreadCount()) + "\n" +
                                                tr("Please kill process \"%1\" manually").arg(qApp->applicationFilePath()));
                    pool_warn->setWindowFlags(pool_warn->windowFlags() | Qt::WindowStaysOnTopHint);
                }
                pool_warn->show();
                qApp->processEvents();
                SleepExport::msleep2(3);
            }
            if (pool_warn) {
                pool_warn->hide();
                qApp->processEvents();
                pool_warn->deleteLater();
            }
#endif

            qApp->quit();

            return true;
        } else
            return false;
    } else
        return QMainWindow::eventFilter(obj, event);
}

MainWindow* MainWindow::GetIt() noexcept
{
    auto mainWindow = ModuleManager::instance->find1VoidPtr("mainwindow");
    return static_cast<MainWindow*>(mainWindow);
}
