#include "networkserver.hpp"

#include <QDataStream>
#include <QTcpSocket>

NetworkServer::NetworkServer(int port, pINC client) : clientCtor(client)
{
    setMaxPendingConnections(1);
    connect(this, SIGNAL(newConnection()), this, SLOT(newConnected()));

    listen(QHostAddress::Any, port);
}

NetworkServer::~NetworkServer()
{
    for (int i = 0; i < clients.size(); i++)
        clients.at(i)->deleteLater();
}

void NetworkServer::newConnected()
{
    while (hasPendingConnections()) {
        QTcpSocket *socket = nextPendingConnection();
        if (!socket)
            return;

        clients.append(clientCtor(socket));
        connect(clients.last(), SIGNAL(closed()), SLOT(newError()));
        emit newClient(clients.last());
    }
}

void NetworkServer::newError()
{
    INetworkClient *client = qobject_cast<INetworkClient *>(sender());
    if (!client)
        return;

    clients.remove(clients.indexOf(client));
}


INetworkClient::INetworkClient(QTcpSocket *client) : socket(nullptr)
{
    setSocket(client);
}

INetworkClient::~INetworkClient()
{
    setSocket(nullptr);
}

void INetworkClient::setSocket(QTcpSocket *socket_)
{
    if (socket) {
        socket->abort();
        socket->deleteLater();
    }

    socket = socket_;
    lastError.clear();
    if (!socket)
        return;

    connect(socket, SIGNAL(readyRead()), SLOT(newRead()));
    connect(socket, SIGNAL(disconnected()), SLOT(newError()));
    connect(socket, SIGNAL(error(QAbstractSocket::SocketError)), SLOT(newError()));
}

void INetworkClient::newError()
{
    disconnect(socket, nullptr, this, nullptr);

    lastError = socket->errorString();

    socket->abort();
    socket->deleteLater();
    socket = nullptr;

    emit closed();
}

void INetworkClient::newRead()
{
    newError();
}


INetworkClient *baseClientCtor(QTcpSocket *socket)
{
    return new BaseNetworkClient(socket);
}

BaseNetworkClient::BaseNetworkClient(QTcpSocket *client) : INetworkClient(client),
    clientsPacketStarted(false), clientsPacketLength(0)
{
    version = QDataStream().version();
}

void BaseNetworkClient::setSocket(QTcpSocket *socket)
{
    clientsPacketStarted = false;
    clientsPacketLength = 0;
    version = QDataStream().version();

    address.clear();
    data.clear();

    if (socket) {
        socket->setSocketOption(QAbstractSocket::LowDelayOption, 1);
        socket->setSocketOption(QAbstractSocket::KeepAliveOption, 1);
    }

    INetworkClient::setSocket(socket);
}

void BaseNetworkClient::newRead()
{
    bool check = true;
    bool newD = false;
    while (check) {
        check = false;
        if (!clientsPacketStarted && socket->bytesAvailable() >= sizeof(quint32)) {
            QDataStream stream(socket);
            stream.setVersion(version);

            quint32 size;
            stream >> size;

            clientsPacketStarted = true;
            clientsPacketLength = size;
        }
        if (clientsPacketStarted && socket->bytesAvailable() >= clientsPacketLength + sizeof(quint32))
        {
            quint32 addr;
            {
                QDataStream stream(socket);
                stream.setVersion(version);
                stream >> addr;
            }

            QByteArray newData = socket->read(clientsPacketLength);
            if (newData.size() != clientsPacketLength) {
                newError();
                return;
            }

            address.append(addr);
            data.append(newData);

            clientsPacketStarted = false;
            clientsPacketLength = 0;

            newD = true;
            check = true;
        }
    }
    if (newD)
        emit newData();
}

void BaseNetworkClient::sendData(quint32 addr, quint8 *data, quint32 size)
{
    if (!socket || socket->state() != QAbstractSocket::ConnectedState)
        return;

    QByteArray outgoingSize;
    QDataStream streamS(&outgoingSize, QIODevice::WriteOnly);
    streamS.setVersion(version);
    streamS << size;
    streamS << addr;
    streamS.setDevice(nullptr);

    if (socket->write(outgoingSize) != outgoingSize.size()) qDebug("socket e1");
    if (socket->write((char *)data, size) != size) qDebug("socket e2");
    socket->flush();
}
