#include "modules.hpp"
#include <QDebug>
#include <ttabwidget.hpp>
#include "options_window.h"
#include <QHBoxLayout>
#include "widgets/threedoverlaybuttons.h"
#include "widgets/twodoverlaybuttons.h"

class EG3000Module;
class DataModule;
class CCDMatrixModule;
class CCDPlotModule;
class ScanModule;
class SPMModule;
class MotorsModule;
class OscilloscopeModule;
class VideoModule;
class OptionsModule;
//class MotorsModule;


ModuleManager *ModuleManager::instance = nullptr;

I_MODULE::~I_MODULE()
{
    if (ModuleManager::instance)
        ModuleManager::instance->unload(this);
}

ModuleManager::ModuleManager()
{
    if (instance)
        throw instance;
    instance = this;
    isRunning = false;

    ModuleManager::instance->addToFactoryGui<OptionsModule>();
    ModuleManager::instance->addToFactory<EG3000Module>();
    ModuleManager::instance->addToFactoryGui<DataModule>();
    ModuleManager::instance->addToFactory<CCDMatrixModule>();
    ModuleManager::instance->addToFactoryGui<CCDPlotModule>();
    ModuleManager::instance->addToFactoryGui<ScanModule>();
    ModuleManager::instance->addToFactoryGui<SPMModule>();
    ModuleManager::instance->addToFactoryGui<MotorsModule>();
    ModuleManager::instance->addToFactoryGui<VideoModule>();
    ModuleManager::instance->addToFactoryGui<OscilloscopeModule>();
}

ModuleManager::~ModuleManager()
{
    instance = nullptr;

    if (isRunning)
        foreach (I_MODULE *m, modules)
            m->setRunning(false);

    foreach (I_MODULE *m, modules)
        delete m;

    modules.clear();
}

I_MODULE *ModuleManager::load(const QLatin1String name, QWidget *parent)
{
    if (!factory.contains(name))
        return nullptr;

    QList<QLatin1String> l;
    if (factory[name].requires)
        l = factory[name].requires();
    for (int i = 0; i < l.size(); i++)
        if (!modules.contains(l.at(i)))
            if (!load(l.at(i), parent))
                qCritical() << "can't create module" << QString(l.at(i)) << "required for" << QString(name);

    I_MODULE *m;
    if (factory[name].creategui)
        m = factory[name].creategui(parent);
    else
        m = factory[name].create();
    add(name, m);
    return m;
}

void ModuleManager::add(const QLatin1String name, I_MODULE *module)
{
    QMapIterator<QLatin1String, I_MODULE *> i(modules);
    while (i.hasNext()) {
        i.next();
        i.value()->connected(name, module);
        module->connected(i.key(), i.value());
    }

    if (isRunning)
        module->setRunning(true); //TODO fix order for dependencies

    modules.insert(name, module);
}

void *ModuleManager::find1VoidPtr(const QLatin1String name) const noexcept
{
    I_MODULE *module = ModuleManager::instance->find1(name);

    if (module == nullptr)
        return nullptr;

    return module->basecast();
}

void *ModuleManager::find1Safety(const QLatin1String name, bool throwExeption) const
{
    auto module = find1VoidPtr(name);

    if (module == nullptr)
    {
        if (!throwExeption)
            return nullptr;

        //assert("Error 0x0001. \"" + name + "\" is not exists or not inited yet!");
        //throw "Error 0x0001. \"" + name + "\" is not exists or not inited yet!";
        //qDebug()<<")";
        qFatal(QString("Error 0x0001. \"" + name + "\" is not exists or not inited yet!").toLatin1());
    }

    return module;
}

I_MODULE *ModuleManager::find1(const QLatin1String name) const
{
    return modules.value(name);
}
const QList<I_MODULE *> ModuleManager::find(const QLatin1String name) const
{
    return modules.values(name);
}

const QLatin1String ModuleManager::find(I_MODULE *module) const
{
    foreach (QLatin1String name, modules.keys(module))
        return name;

    return QLatin1String(nullptr);
}

void ModuleManager::unload(const QLatin1String name)
{
    QList<I_MODULE *> l = modules.values(name);
    if (isRunning)
        foreach (I_MODULE *m0, l)
            m0->setRunning(false);

    foreach (I_MODULE *m0, l)
        foreach (I_MODULE *m, modules)
            if (m != m0)
                m->disconnected(name, m0);

    modules.remove(name);

    foreach (I_MODULE *m0, l)
        delete m0;
}

void ModuleManager::unload(I_MODULE *module)
{
    QLatin1String name = find(module);
    if (!name.latin1())
        return;

    if (isRunning)
        module->setRunning(false);

    foreach (I_MODULE *m, modules)
        if (m != module)
            m->disconnected(name, module);

    modules.remove(name, module);
}

void ModuleManager::setRunning(bool running)
{
    if (isRunning == running)
        return;

    isRunning = running;

    foreach (I_MODULE *m, modules)
        m->setRunning(isRunning);
}

//---------------------------------------------------------------------------------------
#include <QSettings>
#include "mainwindow.h"
#include "mobile/remote_mobile.hpp"
#include "gui/status.hpp"

#include "options/options.h"
class OptionsModule : public I_GUI_MODULE<options>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(OptionsModule) { }

    static const QLatin1String name() { return QLatin1String("options"); }

    void *basecast(const QLatin1String name) override { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<options>::basecast(name); }

    //gui
    QPushButton *createPanelButton() override
    {
        auto btn = new QPushButton(makeHugeIcon(":/MainToolbar/tools.png"), MainWindow::tr("Options"));
        QSettings settings2("Param/gui.conf", QSettings::IniFormat);
        if (settings2.value("gui/Ratis_xyz_stage_mode", false).toBool())
            btn->hide();
        return btn;
    }
};

#include "hardware/EG3000.h"
class EG3000Module : public EG3000_CONTROL, public I_GUI_MODULE_BASE
{
public:
    EG3000Module() {
#ifdef USE_EG3K_EMULATOR
        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        if (settings.value("usb", false).toBool()) {
            SWusb1->setEmulator(true);
            return;
        }
#endif

        start();
    }
    static const QLatin1String name() { return QLatin1String("eg3000"); }
    void *basecast(const QLatin1String name) override {
        if (name == QLatin1String("gui"))
            return static_cast<I_GUI_MODULE_BASE *>(this);
        return name.latin1() != nullptr ? nullptr : static_cast<EG3000_CONTROL *>(this); }
    void connected(const QLatin1String name, I_MODULE *module) override;
    void setRunning(bool on) override {
        if (on) {
            m_waitMainInit = false;
#ifdef USE_EG3K_EMULATOR
            QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
            if (settings.value("usb", false).toBool())
                emit usbStatus(EDok);
#endif
        } else
            thread_close();
    }

    QList<QWidget *> getStatusWidgets() override {
        QList<QWidget *> l;
        l << (QWidget *)nullptr << SWusb1;
        return l;
    }
};

#include "dataplot.h"
class DataModule : public I_GUI_MODULE<DataPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(DataModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        Container->SetAutoSaveTime(settings.value("autosave_time", 30).toInt() * 60 * 1000); //to msec
    }
    static const QLatin1String name() { return QLatin1String("data"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<DataPlot>::basecast(name); }
    void setRunning(bool on) override { if (!on) Container->thread_close(); }

    //gui
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Data_plot.png"), MainWindow::tr("Data")); }

    //net
    void initialSend(MobileInstance *client) override;
    void setValue(int id, QVariant val, QObject *sender);
};

#include "ccd/ccdmatrix.h"
#include "ccd/i_spectr.hpp"
class CCDMatrixModule : public CCDMatrix, public I_MODULE
{
public:
    static const QLatin1String name() { return QLatin1String("ccdmatrix"); }
    void *basecast(const QLatin1String name) override { return name.latin1() != nullptr ? nullptr : static_cast<CCDMatrix *>(this); }
    void setRunning(bool on) override {
        if (on)
            m_waitMainInit = false;
        else {
            thread_close();

            if (m_spectrInterface)
                m_spectrInterface->shutdown();
        }
    }
};

#include "ccd/ccdplot.h"
class CCDPlotModule : public I_GUI_MODULE<CCDPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(CCDPlotModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        AutoSend = settings.value("scan/autosend_data", true).toBool();
    }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<CCDPlot>::basecast(name); }
    static const QLatin1String name() { return QLatin1String("ccdplot"); }
    static const QList<QLatin1String> requires() { return QList<QLatin1String>() << QLatin1String("ccdmatrix"); }
    void connected(const QLatin1String name, I_MODULE *module) override;
    void setRunning(bool on) override { if (on) m_BStart->click(); }

    void initialSend(MobileInstance *client) override;
    void setValue(int id, QVariant val, QObject *sender) override;
    
    QPushButton *createPanelButton() override { return new QPushButton(makeHugeIcon(":/MainToolbar/ccd.png"), MainWindow::tr("Spectroscopy")); }
    QList<QWidget *> getStatusWidgets() override { return QList<QWidget *>() << getDeviceStatusList(); }
};

#include "scanplot.h"
class ScanModule : public I_GUI_MODULE<ScanPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(ScanModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        Autosend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("scan"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<ScanPlot>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);

    //gui
    QPushButton *createPanelButton()
    {
        QSettings settings2("Param/gui.conf", QSettings::IniFormat);
        if (settings2.value("gui/Ratis_xyz_stage_mode", false).toBool()) return nullptr;
        return new QPushButton(makeHugeIcon(":/MainToolbar/Scan.png"), MainWindow::tr("Scan"));
    }

    //net
    void initialSend(MobileInstance *client) override;
    void setValue(int id, QVariant val, QObject *sender) override;
};

#include "motors.hpp"
class MotorsModule : public I_GUI_MODULE<Motors>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(MotorsModule) { }

    static const QLatin1String name() { return QLatin1String("motors"); }
    void connected(const QLatin1String name, I_MODULE *module) override;

    void *basecast(const QLatin1String name) override { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<Motors>::basecast(name); }

    //gui
    QList<QWidget *> getStatusWidgets()
    {
        QSettings settings2("Param/gui.conf", QSettings::IniFormat);
        if (settings2.value("gui/Ratis_xyz_stage_mode", false).toBool()) return QList<QWidget *>();
        return QList<QWidget *>() << (QWidget *)nullptr << new QLabel(MainWindow::tr("Motors:")) << SWmotor1;
    }

    //net
    void setValue(int id, QVariant val, QObject *sender) override;
};

#include "spmtunning.h"
#include "deflect_widget.hpp"
class SPMModule : public I_GUI_MODULE<SPMTunning>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(SPMModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        AutoSend = settings.value("scan/autosend_data", true).toBool();
        isSPMHidded = !settings.value("display/spm", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("spm"); }
    void *basecast(const QLatin1String name) override { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<SPMTunning>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module) override;
    void setRunning(bool on) override {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (on) {
            BLaserOnOff->change_state_man_with_sig(
                        settings.value("state/spm/laser_on", true).toBool());

            bool head, stage;
            {
                QSettings settings("Param/gui.conf", QSettings::IniFormat);
                head = settings.value("display/oscilloscope/head", true).toBool();
                stage = settings.value("display/oscilloscope/stage", true).toBool();
            }
            int fb_en = 0;
            if (!head) fb_en = 2;
            if (!stage) fb_en = 1;
            if (fb_en != 0) {
                CBFBScanner->setEnabled(false);
                CBFBScanner->hide();
                disconnect(CBFBScanner, nullptr, LaserSpot, nullptr);
                disconnect(CBFBScanner, nullptr, LaserSpotWindow, nullptr);
                LaserSpot->setFbscanner(-1);
                LaserSpotWindow->setFbscanner(-1);
            }
            if (fb_en == 2) CBFBScanner->setCurrentIndex(1);
            if (fb_en == 0) {
                QSettings settings("Param/gui.conf", QSettings::IniFormat);
                CBFBScanner->setCurrentIndex(settings.value("state/spm/fb_scanner", 0).toInt());
            }
        } else {
            settings.setValue("state/spm/laser_on", BLaserOnOff->isOn);
            settings.setValue("state/spm/fb_scanner", CBFBScanner->currentIndex());
        }
    }

    //gui
    QPushButton *createPanelButton() override
    {
        if (isSPMHidded) { this->setVisible(false); return nullptr;}
        return new QPushButton(makeHugeIcon(":/MainToolbar/Spm.png"), MainWindow::tr("SPM"));
    }
    QList<QWidget *> getStatusWidgets() override
    {
        QList<QWidget *> l;
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (isSPMHidded) return l;

        l << (QWidget *)nullptr << SWspmfb;
        if (settings.value("display/shuttle_status", true).toBool())
            l << (QWidget *)1 << shuttleStatus;
        return l;
    }

    //net
    void initialSend(MobileInstance *client) override;
    void setValue(int id, QVariant val, QObject *sender) override;
};

#include "motors.hpp"

#include "oscilloscope.h"
#include "autocalib.hpp"
class OscilloscopeModule : public I_GUI_MODULE<Oscilloscope>
{
public:
    GUI_MODULE_CTOR(OscilloscopeModule) { }
    static const QLatin1String name() { return QLatin1String("oscilloscope"); }
    void connected(const QLatin1String name, I_MODULE *module) override;
    void setRunning(bool on) override {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        auto scan = ScanPlot::GetIt();
        if (scan)
        {
            scan->sl_ScanCorrection();
            scan->setScanner(0);
        }

        if (on) {
            for (int i = 0; i < 4; i++)
                CBChannel[i]->setCurrentIndex(
                            qMax(0, CBChannel[i]->findData(
                                     settings.value("state/osc/ch" + QString::number(i), 0))));

            DacsWidget->DacsChannel->setCurrentIndex(
                        qMax(0, DacsWidget->DacsChannel->findText(
                                 settings.value("state/osc/ch_control").toString())));

            CalibWidget->LVMLoad();
            CalibWidget->read_calib_par(CalibWidget->defaultFileName);
        } else {
            for (int i = 0; i < 4; i++)
                settings.setValue("state/osc/ch" + QString::number(i), CBChannel[i]->itemData(CBChannel[i]->currentIndex()));

            settings.setValue("state/osc/ch_control", DacsWidget->DacsChannel->currentText());
        }
    }

    /*QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("display/oscilloscope", true).toBool())
            return 0;
        return new QPushButton(makeHugeIcon(":/MainToolbar/Osc.png"), MainWindow::tr("Oscilloscope"));
    }*/
};

#include "USBVideo/videoplayer.h"
class VideoModule : public I_GUI_MODULE<VideoPlayer>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(VideoModule) { }
    static const QLatin1String name() { return QLatin1String("video"); }
    static const QList<QLatin1String> requires() { return QList<QLatin1String>() << QLatin1String("managedthreadedcleaner"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<VideoPlayer>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module) override;
    void setRunning(bool on) override { if (on) resetCameras(); }

    //gui
    //QPushButton *detachButton() { return PBToNewWindow; }
    //QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/SPMTuning_window/Video.png"), MainWindow::tr("Video") + (inst_idx == 0 ? QString() : " 2")); }
    QList<QWidget *> getStatusWidgets() override { return QList<QWidget *>() << (QWidget *)5 << deviceStatus; } //casting number 5 to... widget?

    //net
    void initialSend(MobileInstance *client) override {}
    void setValue(int id, QVariant val, QObject *sender) override;
};

//---------------------------------------------------------------------------------------
//guidelines:
//connect from user module to common service provider module
//connect from user module to hook/notify module
//be wary of possible order dependencies

#include "hardware/relay.hpp"
void EG3000Module::connected(const QLatin1String name, I_MODULE *module)
{
#ifdef USE_RELAY
    if (name == QLatin1String("relay")) {
        RelayModule *m = (RelayModule *)module;
        if (m->relayCount() < 1)
            return;

        m->enableAll();
        QAction *a;
        SWusb1->addAction(a = new QAction(tr("Power up controller"), SWusb1));
        connect(a, SIGNAL(triggered()), m, SLOT(enableAll()));
        SWusb1->addAction(a = new QAction(tr("Power down controller"), SWusb1));
        connect(a, SIGNAL(triggered()), m, SLOT(disableAll()));
    }
#endif
}

#include "gui/nstslider.h"
void CCDPlotModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
        connect(this, SIGNAL(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)));
    }
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        IExternalScan *externalScan = getNewCCDIExtScan(this);
        m->addExternalScan(MainWindow::tr("CCD"), externalScan);
        if (auto data = DataPlot::GetIt())
            connect(externalScan, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                    data->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));

        connect(hbMotorZFocusUp, SIGNAL(holdingPressed(bool)), m, SLOT(motorZFocusUp(bool)));
        connect(hbMotorZFocusDown, SIGNAL(holdingPressed(bool)), m, SLOT(motorZFocusDown(bool)));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        linkDoubleChangers(m->DacsWidget->Zobj, zPos);
        connect(m->DacsWidget->Zobj->slider, SIGNAL(valueChanged(double)), zPos, SLOT(setValue(double))); //sync value change on setValue
        QObject::connect(zPos, SIGNAL(newValue(double)), this, SLOT(focusChanged(double)));
    }
}

#include <QCheckBox>
#include <QTimer>
void ScanModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(m->Container, SIGNAL(restoreAttributes(plot::TMetaStorage)),
                this, SLOT(restoreAttributes(plot::TMetaStorage)));

        //TODO cant connect externalScans here, race. need modules for extscanmodules
        //for now connect individually
        /*foreach (IExternalScan *sc, externalScans)
            if (sc->metaObject()->indexOfSignal("AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)") != -1)
                connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                        m->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));*/

        connect(this, SIGNAL(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)));
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("spm")) {
        SPMTunning *m = (SPMTunning *)module->basecast();

        #ifdef _OLD_EXTRA_CONTROL
        if (SPMTunning::getLandingMode() != lm01 && SPMTunning::getLandingMode() != lm05)
            BMotorLevel->show();
#endif

        connect(m->DCfbPrec, SIGNAL(newValue()), this, SLOT(check_speed()));
        connect(m->DCfbHeight, SIGNAL(newValue()), this, SLOT(check_speed()));
        connect(m->BFeedbackOnOffMode, SIGNAL(state_changed(bool)), this, SLOT(check_speed()));

        connect(PBSpectr_start_stop, SIGNAL(state_changed(bool)), m, SLOT(s_spectr_start(bool)));
        connect(ev_Speed_spectr, SIGNAL(newValue(double)), m->ev_Speed_spectr, SLOT(setValue(double)));
        connect(CB_Points_spectr, SIGNAL(currentIndexChanged(int)), m->CB_Points_spectr, SLOT(setCurrentIndex(int)));
        connect(m->ev_Z1_spectr, SIGNAL(newValue(double)), ev_Z1_spectr, SLOT(setValue(double)));
        connect(ev_Z1_spectr, SIGNAL(newValue(double)), m, SLOT(s_set_Z1_spectr(double)));
        //connect(ev_Z1_spectr, SIGNAL(newValue(double)), m->ev_Z1_spectr, SLOT(setValue(double)));
        connect(ev_Z1_spectr, SIGNAL(editingFinished()), m, SLOT(sl_up_lines_on()));
        connect(ev_Z1_spectr, SIGNAL(editingStarted()), m, SLOT(sl_up_lines_of()));
        connect(m->NS_Z2_spectr, SIGNAL(newValue(double)), ev_Z2_spectr, SLOT(setValue(double)));
        connect(ev_Z2_spectr, SIGNAL(newValue(double)), m, SLOT(s_set_Z2_spectr(double)));
        connect(ev_Z2_spectr, SIGNAL(editingFinished()), m, SLOT(sl_up_lines_on()));
        connect(ev_Z2_spectr, SIGNAL(editingStarted()), m, SLOT(sl_up_lines_of()));
        for (int i=0; i<m->CB_Channel->count(); i++)
        {
            CB_Channel->addItem(m->CB_Channel->itemText(i));
        }
        connect(CB_Channel, SIGNAL(currentIndexChanged(int)), m->CB_Channel, SLOT(setCurrentIndex(int)));

        //TODO add tab from spm::connected?
        linkDoubleChangers(m->DCfbP, DCfbP, false);
        //todo Oplkill Infinity loop, fix it
        //connect(m->DCfbP, SIGNAL(newValue(double)), this, SLOT(DCfbPNewValue(double)));
        //linkDoubleChangers(m->DCfbP, DCfbPPerc, false);
        linkDoubleChangers(m->DCfbI, DCfbI, false);
        linkDoubleChangers(m->DCfbSetPoint, DCfbSetPoint);
        linkDoubleChangers(m->driveInt, driveInt, true);
        connect(m->DCfbSetPoint->slider, SIGNAL(valueChanged(double)), DCfbSetPoint, SLOT(setValue(double))); //sync value change on setValue - gains, etc.
        connect(m->driveInt->slider, SIGNAL(valueChanged(double)), driveInt, SLOT(setValue(double)));
        connect(deviceSchemeWid->setPointEdit, SIGNAL(valueChanged(double)), m, SLOT(setPoint_DCfbSetPoint_newValue(double)));

        connect(m->BFeedbackOnOff, SIGNAL(state_changed(bool)), this, SLOT(FeedBackChangedUpdateMag(bool)));
        connect(deviceSchemeWid->feedBack, SIGNAL(changedState(bool)), this, SLOT(FeedBackChangedUpdateMag(bool)));

        double maxTipDriveAmplitude = 3.0;
        {
            QSettings settings("Param/tipMonitor.conf", QSettings::IniFormat);
            maxTipDriveAmplitude = settings_double(settings, "max_tip_drive_amplitude", 3.0);
        }
        rectAmplitude->setValue_man((m->driveInt->value()/maxTipDriveAmplitude)*100.0);
        connect(m->driveInt, SIGNAL(newValue(double)), this, SLOT(tipDriveNewValue(double)));
        connect(rectAmplitude, SIGNAL(valueChanged(double)), this, SLOT(tipDriveNewValue_FromSlider(double)));

        //connect(m->DCfbSetPoint, SIGNAL(newValue(double)), this, SLOT(DCfbSetPointChanged(double)));

        DCfbPPerc->setValue(m->DCfbP->value()*10.0);
        rectSlVelocity->setValue_man(m->DCfbP->value()*10.0);
        DCfbPPerc->setNewValue();

        connect(spmMonitorFeedback, SIGNAL(state_changed(bool)), m, SLOT(feedBackClickedInScan(bool)));
        spmMonitorFeedback->change_state_man(m->BFeedbackOnOff->isOn);

        connect(deviceSchemeWid->oscilator->btnAutoFreqScan, SIGNAL(clicked()), m, SLOT(s_FreqScanStartAuto()));
        deviceSchemeWid->oscilator->btnFreqStartScan->setProperty("id", DataID::SPM_freq_scan);
        connect(deviceSchemeWid->oscilator->btnFreqStartScan, SIGNAL(state_changed(bool)), m, SLOT(s_FreqScanStart(bool)));
        fillCBchannels_(deviceSchemeWid->oscilator->cbChannel, false);
        deviceSchemeWid->oscilator->cbChannel->setCurrentIndex(m->CB_fr_Channel->findData(srcMag1));
        deviceSchemeWid->setPointEdit->setProperty("id", DataID::SPM_SetPoint);
        //connect(deviceSchemeWid->setPointEdit, SIGNAL(newValue()), m, SLOT(s_sendSP()));
        connect(deviceSchemeWid->psd, SIGNAL(ChangedValue(bool)), m->BLaserOnOff, SLOT(change_state_man_with_sig(bool)));
        connect(deviceSchemeWid->feedBack, SIGNAL(changedState(bool)), this, SLOT(feedBackClickedInSPM(bool)));
        connect(deviceSchemeWid->feedBack, SIGNAL(changedState(bool)), m, SLOT(s_setFBOnOff(bool)));
        connect(deviceSchemeWid->feedBack, SIGNAL(changedState(bool)), m, SLOT(Zpos_on_off(bool)));
        connect(deviceSchemeWid->oscilator->chbSecHarmonic, SIGNAL(toggled(bool)), m->capMode, SLOT(setChecked(bool)));
        connect(rectSlFbi->centralWidget, SIGNAL(clicked(bool)), deviceSchemeWid->feedBack, SLOT(setState(bool)));
        deviceSchemeWid->ZPiezoPosEdit[0]->setProperty("id", DataID::HV_DAC_X2);
        deviceSchemeWid->ZPiezoPosEdit[1]->setProperty("id", DataID::HV_DAC_Y2);
        deviceSchemeWid->ZPiezoPosEdit[2]->setProperty("id", DataID::HV_DAC_Z2);
        connect(deviceSchemeWid->PIDSwitch, SIGNAL(indexChanged(int)), m, SLOT(pidSwitch_newValue(int)));
        connect(deviceSchemeWid->plusSwitch, SIGNAL(indexChanged(int)), m, SLOT(plusSwitch_newValue(int)));
        connect(deviceSchemeWid->bias->bias, SIGNAL(valueChanged(double)), m, SLOT(edBias_Vtip_newValue(double)));
        connect(deviceSchemeWid->LFLSwitch, SIGNAL(indexChanged(int)), m, SLOT(lflSwitch_lflCurrent_newValue(int)));
        connect(deviceSchemeWid->DFLSwitch, SIGNAL(indexChanged(int)), m, SLOT(dflSwitch_extCurrent_newValue(int)));
        connect(deviceSchemeWid->pid->tsbFBSign, SIGNAL(state_changed(bool)), m->B_PID_sign, SLOT(change_state_man_with_sig(bool)));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        connect(m->CalibWidget, SIGNAL(sigScanError()), this, SLOT(s_scanError()));
        connect(this, SIGNAL(recalibrateSensors(int)), m->acf, SLOT(doXYcalib(int)));
        connect(m->acf, SIGNAL(calibDone(bool)), this, SLOT(recalibDone(bool)));

        connect(m->DacsWidget, SIGNAL(xUpdate(double)), this, SLOT(sl_setCurrentX(double)));
        connect(m->DacsWidget, SIGNAL(yUpdate(double)), this, SLOT(sl_setCurrentY(double)));
        connect(m->DacsWidget, SIGNAL(x2Update(double)), this, SLOT(sl_setCurrentX2(double)));
        connect(m->DacsWidget, SIGNAL(y2Update(double)), this, SLOT(sl_setCurrentY2(double)));

        connect(this, SIGNAL(sync_scan_x(double)), m->DacsWidget, SIGNAL(xUpdate(double)));
        connect(this, SIGNAL(sync_scan_y(double)), m->DacsWidget, SIGNAL(yUpdate(double)));
        connect(this, SIGNAL(sync_scan_x2(double)), m->DacsWidget, SIGNAL(x2Update(double)));
        connect(this, SIGNAL(sync_scan_y2(double)), m->DacsWidget, SIGNAL(y2Update(double)));

        connect(this, SIGNAL(sig_XYPosition(double, double)), m->DacsWidget, SLOT(Go_to_XY(double, double)));
        connect(this, SIGNAL(sig_XYPosition2(double, double)), m->DacsWidget, SLOT(Go_to_XY2(double, double)));

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));

        connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(sl_CB_update()));
        connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(SetSectionUnits()));

        connect(deviceSchemeWid->ZPiezoPosEdit[0], SIGNAL(newValue(double)), m->DacsWidget, SLOT(Go_to_X2(double)));
        connect(deviceSchemeWid->ZPiezoPosEdit[1], SIGNAL(newValue(double)), m->DacsWidget, SLOT(Go_to_Y2(double)));
        connect(deviceSchemeWid->ZPiezoPosEdit[2], SIGNAL(newValue(double)), m->DacsWidget, SLOT(Go_to_Z2(double)));

        connect(deviceSchemeWid->ZPiezoPosEdit[0], SIGNAL(newValue(double)), m->DacsWidget, SLOT(zPiezoPosX_X2_newValue(double)));
        connect(deviceSchemeWid->ZPiezoPosEdit[1], SIGNAL(newValue(double)), m->DacsWidget, SLOT(zPiezoPosY_Y2_newValue(double)));
        connect(deviceSchemeWid->ZPiezoPosEdit[2], SIGNAL(newValue(double)), m->DacsWidget, SLOT(zPiezoPosZ_Z2_newValue(double)));

        connect(m->DacsWidget, SIGNAL(x2Update(double)), deviceSchemeWid->ZPiezoPosEdit[0], SLOT(setValue(double)));
        connect(m->DacsWidget, SIGNAL(y2Update(double)), deviceSchemeWid->ZPiezoPosEdit[1], SLOT(setValue(double)));
        connect(m->DacsWidget, SIGNAL(z2Update(double)), deviceSchemeWid->ZPiezoPosEdit[2], SLOT(setValue(double)));

        connect(m->CalibWidget->CBlvm[0], SIGNAL(toggled(bool)), this, SLOT(setLowVoltageMode(bool)));
        connect(m->CalibWidget->CBlvm[1], SIGNAL(toggled(bool)), this, SLOT(setLowVoltageMode(bool)));
        connect(m->CalibWidget->Scanner_type, SIGNAL(currentChanged(int)), this, SLOT(setScannerLowVoltageMode(int)));
    }
    if (name == QLatin1String("video")) {
        VideoPlayer *v = (VideoPlayer *)module->basecast();

        CBChannel[videoChannelId]->setItemText(0, tr("VideoCam"));

        auto list = v->getVideoProfiles();
        for (int i = 0; i < list->count(); i++)
        {
            comboVideoProfile->addItem(list->at(i));
            videoChannelMagnification->addItem(list->at(i));
        }
        connect(comboVideoProfile, SIGNAL(currentIndexChanged(int)), v, SLOT(sl_select_config(int)));
        connect(videoChannelMagnification, SIGNAL(currentIndexChanged(int)), v, SLOT(sl_select_config(int)));
        //connect(v, SIGNAL(VideoProfileChanged(int)), this, SLOT(videoProfileChanged(int)));

        connect(BAutoFocus, SIGNAL(state_changed(bool)), this, SLOT(autoFocus_ChangedState(bool)));
        connect(BAutoFocus2, SIGNAL(state_changed(bool)), this, SLOT(autoFocus2_ChangedState(bool)));
        connect(videoChannelAutoFocusButtonPAF, SIGNAL(state_changed(bool)), this, SLOT(autoFocus2_ChangedState(bool)));
        connect(v, SIGNAL(AutoFocus2Ended(bool)), this, SLOT(autoFocus2Ended(bool)));

        connect(Plot->overlayButtons, SIGNAL(replacedGUIPlotResized(int, QSize)),
                this, SLOT(videoChannelResized(int, QSize)));
        connect(Plot->overlayButtons, SIGNAL(replacedGUIPlotSetVisible(int, bool)),
                this, SLOT(videoChannelSetVisibleGUI(int, bool)));
        onVideoCBChannelChanged(CBChannel[videoChannelId]->currentIndex());
    }
    if (name == QLatin1String("options")) {
        options *opt = (options *)module->basecast();

        opt->optWin->applyFilterSettings();
        opt->optWin->applyFilter2DSettings();
    }
}

void SPMModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(m->Container, SIGNAL(restoreAttributes(plot::TMetaStorage)),
                this, SLOT(restoreAttributes(plot::TMetaStorage)));

        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        m->Plot->TwoDPlotSPM = MyPlScan;
        m->Plot->ControlLayout->addWidget(MyPlScan->PlotControl);
        MyPlScan->PlotControl->hide();
        connect(MyPlScan->Pure2dPlot, SIGNAL(doubleClicked()), m->Plot, SLOT(plot2dDoubleClicked()));
        connect(MyPlScan->Pure2dPlot, SIGNAL(ShowPlotControl()), m->Plot, SLOT(Show2dPlotControl()));

        IExternalScan *sc;
        m->addExternalScan(MainWindow::tr("JM frce crv"), sc = getNewJMIExtScan(this));
        if (auto data = DataPlot::GetIt())
            connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                    data->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));
        m->spmMonitorGBHlay->addWidget(LaserSpotWindow, 5);
        m->LaserSpotWindow = LaserSpotWindow;
        //connect(LaserSpotWindow, SIGNAL(doubleClicked()), m, SLOT(doubleClickedLaserSpotWindow()));

        connect(BFeedbackOnOff, SIGNAL(state_changed(bool)), m, SLOT(feedBackClickedInSPM(bool)));
        connect(m->rectSlFbi->centralWidget, SIGNAL(clicked(bool)), BFeedbackOnOff, SLOT(change_state_man_with_sig(bool)));
        connect(DCfbP, SIGNAL(newValue(double)), m, SLOT(DCfbPNewValue(double)));
        dynamic_cast<QPushButton*>(m->rectSlFbi->centralWidget)->setChecked(BFeedbackOnOff->isOn);
        m->spmMonitorFeedback->change_state_man(BFeedbackOnOff->isOn);
        connect(m->deviceSchemeWid->oscilator->cbChannel, SIGNAL(currentIndexChanged(int)), CB_fr_Channel, SLOT(setCurrentIndex(int)));

        // auto freq scan buttons
        connect(LaserSpot->probeCheckScanBtn, SIGNAL(clicked()), this, SLOT(s_FreqScanStartAuto()));
        connect(LaserSpotWindow->probeCheckScanBtn, SIGNAL(clicked()), this, SLOT(s_FreqScanStartAuto()));
    }
    if (name == QLatin1String("motors")) {
        Motors *m = (Motors *)module->basecast();
        motors = m;

        m->addLandingSett(OptionsForMotor);
        m->addShuttleSett(OptionsForShuttle->GrBUp, 0);
        m->addShuttleSett(OptionsForShuttle->GrBMap, 1);

        active_shuttle_motors = getLandingMotors();

        connect(m, SIGNAL(mland_stopped()), this, SLOT(slot_mland_stopped()));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        connect(m->acf, SIGNAL(calibInProgress(bool)), this, SLOT(s_setFBmasked(bool)));

        DoubleChanger *Head_Dac = new DoubleChanger(-10, 10);
        Head_Dac->setProperty("id", DataID::SPM_Vtip);
        linkDoubleChangers(Vtip, Head_Dac);
        QSettings settingsGui("Param/gui.conf", QSettings::IniFormat);
        if (!settingsGui.value("gui/Ratis_xyz_stage_mode", false).toBool())
            m->DacsWidget->insertChannel(tr("V Tip"), tr("V"), Head_Dac, 0);

        connect(this, SIGNAL(sig_osc_start()), m, SLOT(sl_osc_start()));
        connect(this, SIGNAL(sig_osc_stop()), m, SLOT(sl_osc_stop()));
        connect(this, SIGNAL(sig_osc_channel(int)), m, SLOT(setChA(int)));
        connect(this, SIGNAL(sig_osc_channelB(int)), m, SLOT(setChB(int)));
        connect(m->ePeriod, SIGNAL(newValue(int)), ePeriod, SLOT(setValue(int)));
        connect(ePeriod, SIGNAL(newValue(int)), m->ePeriod, SLOT(setValue(int)));
        connect(m, SIGNAL(oscRefreshed()), this, SLOT(osc_refresh()));


        connect(m->DacsWidget, SIGNAL(zUpdate(double)), NS_Z_spectr, SLOT(setValue(double)));

        connect(this, SIGNAL(sig_Zpos(double)), m->DacsWidget, SLOT(Go_to_Z(double)));
        connect(this, SIGNAL(sig_Z2pos(double)), m->DacsWidget, SLOT(Go_to_Z2(double)));

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));

        connect(this, SIGNAL(input_calib_updated()), m->CalibWidget, SIGNAL(input_calib_updated()));
        connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(sl_setPlotMode()));

        connect(m, SIGNAL(newOscMAGData(QList<double>, int, int)), LaserSpot, SLOT(newOscMAGData(QList<double>, int, int)));
        connect(m, SIGNAL(newOscMAGData(QList<double>, int, int)), LaserSpotWindow, SLOT(newOscMAGData(QList<double>, int, int)));

        connect(m, SIGNAL(newOscHeightRMS(double, QString)), LaserSpot, SLOT(newOscHeightRMS(double, QString)));
        connect(m, SIGNAL(newOscHeightRMS(double, QString)), LaserSpotWindow, SLOT(newOscHeightRMS(double, QString)));
    }
}
void MotorsModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
    }
    if (name == QLatin1String("options")) {
        options *m = (options *)module->basecast();

        QSettings settings2("Param/gui.conf", QSettings::IniFormat);
        if (!settings2.value("gui/Ratis_xyz_stage_mode", false).toBool())
            m->addNewTab(mainw, QWidget::tr("MotorsAll2"), 2);
        else
            mainw->setVisible(false);
    }
}

void OscilloscopeModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        QSettings settings2("Param/gui.conf", QSettings::IniFormat);
        if (settings2.value("gui/Ratis_xyz_stage_mode", false).toBool())
        {
            m->Plot->ThreeDPlots[0]->setParent(this);
            CalibWidget->threeDPlot = m->Plot->ThreeDPlots[0];
            CalibWidget->threeDPlot->secondParent = nullptr;
            CalibWidget->threeDPlot->singleInTabPage = true;
            CalibWidget->threeDPlot->SetTitle("");
            CalibWidget->threeDPlotGB = new QGroupBox(tr("Scan plot"));
            QHBoxLayout *lay = new QHBoxLayout(CalibWidget->threeDPlotGB);
            lay->addWidget(CalibWidget->threeDPlot, 1);
            CalibWidget->threeDPlotGB->setMinimumWidth(440);
            CalibWidget->vertCalibLay->addWidget(CalibWidget->threeDPlotGB, 1);

            disconnect(CalibWidget->threeDPlot, SIGNAL(doubleClicked()), m->Plot, SLOT(plotDoubleClicked()));

            m->pointSync->setParent(this);
            CalibWidget->pointSync = m->pointSync;
            auto gb = new QGroupBox(tr("Synchronization"));
            auto lay2 = new QHBoxLayout(gb);
            CalibWidget->stageLayout->addWidget(gb);
            lay2->addWidget(CalibWidget->pointSync);

            CalibWidget->eventFilter(nullptr, new QEvent(QEvent::Resize));
            CalibWidget->threeDPlot->Child->on_BRestoreScale_clicked();
        }
    }
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->rThread.readers.add(acf);
        m->wThread.writers.add(this);
        m->wThread.writers.add(DacsWidget);

        connect(&m->wThread, SIGNAL(started()), acf, SLOT(reset()));
        connect(&m->rThread, SIGNAL(capBits(int,long)), acf, SLOT(capBits(int,long)));

        connect(m, SIGNAL(device_status(int)), this, SLOT(device_status(int)));

        //calib
        m->rThread.readers.add(CalibWidget);
        m->wThread.writers.add(CalibWidget);

        connect(&m->rThread, SIGNAL(capBits(int,long)), CalibWidget, SLOT(capBits(int,long)));

        connect(CalibWidget, SIGNAL(program_controller(QString)), m, SLOT(program(QString)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("options")) {
        options *m = (options *)module->basecast();

        m->addNewTab(mainWidget, tr("Oscilloscope"), 4);
    }
}

void VideoModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        linkDoubleChangers(m->DacsWidget->Zobj, zPos);
        connect(m->DacsWidget->Zobj->slider, SIGNAL(valueChanged(double)), zPos, SLOT(setValue(double))); //sync value change on setValue

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));
    }
    if (name == QLatin1String("options")) {
        options *m = (options *)module->basecast();

        m->addNewTab(mainWidget, tr("Video"), 3);
    }
}


//---------------------------------------------------------------------------------------
#include "container/datalist.hpp"
void DataModule::initialSend(MobileInstance *client)
{
    QDir dir(TContainer::getDataDir());
    client->serv->setPropIdVarS(client->socket, DataID::DataOpenList, dir.entryList(QStringList() << "*.nstdat"));
}
void DataModule::setValue(int id, QVariant val, QObject *)
{
//    if (id == DataID::DataOpenName) {
//        QString fname = TContainer::getDataDir() + "/" + val.toString();
//        if (val.toString().startsWith("NST_TAB_IDX/")) {
//            bool ok = false;
//            int idx = val.toString().split("/").last().toInt(&ok);
//            if (!ok) return;
//            Container->CBDataLists->setCurrentIndex(idx);
//            Container->SendContainerData();
//            return;
//        }
//        for (int i = 1; i < Container->CBDataLists->count(); i++) {
//            TListWidget *w = ((TListWidget *)Container->CBDataLists->widget(i));
//            if (QFileInfo(w->filename).absoluteFilePath() == QFileInfo(fname).absoluteFilePath()) {
//                Container->CBDataLists->setCurrentIndex(i);
//                Container->SendContainerData();
//                return;
//            }
//        }
//        Container->OpenFile(fname, true);
//    }
//    if (id == DataID::ActiveData) {
//        int id = qMin(Container->CurrentDataList->findItems("*", Qt::MatchWildcard).size()-1, val.toInt());
//        if (id < 0) return;
//        TListItem *it = Container->CurrentDataList->FindItem(Container->CurrentDataList->findItems("*", Qt::MatchWildcard).at(id));
//        if (!it) return;
//        Container->CurrentDataList->on_itemPressed(it->Item);
//    }
}

void ScanModule::initialSend(MobileInstance *client)
{
    QStringList chnames;
    QList<QVariant> chdata;
    for (int i = 0; i < CBChannel[1]->count(); i++) {
        chnames.append(CBChannel[1]->itemText(i));
        chdata.append(CBChannel[1]->itemData(i));
    }
    QList<QVariant> chIndexes;
    chIndexes << CBChannel[0]->currentIndex()
              << CBChannel[1]->currentIndex()
              << CBChannel[2]->currentIndex()
              << CBChannel[3]->currentIndex();
    client->serv->setPropIdVarS(client->socket, DataID::ChannelList,
                                QList<QVariant>() << chnames << QVariant(chdata) << QVariant(chIndexes) << srcCCD);

    client->serv->setPropIdVarS(client->socket, DataID::ScanNxNy, qSqrt(Nx_points * Ny_points));

    QList<QVariant> sects;
    sects << CBSec[0]->isChecked()
          << CBSec[1]->isChecked()
          << CBSec[2]->isChecked()
          << CBSec[3]->isChecked();
    client->serv->setPropIdVarS(client->socket, DataID::Channel_sectionsUpdate, QVariant(sects));

    if (!m_BStart->on_off) {
        Send2dPlotData();
        Send3dPlotData1();
        Send3dPlotData2();
        Send3dPlotData3();
        Send3dPlotData4();
    }
}
void ScanModule::setValue(int id, QVariant val, QObject *sender)
{
    if (id == DataID::ActivePlot_)
    {
        int plotId = val.toInt();
        if (plotId >= 0 && plotId < numScanChannel)
            Plot->SetActivePlot(plotId);
    }
    
    if (id == DataID::plot3DCCDColorPick)
    {
        int NOfPalette = val.toInt();
        Plot->SetActivePlot(1);
        Plot->PlotControl->AGPalettes->actions()[NOfPalette]->trigger();
        Plot->PlotControl->AGPalettes->actions()[NOfPalette]->setChecked(true);
    }
    
    if (id == DataID::ScanChB)
    {
        if (val.toInt() == srcCCD && CBExternalMode->currentIndex() != 1)
            CBExternalMode->setCurrentIndex(1);
        else if (CBExternalMode->currentIndex() == 1)
            CBExternalMode->setCurrentIndex(0);
    }
    
    if (id == DataID::scan3DSave)
    {
        int plotId = val.toInt();
        
        plot::T3dDataOwner src;
        plot::TMetaStorage attributes;
        if (!Plot->SavePlot(plotId, &src, &attributes))
            return;
        
        FillContainer(plotId, &src, &attributes);
        
        auto mob = MobileServer::GetIt();
        if (mob && mob->need_send())
        {
            QList<QVariant> list;
            list << plotId
                 << QVariant::fromValue(attributes);
            
            send3dPlot(PlotID::Scan1+plotId, Plot, plotId, false, true);
            mob->setPropIdVar(DataID::addedScanDataToContainer_3d, QVariant::fromValue(list));
        }
    }
    
    if (id == DataID::scanConvFuncRangeChanged)
    {
        auto rect = val.toRectF();
        double Min = rect.left();
        double Max = rect.right();
        
        currExtScan->setSelectionRange(Min, Max);
    }

    if (id >= DataID::needPlot3D_A && id <= DataID::needPlot3D_D)
        if (sender)
        {
            int numPlot = id - DataID::needPlot3D_A;
            QString property = "scan_plot_" + QString::number(numPlot);
            sender->setProperty(property.toLocal8Bit().constData(), val);
            if (val.toBool())
                send3dPlot(PlotID::Scan1+numPlot, Plot, numPlot, false, true);
        }

    if (id == DataID::VideoOn)
        BMakeScreen->change_state_man_with_sig(val.toBool());
    if (id == DataID::needPlot2D)
        if (sender)
        {
            sender->setProperty(QString("need_2dPlot").toLocal8Bit().constData(), val);
            send2dPlot(PlotID::ScanSection, Plot->TwoDPlot);
        }

    if (id == DataID::plot2DSaveToData)
        Plot->TwoDPlot->overlayButtons->B2DSendToData->click();

    if (id == DataID::plot3DChangeSize)
    {
        auto list = val.toList();
        int indexPoints = list[0].toInt();
        double deltaValue = list[1].toDouble();
        if (CBPointsOX->currentIndex() != indexPoints)
            CBPointsOX->setCurrentIndex(indexPoints);
        if (LEdDeltaX->isEnabled() && LEdDeltaX->value() != deltaValue)
            LEdDeltaX->setValue(deltaValue);
    }

    if (id == DataID::plot3DGetSizes)
    {
        auto plot = Plot->ThreeDPlots[0];
        QList<QVariant> list;
        list << CBPointsOX->currentIndex();
        list << LEdDeltaX->value();
        list << LEdDeltaX->isEnabled();
        list << plot->XYAxisUnit.Legend;
        list << plot->XYAxisUnit.Multiplier;

        auto mob = MobileServer::GetIt();
        if (mob && mob->need_send())
            mob->setPropIdVar(DataID::plot3DChangeSize, list);
    }

    if (id == DataID::plot3DAreaSelected)
    {
        auto list = val.toList();
        double x1 = list[0].toDouble();
        double x2 = list[1].toDouble();
        double y1 = list[2].toDouble();
        double y2 = list[3].toDouble();
        double selSize = list[4].toDouble();

        Plot->SetPlotsXScales(x1, x2);
        Plot->SetPlotsYScales(y1, y2);
        LEdDeltaX->setValue(selSize);

        for (int i=0; i<Plot->ThreeDPlots.size(); i++)
            Plot->ThreeDPlots[i]->ResizeQPlot();
    }

    if (id == DataID::motorZFocusMove)
    {
        int speed = val.toInt();
        auto m = Motors::GetIt();
        auto v = VideoPlayer::GetIt();
        auto mZfocus = m->getRoleMotor(Motors::MotorRoles::RoleZ2Focus);
        if (speed == 1)
            m->goDownE(mZfocus, v->motorZFocusSpeed);
        else if (speed == -1)
            m->goUpE(mZfocus, v->motorZFocusSpeed);
        else
            m->goStop(mZfocus);
    }

    if (id == DataID::motorXStageMove)
    {
        int speed = val.toInt();
        auto m = Motors::GetIt();
        auto v = VideoPlayer::GetIt();
        auto mXStage = m->getRoleMotor(Motors::MotorRoles::RoleStageX);
        if (speed == 1)
            m->goDownE(mXStage, v->motorXYStageSpeed);
        else if (speed == -1)
            m->goUpE(mXStage, v->motorXYStageSpeed);
        else
            m->goStop(mXStage);
    }

    if (id == DataID::motorYStageMove)
    {
        int speed = val.toInt();
        auto m = Motors::GetIt();
        auto v = VideoPlayer::GetIt();
        auto mYStage = m->getRoleMotor(Motors::MotorRoles::RoleStageY);
        if (speed == 1)
            m->goDownE(mYStage, v->motorXYStageSpeed);
        else if (speed == -1)
            m->goUpE(mYStage, v->motorXYStageSpeed);
        else
            m->goStop(mYStage);
    }

    if (id == DataID::videoChangedMagnification)
    {
        int index = val.toInt();
        if (index >= 0 && index < videoChannelMagnification->count())
            videoChannelMagnification->setCurrentIndex(index);
        else
            updateVideoChannel();
    }

    if (id == DataID::plot3DRestoreScale)
    {
        Plot->PlotControl->BRestoreScale->click();
    }

    if (id == DataID::videoMafPrecise)
    {
        bool on = val.toBool();
        if (on)
        {
            QTimer *timer = new QTimer();
            timer->setSingleShot(true);
            connect(timer, SIGNAL(timeout()), this, SLOT(startMAFPrcs()));
            timer->start(10); //starting in another thread
        }
        else
            videoChannelAutoFocusButtonPrec->change_state_man_with_sig(false);
    }

    if (id == DataID::videoMafFast)
    {
        bool on = val.toBool();
        if (on)
        {
            QTimer *timer = new QTimer();
            timer->setSingleShot(true);
            connect(timer, SIGNAL(timeout()), this, SLOT(startPAFFast()));
            timer->start(10); //starting in another thread
        }
        else
            videoChannelAutoFocusButtonFast->change_state_man_with_sig(false);
    }

    if (id == DataID::videoPaf)
    {
        bool on = val.toBool();
        if (on)
        {
            QTimer *timer = new QTimer();
            timer->setSingleShot(true);
            connect(timer, SIGNAL(timeout()), this, SLOT(startPAF()));
            timer->start(10); //starting in another thread
        }
        else
            videoChannelAutoFocusButtonPAF->change_state_man_with_sig(false);
    }

    if (id == DataID::videoFocusUpDown)
    {
        int newSt = val.toInt();
        if (newSt == -1)
        {
            auto mob = MobileServer::GetIt();
            if (mob && mob->need_send())
            {
                //mob->setPropIdVar(DataID::videoFocusUpDown, videoChannelAutoFocusUpDownBtn->state);
            }
        }
        else
        {
            //videoChannelAutoFocusUpDownBtn->change_state(newSt);
        }
    }

    #ifdef _SCAN_SLICE
    if (id == DataID::plot3DSliceScales)
    {
        int plotId = val.toInt();
        auto plot = Plot->ThreeDPlots.at(plotId);
        if (plot->sliceBounds)
        {
            QList<QVariant> list;
            list << plotId;
            list << plot->sliceX.arr[0];
            list << plot->sliceX.arr[1];
            list << plot->sliceY.arr[0];
            list << plot->sliceY.arr[1];
            list << plot->XYAxisUnit.Multiplier;
            auto mob = MobileServer::GetIt();
            if (mob && mob->need_send())
                mob->setPropIdVar(DataID::plot3DSliceScales, list);
        }
    }
#endif

    if (id == DataID::scanFilterOnOff)
        if (sender)
        {
            auto list = val.toList();
            int plotId = list.at(0).toInt();
            if (list.size() == 1)
            {
                QList<QVariant> list;
                list << plotId
                     << Plot->ThreeDPlots[plotId]->isFiltering;
                auto mob = MobileServer::GetIt();
                if (mob && mob->need_send())
                    mob->setPropIdVar(DataID::scanFilterOnOff, list);
            }
            else
            {
                bool pressed = list.at(1).toInt();
                Plot->ThreeDPlots[plotId]->setFiltering(pressed);
                if (Plot->ActivePlot == plotId)
                {
                    Plot->overlayButtons->setButtonActive(Plot->overlayButtons->BFilters, pressed);
                }
            }
        }

    if (id == DataID::plot2DFilterOnOff)
        if (sender)
        {
            if (val.isNull())
            {
                auto mob = MobileServer::GetIt();
                if (mob && mob->need_send())
                    mob->setPropIdVar(DataID::plot2DFilterOnOff, Plot->TwoDPlot->Pure2dPlot->isFiltering);
            }
            else
            {
                bool pressed = val.toBool();
                Plot->TwoDPlot->overlayButtons->B2DFilter->blockSignals(true);
                Plot->TwoDPlot->overlayButtons->B2DFilter->change_state_man(pressed);
                Plot->TwoDPlot->setFiltering(pressed);
                Plot->TwoDPlot->overlayButtons->setButtonActive(Plot->TwoDPlot->overlayButtons->B2DFilter, pressed);
                Plot->TwoDPlot->overlayButtons->B2DFilter->blockSignals(false);
            }
        }

    if (id == DataID::scanDirectionChanged)
        if (sender)
        {
            m_BStart->changeDirection(static_cast<swapDir::Direction>(val.toInt()), true);
        }

    if (id == DataID::rectSpeed)
    {
        rectSlFbp->setValue(val.toDouble());
    }

    if (id == DataID::plot2DDefaultFilterSettings)
        if (sender)
        {
            auto mob = MobileServer::GetIt();
            auto scan = ScanPlot::GetIt();
            if (mob && scan)
            {
                int fitOrder = FindCheckedItemIn(scan->Plot->TwoDPlot->PlotControl->AGFitOrders);
                mob->setPropIdVar(DataID::plot2DDefaultFilterSettings, fitOrder);
            }
        }

    if (id == DataID::needRectSpeed && sender)
    {
        auto mob = MobileServer::GetIt();
        if (mob && mob->need_send())
        {
            mob->setPropIdVar(DataID::rectMaxSpeed, rectSlFbp->GetMaximum());
            mob->setPropIdVar(DataID::rectSpeed, rectSlFbp->Value());
        }
    }

    if (id == DataID::rectTipPush)
    {
        rectSlFbi->setValue(val.toDouble());
    }

    if (id == DataID::rectTipDrive)
    {
        rectAmplitude->setValue(val.toDouble());
    }

    if (id == DataID::needRectTipDrive && sender)
    {
        auto mob = MobileServer::GetIt();
        if (mob && mob->need_send())
            mob->setPropIdVar(DataID::rectTipDrive, rectAmplitude->Value());
    }

    if (id == DataID::needRectTipPush && sender)
    {
        auto mob = MobileServer::GetIt();
        if (mob && mob->need_send())
            mob->setPropIdVar(DataID::rectTipPush, rectSlFbi->Value());
    }

    if (id == DataID::rectSensitivity)
    {
        rectSlVelocity->setValue(val.toDouble());
    }

    if (id == DataID::smartLandingChanged)
    {
        smartLanding->animateClick();
    }

    if (id == DataID::approachSOff)
    {
        smartLandingBtn1->animateClick();
    }

    if (id == DataID::approachSOn)
    {
        smartLandingBtn2->animateClick();
    }

    if (id == DataID::plot2DSelectLine)
    {
        Plot->TwoDPlot->overlayButtons->CBLineSelector->setCurrentIndex(val.toInt());
    }

    if (id == DataID::plot2DLineFontSettings)
    {
        QList<QVariant> list = val.toList();
        QColor lineColor = list.at(0).value<QColor>();
        int lineWidth = list.at(1).toInt();
        int font = list.at(2).toInt();
        //Plot->TwoDPlot->overlayButtons->CBLineColor->setCurrentIndex(lineColor);
        Plot->TwoDPlot->overlayButtons->CBLineWidth->setCurrentIndex(lineWidth);
        Plot->TwoDPlot->overlayButtons->CBFont->setCurrentIndex(font);

        Plot->TwoDPlot->setCurveColor(lineColor);
    }

    if (id == DataID::rectTipDriveAutoControll)
    {
        auto btn = dynamic_cast<QPushButton*>(rectAmplitude->centralWidget);
        if (val.isNull())
        {
            auto mob = MobileServer::GetIt();
            if (mob && mob->need_send())
                mob->setPropIdVar(DataID::rectTipDriveAutoControll, btn->isChecked());
        }
        else
        {
            bool checked = val.toBool();
            btn->setChecked(checked);
            btn->setText(checked ? tr("auto") : tr("mnl"));
        }
    }

    if (id == DataID::ChannelCurrentIds_update_send && sender)
    {
        CBChannel[0]->setCurrentIndex(CBChannel[0]->currentIndex());
    }

    if (id == DataID::ScanOnOff)
    {
        if (val.isNull())
        {
            auto mob = MobileServer::GetIt();
            if (mob && mob->need_send())
            {
                bool saveToData = false;
                QList<QVariant> list;
                list << m_BStart->on_off << saveToData;
                mob->setPropIdVar(DataID::ScanOnOff, list);
            }
        }
        else
            m_BStart->change_state_man_with_sig(val.toBool());
    }

    if (id == DataID::ScanNxNy) {
        int i = val.toInt();
        i += 3;

        CBPointsOX->blockSignals(true);
        CBPointsOY->blockSignals(true);
        CBPointsOXext->blockSignals(true);
        CBPointsOYext->blockSignals(true);
        CBPointsOX->setCurrentIndex(i);
        CBPointsOY->setCurrentIndex(i);
        CBPointsOXext->setCurrentIndex(i);
        CBPointsOYext->setCurrentIndex(i);
        CBPointsOX->blockSignals(false);
        CBPointsOY->blockSignals(false);
        CBPointsOXext->blockSignals(false);
        CBPointsOYext->blockSignals(false);
        check_nx_ny();
        sl_ext_mode();
    }
    if (id == DataID::ScanSaveCh) {
        if (val.toInt() == 4) {
            Plot->TwoDPlot->PlotControl->BSave->animateClick();
        } else {
            Plot->SetActivePlot(val.toInt());
            Plot->PlotControl->BSave->animateClick();
        }
    }

    if (id == DataID::PlotScalesChanged) {
        QVariantList lst = val.toList();
        plot::TScalesChangedEvent e(Range(), Range(), (plot::TScalesChangedEvent::TReason)lst.at(0).toInt());
        e.XLimits[0] = lst.at(1).toDouble();
        e.XLimits[1] = lst.at(2).toDouble();
        e.YLimits[0] = lst.at(3).toDouble();
        e.YLimits[1] = lst.at(4).toDouble();
        Plot->SetPlotsXScales(e.XLimits[0], e.XLimits[1]);
        Plot->SetPlotsYScales(e.YLimits[0], e.YLimits[1]);
        eventFilter(this, &e);
    }
}

void SPMModule::initialSend(MobileInstance *client)
{
    QStringList chnames;
    for (int i = 0; i < fbs->count(); i++)
        chnames.append(fbs->itemText(i));
    client->serv->setPropIdVarS(client->socket, DataID::SPM_FBch_list, chnames);
}
void SPMModule::setValue(int id, QVariant val, QObject *sender)
{
    if (id == DataID::Deflect_send)
        if (sender)
            sender->setProperty("deflect_need", val);

    if (id == DataID::Deflect_probeCheck)
        PBFreqStart->change_state_man_with_sig(true);
}

void MotorsModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::Motor_123)
        //TODO landing motors?
        goShuttle(val.toInt(), 250);
    if (id == DataID::Motor_1)
        goShuttle(val.toInt(), mapPhysToVirt(1));
    if (id == DataID::Motor_2)
        goShuttle(val.toInt(), mapPhysToVirt(2));
    if (id == DataID::Motor_3)
        goShuttle(val.toInt(), mapPhysToVirt(3));
    if (id == DataID::Motor_4)
        goShuttle(val.toInt(), mapPhysToVirt(4));
}

void CCDPlotModule::initialSend(MobileInstance *client)
{
    client->serv->setPropIdVarS(client->socket, DataID::spectrSetRelativeScales, currentRelativeScale);
    client->serv->setPropIdVarS(client->socket, DataID::spectrStartOnOff, m_BStart->isOn);
    client->serv->setPropIdVarS(client->socket, DataID::spectrSetInfMode, One_Cont->isOn);
    client->serv->setPropIdVarS(client->socket, DataID::spectrSetFocusMax, zPos->maximum());
    client->serv->setPropIdVarS(client->socket, DataID::spectrSetFocus, zPos->value());
    client->serv->setPropIdVarS(client->socket, DataID::spectrSetCutSpikes, Cut_Pikes->isOn);
    client->serv->setPropIdVarS(client->socket, DataID::spectrSetBaselineEnable, Base_line->isOn);
    client->serv->setPropIdVarS(client->socket, DataID::PlotTitle2d_Spectr, MyPl->TwoDPlot->Pure2dPlot->Title);
    
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrWaveLength, lineCentralWlMonoch->value());
    QList<QVariant> list;
    for (int i=0; i<CBgrating_select->count(); i++)
        list << CBgrating_select->itemText(i);
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrGratingList, list);
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrGrating, CBgrating_select->currentIndex());
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrEntrance, lineSlitInMonoch->value());
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrExisSlit, lineSlitOutMonoch->value());
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrExitMirror, CBexitport->currentIndex());
    client->serv->setPropIdVarS(client->socket, DataID::setMonochrEnabled, monochromatorWidgets.first()->isEnabled());
    
    client->serv->setPropIdVarS(client->socket, DataID::setCCDExposure, lineExpTime->value());
    client->serv->setPropIdVarS(client->socket, DataID::setCCDTemperature, lineSetTemp->value());
}

void CCDPlotModule::setValue(int id, QVariant val, QObject *sender)
{
    if (id == DataID::needPlot2D_Spectr)
        if (sender)
        {
            sender->setProperty(QString("need_2dPlotSpectr").toLocal8Bit().constData(), val);
            Send2dPlotData();
        }
    
    if (id == DataID::spectrStartOnOff)
    {
        bool state = val.toBool();
        m_BStart->change_state_man_with_sig(state);
    }
    
    if (id == DataID::spectrSetInfMode)
    {
        bool state = val.toBool();
        One_Cont->change_state_man_with_sig(state);
    }
    
    if (id == DataID::spectrSetRelativeScales)
    {
        int scale = val.toInt();
        AGScales->actions()[scale]->setChecked(true);
        AGScales->actions()[scale]->trigger();
    }
    
    if (id == DataID::spectrSetFocus)
    {
        double focus = val.toDouble();
        zPos->setValue(focus);
        zPos->setNewValue(true);
    }
    
    if (id == DataID::spectrSetCutSpikes)
    {
        bool state = val.toBool();
        Cut_Pikes->change_state_man_with_sig(state);
    }

    if (id == DataID::spectrStoreBaseline)
    {
        base_line_save->click();
    }
    
    if (id == DataID::spectrSetBaselineEnable)
    {
        bool state = val.toBool();
        Base_line->change_state_man_with_sig(state);
    }
    if (id == DataID::plot2DSaveToDataSpectr)
        MyPl->TwoDPlot->overlayButtons->B2DSendToData->click();
    
    if (id == DataID::setMonochrWaveLength)
    {
        lineCentralWlMonoch->setValue(val.toDouble());
        lineCentralWlMonoch->setNewValue();
    }
    if (id == DataID::setMonochrGrating)
    {
        CBgrating_select->setCurrentIndex(val.toInt());
    }
    if (id == DataID::setMonochrEntrance)
    {
        lineSlitInMonoch->setValue(val.toDouble());
        lineSlitInMonoch->setNewValue();
    }
    if (id == DataID::setMonochrExisSlit)
    {
        lineSlitOutMonoch->setValue(val.toDouble());
        lineSlitOutMonoch->setNewValue();
    }
    if (id == DataID::setMonochrExitMirror)
    {
        CBexitport->setCurrentIndex(val.toInt());
    }
    if (id == DataID::setCCDTemperature)
    {
        lineSetTemp->setValue(val.toDouble());
        lineSetTemp->setNewValue();
    }
    if (id == DataID::setCCDExposure)
    {
        lineExpTime->setValue(val.toDouble());
        lineExpTime->setNewValue();
    }
}
void VideoModule::setValue(int id, QVariant val, QObject *sender)
{
    //if (inst_idx != 0)
    //    return;

    if (id == DataID::Video_send)
        if (sender)
            sender->setProperty("video_need", val);

    if (id == DataID::VideoOnToggle)
        m_BStart->animateClick();
    if (id == DataID::VideoXYjoy)
        goXYjoy(val.toPointF());
    if (id == DataID::VideoZjoy)
        goZjoy(val.toDouble());
}
