#include "spmtunning.h"
#include "defs.h"

#include "gui/status.hpp"
#include "modules.hpp"
#include "remote_nst_dataid.hpp"
#include "mobile/remote_mobile.hpp"
#include "oscilloscope.h"
#include "nstslider.h"
#include "hardware/EG3000.h"
#include "ccd/ccdplot.h"

#include <QCheckBox>
#include <QFormLayout>
#include <QSettings>
#include <QTimer>
#include <QApplication>
#include <QDialog>
#include <QProgressBar>
#include "deflect_widget.hpp"
#include "scanplot.h"


typedef QPair<int, int> FBinfo;
Q_DECLARE_METATYPE(FBinfo)


void SPMTunning::sl_get_data(int ch, long val)
{
    if (cz_updated)
        return;

    if (ch == (FB_Scanner ? srcCZ2 : srcCZ)) {
        auto osc = Oscilloscope::GetIt();
        if (!osc)
            return;

        double zsp = osc->CalibWidget->reverse_nlc_sensor(FB_Scanner ? AxisZ2 : AxisZ, val);
        NS_Z_spectr->setValue(zsp);
        cz_updated = true;
    }
}

void SPMTunning::s_spectr_start(bool st)
{
    Spectr_Scan_On = st;
    bool jm_on = st && BFeedbackOnOff->isOn && BFeedbackOnOffMode->isOn && BSpectrJM->isOn;

    destructStatus(Spect_progBar);
    if (st) {
        if (modeBar->currentIndex() != 2)
            modeBar->setCurrentIndex(2);
        Spect_progBar = new StatusBar(tr("SPM Z scan"), jm_on ? 0 : 10000);
        Spect_progBar->bar->setFormat("%p%");
        connect(Spect_progBar, SIGNAL(canceled()), PBSpectr_start_stop, SLOT(click()));
        Spect_progBar->add();
    }

    auto osc = Oscilloscope::GetIt();

    bool cl_fix = BFeedbackOnOff->isOn && osc &&
            (FB_Scanner ? osc->CalibWidget->CL2_status_z :
                          osc->CalibWidget->CL_status_z);
    if (jm_on) cl_fix = false;
    if (st && cl_fix) {
        cz_updated = false;
        doGetCZ = true;

        QElapsedTimer t;
        t.start();
        while (!cz_updated && t.elapsed() < 2000) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }

        cz_updated = true;
        if (t.elapsed() >= 2000) {
            new Notification(tr("SPM Z scan") + ": " + tr("Can't get Cz"), Notification::Bad, 10);
            doBeep();
            PBSpectr_start_stop->change_state_man(false);

            auto scan = ScanPlot::GetIt();
            if (scan)
                scan->PBSpectr_start_stop->change_state_man(false);

            destructStatus(Spect_progBar);
            return;
        }
    }

    double z = (BFeedbackOnOff->isOn && !cl_fix) ? Z_scanner : NS_Z_spectr->value();
    Z1spectr = z - Z1;
    Z2spectr = z + Z2;
    SpeedSpectr = ev_Speed_spectr->value();
    SpectrChannel = CB_Channel->itemData(CB_Channel->currentIndex()).toInt();
    SpectrPoints = CB_Points_spectr->currentText().toInt();

    spect_onn_off = st ? (jm_on ? 2 : 1) : 0;
    sendSpectroscopy = true;

    if (st) {
        int div = (Z1 * 2 + Z2) * 2;
        if (div <= 0) div = 1;
        progDelta = ev_Speed_spectr->value() * 1e3 / div;
        Progress_timer->start(100);
    } else
        Progress_timer->stop();

#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (st && settings.value("usb", false).toBool()) {
        EG_ForceCurveData ForceCurve;
        ForceCurve.Size = 1000;
        ForceCurve.Source = srcMag1;
        for (int i = 0; i < 1000; i++) {
            ForceCurve.Data[i] = 65535 * exp(-0.05 * i);
            ForceCurve.DataBack[i] = 65535 * (1 - exp(-0.05 * (1000 - i - 1)));
        }

        processPacket(ADR_CPU, CPU_FORCE_CURVE_DATA, &ForceCurve, 0);
    }
#endif
}
void SPMTunning::s_spectr_jm(bool st)
{
    ev_Speed_spectr->setEnabled(!st);
    NS_Z1_spectr->setEnabled(!st);NS_Z1_spectr->update();
    ev_Z1_spectr->setEnabled(!st);
    NS_Z2_spectr->setEnabled(!st);
}

void SPMTunning::s_updSpectrProgBar()
{
    if (!Spect_progBar) //|| !Spectr_Scan_On || !PBSpectr_start_stop->on_off
        return;

	auto bar = Spect_progBar->bar;
    if (bar->maximum() && bar->value() < bar->maximum())
        Spect_progBar->setValue(qMin(bar->value() + progDelta, bar->maximum()));
    else {
        Progress_timer->stop();
        bar->setMaximum(0);
    }
}

void SPMTunning::s_setForceCurveData(QByteArray data)
{
    curveData_force[0]->clear();
    curveData_force[1]->clear();

    const EG_ForceCurveData *sdata = (const EG_ForceCurveData *)data.constData();
    int size = qMin<int>(FORCE_CURVE_LINE_SIZE, sdata->Size);
    bool jm_on = BFeedbackOnOff->isOn && BFeedbackOnOffMode->isOn && BSpectrJM->isOn;
    if (!Spectr_Scan_On || jm_on) { //jm
        Z1spectr = 0;
        Z2spectr = qMax(0.1, DCfbPrec->value() * JMdiv->value()) * size * 1e-3;
    }

    curveData_force[0]->setXCalib(Z1spectr * 1e-6, (Z2spectr - Z1spectr) / double(size - 1) * 1e-6);
    curveData_force[1]->setXCalib(Z1spectr * 1e-6, (Z2spectr - Z1spectr) / double(size - 1) * 1e-6);
    curveData_force[0]->YSeq.resize(0);
    curveData_force[0]->YSeq.reserve(size);
    curveData_force[1]->YSeq.resize(0);
    curveData_force[1]->YSeq.reserve(size);
    for (int i = 0; i < size; i++) {
        curveData_force[0]->AddY(input_calib(sdata->Source, sdata->Data[i]));
        curveData_force[1]->AddY(input_calib(sdata->Source, sdata->DataBack[i]));
    }

    Progress_timer->stop();
    PBSpectr_start_stop->change_state_man(false);

    auto scan = ScanPlot::GetIt();
    if (scan)
        scan->PBSpectr_start_stop->change_state_man(false);

    destructStatus(Spect_progBar);

    if (PlotMode == mode_Spectroscopy) {
        MyPl->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(CB_Channel->currentText(), input_units(CB_Channel->itemData(CB_Channel->currentIndex()).toInt()));
        MyPl->Refresh();
        MyPlScan->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(CB_Channel->currentText(), input_units(CB_Channel->itemData(CB_Channel->currentIndex()).toInt()));
        MyPlScan->Refresh();
        if (MyPl->Pure2dPlot->QwtCurves.size() >= 2) {
            MyPl->Pure2dPlot->QwtCurves[0]->setTitle(tr("Approach"));
            MyPl->Pure2dPlot->QwtCurves[1]->setTitle(tr("Retract"));
            MyPlScan->Pure2dPlot->QwtCurves[0]->setTitle(tr("Approach"));
            MyPlScan->Pure2dPlot->QwtCurves[1]->setTitle(tr("Retract"));
        }

        if (Spectr_Scan_On && AutoSend) {
            MyPl->on_SyncTimer_timeout();
            MyPl->PlotControl->BSave->animateClick();
            MyPlScan->on_SyncTimer_timeout();
            MyPlScan->PlotControl->BSave->animateClick();
        }

        if (Spectr_Scan_On)
            doBeep();
    }

    Spectr_Scan_On = false;

    if (auto_land_state == 2)
        doAutoLand(); //to next step
}


struct ResPeakInfo {
    bool valid;
    double resfreq;
    double Pres; //freq/q-normalized power
    double deltafreq; //width
    double qfactor;
};
ResPeakInfo SPMTunning::findPeakInfo(bool extended_slopes)
{
    ResPeakInfo result;
    result.valid = false;

    result.resfreq = curFrequency->value() * 1e3; //todo fit?
    int residx = 0;
	auto firstCurve = curveData_freq.at(0);
    do {
        if (firstCurve->getXat(residx) > result.resfreq)
            break;
        residx++;
    } while (residx < firstCurve->size());
    residx = qMax(0, residx-1);

    double Vres = firstCurve->YSeq.at(residx); //todo better fit
    int peakrange = qMax<int>(1, result.resfreq / 2000 / firstCurve->XSeq.at(1));
    for (int i = qBound(0, residx - peakrange, firstCurve->size()-1);
         i <= qBound(0, residx + peakrange, firstCurve->size()-1); i++)
        Vres = qMax(Vres, firstCurve->YSeq.at(i));
    double Vnoise = min(firstCurve->YSeq);
    if (Vres / Vnoise < 2)
        return result;
    if (Vnoise < Vres / 3) Vnoise = 0; //ignore low background noise variation - more stable q factor

    result.Pres = pow(Vres, 2) - pow(Vnoise, 2); //need to normalize by qfactor^2 later
    result.Pres /= result.resfreq; //passband normalization?

    QVector<bool> dir; QVector<int> pos;
    ccd::GetIntersections(firstCurve->YSeq, 0, firstCurve->YSeq.size()-1, 0.7*(Vres - Vnoise) + Vnoise, dir, pos);
    if (pos.size() < 2)
        return result;
    int resintidx = 0; //right slope idx
    while (pos.at(resintidx) < residx && resintidx < pos.size()-1) resintidx++;
    int resintidx0 = extended_slopes ? qMax(0, resintidx-1) : resintidx; //left slope idx
    while (!dir.at(resintidx0) && resintidx0 > 0) resintidx0--;
    while (dir.at(resintidx) && resintidx < pos.size()-1) resintidx++;
    if (!dir.at(resintidx0) || dir.at(resintidx))
        return result;
    result.deltafreq = (pos.at(resintidx) - pos.at(resintidx0)) * firstCurve->XSeq.at(1);
    result.qfactor = result.resfreq / result.deltafreq;
    if (result.qfactor < 10)
        return result;

    result.Pres = result.Pres / result.qfactor / result.qfactor; //normalize
    result.valid = true;
    return result;
}


bool SPMTunning::s_FreqScanStartAuto(bool use_current_range)
{
    auto scan = ScanPlot::GetIt();
    modeBar->setCurrentIndex(0);

    //reseting Q to "none"
    LaserSpot->setQ(true);
    LaserSpotWindow->setQ(true);
    if (scan)
        scan->LaserSpotWindow->setQ(true);

    double mino = EFreqScanMin->value();
    double maxo = EFreqScanMax->value();
    bool set = freqScanSet->isChecked();
    freqScanSet->setChecked(true);
    if (scan)
        scan->deviceSchemeWid->oscilator->chbSetOn->setChecked(true);

    double minf0 = lowPassFilter->value() * 1.1 + lowPassFilter->edit->minimum() * 1.5;
    double minf = minf0;
    double maxf = curFrequency->edit->maximum() * 0.9;
    double prev_freq;
    int check_low_freq = 0;
    bool last = false;
    if (use_current_range) {
        minf = mino;
        maxf = maxo;
        last = true;
    }

redo:
    EFreqScanMin->setValue(minf);
    EFreqScanMax->setValue(maxf);

    if (scan)
    {
        scan->deviceSchemeWid->oscilator->edFreqScanMin->setValue(minf);
        scan->deviceSchemeWid->oscilator->edFreqScanMax->setValue(maxf);
    }

    //measure resonance
    PBFreqStart->change_state_man_with_sig(true);
    if (scan)
        scan->deviceSchemeWid->oscilator->btnFreqStartScan->change_state_man(true);
    QElapsedTimer t; t.start();
    bool progbar_cancel = false;
    while (PBFreqStart->isOn && ModuleManager::instance->isRunning && t.elapsed() < 10000 && !(progbar_cancel = !progBar)) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }

    if (!ModuleManager::instance->isRunning || t.elapsed() >= 10000 || progbar_cancel) {
exit:
        EFreqScanMin->setValue(mino);
        EFreqScanMax->setValue(maxo);
        freqScanSet->setChecked(set);
        if (scan)
        {
            scan->deviceSchemeWid->oscilator->edFreqScanMin->setValue(mino);
            scan->deviceSchemeWid->oscilator->edFreqScanMax->setValue(maxo);
            scan->deviceSchemeWid->oscilator->chbSetOn->setChecked(set);
        }

        new Notification(tr("Frequency scan") + ": " + tr("Can't find peak"), Notification::Warn, 10);
        doBeep();
        return false;
    }


    if (((curFrequency->value() - minf) < 0.1 * (maxf - minf) && !check_low_freq) ||
            (check_low_freq && fabs(prev_freq - curFrequency->value()) > 0.01 * curFrequency->value())) {
        prev_freq = curFrequency->value();
        check_low_freq++;
        if (check_low_freq > 3)
            goto exit;
        goto redo;
    }
    check_low_freq = 0;


    ResPeakInfo info = findPeakInfo(false);
    if (!info.valid)
        goto exit;

    if (!last) {
        double basefreq = info.deltafreq * 6.5; //gaussian at 0.1A base width
        const double width_percent = 0.2;
        double width = (maxf - minf) * 1e3;
        width /= 2.7;
        if (basefreq > width_percent * width)
            last = true;
        double minfn = (info.resfreq - width / 2) / 1e3;
        double maxfn = (info.resfreq + width / 2) / 1e3;
        if (minfn < minf0) {
            maxfn += minf0 - minfn;
            minfn = minf0;
        }
        minf = minfn;
        maxf = maxfn;
        goto redo;
    }

    LaserSpot->setQ(false, EFreqScanMin->value(), EFreqScanMax->value());
    LaserSpotWindow->setQ(false, EFreqScanMin->value(), EFreqScanMax->value());
    if (scan)
        scan->LaserSpotWindow->setQ(false, EFreqScanMin->value(), EFreqScanMax->value());

    doBeep();
    return true;
}


void SPMTunning::s_FreqScanStart(bool st)
{
    auto scan = ScanPlot::GetIt();

    PBFreqStart->setEnabled(!st);
    PBFreqStartAuto->setEnabled(!st);
    if (scan)
    {
        scan->deviceSchemeWid->oscilator->btnAutoFreqScan->setEnabled(!st);
        scan->deviceSchemeWid->oscilator->btnFreqStartScan->setEnabled(!st);
    }
    destructStatus(progBar);
    if (st) {
        fr_start = EFreqScanMin->value()*1000.0;
        fr_end = EFreqScanMax->value()*1000.0;
        fr_channel = CB_fr_Channel->itemData(CB_fr_Channel->currentIndex()).toInt();
        sendFreqScan = true;

        progBar = new StatusBar(tr("Frequency scan"), 10000);
        progBar->bar->setFormat("%p%");
        //connect(progBar, SIGNAL(canceled()), PBFreqStart, SLOT(click()));
        progBar->add();
        progDelta = 200;
        if (l2k_controller)
            progDelta *= 3;
        Progress_timer->start(100);
        sl_save_to_file();
    } else
        Progress_timer->stop();

#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (st && settings.value("usb", false).toBool()) {
        int resfr = 250 + rand() % 500;
        EG_LockInScanData FreqData;
        FreqData.Size = 1000;
        FreqData.Source = srcMag1;
        for (int i = 0, resI = 0; i < 1000; i++)
        {
            FreqData.Data[i] = (WORD)(30000*exp(-0.0001*(i-resfr)*(i-resfr))+32768);
            if (FreqData.Data[i] > FreqData.Data[resI])
            {
                FreqData.res_point = (float)i*1000.0 + rand()%999;
                resI = i;
            }
        }

        processPacket(ADR_CPU, CPU_FREQSCAN_DATA, &FreqData, 0);
    }
#endif
}
void SPMTunning::s_updFreqProgBar()
{
    if (!PBFreqStart->isOn || !progBar)
        return;

	auto bar = progBar->bar;
    if (bar->maximum() && bar->value() < bar->maximum())
        progBar->setValue(qMin(bar->value() + progDelta, bar->maximum()));
    else {
        Progress_timer->stop();
        bar->setMaximum(0);
    }
}

void SPMTunning::s_setFrequencyScanData(QByteArray data)
{
    auto scan = ScanPlot::GetIt();

    curveData_freq[0]->clear();

    const EG_LockInScanData *sdata = (const EG_LockInScanData *)data.constData();
    int size = qMin<int>(LOCKIN_LINE_SIZE, sdata->Size);

    curveData_freq[0]->setXCalib(EFreqScanMin->value() * 1e3, (EFreqScanMax->value() - EFreqScanMin->value()) / double(size - 1) * 1e3);
    curveData_freq[0]->YSeq.resize(0);
    curveData_freq[0]->YSeq.reserve(size);
    QList<QVariant> freqData;
    for (int i = 0; i < size; i++)
    {
        qreal y = input_calib(sdata->Source, sdata->Data[i]);
        curveData_freq[0]->AddY(y);
        LaserSpot->AddFreqCurveData(i, y);
        LaserSpotWindow->AddFreqCurveData(i, y);
        freqData << y;
        if (scan)
            scan->deviceSchemeWid->oscilator->AddFreqCurveData(i, y);
    }
    LaserSpot->SetFreqCurveDataMaximum(sdata->res_point/1000.0f);
    LaserSpotWindow->SetFreqCurveDataMaximum(sdata->res_point/1000.0f);

    if (auto mob = MobileServer::GetIt()) {
        if (mob->any_need_data("deflect_need"))
        {
            auto list = QVariantList();
            list << QVariant(freqData);
            list << sdata->res_point/1000.0f;
            mob->sendData(DataID::Deflect_dataFreq, list);
        }
    }

    if (scan)
        scan->deviceSchemeWid->oscilator->SetFreqCurveDataMaximum(sdata->res_point/1000.0f);

    if (freqScanSet->isChecked()) {
        double value = ((float)((unsigned long)(sdata->res_point * 1000))) / 1000000.;

        curFrequency->setValue(value);
        if (scan)
            scan->deviceSchemeWid->oscilator->edFreq->setValue(value);
        MobileServer::sendData(DataID::SPM_freq, value);

        MyPl->Pure2dPlot->MoveVertLine(value * 1000, 0);
        MyPlScan->Pure2dPlot->MoveVertLine(value * 1000, 0);
    } else {
        curFrequency->setNewValue();
    }

    if (PlotMode == mode_ResCurve) {
        MyPl->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(CB_fr_Channel->currentText(), input_units(CB_fr_Channel->itemData(CB_fr_Channel->currentIndex()).toInt()));
        MyPl->Refresh();
        MyPlScan->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(CB_fr_Channel->currentText(), input_units(CB_fr_Channel->itemData(CB_fr_Channel->currentIndex()).toInt()));
        MyPlScan->Refresh();
    }

    PBFreqStart->change_state_man_with_sig(false);
    if (scan)
        scan->deviceSchemeWid->oscilator->btnFreqStartScan->change_state_man(false);
}

bool SPMTunning::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD )
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_GET_DATA) {
        EG_GetData *Get_data = (EG_GetData *)data;
        BYTE src = Get_data->Source;
        DWORD data = Get_data->Data;

        if (!cz_updated && (src == srcCZ || src == srcCZ2))
            QMetaObject::invokeMethod(this, "sl_get_data", Q_ARG(int, src), Q_ARG(long, data));

        return false;
    }

    if (sub_addr == CPU_LANDING_STATUS) {
        WORD landing_status;
        memcpy(&landing_status, data, sizeof(WORD));
        QMetaObject::invokeMethod(this, "slot_set_Landing_status", Q_ARG(int, (int)landing_status));

        return true;
    }


    if (sub_addr == CPU_STATUS) {
        EG_Status *Deflectomter = (EG_Status *)data;

        nDFLr = Deflectomter->DFL / 65535. - 0.5;
        nLFLr = Deflectomter->LFL / 65535. - 0.5;
        nDFL = input_calib(srcDFL, Deflectomter->DFL);
        nLFL = input_calib(srcLFL, Deflectomter->LFL);
        nMAG = input_calib(srcMag1, Deflectomter->MAG);

        nInt = Deflectomter->DInt / 32767. - 1;
        nZ = Deflectomter->Zpos / 65535.;

        QMetaObject::invokeMethod(this, "set_spot_data");

        return false;//true;
    }

    if (sub_addr == CPU_FREQSCAN_DATA) {
        EG_LockInScanData *FreqData = (EG_LockInScanData *)data;

        int dataSize = sizeof(EG_LockInScanData) - (LOCKIN_LINE_SIZE - FreqData->Size)*sizeof(WORD);
        QMetaObject::invokeMethod(this, "s_setFrequencyScanData", Q_ARG(QByteArray, QByteArray((char *)FreqData, dataSize)));

        return true;
    }

    if (sub_addr == CPU_FORCE_CURVE_DATA) {
        EG_ForceCurveData *ForceCurve = (EG_ForceCurveData *)data;

        int dataSize = sizeof(EG_ForceCurveData) - (FORCE_CURVE_LINE_SIZE - ForceCurve->Size)*sizeof(WORD);
        QMetaObject::invokeMethod(this, "s_setForceCurveData", Q_ARG(QByteArray, QByteArray((char *)ForceCurve, dataSize)));

        return true;
    }
    
    #ifndef _OLD_AUTO_FINDING_MAG
    
    if (sub_addr == CPU_AUTO_MAG_ENDED) {
        autoFindingMAGEnded = true;
        return true;
    }
    
#endif
    
    if (sub_addr == CPU_SET_EXCITATION_A)
    {
        float *t = (float *)data;
        newTipDriveValue = *t;
        QMetaObject::invokeMethod(this, "updateTipDrive");
        return true;
    }

    return false;
}

/*
void SPMTunning::doVoltageScan()
{
    setEnabled(false);
    QVector<double> y;
    double drive1 = driveInt->value();
    for (int i = 0; i < 100; i++) {
        driveInt->setValue(drive1/100*i);
        driveInt->setNewValue();
        QElapsedTimer t;t.start();
        while (t.elapsed() < 10) {
            qApp->processEvents();
            SleepExport::msleep2(2);
        }
        double v;
        _mainw->osc->acp->read_osc(&v, srcMag1);
        y << v;
    }

    curveData_freq[0]->clear();

    curveData_freq[0]->setXCalib(0, drive1/100);
    curveData_freq[0]->YSeq.resize(0);
    curveData_freq[0]->YSeq.reserve(y.size());
    for (int i = 0; i < y.size(); i++)
        curveData_freq[0]->AddY(input_calib(srcMag1, y.at(i)));

    MyPl->Refresh();

    setEnabled(true);
    return;
}
*/



void SPMTunning::doForceCalib()
{
    int trygain = gain1->count(); //todo config?
redo:
    trygain--;
    if (trygain < 0) {
        new Notification(tr("Frequency scan") + ": " + tr("Can't find peak"), Notification::Warn, 10);
        doBeep();
        return;
    }

    setEnabled(false);

    auto scan = ScanPlot::GetIt();

    int previdx = fbs->currentIndex();
    if (srcMag1 != fbs->itemData(previdx).value<QPair<int, int> >().second)
    {
        for (int i = 0; i < fbs->count(); i++)
            if (srcMag1 == fbs->itemData(i).value<QPair<int, int> >().second) {
                fbs->setCurrentIndex(i);
                break;
            }
        if (scan)
            scan->deviceSchemeWid->PIDSwitch->currentConnectedIn = 3;
    }

    double drive = driveInt->value();
    driveInt->setValue(0);
    driveInt->setNewValue();
    int gainidx = gain1->currentIndex();
    gain1->setCurrentIndex(trygain);
    if (scan)
    {
        scan->deviceSchemeWid->oscilator->edAmp->setValue(0);
        scan->deviceSchemeWid->inputGain->cbDFL->setCurrentIndex(trygain);
        scan->deviceSchemeWid->oscilator->chbSetOn->setChecked(false);
    }
    CBcalibDFL_F->setChecked(false);
    freqScanSet->setChecked(false);
    double qc = EQControlGain->value();
    EQControlGain->setValue(0);
    EQControlGain->setNewValue();


    //measure thermal noise spectrum
    PBFreqStart->change_state_man_with_sig(true);
    if (scan)
        scan->deviceSchemeWid->oscilator->btnFreqStartScan->change_state_man(true);
    QElapsedTimer t; t.start();
    bool progbar_cancel = false;
    while (PBFreqStart->isOn && ModuleManager::instance->isRunning && t.elapsed() < 10000 && !(progbar_cancel = !progBar)) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }

    double gain = 1e-9 / gain1->itemData(gain1->currentIndex()).toDouble();
    gain1->setCurrentIndex(gainidx);
    if (scan)
    {
        scan->deviceSchemeWid->inputGain->cbDFL->setCurrentIndex(gainidx);
        scan->deviceSchemeWid->oscilator->edAmp->setValue(drive);
        scan->deviceSchemeWid->oscilator->chbSetOn->setChecked(true);
    }
    driveInt->setValue(drive);
    driveInt->setNewValue();
    fbs->setCurrentIndex(previdx);
    if (scan)
    {
        if (srcMag1 == fbs->itemData(previdx).value<QPair<int, int> >().second)
            scan->deviceSchemeWid->PIDSwitch->currentConnectedIn = 3;
        else if (srcPhase1 == fbs->itemData(previdx).value<QPair<int, int> >().second)
            scan->deviceSchemeWid->PIDSwitch->currentConnectedIn = 0;
        else if (srcDFL == fbs->itemData(previdx).value<QPair<int, int> >().second)
            scan->deviceSchemeWid->PIDSwitch->currentConnectedIn = 4;
        else if (srcLFL == fbs->itemData(previdx).value<QPair<int, int> >().second)
            scan->deviceSchemeWid->PIDSwitch->currentConnectedIn = 5;
    }
    freqScanSet->setChecked(true);
    EQControlGain->setValue(qc);
    EQControlGain->setNewValue();
    setEnabled(true);

    if (!ModuleManager::instance->isRunning || t.elapsed() >= 10000 || progbar_cancel)
        return;


    //fit/set model params
    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    double width = 1e-6 * settings_double(settings, "spm/force_calib/tip_width", 30);
    double length = 1e-6 * settings_double(settings, "spm/force_calib/tip_length", 125);

    ResPeakInfo info = findPeakInfo();
    if (!info.valid)
        goto redo;

    if (l2k_controller) //l2k different gain for low voltage?
        info.Pres *= 6.2*6.2;

    //some calculus
    const double air_density = 1.18;
    const double air_viscosity = 1.86e-5;
    const double water_density = 997.7;
    const double water_viscosity = 9.98e-4;
    bool in_liquid = settings.value("spm/force_calib/in_liquid", false).toBool();
    double calib_density = in_liquid ? water_density : air_density;
    double calib_viscosity = in_liquid ? water_viscosity : air_viscosity;
    double reynolds = calib_density * width * width * (2 * M_PI * info.resfreq) / 4 / calib_viscosity;
    double a0 = 0.9;//0.8-1.0
    bool triangle_tip = settings.value("spm/force_calib/tip_triangular", false).toBool();
    if (triangle_tip)
        a0 = 1.3; //1.2-1.5
    double a1 = -0.7;
    double a2 = 0.035 * log(reynolds);
    double somefluffybullshit = a0 * pow(reynolds, a1 + a2); //0.1906* imaginary part of hydrodynamic function at res freq...
    double k = calib_density * width * width * length * info.qfactor * (2 * M_PI * info.resfreq) * (2 * M_PI * info.resfreq) * somefluffybullshit; //generalised sader 2012
    qDebug() << "k Q Re";
    qDebug() << k << info.qfactor << reynolds;

    const double kbolz = 1.380648813e-23;
    const double Temp = 273.16 + 22;
    double invols = sqrt(2 * kbolz * Temp / M_PI / k / info.resfreq / info.Pres / info.qfactor); //equipartition theorem - afm ols 2006
    qDebug() << invols;


    //apply
    double voltagetoforce = invols * k / gain;
    qDebug() << voltagetoforce;
    if (voltagetoforce > 0.1) {
        EcalibDFL_F->setValue(voltagetoforce);
        EcalibDFL_F->setNewValue();
        CBcalibDFL_F->setChecked(true);
    }

    doBeep();
}
void SPMTunning::doForceCalibConfig()
{
    QSettings settings("Param/gui.conf", QSettings::IniFormat);

    QDialog *d = new QDialog;
    d->setWindowTitle(tr("Force calib settings"));
    QFormLayout *lay = makeFormLayout(d);
    Edit *width;
    lay->addRow(tr("Tip width, ")+tr("mkm"), width = new Edit(0, 1000, 1));
    Edit *length;
    lay->addRow(tr("Tip length, ")+tr("mkm"), length = new Edit(0, 1000, 1));
    QCheckBox *triangular;
    lay->addRow(triangular = new QCheckBox(tr("Triangular tip")));
    QCheckBox *liquid;
    lay->addRow(liquid = new QCheckBox(tr("In liquid")));

    lay->addItem(new QSpacerItem(0, 0));
    lay->addRow(new QLabel("References: Sader et al."));
    QLabel *l;
    lay->addRow(l = new QLabel("<a href='http://dx.doi.org/10.1063/1.2162455'>Determination of optical lever sensitivity...</a>"));
    l->setOpenExternalLinks(true);
    lay->addRow(l = new QLabel("<a href='http://dx.doi.org/10.1063/1.4757398'>Spring constant calibration...</a>"));
    l->setOpenExternalLinks(true);

    double widthv = settings_double(settings, "spm/force_calib/tip_width", 30);
    double lengthv =settings_double(settings, "spm/force_calib/tip_length", 125);
    bool in_liquid = settings.value("spm/force_calib/in_liquid", false).toBool();
    bool triangle_tip = settings.value("spm/force_calib/tip_triangular", false).toBool();
    width->setValue(widthv);
    length->setValue(lengthv);
    liquid->setChecked(in_liquid);
    triangular->setChecked(triangle_tip);

    d->exec();

    if (widthv != width->value())
        settings.setValue("spm/force_calib/tip_width", width->value());
    if (lengthv != length->value())
        settings.setValue("spm/force_calib/tip_length", length->value());
    if (triangle_tip != triangular->isChecked())
        settings.setValue("spm/force_calib/tip_triangular", triangular->isChecked());
    if (in_liquid != liquid->isChecked())
        settings.setValue("spm/force_calib/in_liquid", liquid->isChecked());

    d->deleteLater();
}


//qselect
template<class T>
int partition(T *list, int l, int r, int pv)
{
     T pivotValue = list[pv];
     qSwap(list[pv], list[r]); // Move pivot to end
     int storeIndex = l;
     for (int i = l; i < r; i++)
         if (list[i] < pivotValue) {
             qSwap(list[storeIndex], list[i]);
             storeIndex++;
         }

     qSwap(list[r], list[storeIndex]); // Move pivot to its final place
     return storeIndex;
}
// Returns the n-th smallest element of list within left..right inclusive
template<class T>
double qselect(T *list, int l, int r, int n)
{
redo:
    if (l == r)
        return list[l];

    int pivotIndex  = (l + r) / 2;
    pivotIndex = partition(list, l, r, pivotIndex);
    if (n == pivotIndex)
        return list[n];

    if (n < pivotIndex)
        r = pivotIndex - 1;
    else
        l = pivotIndex + 1;
    goto redo;
}


#include <algorithm>
bool SPMTunning::doPICalib()
{
    auto scan = ScanPlot::GetIt();
    const int msize = 3;

    if (curveData_force.at(0)->size() < 2*msize) {
        new Notification(tr("PI calib") + ": " + tr("No data"), Notification::Warn, 10);
        doBeep();
        return false;
    }

    //median filter
    QVector<double> YSeq_filtered = curveData_force.at(0)->YSeq;
    double tmp[msize*2+1];
    for (int i = msize; i < YSeq_filtered.size() - msize; i++) {
        memcpy(tmp, curveData_force.at(0)->YSeq.constData() + i - msize, sizeof(double) * (msize*2+1));
        YSeq_filtered[i] = qselect(tmp, 0, (msize*2+1)-1, (msize+1)-1);
    }

    std::pair<QVector<double>::const_iterator, QVector<double>::const_iterator> minmax;
    minmax = std::minmax_element(YSeq_filtered.constBegin()+msize, YSeq_filtered.constEnd()-msize);

    double minv = *minmax.first;
    double maxv = *minmax.second;
    double setpoint = DCfbSetPoint->value();
    if (setpoint <= (minv*0.9 + maxv*0.1) || setpoint >= (minv*0.1 + maxv*0.9)) {
        setpoint = (minv + maxv) / 2;
        if (!BSpectrJM->isOn)
            qDebug() << "pi setpoint not in range";
    }

    QVector<bool> dir; QVector<int> pos;
    ccd::GetIntersections(YSeq_filtered, msize, YSeq_filtered.size()-1-msize, setpoint, dir, pos);
    if (!pos.size() || pos.size() > 15) {
        new Notification(tr("PI calib") + ": " + tr("No data for setpoint"), Notification::Warn, 10);
        doBeep();
        return false;
    }

    int minpos = std::distance(YSeq_filtered.constBegin(), minmax.first);
    int maxpos = std::distance(YSeq_filtered.constBegin(), minmax.second);
    int dz_pos = pos.at(pos.size() / 2);

    bool sign = Mode_data[currModeIdx][4];
    double dVz = (setpoint - (sign ? maxv : minv)) /
            (curveData_force.at(0)->getXat(dz_pos) - curveData_force.at(0)->getXat(sign ? maxpos : minpos));
    dVz = qFabs(dVz / input_gain(FB_channel)) * (FB_Scanner ? Zrange2 : Zrange1) * 1e-6 / 65535. / 20. * 65535.; //z->16b, spm_sig->16b
    qDebug() << "dV/dZ" << dVz;

    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    double ifactor = settings.value("pi_autocalib_spm_i_scale", 2).toDouble();
    double pfactor = settings.value("pi_autocalib_spm_p_scale", 0.44).toDouble();
    DCfbP->setValue(qMax(0.1, dVz * pfactor));
    //1 bit correction takes 1 second = (20000 cycles)
    //100 = spm I prescale
    DCfbI->setValue(qMax(0.001, dVz / ifactor / 20000 * 100));
    DCfbP->setNewValue();
    DCfbI->setNewValue();
    if (scan)
    {
        scan->deviceSchemeWid->pid->edPr->setValue(qMax(0.1, dVz * pfactor));
        scan->deviceSchemeWid->pid->edInt->setValue(qMax(0.001, dVz / ifactor / 20000 * 100));
    }

    doBeep();
    return true;
}

bool SPMTunning::doCheckFixFBosc()
{
    if (!BFeedbackOnOff->isOn || BFeedbackOnOffMode->isOn)
        return false;

    auto scan = ScanPlot::GetIt();

    setEnabled(false);

    int oldT = ePeriod->value();
    ePeriod->setValue(30);

    int c = 0;
    QElapsedTimer t;
    auto osc = Oscilloscope::GetIt();
    if (!osc)
        return false;

    //fb off
    if (scan)
        scan->deviceSchemeWid->feedBack->setState(false);
    BFeedbackOnOff->change_state_man_with_sig(false);

    t.start();
    while (t.elapsed() < 1500) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }


    //full osc restart, clears old curve
    PlotMode = mode_Osc;
    sl_setPlotMode();
    PlotMode = mode_ResCurve;
    sl_setPlotMode();
    PlotMode = mode_Osc;
    sl_setPlotMode();


    t.start();
    while (t.elapsed() < 400) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }

    double floor_noise = qMax(0.002, rms(osc->curveData.at(0)->YSeq) / osc_y_postscale);


    //fb on
    if (scan)
        scan->deviceSchemeWid->feedBack->setState(true);
    BFeedbackOnOff->change_state_man_with_sig(true);

    t.start();
    while (t.elapsed() < 1500) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }


redo:
    //full osc restart, clears old curve
    PlotMode = mode_Osc;
    sl_setPlotMode();
    PlotMode = mode_ResCurve;
    sl_setPlotMode();
    PlotMode = mode_Osc;
    sl_setPlotMode();


    t.start();
    while (t.elapsed() < 400) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }


    bool ok = true;
    double vavg = avg(osc->curveData.at(0)->YSeq) / osc_y_postscale;
    double vrms = rms(osc->curveData.at(0)->YSeq) / osc_y_postscale;
    if (vrms > qMax(0.04 * qFabs(vavg), floor_noise)) { //4%
        qDebug() << "FB osc" << vavg << vrms << floor_noise << DCfbP->value();
        ok = false;
        DCfbP->setValue(qMax(0.1, DCfbP->value() * 0.7));
        DCfbI->setValue(qMax(0.001, DCfbI->value() * 0.7));
        DCfbP->setNewValue();
        DCfbI->setNewValue();
        if (scan)
        {
            scan->deviceSchemeWid->pid->edPr->setValue(qMax(0.1, DCfbP->value() * 0.7));
            scan->deviceSchemeWid->pid->edInt->setValue(qMax(0.001, DCfbI->value() * 0.7));
        }
        c++;
        if (DCfbP->value() > 0.1 && c < 10) {
            t.start();
            while (t.elapsed() < 70) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            goto redo;
        }
    }


    ePeriod->setValue(oldT);

    setEnabled(true);
    return ok;
}


void SPMTunning::createSpectroscopyWorkSpace()
{
    PBSpectr_start_stop = new TwoStateButton(":start.png",":stop.png");
    connect(PBSpectr_start_stop, SIGNAL(state_changed(bool)), this, SLOT(s_spectr_start(bool)));

    BSpectrJM = new TwoStateButton(":/SPMTuning_window/JmOf.png", ":/SPMTuning_window/JmOn.png");
    BSpectrJM->setToolTip(tr("Get data from JM feedback, when both are enabled"));
    connect(BSpectrJM, SIGNAL(state_changed(bool)), this, SLOT(s_spectr_jm(bool)));

    JMdiv = new Edit(1, 255);
    connect(JMdiv, SIGNAL(newValue(int)), this, SLOT(spm_jm_div(int)));

    NS_Z1_spectr = new NSTSlider(Zrange, 0, 4);
    connect(NS_Z1_spectr, SIGNAL(valueChanged(double)), this, SLOT(s_set_Z1_spectr(double)));
    connect(NS_Z1_spectr, SIGNAL(sliderReleased()), this, SLOT(sl_up_lines_on()));
    connect(NS_Z1_spectr, SIGNAL(sliderPressed()), this, SLOT(sl_up_lines_of()));

    ev_Z1_spectr = new Edit(0, 500, 4);
    connect(ev_Z1_spectr, SIGNAL(newValue(double)), NS_Z1_spectr, SLOT(setValue(double)));
    connect(ev_Z1_spectr, SIGNAL(editingFinished()), this, SLOT(sl_up_lines_on()));
    connect(ev_Z1_spectr, SIGNAL(editingStarted()), this, SLOT(sl_up_lines_of()));

    NS_Z2_spectr = new DoubleChanger(0, Zrange, 4);
    connect(NS_Z2_spectr->slider, SIGNAL(valueChanged(double)), this, SLOT(s_set_Z2_spectr(double)));
    connect(NS_Z2_spectr, SIGNAL(newValue(double)), this, SLOT(s_set_Z2_spectr(double)));
    connect(NS_Z2_spectr, SIGNAL(editingFinished()), this, SLOT(sl_up_lines_on()));
    connect(NS_Z2_spectr, SIGNAL(editingStarted()), this, SLOT(sl_up_lines_of()));

    NS_Z_spectr = new DoubleChanger(0, Zrange, 4);
    connect(NS_Z_spectr->slider, SIGNAL(valueChanged(double)), this, SLOT(s_set_Z_spectr(double)));
    connect(NS_Z_spectr, SIGNAL(newValue(double)), this, SLOT(s_set_Z_spectr(double)));
    connect(NS_Z_spectr, SIGNAL(newValue()), this, SLOT(s_send_Zpos()));
    connect(NS_Z_spectr, SIGNAL(editingFinished()), this, SLOT(sl_up_lines_on()));
    connect(NS_Z_spectr, SIGNAL(editingStarted()), this, SLOT(sl_up_lines_of()));

    CB_Points_spectr = new QComboBox;
    CB_Points_spectr->addItem("50");  CB_Points_spectr->addItem("100");
    CB_Points_spectr->addItem("200"); CB_Points_spectr->addItem("300");
    CB_Points_spectr->addItem("400"); CB_Points_spectr->addItem("500");
    CB_Points_spectr->addItem("600"); CB_Points_spectr->addItem("700");
    CB_Points_spectr->addItem("800"); CB_Points_spectr->addItem("900");
    CB_Points_spectr->addItem("1000");
    CB_Points_spectr->setCurrentIndex(10);

    CB_Channel = new QComboBox;
    fillCBchannels_(CB_Channel, false);

    ev_Speed_spectr = new Edit(0.1, 1000, 1);
    ev_Speed_spectr->setImmediateMode();
    ev_Speed_spectr->setValue(10);
    NS_Z1_spectr->setValue(Z1);
    NS_Z2_spectr->setValue(Z2);

    PBPIDCalib = new QPushButton(tr("Find PI"));
    PBPIDCalib->setToolTip(tr("From force curve slope"));
    connect(PBPIDCalib, SIGNAL(clicked()), this, SLOT(doPICalib()));
}

void SPMTunning::createFreqWorkSpace()
{
    auto scan = ScanPlot::GetIt();

    CB_fr_Channel = new QComboBox;
    fillCBchannels_(CB_fr_Channel, false);
    CB_fr_Channel->setCurrentIndex(CB_fr_Channel->findData(srcMag1));
    if (scan)
        scan->deviceSchemeWid->oscilator->cbChannel->setCurrentIndex(CB_fr_Channel->findData(srcMag1));

    EFreqScanMin = new Edit(0, 20000, 1);
    EFreqScanMin->setImmediateMode();
    EFreqScanMin->setValue(50);

    EFreqScanMax = new Edit(0, 20000, 1);
    EFreqScanMax->setImmediateMode();
    EFreqScanMax->setValue(500);

    PBFreqStart = new TwoStateButton(":start.png",":stop.png");
    PBFreqStart->setProperty("id", DataID::SPM_freq_scan);
    PBFreqStart->setToolTip(tr("Frequency scan start"));
    connect(PBFreqStart, SIGNAL(state_changed(bool)), this, SLOT(s_FreqScanStart(bool)));

    PBFreqStartAuto = new QPushButton(makeHugeIcon(":AutoStart.png"), "");
    PBFreqStartAuto->setToolTip(tr("Auto freq scan"));
    connect(PBFreqStartAuto, SIGNAL(clicked()), this, SLOT(s_FreqScanStartAuto()));

    PBForceCalib = new QPushButton(tr("Calibrate Force"));
    connect(PBForceCalib, SIGNAL(clicked()), this, SLOT(doForceCalib()));
    PBForceCalib->setToolTip(tr("See settings window for reference papers information."));

    PBForceCalibConfig = new QPushButton(makeHugeIcon(":/MainToolbar/tools.png"), "");
    PBForceCalibConfig->setToolTip(tr("Tip geometry"));
    connect(PBForceCalibConfig, SIGNAL(clicked()), this, SLOT(doForceCalibConfig()));
}
